import { BABYLON } from "../render3d/babylon";

const PI_ALPHA_SRCCOLOR_ONE = 20;
const PI_ALPHA_ONE_ONEMINUSSRCALPHA = 21;
const PI_ALPHA_SRCCOLOR_ONEMINUSSRCALPHA = 22;
const PI_ALPHA_SRCCOLOR_ONEMINUSSRCCOLOR = 23;

export function initialAlphaModeExtends() {

    BABYLON.ThinEngine.prototype.setAlphaMode = function(mode: number, noDepthWriteChange: boolean = false): void {
        if (this._alphaMode === mode) {
            return;
        }

        const gl: WebGLRenderingContext = this._gl;

        switch (mode) {
            case BABYLON.Constants.ALPHA_DISABLE:
                this._alphaState.alphaBlend = false;
                break;
            case PI_ALPHA_ONE_ONEMINUSSRCALPHA:
                this._alphaState.setAlphaBlendFunctionParameters(gl.ONE, gl.ONE_MINUS_SRC_ALPHA, gl.ONE, gl.ONE_MINUS_SRC_ALPHA);
                this._alphaState.alphaBlend = true;
                break;
            case BABYLON.Constants.ALPHA_PREMULTIPLIED_PORTERDUFF:
                this._alphaState.setAlphaBlendFunctionParameters(gl.ONE, gl.ONE_MINUS_SRC_ALPHA, gl.ONE, gl.ONE_MINUS_SRC_ALPHA);
                this._alphaState.alphaBlend = true;
                break;
            case BABYLON.Constants.ALPHA_ONEONE:
                // this._alphaState.setAlphaBlendFunctionParameters(gl.ONE, gl.ONE, gl.ZERO, gl.ONE);
                this._alphaState.setAlphaBlendFunctionParameters(gl.ONE, gl.ONE, gl.ONE, gl.ONE_MINUS_SRC_ALPHA);
                this._alphaState.alphaBlend = true;
                break;
            case BABYLON.Constants.ALPHA_SCREENMODE:
                this._alphaState.setAlphaBlendFunctionParameters(gl.ONE, gl.ONE_MINUS_SRC_COLOR, gl.ONE, gl.ONE_MINUS_SRC_ALPHA);
                this._alphaState.alphaBlend = true;
                break;
            case PI_ALPHA_SRCCOLOR_ONE:
                this._alphaState.setAlphaBlendFunctionParameters(gl.SRC_COLOR, gl.ONE, gl.ONE, gl.ONE_MINUS_SRC_ALPHA);
                this._alphaState.alphaBlend = true;
                break;
            case PI_ALPHA_SRCCOLOR_ONEMINUSSRCALPHA:
                this._alphaState.setAlphaBlendFunctionParameters(gl.SRC_COLOR, gl.ONE_MINUS_SRC_ALPHA, gl.ONE, gl.ONE_MINUS_SRC_ALPHA);
                this._alphaState.alphaBlend = true;
                break;
            case PI_ALPHA_SRCCOLOR_ONEMINUSSRCCOLOR:
                this._alphaState.setAlphaBlendFunctionParameters(gl.SRC_COLOR, gl.ONE_MINUS_SRC_COLOR, gl.ONE, gl.ONE_MINUS_SRC_ALPHA);
                this._alphaState.alphaBlend = true;
                break;
            case BABYLON.Constants.ALPHA_COMBINE:
                // this._alphaState.setAlphaBlendFunctionParameters(gl.SRC_ALPHA, gl.ONE_MINUS_SRC_ALPHA, gl.ONE, gl.ONE);
                this._alphaState.setAlphaBlendFunctionParameters(gl.SRC_ALPHA, gl.ONE_MINUS_SRC_ALPHA, gl.ONE, gl.ONE_MINUS_SRC_ALPHA);
                this._alphaState.alphaBlend = true;
                break;
            case BABYLON.Constants.ALPHA_MAXIMIZED:
                // this._alphaState.setAlphaBlendFunctionParameters(gl.SRC_ALPHA, gl.ONE_MINUS_SRC_COLOR, gl.ONE, gl.ONE);
                this._alphaState.setAlphaBlendFunctionParameters(gl.SRC_ALPHA, gl.ONE_MINUS_SRC_COLOR, gl.ONE, gl.ONE_MINUS_SRC_ALPHA);
                this._alphaState.alphaBlend = true;
                break;
            case BABYLON.Constants.ALPHA_ADD:
                // this._alphaState.setAlphaBlendFunctionParameters(gl.SRC_ALPHA, gl.ONE, gl.ZERO, gl.ONE);
                this._alphaState.setAlphaBlendFunctionParameters(gl.SRC_ALPHA, gl.ONE, gl.ONE, gl.ONE_MINUS_SRC_ALPHA);
                this._alphaState.alphaBlend = true;
                break;

            case BABYLON.Constants.ALPHA_PREMULTIPLIED:
                this._alphaState.setAlphaBlendFunctionParameters(gl.ONE, gl.ONE_MINUS_SRC_ALPHA, gl.ONE, gl.ONE);
                this._alphaState.alphaBlend = true;
                break;
            case BABYLON.Constants.ALPHA_SUBTRACT:
                this._alphaState.setAlphaBlendFunctionParameters(gl.ZERO, gl.ONE_MINUS_SRC_COLOR, gl.ONE, gl.ONE);
                this._alphaState.alphaBlend = true;
                break;
            case BABYLON.Constants.ALPHA_MULTIPLY:
                this._alphaState.setAlphaBlendFunctionParameters(gl.DST_COLOR, gl.ZERO, gl.ONE, gl.ONE);
                this._alphaState.alphaBlend = true;
                break;
            case BABYLON.Constants.ALPHA_INTERPOLATE:
                this._alphaState.setAlphaBlendFunctionParameters(gl.CONSTANT_COLOR, gl.ONE_MINUS_CONSTANT_COLOR, gl.CONSTANT_ALPHA, gl.ONE_MINUS_CONSTANT_ALPHA);
                this._alphaState.alphaBlend = true;
                break;
            case BABYLON.Constants.ALPHA_ONEONE_ONEONE:
                this._alphaState.setAlphaBlendFunctionParameters(gl.ONE, gl.ONE, gl.ONE, gl.ONE);
                this._alphaState.alphaBlend = true;
                break;
            case BABYLON.Constants.ALPHA_ALPHATOCOLOR:
                this._alphaState.setAlphaBlendFunctionParameters(gl.DST_ALPHA, gl.ONE, gl.ZERO, gl.ZERO);
                this._alphaState.alphaBlend = true;
                break;
            case BABYLON.Constants.ALPHA_REVERSEONEMINUS:
                this._alphaState.setAlphaBlendFunctionParameters(gl.ONE_MINUS_DST_COLOR, gl.ONE_MINUS_SRC_COLOR, gl.ONE_MINUS_DST_ALPHA, gl.ONE_MINUS_SRC_ALPHA);
                this._alphaState.alphaBlend = true;
                break;
            case BABYLON.Constants.ALPHA_SRC_DSTONEMINUSSRCALPHA:
                this._alphaState.setAlphaBlendFunctionParameters(gl.ONE, gl.ONE_MINUS_SRC_ALPHA, gl.ONE, gl.ONE_MINUS_SRC_ALPHA);
                this._alphaState.alphaBlend = true;
                break;
            case BABYLON.Constants.ALPHA_ONEONE_ONEZERO:
                this._alphaState.setAlphaBlendFunctionParameters(gl.ONE, gl.ONE, gl.ONE, gl.ZERO);
                this._alphaState.alphaBlend = true;
                break;
            case BABYLON.Constants.ALPHA_EXCLUSION:
                this._alphaState.setAlphaBlendFunctionParameters(gl.ONE_MINUS_DST_COLOR, gl.ONE_MINUS_SRC_COLOR, gl.ZERO, gl.ONE);
                this._alphaState.alphaBlend = true;
                break;
        }
        if (!noDepthWriteChange) {
            this.depthCullingState.depthMask = (mode === BABYLON.Constants.ALPHA_DISABLE);
        }
        this._alphaMode = mode;
    };
}