
import { BABYLON } from "./render3d/babylon";
import { Scene } from "./render3d/scene";
import { Engine } from "./render3d/engine";
import { ExtensionTools, ResTab } from './interface/index';

declare let module;
let logger;
ExtensionTools.RegistStartUp(() => { logger = ExtensionTools.Logger(module.name, "BABYLON Factory"); });

export const factorySceneGUI = (engine: BABYLON.Engine) => {
    return new BABYLON.Scene(engine);
};

export const factoryScene3D = (engine: Engine) => {
    return new Scene(engine);
};

export const factoryEngineOnlyGUI = (canvas: any) => {
    // 引擎创建 - 注意参数配置
    return new BABYLON.Engine(canvas, false, {
        disableWebGL2Support: false
    }, false);
};

export const factoryEngine = (canvas: HTMLCanvasElement) => {
    // 资源管理
    const resTab = ExtensionTools.CreateResTab();
    // resTab.timeout = 5000;
    Engine.prepare(resTab);

    const width = canvas.width;
    const height = canvas.height;
    // 引擎创建 - 注意参数配置
    const engine = new Engine(canvas, false, {
        disableWebGL2Support: false
        // stencil: true
    }, false);
    engine.setSize(width, height);

    engine.doNotHandleContextLost = true; // Handling WebGL context lost
    engine.enableOfflineSupport = true; // 该值设置为true时才能支持外部脱机资源接管
    engine.disableManifestCheck = true;

    logger.warn("Factory Engine Success.");

    return engine;
};
