import { ICustomGeometry } from "../interface/geometry";
import { Constants } from "../interface/constants";
import { BABYLON } from "../render3d/babylon";
import { CommonVertexBuffer, ICommonBufferScene } from "../render3d/commonbuffer";

export class CustomGeometry implements ICustomGeometry {
    private kinds: (typeof Constants.VERTEX_ATTR_POSITION | typeof Constants.VERTEX_ATTR_COLOR4 | typeof Constants.VERTEX_ATTR_UV)[] = [];
    private dirty: boolean[] = [];
    // private kindSize: number[] = [];
    // private kindOffset: number[] = [];
    private mPositions: Map<(typeof Constants.VERTEX_ATTR_POSITION | typeof Constants.VERTEX_ATTR_COLOR4 | typeof Constants.VERTEX_ATTR_UV), CommonVertexBuffer> = new Map();
    private mIndices: Uint16Array;
    private indicesDirty: boolean = true;
    constructor(private mesh: BABYLON.Mesh, private unique: string, kinds: (typeof Constants.VERTEX_ATTR_POSITION | typeof Constants.VERTEX_ATTR_COLOR4 | typeof Constants.VERTEX_ATTR_UV)[], public readonly maxVertex: number, private maxtriangle: number, useIndices: boolean = true) {
        this.kinds = kinds;
        let offset = 0;
        let scene: ICommonBufferScene = <any>mesh._scene;
        kinds.forEach((kind, idx) => {
            let size = 0;
            let kindkey;
            if (kind == Constants.VERTEX_ATTR_POSITION)    { size = 3; kindkey = "position"; }
            if (kind == Constants.VERTEX_ATTR_UV)          { size = 2; kindkey = "uv"; }
            if (kind == Constants.VERTEX_ATTR_COLOR4)      { size = 4; kindkey = "color"; }
            // attrs.push(CustomGeometry.kind(kind));

            // this.kindOffset[idx] = offset;
            // this.kindSize[idx] = size;
            offset += size;

            let vbbuffer = scene.__commonBuffer.createVertexBuffer(kindkey, maxVertex * size);
            let data = vbbuffer.commonData;
            if (kind == Constants.VERTEX_ATTR_COLOR4)   {
                for (let i = 0; i < maxVertex * size; i++) {
                    data[i] = 1.;
                }
                vbbuffer.commonDirty();
            }
            this.mesh.setVerticesBuffer(vbbuffer)
            this.mPositions.set( kind, vbbuffer );
        });

        if (useIndices) {
            this.mIndices = new Uint16Array(maxtriangle * 3);
            this.mesh.setIndices(this.mIndices, maxtriangle * 3, true);
            this.indicesDirty = false;
        }
        this.mesh.hasVertexAlpha = true;
    }
    appyToMesh(mesh: BABYLON.Mesh) {
        this.mesh = mesh;
        this.mPositions.forEach( (vbbuffer, kind) => {
            this.mesh.setVerticesBuffer(vbbuffer);
        } );
        if (this.mIndices) {
            this.mesh.setIndices(this.mIndices, this.maxtriangle * 3, true);
        }
    }
    uninstallMesh() {
        this.mesh = undefined;
    }
    update(pointCount: number): void {
        if (pointCount > 0) {
            if (this.mIndices) {
                pointCount = Math.ceil(pointCount / 2) * 2;
                if (this.indicesDirty) {
                    this.indicesDirty = false;
                    this.mesh.setIndices(this.mIndices);
                }
                if (this.mesh.subMeshes[0]) {
                    this.mesh.subMeshes[0].indexCount = pointCount;
                }
                this.kinds.forEach((kind, idx) => {
                    if (this.dirty[idx]) {
                        let size = 0;
                        let kindkey;
                        if (kind == Constants.VERTEX_ATTR_POSITION)    { size = 3; kindkey = "position"; }
                        if (kind == Constants.VERTEX_ATTR_UV)          { size = 2; kindkey = "uv"; }
                        if (kind == Constants.VERTEX_ATTR_COLOR4)      { size = 4; kindkey = "color"; }
                        this.mPositions.get(kind).commonDirty();
                        // this.mesh.updateVerticesData(kindkey, this.mPositions.get(kind));
                    }
                });
            } else {
                this.kinds.forEach((kind, idx) => {
                    let size = 0;
                    let kindkey;
                    if (kind == Constants.VERTEX_ATTR_POSITION)    { size = 3; kindkey = "position"; }
                    if (kind == Constants.VERTEX_ATTR_UV)          { size = 2; kindkey = "uv"; }
                    if (kind == Constants.VERTEX_ATTR_COLOR4)      { size = 4; kindkey = "color"; }
                    this.mPositions.get(kind).commonDirty();
                    if (this.mesh.subMeshes[0]) {
                        this.mesh.subMeshes[0].verticesCount = pointCount;
                    }
                    // this.mesh.setVerticesData(kindkey, this.mPositions.get(kind).slice(0, pointCount * size));
                });
            }
            this.mesh.subMeshes[0].setBoundingInfo(this.mesh.getBoundingInfo());
        }
        this.dirty.length = 0;
    }
    vertices(kind: typeof Constants.VERTEX_ATTR_POSITION | typeof Constants.VERTEX_ATTR_COLOR4 | typeof Constants.VERTEX_ATTR_UV): [Float32Array, number, number] {

        let idx = this.kinds.indexOf(kind);
        if (idx >= 0) {
            this.dirty[idx] = true;
            return [this.mPositions.get(kind).commonData, undefined, 0]
        }
        return undefined;
    }
    indices() {
        this.indicesDirty = true;
        return this.mIndices;
    }
    dispose() {
        this.mesh?.dispose();
    }
}