import { Sprite } from "../sprite/sprite";
import { ISprite } from "../interface/sprite";
import { SceneStruct } from "../scene_struct";
import { BABYLON } from "../render3d/babylon";
import { MainOpacityMaterial } from "material/main_opacity_material";
import { MathDataPool } from "../render3d/extends/math";

const tempVec = new BABYLON.Vector3();
export function elementToSprite(clientWidth: number, clientHeight: number, eleWidth: number, eleHeight: number, eleLeft: number, eleTop: number, scene: SceneStruct, spriteAtlas: string, frame: number | string, distanceToCamera: number = 10): Sprite {
    let node = new Sprite(spriteAtlas, frame, scene.impl);
    let w = eleWidth  / clientWidth * (scene.camera.orthoRight - scene.camera.orthoLeft);
    let h = eleHeight / clientHeight * (scene.camera.orthoTop - scene.camera.orthoBottom);
    tempVec.x = (eleLeft + eleWidth / 2 - clientWidth / 2) / clientWidth * (scene.camera.orthoRight - scene.camera.orthoLeft);
    tempVec.y = (clientHeight/2 -(eleTop + eleHeight / 2)) / clientHeight * (scene.camera.orthoTop - scene.camera.orthoBottom);
    tempVec.z = distanceToCamera;

    BABYLON.Vector3.TransformCoordinatesToRef(tempVec, scene.impl.activeCamera.getWorldMatrix(), tempVec);

    MathDataPool.Recycle(node.scaling);
    MathDataPool.Recycle(node.position);
    MathDataPool.Recycle(node.rotation);

    node.scaling = MathDataPool.Vector3(w, h, 1);
    node.position = MathDataPool.Vector3(tempVec.x, tempVec.y, tempVec.z);
    node.rotation = MathDataPool.Vector3(0, 0, 0); 
    scene.camera.absoluteRotation.toEulerAnglesToRef(node.rotation);

    return node;
}

export function elementToQuad(clientWidth: number, clientHeight: number, eleWidth: number, eleHeight: number, eleLeft: number, eleTop: number, scene: SceneStruct, image: string, distanceToCamera: number = 10): BABYLON.Mesh {
    let node = BABYLON.PlaneBuilder.CreatePlane(image, { size: 1 }, scene.impl);
    let mat = new MainOpacityMaterial(image, scene.impl);
    let texture = new BABYLON.Texture(null, scene.impl, true, false, BABYLON.Texture.LINEAR_LINEAR);
    texture.updateURL(image);
    texture.wrapU = BABYLON.Texture.CLAMP_ADDRESSMODE;
    texture.wrapV = BABYLON.Texture.CLAMP_ADDRESSMODE;
    if (mat.diffuseTexture) {
        mat.diffuseTexture.dispose();
    }
    mat.diffuseTexture = texture;
    node.material = mat;

    let w = eleWidth  / clientWidth * (scene.camera.orthoRight - scene.camera.orthoLeft);
    let h = eleHeight / clientHeight * (scene.camera.orthoTop - scene.camera.orthoBottom);
    tempVec.x = (eleLeft + eleWidth / 2 - clientWidth / 2) / clientWidth * (scene.camera.orthoRight - scene.camera.orthoLeft);
    tempVec.y = (clientHeight/2 -(eleTop + eleHeight / 2)) / clientHeight * (scene.camera.orthoTop - scene.camera.orthoBottom);
    tempVec.z = distanceToCamera;

    BABYLON.Vector3.TransformCoordinatesToRef(tempVec, scene.impl.activeCamera.getWorldMatrix(), tempVec);

    MathDataPool.Recycle(node.scaling);
    MathDataPool.Recycle(node.position);
    MathDataPool.Recycle(node.rotation);

    node.scaling = MathDataPool.Vector3(w, h, 1);
    node.position = MathDataPool.Vector3(tempVec.x, tempVec.y, tempVec.z);
    scene.camera.absoluteRotation.toEulerAnglesToRef(tempVec);
    node.rotation = MathDataPool.Vector3(tempVec.x, tempVec.y, tempVec.z);
    return node;
}