import { ETempNodeAnimationDataKey, LoaderRecord } from "./base";
import { INodeNew } from "./interface";
import { BABYLON } from "../render3d/babylon";
import { CameraType } from "./glTF2Interface";

export class CameraLoader {
    public static loadNodeAsync(nodeIndex: number, nodeInfo: INodeNew, loaders: LoaderRecord, babylonNode: BABYLON.TransformNode) {
        if (nodeInfo.camera != undefined) {
            let camera = loaders.gltf.cameras[nodeInfo.camera];

            const babylonCamera = new BABYLON.FreeCamera(camera.name || `camera${nodeIndex}`, BABYLON.Vector3.Zero(), loaders.babylonScene, false);
            babylonCamera.ignoreParentScaling = true;
            babylonCamera.parent = babylonNode;

            // PI_BEGIN - 导入的相机通过 target 调整方向
            // babylonCamera.rotation = new Vector3(0, Math.PI, 0);
            // ////////////////////
            if (loaders.babylonScene.importCameraUseTarget) {
                babylonCamera.setTarget(BABYLON.Vector3.Forward());
            }
            else {
                babylonCamera.rotation = new BABYLON.Vector3(0, Math.PI, 0);
            }
            // PI_END

            switch (camera.type) {
                case CameraType.PERSPECTIVE: {
                    const perspective = camera.perspective;
                    if (!perspective) {
                        throw new Error(`Camera perspective properties are missing`);
                    }
    
                    babylonCamera.fov = perspective.yfov;
                    babylonCamera.minZ = perspective.znear;
                    babylonCamera.maxZ = perspective.zfar || Number.MAX_VALUE;
                    break;
                }
                case CameraType.ORTHOGRAPHIC: {
                    if (!camera.orthographic) {
                        throw new Error(`Camera orthographic properties are missing`);
                    }
    
                    babylonCamera.mode = BABYLON.Camera.ORTHOGRAPHIC_CAMERA;
                    // PI_BEGIN - 正交相机设置简化
                    // babylonCamera.orthoLeft = -camera.orthographic.xmag;
                    // babylonCamera.orthoRight = camera.orthographic.xmag;
                    // babylonCamera.orthoBottom = -camera.orthographic.ymag;
                    // babylonCamera.orthoTop = camera.orthographic.ymag;
                    // babylonCamera.minZ = camera.orthographic.znear;
                    // babylonCamera.maxZ = camera.orthographic.zfar;
                    // //////////////////////////////////
                    if (camera.orthographic.size) {
                        babylonCamera.piOrthSize = camera.orthographic.size;
                    }
                    else {
                        babylonCamera.orthoLeft = -camera.orthographic.xmag;
                        babylonCamera.orthoRight = camera.orthographic.xmag;
                        babylonCamera.orthoBottom = -camera.orthographic.ymag;
                        babylonCamera.orthoTop = camera.orthographic.ymag;
                        babylonCamera.minZ = camera.orthographic.znear;
                        babylonCamera.maxZ = camera.orthographic.zfar;
                    }
                    // PI_END
                    break;
                }
                default: {
                    throw new Error(`Invalid camera type (${camera.type})`);
                }
            }

            if (babylonCamera) {
                let animeAbout = loaders.nodeAnimationAbout.get(nodeIndex);
                animeAbout[ETempNodeAnimationDataKey.Camera] = babylonCamera;
                
                // PI_BEGIN - 20200909 导入时自动处理 layerMask
                if (camera.cullingMask != undefined && loaders.babylonScene.pi_autoApplyLayerMask) {
                    babylonCamera.layerMask = camera.cullingMask;
                }
                // PI_END
            }
        }
    }
}