import { ETempNodeAnimationDataKey, GLTFTool, LoaderRecord } from "../base";
import { BABYLON } from "../../render3d/babylon";
import { INodeNew } from "../interface";
import { KHRLightsPunctual_LightType } from "../glTF2Interface";

const NAME = "KHR_lights_punctual";

/**
 * [Specification](https://github.com/KhronosGroup/glTF/blob/master/extensions/2.0/Khronos/KHR_lights_punctual)
 */
export class KHR_lights {

    /** @hidden */
    public static loadNodeAsync(nodeIndex: number, node: INodeNew, loaders: LoaderRecord, babylonNode: BABYLON.TransformNode): BABYLON.Light {
        let light = GLTFTool.light(node, loaders.gltf);
        let babylonLight: BABYLON.Light;
        if (light) {
            let name = light.name || babylonNode.name;

            switch (light.type) {
                case KHRLightsPunctual_LightType.DIRECTIONAL: {
                    // PI_BEGIN - 灯光方向调整
                    babylonLight = new BABYLON.DirectionalLight(name, BABYLON.Vector3.Forward(), loaders.scene, !!light.isLocal);
                    // PI_END
                    break;
                }
                case KHRLightsPunctual_LightType.POINT: {
                    babylonLight = new BABYLON.PointLight(name, BABYLON.Vector3.Zero(), loaders.scene, !!light.isLocal);
                    break;
                }
                case KHRLightsPunctual_LightType.SPOT: {
                    // PI_BEGIN - 灯光方向调整
                    const babylonSpotLight = new BABYLON.SpotLight(name, BABYLON.Vector3.Zero(), BABYLON.Vector3.Forward(), 0, 1, loaders.babylonScene);
                    // PI_END
                    babylonSpotLight.angle = ((light.spot && light.spot.outerConeAngle) || Math.PI / 4) * 2;
                    babylonSpotLight.innerAngle = ((light.spot && light.spot.innerConeAngle) || 0) * 2;
                    babylonLight = babylonSpotLight;
                    break;
                }
                default: {
                    loaders.babylonScene._blockEntityCollection = false;
                    throw new Error(`Invalid light type (${light.type})`);
                }
            }

            if (babylonLight) {
                let animeAbout = loaders.nodeAnimationAbout.get(nodeIndex);
                animeAbout[ETempNodeAnimationDataKey.Light] = babylonLight;

                babylonLight.falloffType    = BABYLON.Light.FALLOFF_GLTF;
                babylonLight.diffuse        = light.color ? BABYLON.Color3.FromArray(light.color) : BABYLON.Color3.White();
                babylonLight.specular       = babylonLight.diffuse;
                babylonLight.intensity      = light.intensity == undefined ? 1 : light.intensity;
                babylonLight.range          = light.range == undefined ? Number.MAX_VALUE : light.range;
                babylonLight.lightmapMode   = light.shadowOnly ? BABYLON.Light.LIGHTMAP_SHADOWSONLY : BABYLON.Light.LIGHTMAP_DEFAULT;
                
                // PI_BEGIN - 20200909 导入时自动处理 layerMask
                if (light.cullingMask != undefined && loaders.babylonScene.pi_autoApplyLayerMask) {
                    babylonLight.includeOnlyWithLayerMask = light.cullingMask;
                }
                // PI_END

                babylonLight.parent = babylonNode;
            }
        }

        return babylonLight;
    }
}
