import { Nullable } from "./types";
import { Constants } from "./constants";
import { ITransformNode } from "./transform";
import { ICollisionResponse, ILayerMask } from "./base";
import { IVector3 } from "./math";

export interface IBoundingBox {
    min: IVector3;
    max: IVector3;
}

export interface IAbstractMesh extends ITransformNode, ILayerMask {
    boundingBox: IBoundingBox;
    isVisible: boolean;
    scalingMode: Constants.ScalingMode;
    alwaysSelectAsActiveMesh: boolean;
    cullingStrategy: Constants.CULLING_STRATEGY;
    lookAt(target: IVector3): void;
    /**
     * Gets or sets a collision mask used to mask collisions (default is -1).
     * A collision between A and B will happen if A.collisionGroup & b.collisionMask !== 0
     */
    collisionMask: number;
    /**
     * Gets or sets a collision response flag (default is true).
     * when collisionResponse is false, events are still triggered but colliding entity has no response
     * This helps creating trigger volume when user wants collision feedback events but not position/velocity
     * to respond to the collision.
     */
    collisionResponse: boolean;
    /**
     * Gets or sets the current collision group mask (-1 by default).
     * A collision between A and B will happen if A.collisionGroup & b.collisionMask !== 0
     */
    collisionGroup: number;
    /**
     * Gets or sets current surrounding meshes (null by default).
     *
     * By default collision detection is tested against every mesh in the scene.
     * It is possible to set surroundingMeshes to a defined list of meshes and then only these specified
     * meshes will be tested for the collision.
     *
     * Note: if set to an empty array no collision will happen when this mesh is moved.
     */
    surroundingMeshes: Nullable<ICollisionResponse[]>;
}