
import { BABYLON } from "../render3d/babylon";
import { Constants } from "./constants";
import { Nullable } from "./types";
import { Observable } from "./observable";

export interface IInitialOption {
    pbr: boolean;
    KHR_lights_punctual: boolean;
    PI_environment: boolean;
    PI_particle: boolean;
    onoffCollectInfo: boolean;
    DefaultWhitePath: string;
    width: number;
    height: number;
    optimizeMaterialWithResTab?: boolean;
    optimizeBoundingInfoOnlySphere?: boolean;
    cacheCustomGeometryDynamicBufferSize?: number;
}

export interface GlContext {
    canvas: {
        setBrush(value: number): void;
        width: number;
        height: number;
    }
}

export interface ITempFrameBuffer {
    rect: [number, number, number, number],
    frameBuffer?: WebGLFramebuffer,
    ctx?: GlContext,
}

export type DefaultKey = number;

/**
 * Define an interface for all classes that will hold resources
 */
export interface IDisposable {
    /**
     * Releases all held resources
     */
    dispose(): void;
    _isDisposed: boolean;
    onDisposeObservable: any; // Observable<any>
    name?: string;
}

export interface ICollisionResponse {
    /**
     * Gets or sets a collision mask used to mask collisions (default is -1).
     * A collision between A and B will happen if A.collisionGroup & b.collisionMask !== 0
     */
    collisionMask: number;

    /**
     * Gets or sets a collision response flag (default is true).
     * when collisionResponse is false, events are still triggered but colliding entity has no response
     * This helps creating trigger volume when user wants collision feedback events but not position/velocity
     * to respond to the collision.
     */
    collisionResponse: boolean;
    /**
     * Gets or sets the current collision group mask (-1 by default).
     * A collision between A and B will happen if A.collisionGroup & b.collisionMask !== 0
     */
    collisionGroup: number;

    /**
     * Gets or sets current surrounding meshes (null by default).
     *
     * By default collision detection is tested against every mesh in the scene.
     * It is possible to set surroundingMeshes to a defined list of meshes and then only these specified
     * meshes will be tested for the collision.
     *
     * Note: if set to an empty array no collision will happen when this mesh is moved.
     */
    surroundingMeshes: Nullable<ICollisionResponse[]>;
}

export interface ICustomRenderTarget {
    key: number,
    dispose(): void;
}
export interface TAutoRenderTarget {
    autoTarget: true;
    width: number,
    height: number,
    colorFormat: Constants.RenderFormat,
    depthStencilFormat: Constants.DepthStencilFormat,
}

export interface ILayerMask {
    layerMask: number,
}
