import { BABYLON } from "../render3d/babylon";
import { Observable } from "./observable";

export interface IRay {}
export interface IPiCollider {
    /**
     * Recreates the entire bounding info from scratch as if we call the constructor in place
     * @param min defines the new minimum vector (in local space)
     * @param max defines the new maximum vector (in local space)
     * @param worldMatrix defines the new world matrix
     */
    reConstruct(min: BABYLON.DeepImmutable<BABYLON.Vector3>, max: BABYLON.DeepImmutable<BABYLON.Vector3>, worldMatrix?: BABYLON.DeepImmutable<BABYLON.Matrix>);

    /**
     * Updates the bounding sphere and box
     * @param world world matrix to be used to update
     */
    update(world: BABYLON.DeepImmutable<BABYLON.Matrix>);
    intersects(ray: IRay);
}

export interface IColliderItem {
    alphaIndex: number,
    getDistanceToCamera(camera: BABYLON.Nullable<BABYLON.Camera>): number,
    _colliderBoundle?: IPiCollider,
    _boundingInfo?: BABYLON.BoundingInfo;
    onDisposeObservable: Observable<IColliderItem>;
    worldMatrixFromCache?: BABYLON.Matrix;
    isEnabled(): boolean;
}
