import { IInitialOption } from "./base";

export * from "./abstractmesh";
export * from "./animation";
export * from "./base";
export * from "./camera";
export * from "./collider";
export * from "./constants";
export * from "./entity";
export * from "./event";
export * from "./geometry";
export * from "./instancedmesh";
export * from "./light";
export * from "./material";
export * from "./mesh";
export * from "./node";
export * from "./nodematerial";
export * from "./observable";
export * from "./particlesystem";
export * from "./renderer";
export * from "./scene";
export * from "./shadow";
export * from "./skeleton";
export * from "./spine_baked";
export * from "./sprite";
export * from "./state";
export * from "./transform";
export * from "./types";
export * from "./texturepacker";

export * from "./postprocess";
export * from "./modelobj";
export * from "./modelprefab";

export interface Logger {
    warn(msg: string, val?: any): void;
    error(msg: string, val?: any): void;
    info(msg: string, val?: any): void;
}
export interface Res<A, D> {
    free?(): void;
    /**
     * 资源参数
     */
    args: A;
    name: string;
    link: D;
    key: any;
}
export type TLoadCall<T, A, D> = (resTab: ResTab, type: T, filePath: string, ...args: A[]) => Promise<D>;
export interface ResTab {
    load: (type: any, name: string, loadArgs: any[], hasTabRes?: boolean) => Promise<Res<any, any>>;
    clear(): void;
    delete<A, D>(res: Res<A, D>): void;
    get<A, D>(vtype: any, url: any): Res<A, D>;
    createRes<A, D>(name: string, vtype: string, data: D, args: A): Res<A, D>;
}

export interface IAdaptiveResult {
    layoutSize: any;
    renderSize: { width: number; height: number; };
    styleSize: { width: number; height: number; };
    styleLeft: number;
    styleRight: number;
    styleTop: number;
    styleBottom: number;
}

export interface KTXInfo {
    mipmaps: {
        data: Uint8Array,
        width: number,
        height: number
    }[];
    width: number;
    height: number;
    format: number;
    isCubemap: boolean;
    mipmapCount: number;
}

/**
 * 日志采集器 接口
 */
export interface ILogger {

    /**
     * 详细 信息，最高等级，类似于 js -- console.debug | rust -- debug!
     * 
     * @param format 格式化字符串，类似于 NodeJS 环境的 #[util.format](http://nodejs.cn/api/util.html#util_util_format_format_args)
     * 
     * @note 格式化字符如下
     *      + %s 字符串，实现：String(value);
     *      + %d 数字，实现： Number(value);
     *      + %i 十进制整数，实现： parseInt(value, 10);
     *      + %f 浮点数，实现： parseFloat(value);
     *      + %o JSON字符串，实现：JSON.stringify(value);
     *      + %% 输出一个百分号；
     */
    logV(format: string, ...args: any[]): void;

    /**
     * 详细 信息，类似于 js -- console.debug | rust -- debug!
     * 
     * @param format 格式化字符串，类似于 NodeJS 环境的 #[util.format](http://nodejs.cn/api/util.html#util_util_format_format_args)
     * 
     * @note 格式化字符如下
     *      + %s 字符串，实现：String(value);
     *      + %d 数字，实现： Number(value);
     *      + %i 十进制整数，实现： parseInt(value, 10);
     *      + %f 浮点数，实现： parseFloat(value);
     *      + %o JSON字符串，实现：JSON.stringify(value);
     *      + %% 输出一个百分号；
     */
    logD(format: string, ...args: any[]): void;

    /**
     * 关键 信息，类似于 js -- console.info | rust -- info!
     * 
     * @param format 格式化字符串，类似于 NodeJS 环境的 #[util.format](http://nodejs.cn/api/util.html#util_util_format_format_args)
     * 
     * @note 格式化字符如下
     *      + %s 字符串，实现：String(value);
     *      + %d 数字，实现： Number(value);
     *      + %i 十进制整数，实现： parseInt(value, 10);
     *      + %f 浮点数，实现： parseFloat(value);
     *      + %o JSON字符串，实现：JSON.stringify(value);
     *      + %% 输出一个百分号；
     */
    logI(format: string, ...args: any[]): void;

    /**
     * 警告 信息，类似于 js -- console.warn | rust -- warn!
     * 
     * @param format 格式化字符串，类似于 NodeJS 环境的 #[util.format](http://nodejs.cn/api/util.html#util_util_format_format_args)
     * 
     * @note 格式化字符如下
     *      + %s 字符串，实现：String(value);
     *      + %d 数字，实现： Number(value);
     *      + %i 十进制整数，实现： parseInt(value, 10);
     *      + %f 浮点数，实现： parseFloat(value);
     *      + %o JSON字符串，实现：JSON.stringify(value);
     *      + %% 输出一个百分号；
     */
    logW(format: string, ...args: any[]): void;

    /**
     * 错误 信息，类似于 js -- console.error | rust -- error!
     * 
     * @param format 格式化字符串，类似于 NodeJS 环境的 #[util.format](http://nodejs.cn/api/util.html#util_util_format_format_args)
     * 
     * @note 格式化字符如下
     *      + %s 字符串，实现：String(value);
     *      + %d 数字，实现： Number(value);
     *      + %i 十进制整数，实现： parseInt(value, 10);
     *      + %f 浮点数，实现： parseFloat(value);
     *      + %o JSON字符串，实现：JSON.stringify(value);
     *      + %% 输出一个百分号；
     */
    logE(format: string, ...args: any[]): void;
}

export class ExtensionTools {
    public static Option: IInitialOption;
    public static RES_FILE: string;
    public static RES_IMAGE: string;
    public static GetPureResPath(url: string): string {
        throw new Error("Not Implement");
    }
    public static QueryRuntimeFullName(name: string): string {
        throw new Error("Not Implement");
    }
    public static Logger(module: string, tag: string): Logger {
        throw new Error("Not Implement");
    }
    public static GetLogger(module: string, tag: string): ILogger {
        throw new Error("Not Implement");
    }
    public static CreateResTab(): ResTab {
        throw new Error("Not Implement");
    }
    public static LoadFile(resTab: ResTab, url: string): Promise<ArrayBuffer | ArrayBufferView> {
        throw new Error("Not Implement");
    }
    /**
     * Env.get("adaptiveResult")
     */
    public static AdaptiveResult(): IAdaptiveResult {
        throw new Error("Not Implement");
    }
    /**
     * 注册：类型名，以及该类型的加载函数，还有销毁函数
     * @param load 加载函数，返回Promise<link>，通过返回值，可以取到 将要 放到res.link的数据
     * @param destroy 销毁函数，用于销毁res.link
     * @param sizeOf 查询资源的字节大小方法
     * @param replace load 方法是否明确为重载
     * @param timeout 资源缓存超时时间
     * @param min_capacity 资源缓存尺寸下限
     * @param max_capacity 资源缓存尺寸上限
     */
    public static ResLoadRegist<T, A, D>(type: string, load: TLoadCall<T, A, D>, destroy: (link: any) => void, sizeOf?: (link: any) => number, replace: boolean = false, timeout?: number, min_capacity?: number, max_capacity?: number) {
        throw new Error("Not Implement");
    }
    /**
     * @description 设置任务，任务的调用函数和参数（参数必须为数组），任务的优先级和类型（异步0，同步顺序1，同步立即2）
     * @param func 目标函数
     * @param args 函数参数数组
     * @param priority 任务的优先级
     * @param type 任务类型（异步0，同步顺序1，同步立即2）
     */
    public static AddTask(func: Function, args: any[], priority: number, type: 0 | 1 | 2) {
        throw new Error("Not Implement");
    }
    public static UTF8Decode(data: ArrayBuffer): string {
        throw new Error("Not Implement");
    }
    private static startups: (() => void)[] = [];
    public static RegistStartUp(func: () => void) {
        ExtensionTools.startups.push(func);
    }
    public static Ready(): void {
        ExtensionTools.startups.forEach(func => {
            func();
        });
    }
    public static FileNameWithoutDirectory(path: string): string {
        throw new Error("Not Implement");
    }
    /**
     * Env.get('compressed_tex_format')
     */
    public static QuerySupportCompressedTextureFormat(): string {
        throw new Error("Not Implement");
    }
    /**
     * Env.get("domains")
     */
    public static Domains(): string[] {
        throw new Error("Not Implement");
    }
    public static setPermanentBefore(func: () => void) {
        throw new Error("Not Implement");
    }
    public static setPermanent(func: () => void) {
        throw new Error("Not Implement");
    }
    public static clearPermanent(func: () => void) {
        throw new Error("Not Implement");
    }
    public static getAtom(v: number | string | any) {
        throw new Error("Not Implement");
    }
    public static getFileSize(v: string): number {
        throw new Error("Not Implement");
    }}