import { IColor3, IVector3 } from "./math";
import { IDisposable } from "./base";
import { INode } from "./node";

export interface ILight extends IDisposable, INode {
    diffuse: IColor3;
    specular: IColor3;
    intensity: number;
}

export interface IDirectLight extends ILight {
    direct: IVector3;
    /**
     * Anguler diameter of the sun as seen from the Earth
     */
    angle: number;
}

export interface IPointLight extends ILight {
    radius: number;
    power: number;
}

export interface ISpotLight extends ILight {
    radius: number;
    power: number;
    /**
     * Angle of the beam
     */
    size: number;
    /**
     * Softless of the edge
     */
    blend: number;
    inner: number;
    outer: number;
}
