import { BABYLON } from "../render3d/babylon";
import { Constants, CustomGeometry, CustomScene, MainOpacityMaterial } from "../index";
import { ILine, ILineRenderer } from "./interface";
import { TextureHSVMaterial } from "../material/hsv";
import { TextureHueSaturateMaterial } from "../material/hue_saturate";
import { Scene } from "render3d/scene";

export class Line implements ILine {
    protected static TempNormal: BABYLON.Vector3 = BABYLON.Vector3.Zero();
    protected static TempDirect: BABYLON.Vector3 = BABYLON.Vector3.Zero();
    protected static TempExtend: BABYLON.Vector3 = BABYLON.Vector3.Zero();
    private _startx: number;
    private _starty: number;
    private _startz: number;
    private _endx: number;
    private _endy: number;
    private _endz: number;
    private _r: number;
    private _g: number;
    private _b: number;
    private _a: number;
    private _repeatV: number;
    private _uOffset: number = 0;
    private _vOffset: number = 0;
    public set startx(val: number) {
        this._startx = val;
        this._dirty = true;
        this._extendDirty = true;
    }
    public set starty(val: number) {
        this._starty = val;
        this._dirty = true;
        this._extendDirty = true;
    }
    public set startz(val: number) {
        this._startz = val;
        this._dirty = true;
        this._extendDirty = true;
    }
    public set endx(val: number) {
        this._endx = val;
        this._dirty = true;
        this._extendDirty = true;
    }
    public set endy(val: number) {
        this._endy = val;
        this._dirty = true;
        this._extendDirty = true;
    }
    public set endz(val: number) {
        this._endz = val;
        this._dirty = true;
        this._extendDirty = true;
    }
    public set r(val: number) {
        this._r = val;
        this._dirty = true;
    }
    public set g(val: number) {
        this._g = val;
        this._dirty = true;
    }
    public set b(val: number) {
        this._b = val;
        this._dirty = true;
    }
    public set a(val: number) {
        this._a = val;
        this._dirty = true;
    }
    public set repeatV(val: number) {
        this._repeatV = val;
        this._dirty = true;
        this._extendDirty = true;
    }
    public set uOffset(val: number) {
        this._uOffset = val;
        this._dirty = true;
    }
    public set vOffset(val: number) {
        this._uOffset = val;
        this._dirty = true;
    }
    public get startx() { return this._startx }
    public get starty() { return this._starty }
    public get startz() { return this._startz }
    public get endx() { return this._endx }
    public get endy() { return this._endy }
    public get endz() { return this._endz }
    public get r() { return this._r }
    public get g() { return this._g }
    public get b() { return this._b }
    public get a() { return this._a }
    public get repeatV() { return this._repeatV }
    public get uOffset() { return this._uOffset }
    public get vOffset() { return this._uOffset }
    private _nx: number;
    private _ny: number;
    private _nz: number;
    private _width: number;
    private _repeatULen: number;
    public set width(val: number) {
        this._width = val;
        this._dirty = true;
        this._extendDirty = true;
    }
    public set normalx(val: number) {
        this._nx = val;
        this._dirty = true;
        this._extendDirty = true;
    }
    public set normaly(val: number) {
        this._ny = val;
        this._dirty = true;
        this._extendDirty = true;
    }
    public set normalz(val: number) {
        this._nz = val;
        this._dirty = true;
        this._extendDirty = true;
    }
    public set repeatULen(val: number) {
        this._repeatULen = val;
        this._dirty = true;
        this._extendDirty = true;
    }
    ex: number;
    ey: number;
    ez: number;
    image: string;
    repeatU: number;

    private _extendDirty: boolean = false;
    public _dirty: boolean = false;

    constructor(public readonly unique: number, protected renderer: LineMeshRenderer) {

    }
    _onUpdate(): boolean {
        let isDirty = this._dirty || this._uOffset != 0 || this._vOffset != 0;
        if (this._extendDirty) {
            this.extend();
        }
        this._dirty = false;
        this._extendDirty = false;
        return isDirty;
    }
    extend() {
        Line.TempNormal.copyFromFloats(this._nx, this._ny, this._nz).normalize();
        Line.TempDirect.copyFromFloats(this.endx - this.startx, this.endy - this.starty, this.endz - this.startz);
        let distance = Line.TempDirect.length();
        Line.TempDirect.normalize()
        BABYLON.Vector3.CrossToRef(Line.TempDirect, Line.TempNormal, Line.TempExtend);

        this.ex = Line.TempExtend.x * this._width;
        this.ey = Line.TempExtend.y * this._width;
        this.ez = Line.TempExtend.z * this._width;

        this.repeatU = distance / this._repeatULen;
    }
    dispose() {
        this.renderer.del(this);
    }
}


class LineMeshRenderer {
    protected counter = Number.MIN_SAFE_INTEGER;
    protected lines: Map<number, Line> = new Map();
    protected mesh: BABYLON.Mesh;
    protected material: BABYLON.Material;
    protected texture: BABYLON.Texture;


    protected _dirty: boolean = false;
    protected customGeometry: CustomGeometry;

    public static CreateMaterial(scene: Scene, atlasImage: string, texture: BABYLON.Texture, alphaMode: number, pass?: Constants.PassTag, renderMode: 0 | 1 | 2 = 0) {
        if (renderMode == 1) {
            let _material = new TextureHSVMaterial(`${atlasImage}`, scene);
            _material.diffuseTexture = texture;
            _material.emissiveColor = new BABYLON.Color3(1, 1, 1);
            _material.disableDepthWrite = true;
            _material.backFaceCulling = false;
            _material.alphaMode = alphaMode;
            _material.transparencyMode = BABYLON.Material.MATERIAL_ALPHATESTANDBLEND;
            return _material;
        } else if (renderMode == 2) {
            let _material = new TextureHueSaturateMaterial(`${atlasImage}`, scene);
            _material.diffuseTexture = texture;
            _material.emissiveColor = new BABYLON.Color3(1, 1, 1);
            _material.disableDepthWrite = true;
            _material.backFaceCulling = false;
            _material.alphaMode = alphaMode;
            _material.transparencyMode = BABYLON.Material.MATERIAL_ALPHATESTANDBLEND;
            return _material;
        } else {
            let _material = new BABYLON.StandardMaterial(`${atlasImage}`, scene);
            _material.diffuseTexture = texture;
            _material.disableLighting = true;
            _material.emissiveColor = new BABYLON.Color3(1, 1, 1);
            _material.disableDepthWrite = true;
            _material.backFaceCulling = false;
            _material.alphaMode = alphaMode;
            _material.transparencyMode = BABYLON.Material.MATERIAL_ALPHATESTANDBLEND;
            _material.useAlphaFromDiffuseTexture = true;
            _material.cutoffEdgeAlphaDiff = 0.0;
            _material.alphaCutOff = 0.0;
            return _material;
        }
    }
    constructor(protected scene: CustomScene, protected image: string, protected maxCount = 128, protected renderMode: 0 | 1 | 2 = 0, noMipmap: boolean = true, samplerMode: number = undefined, alphaMode: number = 2, pass?: Constants.PassTag) {
        let texture = new BABYLON.Texture(null, this.scene.impl, noMipmap, false, samplerMode || BABYLON.Texture.LINEAR_LINEAR);
        texture.updateURL(image);
        texture.hasAlpha = true;
        texture.wrapU = BABYLON.Texture.WRAP_ADDRESSMODE;
        texture.wrapV = BABYLON.Texture.WRAP_ADDRESSMODE;
        texture.uScale = 1.001;

        let material = LineMeshRenderer.CreateMaterial(this.scene.impl, image, texture, alphaMode, pass, renderMode);

        this.material = material;
        this.texture = texture;

        this.mesh = new BABYLON.Mesh("meshSource", this.scene.impl);
        this.customGeometry = new CustomGeometry(this.mesh, `${image}#`, [Constants.VERTEX_ATTR_POSITION, Constants.VERTEX_ATTR_COLOR4, Constants.VERTEX_ATTR_UV], this.maxCount * 4, this.maxCount * 2);
        this.mesh.material = material;
        this.mesh.useVertexColors = true;
        this.mesh.hasVertexAlpha = true;
        this.mesh.alwaysSelectAsActiveMesh = true;

        let indices = this.customGeometry.indices();
        for (let i = 0; i < maxCount; i++) {
            let idx = i * 2 * 3;
            let ind = i * 4;
            indices[idx + 0] = ind + 0;
            indices[idx + 1] = ind + 1;
            indices[idx + 2] = ind + 2;
            indices[idx + 3] = ind + 3;
            indices[idx + 4] = ind + 2;
            indices[idx + 5] = ind + 1;
        }
    }
    set alphaIndex(val: number) {
        this.mesh.alphaIndex = val;
    }
    set alphaMode(val: number) {
        this.material.alphaMode = val;
    }
    set uvSpeedU(val: number) {
        if (this.texture) {
            this.texture.offsetSpeedU = val;
        }
    }
    set uvSpeedV(val: number) {
        if (this.texture) {
            this.texture.offsetSpeedV = val;
        }
    }
    create(image: string, startx: number, starty: number, startz: number, endx: number, endy: number, endz: number, normalx: number, normaly: number, normalz: number, r: number, g: number, b: number, a: number, width: number, repeatULen: number, repeatV: number) {
        let unique = this.counter++;
        let line = new Line(unique, this);
        line.startx = startx;
        line.starty = starty;
        line.startz = startz;
        line.endx = endx;
        line.endy = endy;
        line.endz = endz;
        line.r = r;
        line.g = g;
        line.b = b;
        line.a = a;
        line.image = image;
        line.repeatV = repeatV;
        line.normalx = normalx;
        line.normaly = normaly;
        line.normalz = normalz;
        line.width = width;
        line.repeatULen = repeatULen;

        this.lines.set(unique, line);
        this._dirty = true;
        return line;
    }
    del(line: Line) {
        let item = this.lines.get(line.unique);
        if (item == line) {
            this._dirty = true;
            this.lines.delete(line.unique);
        } else {
            console.error(`Line Dispose Error !`);
        }
    }
    update() {
        this.lines.forEach((line) => {
            this._dirty = line._onUpdate() || this._dirty;
        });
        if (this._dirty) {
            let idx = 0;
            let uvStep = 4 * 2;
            let colorStep = 4 * 4;
            let posStep = 4 * 3;
            let [uvs, uvstrip, uvoffset] = this.customGeometry.vertices(Constants.VERTEX_ATTR_UV);
            let [positions, posstrip, posoffset] = this.customGeometry.vertices(Constants.VERTEX_ATTR_POSITION);
            let [colors, colorstrip, coloroffset] = this.customGeometry.vertices(Constants.VERTEX_ATTR_COLOR4);
            this.lines.forEach((line) => {
                {
                    let off = uvoffset + idx * uvStep;
                    uvs[off + 0 + 0 * 2] = 0 * line.repeatU + line.uOffset;
                    uvs[off + 1 + 0 * 2] = 0 * line.repeatV + line.vOffset;
                    uvs[off + 0 + 1 * 2] = 1 * line.repeatU + line.uOffset;
                    uvs[off + 1 + 1 * 2] = 0 * line.repeatV + line.vOffset;
                    uvs[off + 0 + 2 * 2] = 0 * line.repeatU + line.uOffset;
                    uvs[off + 1 + 2 * 2] = 1 * line.repeatV + line.vOffset;
                    uvs[off + 0 + 3 * 2] = 1 * line.repeatU + line.uOffset;
                    uvs[off + 1 + 3 * 2] = 1 * line.repeatV + line.vOffset;
                }

                {
                    let off = posoffset + idx * posStep;
                    positions[off + 0] = line.startx - line.ex;
                    positions[off + 1] = line.starty - line.ey;
                    positions[off + 2] = line.startz - line.ez;

                    positions[off + 0 + 3] = line.endx - line.ex;
                    positions[off + 1 + 3] = line.endy - line.ey;
                    positions[off + 2 + 3] = line.endz - line.ez;

                    positions[off + 0 + 6] = line.startx + line.ex;
                    positions[off + 1 + 6] = line.starty + line.ey;
                    positions[off + 2 + 6] = line.startz + line.ez;

                    positions[off + 0 + 9] = line.endx + line.ex;
                    positions[off + 1 + 9] = line.endy + line.ey;
                    positions[off + 2 + 9] = line.endz + line.ez;
                }

                {
                    let off = coloroffset + idx * colorStep;
                    colors[off + 0 + 0 * 4] = line.r;
                    colors[off + 1 + 0 * 4] = line.g;
                    colors[off + 2 + 0 * 4] = line.b;
                    colors[off + 3 + 0 * 4] = line.a;
                    colors[off + 0 + 1 * 4] = line.r;
                    colors[off + 1 + 1 * 4] = line.g;
                    colors[off + 2 + 1 * 4] = line.b;
                    colors[off + 3 + 1 * 4] = line.a;
                    colors[off + 0 + 2 * 4] = line.r;
                    colors[off + 1 + 2 * 4] = line.g;
                    colors[off + 2 + 2 * 4] = line.b;
                    colors[off + 3 + 2 * 4] = line.a;
                    colors[off + 0 + 3 * 4] = line.r;
                    colors[off + 1 + 3 * 4] = line.g;
                    colors[off + 2 + 3 * 4] = line.b;
                    colors[off + 3 + 3 * 4] = line.a;
                }
                idx++;
            });

            this._dirty = false;
            this.customGeometry.update(idx * 6);
            this.mesh.setEnabled(idx > 0);
        }
    }
    dispose() {
        this.lines.clear();
        this.mesh.dispose(false, true);
        this.material.dispose();
    }
}

export class LineRenderer implements ILineRenderer {
    protected meshes: Map<string, LineMeshRenderer> = new Map();
    constructor(
        protected scene: CustomScene, protected maxCount: number = 128,
        protected renderMode: 0 | 1 | 2 = 0,
        protected noMipmap: boolean = true,
        protected samplerMode: number = BABYLON.Texture.LINEAR_LINEAR,
        protected pass?: Constants.PassTag
    ) {

        this.scene.registerBeforeRenderCall(this.update);
    }
    protected _alphaIndex: number;
    set alphaIndex(val: number) {
        this.meshes.forEach((mesh) => {
            mesh.alphaIndex = val;
        });
        this._alphaIndex = val;
    }
    protected _alphaMode: number;
    set alphaMode(val: number) {
        this.meshes.forEach((mesh) => {
            mesh.alphaMode = val;
        });
        this._alphaMode = val;
    }
    protected _uvSpeedU: number;
    /**
     * 每多少ms变化一轮
     */
    set uvSpeedU(val: number) {
        this.meshes.forEach((mesh) => {
            mesh.uvSpeedU = val;
        });
        this._uvSpeedU = val;
    }
    protected _uvSpeedV: number;
    /**
     * 每多少ms变化一轮
     */
    set uvSpeedV(val: number) {
        this.meshes.forEach((mesh) => {
            mesh.uvSpeedV = val;
        });
        this._uvSpeedV = val;
    }
    protected update = () => {
        this.meshes.forEach((mesh) => {
            mesh.update();
        });
    }
    create(image: string, startx: number, starty: number, startz: number, endx: number, endy: number, endz: number, normalx: number, normaly: number, normalz: number, r: number, g: number, b: number, a: number, width: number, repeatULen: number, repeatV: number): ILine {
        let mesh = this.meshes.get(image);
        if (mesh == undefined) {
            mesh = new LineMeshRenderer(this.scene, image, this.maxCount, this.renderMode, this.noMipmap, this.samplerMode, this.alphaMode, this.pass);
            mesh.alphaIndex = this._alphaIndex;
            mesh.alphaMode = this._alphaMode;
            mesh.uvSpeedU = this._uvSpeedU;
            mesh.uvSpeedV = this._uvSpeedV;
            this.meshes.set(image, mesh);
        }

        return mesh.create(image, startx, starty, startz, endx, endy, endz, normalx, normaly, normalz, r, g, b, a, width, repeatULen, repeatV);
    }
    dispose() {
        this.meshes.forEach((item) => {
            item.dispose();
        });
        this.scene.unregisterBeforeRenderCall(this.update);
        this.meshes.clear();
    }
}

export class LineRenderer2D extends LineRenderer {
    constructor(scene: CustomScene, protected viewDirection: BABYLON.Vector3, maxCount: number = 128,
        renderMode: 0 | 1 | 2 = 0,
        noMipmap: boolean = true,
        samplerMode: number = BABYLON.Texture.LINEAR_LINEAR,
        pass?: Constants.PassTag
    ) {
        super(scene, maxCount, renderMode, noMipmap, samplerMode, pass);
        this.viewDirection.normalize();
    }

    /**
     * 
     * @param image image 因为可能有UV流动需求, 因此此处不使用精灵图，且图片尺寸需要2的幂次
     * @param startx 
     * @param starty 
     * @param startz 
     * @param endx 
     * @param endy 
     * @param endz 
     * @param r 
     * @param g 
     * @param b 
     * @param a 
     * @param width 线宽度
     * @param repeatULen 每多少距离纹理重复一次
     * @param repeatV 纹理重复多少次
     * @returns 
     */
    create(image: string, startx: number, starty: number, startz: number, endx: number, endy: number, endz: number, r: number, g: number, b: number, a: number, width: number, repeatULen: number, repeatV: number): ILine {
        let mesh = this.meshes.get(image);
        if (mesh == undefined) {
            mesh = new LineMeshRenderer(this.scene, image, this.maxCount, this.renderMode, this.noMipmap, this.samplerMode, this.alphaMode, this.pass);
            mesh.alphaIndex = this._alphaIndex;
            this.meshes.set(image, mesh);
        }

        return mesh.create(image, startx, starty, startz, endx, endy, endz, this.viewDirection.x, this.viewDirection.y, this.viewDirection.z, r, g, b, a, width, repeatULen, repeatV);
    }
}