import { BABYLON } from "../render3d/babylon";
import { MainOpacityMaterial } from "./main_opacity_material";
import { LoadExtraMaterial } from "../render3d/extends/material";
import { OpsPass } from "../renderpass";
import { GLTFLoader } from "../gltf/glTFLoader";


export class MainOpacityEmissiveFresnelMaterial extends MainOpacityMaterial {
    public static ShaderName = "MainOpacityEmissiveFresnelMaterial";
    constructor(name: string, scene: BABYLON.Scene, pass: number = 0, shaderName = MainOpacityEmissiveFresnelMaterial.ShaderName) {
        super(name, scene, pass, shaderName);
        this.emissiveFresnelParameters = new BABYLON.FresnelParameters();
    }

    public static regist() {
        LoadExtraMaterial.loadCalls.set("main_opacity_emissive_fresnel", this.create);
        LoadExtraMaterial.loadPropertyCalls.set("main_opacity_emissive_fresnel", this.load);
    }
    protected static create(context: string, babylonScene: BABYLON.Scene, materialName: string, index: number, materialInfo: any, babylonDrawMode: number) {
        if (materialInfo) {
            let mat = new MainOpacityEmissiveFresnelMaterial(materialName, babylonScene, OpsPass.Transparent);
            let info = materialInfo;

            let numbers = info.diffuseColor;
            // if (numbers) {
            //     mat.diffuseColor = new Color3(numbers[0], numbers[1], numbers[2]);
            // }
            numbers = info.emissionColor;
            if (numbers) {
                mat.mainColor(new BABYLON.Color3(numbers[0], numbers[1], numbers[2]));
            }
            
            mat.emissiveFresnelParameters.bias = info.EFBias;
            mat.emissiveFresnelParameters.power = info.EFPower;
            mat.emissiveFresnelParameters.leftColor = new BABYLON.Color3(...info.EFLeft);
            mat.emissiveFresnelParameters.rightColor = new BABYLON.Color3(...info.EFRight);

            // callback(mat, state);

            return mat;
        } else {
            return undefined;
        }
    }
    protected static load(context: string, material: any, mat: any, _loader: GLTFLoader): Promise<void> {

        let info = material;

        let promises = [];
        if (info.diffuseTexture) {
            promises.push(LoadExtraMaterial.diffuseTexture(mat, _loader, context, info));
        }

        if (info.opacityTexture) {
            promises.push(LoadExtraMaterial.opacityTexture(mat, _loader, context, info));
        }

        LoadExtraMaterial.state(mat, info, _loader);

        return Promise.all(promises).then(() => {});
    }
}