
import { LoadExtraMaterial } from "../render3d/extends/material";
import { BABYLON } from "../render3d/babylon";
import { OpsPass } from "../renderpass";
import { MaterialBase } from "./base";
import { GLTFLoader } from "../gltf/glTFLoader";

export class MainOpacityMaterial extends BABYLON.StandardMaterial {
    public static ShaderName = "MainOpacityMaterial";
    public readonly pass: OpsPass;
    public readonly shaderName: string;
    constructor(name: string, scene: BABYLON.Scene, pass: number = 0, shaderName = MainOpacityMaterial.ShaderName) {
        super(name, scene);
        this.pass = pass;
        this.shaderName = shaderName;
        this.diffuseColor = new BABYLON.Color3(1, 1, 1);
        this.disableLighting = true;
        {
            let texture = new BABYLON.Texture(null, scene, true, true); texture.updateURL(MaterialBase.DefaultWhitePath);
            this.diffuseTexture = texture;
        }
        // {
        //     let texture = new BABYLON.Texture(null, scene, true, true); texture.updateURL(MaterialBase.DefaultWhitePath);
        //     this.opacityTexture = texture;
        // }
        if (pass >= OpsPass.Transparent) {
            this.diffuseTexture.hasAlpha = true;
            // this.opacityTexture.hasAlpha = true;
            this.useAlphaFromDiffuseTexture = true;
        }
    }

    public mainColor(val: BABYLON.Color3): void {
        this.emissiveColor.copyFrom(val);
    }

    public static regist() {
        LoadExtraMaterial.loadCalls.set("main_opacity", this.create);
        LoadExtraMaterial.loadPropertyCalls.set("main_opacity", this.load);
    }
    public static create(context: string, babylonScene: BABYLON.Scene, materialName: string, index: number, materialInfo: any, babylonDrawMode: number) {
        if (materialInfo) {
            let mat = new MainOpacityMaterial(materialName, babylonScene, OpsPass.Transparent);
            let info = materialInfo;

            let numbers = info.diffuseColor;
            // if (numbers) {
            //     mat.diffuseColor = new Color3(numbers[0], numbers[1], numbers[2]);
            // }
            numbers = info.emissionColor;
            if (numbers) {
                mat.mainColor(new BABYLON.Color3(numbers[0], numbers[1], numbers[2]));
            }
            
            // let state = LoadExtraMaterial.state(mat, info);

            // callback(mat, state);

            return mat;
        } else {
            return undefined;
        }
    }
    public static load(context: string, material: any, mat: any, _loader: GLTFLoader): Promise<void> {

        let info = material;

        let promises = [];
        if (info.diffuseTexture) {
            promises.push(LoadExtraMaterial.diffuseTexture(mat, _loader, context, info));
        }

        if (info.opacityTexture) {
            promises.push(LoadExtraMaterial.opacityTexture(mat, _loader, context, info));
        }

        LoadExtraMaterial.state(mat, info, _loader);

        return Promise.all(promises).then(() => {});
    }
}