
import { GLTFLoader } from "../gltf/glTFLoader";
import { BABYLON } from "../render3d/babylon";
import { LoadExtraMaterial } from "../render3d/extends/material";
import { OpsPass } from "../renderpass";

// export class MainTextureMaterial extends Material {
//     constructor(name: string, scene: Scene, pass: number = 0) {
//         super(name, scene, "UnlitShader", pass);
//     }
//     public set mainTexture(val: Texture) {
//         MainTextureBlock.texture(this.entity, val);
//     }
//     public baseColor(r: number, g: number, b: number, a: number) {
//         MainTextureBlock.color(this.entity, r, g, b, a);
//     }
// }

export class MainTextureMaterial extends BABYLON.StandardMaterial {
    public set mainTexture(val: BABYLON.Texture) {
        this.emissiveTexture = val;
    }
    public baseColor(r: number, g: number, b: number, a: number) {
        this.emissiveColor = new BABYLON.Color3(r, g, b);
    }
    public readonly pass: OpsPass;
    constructor(name: string, scene: BABYLON.Scene, pass: number = OpsPass.Transparent) {
        super(name, scene);
        this.pass = pass;
    }

    public static regist() {
        LoadExtraMaterial.loadCalls.set("main_tex", this.create);
        LoadExtraMaterial.loadPropertyCalls.set("main_tex", this.load);
    }
    private static create(context: string, babylonScene: BABYLON.Scene, materialName: string, index: number, materialInfo: any, babylonDrawMode: number) {
        if (materialInfo) {
            let mat = new MainTextureMaterial(materialName, babylonScene);
            let info = materialInfo;

            let numbers = info.diffuseColor;
            // if (numbers) {
            //     mat.diffuseColor = new Color3(numbers[0], numbers[1], numbers[2]);
            // }
            numbers = info.emissionColor;
            if (numbers) {
                mat.diffuseColor = new BABYLON.Color3(numbers[0], numbers[1], numbers[2]);
            }
            
            // let state = LoadExtraMaterial.state(mat, info);

            // callback(mat, state);

            return mat;
        } else {
            return undefined;
        }
    }
    private static load(context: string, material: any, mat: any, _loader: GLTFLoader): Promise<void> {

        let info = material;

        let promises = [];
        if (info.diffuseTexture) {
            promises.push(LoadExtraMaterial.diffuseTexture(mat, _loader, context, info));
        }

        LoadExtraMaterial.state(mat, info, _loader);

        return Promise.all(promises).then(() => {});
    }
}