import { IPITextureInfo } from "../material_info";
import { BABYLON } from "../render3d/babylon";
import { ShaderMaterial } from "./shader_material";
import { LoadExtraMaterial } from "../render3d/extends/material";
const KEY_OVERLAY = "KEY_OVERLAY";
//laserEffect
BABYLON.Effect.ShadersStore[`${KEY_OVERLAY}VertexShader`] = `
    precision mediump float;

    attribute vec3 position;
    attribute vec4 color;
    attribute vec2 uv;

    uniform mat4 worldViewProjection;

    varying vec3 vColor;
    varying float vAlpha;
    varying vec2 vUV;

    void main(void) {
        vec3 positionUpdated = position;
        vec2 uvUpdated = uv;

        gl_Position = worldViewProjection*vec4(positionUpdated, 1.0);

        vUV = uv;

        vColor = color.rgb;
        vAlpha = color.a;
    }
`;

BABYLON.Effect.ShadersStore[`${KEY_OVERLAY}FragmentShader`] =
    `
    precision mediump float;

    varying vec3 vColor;
    varying float vAlpha;
    varying vec2 vUV;

    uniform sampler2D diffuseTexture;
 
    void main(void) {
        vec4 src = texture2D(diffuseTexture, vUV);
        gl_FragColor = vec4(src.rgb * vAlpha + vColor, src.w);
    }
`;

export class OverlayColorMaterial extends ShaderMaterial {
    public static readonly KEY = KEY_OVERLAY;

    private _time: number = Date.now();
    private _distortionInfo: BABYLON.Vector4 = new BABYLON.Vector4(0, 0, 0, 0);
    private _tintColor: BABYLON.Color4 = new BABYLON.Color4(0, 0, 0, 0);
    private _control: BABYLON.Vector4 = new BABYLON.Vector4(0, 0, 0, 0);
    public _emissiveColor: BABYLON.Color3 = new BABYLON.Color3(1, 1, 1);
    public get emissiveColor() {
        return this._emissiveColor;
    }
    public set emissiveColor(value: BABYLON.Color3) {
        this._emissiveColor = value;
    }
    public get alpha() {
        return this._tintColor.a;
    }
    public set alpha(value: number) {
        this._tintColor.a = value;
    }
    constructor(materialName: string, babylonScene: any) {
        super(
            materialName,
            babylonScene,
            `${OverlayColorMaterial.KEY}`,
            {
                attributes: ["position", "uv", "color"],
                uniforms: ["worldViewProjection"],
                samplers: ["diffuseTexture"],
                needAlphaBlending: true
            }
        );
    }
    protected _diffuseTexture: BABYLON.Texture;
    public set diffuseTexture(tex: BABYLON.Texture) {
        this.setTexture('diffuseTexture', tex);
        this._diffuseTexture = tex;
    }
    public get diffuseTexture() {
        return this._diffuseTexture;
    }
    isReady(mesh?: BABYLON.AbstractMesh, useInstances?: boolean): boolean {
        let result = !!this._diffuseTexture;
        return result && super.isReady(mesh, useInstances);
    }
    bind(world: BABYLON.Matrix, mesh?: BABYLON.Mesh, effectOverride?: BABYLON.Effect): void {
        super.bind(world, mesh, effectOverride);
    }
    dispose(forceDisposeEffect?: boolean, forceDisposeTextures?: boolean, notBoundToMesh?: boolean): void {
        this._diffuseTexture?.dispose();
        super.dispose(forceDisposeEffect, forceDisposeTextures, notBoundToMesh);
    }
}
