
import { MainOpacityMaterial } from "./main_opacity_material";
import { BABYLON } from "../render3d/babylon";
import { LoadExtraMaterial } from "../render3d/extends/material";
import { OpsPass } from "../renderpass";
import { GLTFLoader } from "../gltf/glTFLoader";

export class TwoOpacityMixMaterial extends MainOpacityMaterial {
    constructor(name: string, scene: BABYLON.Scene) {
        super(name, scene, OpsPass.Transparent, "TwoOpacityMixMaterial");
    }

    public static KEY_MixOpacityInfo = "uMixOpacityInfo";
    public static KEY_MixOpacityTilloff = "uMixOpacityTilloff";
    public static KEY_MixOpacityControlTilloff = "uMixOpacityControlTilloff";
    public static KEY_MixOpacityUVSpeed = "uMixOpacityUVSpeed";
    public static regist() {
        LoadExtraMaterial.loadCalls.set("two_opacity_mix", this.create);
        LoadExtraMaterial.loadPropertyCalls.set("two_opacity_mix", this.load);
    }
    protected static create(context: string, babylonScene: BABYLON.Scene, materialName: string, index: number, materialInfo: any, babylonDrawMode: number) {
        if (materialInfo) {
            let mat = new TwoOpacityMixMaterial(materialName, babylonScene);
            let info = materialInfo;

            mat.mixOpacityUScale            = info.MixOUS;
            mat.mixOpacityVScale            = info.MixOVS;
            mat.mixOpacityUOffset           = info.MixOUO;
            mat.mixOpacityVOffset           = info.MixOVO;
            mat.mixOpacityUSpeed            = info.MixOUSp;
            mat.mixOpacityVSpeed            = info.MixOVSp;
            mat.mixOpacityIntensity         = info.MixOIns;
            mat.mixOpacityControlUScale     = info.MixCUS;
            mat.mixOpacityControlVScale     = info.MixCVS;
            mat.mixOpacityControlUOffset    = info.MixCUO;
            mat.mixOpacityControlVOffset    = info.MixCVO;
            mat.mixOpacityControlIntensity  = info.MixCIns;
            mat.mixOpacityControlValue      = info.MixCVal;

            let numbers = info.diffuseColor;
            // if (numbers) {
            //     mat.diffuseColor = new Color3(numbers[0], numbers[1], numbers[2]);
            // }
            numbers = info.emissionColor;
            if (numbers) {
                mat.diffuseColor = new BABYLON.Color3(numbers[0], numbers[1], numbers[2]);
            }
            
            // let state = LoadExtraMaterial.state(mat, info);

            // callback(mat, state);

            return mat;
        } else {
            return undefined;
        }
    }
    protected static load(context: string, material: any, mat: any, _loader: GLTFLoader): Promise<void> {

        let info = material;

        let promises = [];
        if (info.diffuseTexture) {
            promises.push(LoadExtraMaterial.diffuseTexture(mat, _loader, context, info));
        }

        if (info.opacityTexture) {
            promises.push(LoadExtraMaterial.opacityTexture(mat, _loader, context, info));
        }

        LoadExtraMaterial.state(mat, info, _loader);

        return Promise.all(promises).then(() => {});
    }
}