import { MathDataPool } from "../../render3d/extends/math";
import { BABYLON } from "../../render3d/babylon";
// import * as BABYLON from "../../pi_babylon/render3d/babylon";
import { EBoxShapeMode, IShapeEmitterType } from "./ishape_emitter_type";

export class BoxShapeEmitter implements IShapeEmitterType {
    clone(): IShapeEmitterType {
        throw new Error("Method not implemented.");
    }
    getClassName(): string {
        throw new Error("Method not implemented.");
    }
    serialize() {
        throw new Error("Method not implemented.");
    }
    parse(serializationObject: any, scene: BABYLON.Scene): void {
        throw new Error("Method not implemented.");
    }
    public dispose() {
        MathDataPool.Recycle(this.minEmitBox);
        MathDataPool.Recycle(this.maxEmitBox);
        MathDataPool.Recycle(this.direction);
        MathDataPool.Recycle(this.position);
        MathDataPool.Recycle(this.rotation);
        MathDataPool.Recycle(this.scaling);
        MathDataPool.Recycle(this.localMatrix);
        this.minEmitBox = undefined;
        this.maxEmitBox = undefined;
        this.direction  = undefined;
        this.position   = undefined;
        this.rotation   = undefined;
        this.scaling    = undefined;
        this.localMatrix= undefined;
    }
    private minEmitBox = MathDataPool.Vector3(-0.5, -0.5, -0.5);
    private maxEmitBox = MathDataPool.Vector3(0.5, 0.5, 0.5);

    public emitMode: EBoxShapeMode = EBoxShapeMode.Volume;
    public direction = MathDataPool.Vector3(0.0, 0.0, 1.0);

    public position: BABYLON.Vector3 = MathDataPool.Vector3(0, 0, 0);
    public rotation: BABYLON.Vector3 = MathDataPool.Vector3(0, 0, 0);
    public scaling: BABYLON.Vector3 = MathDataPool.Vector3(1, 1, 1);
    public localMatrix: BABYLON.Matrix = MathDataPool.Matrix();
    alignDirection: boolean = false;
    randomizeDirection: number = 0;
    spherizeDirection: number = 0;
    randomizePosition: number = 0;
    public rotateByQuaternionAroundPoint = (position: BABYLON.Vector3, rotateQuat: BABYLON.Quaternion, centerOffset: BABYLON.Vector3, orbtialDiff: BABYLON.Vector3) => {
        position.rotateByQuaternionAroundPointToRef(rotateQuat, centerOffset, orbtialDiff);
    }
    /**
     * 
     */
    public startDirectionFunction(worldMatrix: BABYLON.Matrix, directionToUpdate: BABYLON.Vector3, position: BABYLON.Vector3, localPosition: BABYLON.Vector3, isLocal: boolean): void {

        var direction = BABYLON.TmpVectors.Vector3[0];
        direction.copyFrom(this.direction);

        localPosition.normalizeToRef(BABYLON.TmpVectors.Vector3[1]);
        if (this.spherizeDirection != 0) {
            direction.x = direction.x * (1.0 - this.spherizeDirection) + BABYLON.TmpVectors.Vector3[1].x * this.spherizeDirection;
            direction.y = direction.y * (1.0 - this.spherizeDirection) + BABYLON.TmpVectors.Vector3[1].y * this.spherizeDirection;
            direction.z = direction.z * (1.0 - this.spherizeDirection) + BABYLON.TmpVectors.Vector3[1].z * this.spherizeDirection;
            direction.normalize();
        }

        if (this.randomizeDirection != 0) {
            direction.x += BABYLON.Scalar.RandomRange(0, this.randomizeDirection);
            direction.y += BABYLON.Scalar.RandomRange(0, this.randomizeDirection);
            direction.z += BABYLON.Scalar.RandomRange(0, this.randomizeDirection);
            direction.normalize();
        }
        if (isLocal) {
            directionToUpdate.copyFrom(direction);
        }
        else {
            BABYLON.Vector3.TransformNormalFromFloatsToRef(direction.x, direction.y, direction.z, worldMatrix, directionToUpdate);
        }
        directionToUpdate.normalize();
    }

    /**
     * Called by the particle System when the position is computed for the created particle.
     * @param worldMatrix is the world matrix of the particle system
     * @param positionToUpdate is the position vector to update with the result
     * @param particle is the particle we are computed the position for
     * @param isLocal defines if the position should be set in local space
     */
    startPositionFunction(worldMatrix: BABYLON.Matrix, positionToUpdate: BABYLON.Vector3, emissionLoop: number, emissionProgress: number, emissionIndex: number, emissionTotal: number, isLocal: boolean): void {
        var randX = 0;
        var randY = 0;
        var randZ = 0;

        randX = BABYLON.Scalar.RandomRange(-0.5, 0.5);
        randY = BABYLON.Scalar.RandomRange(-0.5, 0.5);
        randZ = BABYLON.Scalar.RandomRange(-0.5, 0.5);

        switch (this.emitMode) {
            case (EBoxShapeMode.Volume): {
                // randX = BABYLON.Scalar.RandomRange(-0.5, 0.5);
                // randY = BABYLON.Scalar.RandomRange(-0.5, 0.5);
                // randZ = BABYLON.Scalar.RandomRange(-0.5, 0.5);
                break;
            }
            case (EBoxShapeMode.Shell): {
                let faceRange = Math.random();
                // YZ
                if (faceRange < 0.3333) {
                    randX = faceRange < 0.1666 ? 0.5 : -0.5;
                    // XZ
                } else if (faceRange < 0.6666) {
                    randY = faceRange < 0.5 ? 0.5 : -0.5;
                    // XY
                } else {
                    randZ = faceRange < 0.8333 ? 0.5 : -0.5;
                }
                break;
            }
            case (EBoxShapeMode.Edge): {
                let faceRange = Math.random();
                let faceRange2 = Math.random();
                let faceRange3 = Math.random();
                // X
                if (faceRange < 0.3333) {
                    randY = faceRange2 < 0.5 ? 0.5 : -0.5;
                    randZ = faceRange3 < 0.5 ? 0.5 : -0.5;
                    // Y
                } else if (faceRange < 0.6666) {
                    randX = faceRange2 < 0.5 ? 0.5 : -0.5;
                    randZ = faceRange3 < 0.5 ? 0.5 : -0.5;
                    // Z
                } else {
                    randX = faceRange2 < 0.5 ? 0.5 : -0.5;
                    randY = faceRange3 < 0.5 ? 0.5 : -0.5;
                }
                break;
            }
        }

        if (this.randomizePosition != 0) {
            randX += BABYLON.Scalar.RandomRange(-this.randomizePosition, this.randomizePosition);
            randZ += BABYLON.Scalar.RandomRange(-this.randomizePosition, this.randomizePosition);
            randY += BABYLON.Scalar.RandomRange(-this.randomizePosition, this.randomizePosition);
        }

        if (isLocal) {
            positionToUpdate.copyFromFloats(randX, randY, randZ);
        }
        else {
            BABYLON.Vector3.TransformCoordinatesFromFloatsToRef(randX, randY, randZ, worldMatrix, positionToUpdate);
        }
    }
}