import { BABYLON } from "../render3d/babylon";
// import * as BABYLON from "../pi_babylon/render3d/babylon";
// import { CUSTOM_MATERIAL_SHADER_NAMES } from "pi_babylon/render3d/extends/custom_material";
import { BoxShapeEmitter } from "./emitter/box_shape_emitter";
import { CircleShapeEmitter } from "./emitter/circle_shape_emitter";
import { ConeShapeEmitter } from "./emitter/cone_shape_emitter";
import { EdgeShapeEmitter } from "./emitter/edge_shape_emitter";
import { PointShapeEmitter } from "./emitter/point_shape_emitter";
import { RectangleShapeEmitter } from "./emitter/rectangle_shape_emitter";
import { SphereShapeEmitter } from "./emitter/sphere_shape_emitter";
import { HemisphereShapeEmitter } from "./emitter/hemisphere_shape_emitter";
import { FloatInterpolation, parseColor4Gradient, parseFloatInterpolation } from "./interpolation";
import { DefaultValue, IParticleSystemConfig, IShape, TEmissionIndexBursts, TEmissionIndexRate, TParamGravity, TParamLimitVelocityOverLifetime, TParamRotationOverLifetime, TParamSizeOverLifetime, TParamStartColor, TParamStartLifetime, TParamStartRotation, TParamStartSize, TParamStartSpeed, TParamTextureSheet, TParamVelocityOverLifetime, TShapeTypeBox, TShapeTypeCircle, TShapeTypeCone, TShapeTypeEdge, TShapeTypeHemisphere, TShapeTypeRectangle, TShapeTypeSphere } from "./iparticle_system_config";
import { EBoxShapeMode, IShapeEmitterType } from "./emitter/ishape_emitter_type";
import { MeshParticleSystem } from "./mesh_particle_system";
import { EMeshParticleSpaceMode } from "./particle_system_tool";
import { SizeOverLifetime } from "./modifier/size_over_lifetime";
import { StartSize } from "./modifier/start_size";
import { StartRotation } from "./modifier/start_rotation";
import { StartColor } from "./modifier/start_color";
import { VelocityOverLifetime } from "./modifier/velocity_over_lifetime";
import { ForceOverLifetime } from "./modifier/force_over_lifetime";
import { Color4Interpolate } from "./modifier/base";
import { RotationOverLifetime } from "./modifier/rotation_over_lifetime";
import { TrailModifier } from "./modifier/trail";
import { GLTFLoader } from "../gltf/glTFLoader";
import { INodeNew } from "../gltf/interface";
import { ETempNodeAnimationDataKey, GLTFTool } from "../gltf/base";
import { LoadExtraMaterial } from "../render3d/extends/material";
import { MeshParticleSystemGeometry } from "./mesh_particle_system_geomtry";
import { Scene } from "../render3d/scene";
import { ExtensionTools } from "../interface/index";
import { StandardMaterialData, StandardMaterialMesh } from "../material/standardmaterial";

// import { ModelObj } from "../model_obj";

export function initial() {

    // (<any>ModelObj.prototype).animEndCB = function(this: ModelObj) {
    //     if (this.animOpt) {
    //         const func = this.animOpt.endCall;
    //         this.animOpt = undefined;
    //         if (func) {
    //             let maxWaitTime = 0;
    //             this.particleSysList.forEach((ps: any) => {
    //                 if ((<MeshParticleSystem>ps).psTool && !(<MeshParticleSystem>ps).psTool.looping && (<MeshParticleSystem>ps).psTool.maxLifetimeOverage) {
    //                     maxWaitTime = Math.max((<MeshParticleSystem>ps).psTool.maxLifetimeOverage, maxWaitTime);
    //                 }
    //             });

    //             if (maxWaitTime > 5) {
    //                 setTimer(func, [], maxWaitTime, 1);
    //             } else {
    //                 func();
    //             }
    //         }
    //     }
    // };

    (<any>BABYLON.Light.prototype)._resyncMeshes = function() {
        if (!this.onlyWorkInLocal) {
            for (var mesh of this.getScene().meshes) {
                mesh._resyncLightSource(this);
            }
        }
    };

    BABYLON.MaterialHelper.BindLights = function (scene, mesh: BABYLON.AbstractMesh, effect, defines, maxSimultaneousLights, rebuildInParallel) {
        if (maxSimultaneousLights === void 0) { maxSimultaneousLights = 4; }
        if (rebuildInParallel === void 0) { rebuildInParallel = false; }
        var len = Math.min(mesh.lightSources.length, maxSimultaneousLights);
        for (var i = 0; i < len; i++) {
            var light = mesh.lightSources[i];
            this.BindLight(light, i, scene, effect, typeof defines === "boolean" ? defines : defines["SPECULARTERM"], rebuildInParallel);
        }
        var len2 = Math.min(mesh.lightSourcesOnlyWorkInLocal.length + len, maxSimultaneousLights) - len;
        for (var i = 0; i < len2; i++) {
            var light = mesh.lightSourcesOnlyWorkInLocal[i];
            this.BindLight(light, i + len, scene, effect, typeof defines === "boolean" ? defines : defines["SPECULARTERM"], rebuildInParallel);
        }
    };

    GLTFLoader.prototype.loadNodePure = function(this: GLTFLoader, context: string, nodeIndex: number, nodeInfo: INodeNew) {
        
        let animationAbout = this.loaderRecord.nodeAnimationAbout.get(nodeIndex);

        if (nodeInfo.mesh != undefined) {
            let meshInfo = this.gltf.meshes[nodeInfo.mesh];
            let node = new BABYLON.TransformNode(nodeInfo.name, this._scene);
            
            let meshParticleConfig: IParticleSystemConfig = GLTFTool.meshParticleCfg(nodeInfo);
            if (meshParticleConfig) {
                let instanceID = meshParticleConfig.instanceID;
                // if (instanceID == undefined &&  meshParticleConfig.looping) {
                //     instanceID = nodeInfo.name;
                // }
                if (instanceID) {
                    instanceID = this.rootUrl + "#" + instanceID;
                    if (!this.babylonScene.metadata.InstanceMeshParticleMap) {
                        this.babylonScene.metadata.InstanceMeshParticleMap = new Map();
                    }
                    let InstanceMeshParticleMap        = this.babylonScene.metadata.InstanceMeshParticleMap;
                    let meshParticle: MeshParticleSystemGeometry = InstanceMeshParticleMap.get(instanceID);
                    if (!meshParticle) {
                        meshParticle = new MeshParticleSystemGeometry(nodeInfo.name, this._scene);
                        meshParticle.metadata = {};
                        InstanceMeshParticleMap.set(instanceID, meshParticle);

                        let assign = (mesh: BABYLON.AbstractMesh) => {
                            mesh.setEnabled(false);
                            formatMeshParticle(meshParticleConfig, <any>meshParticle);
                            meshParticle.setSourceMesh(<BABYLON.Mesh>mesh);
                            
                            let instance = meshParticle.createInstance(nodeInfo.name);
                            instance.parent = node;
                        }
                        let loader = this._loadMeshAsync(``, nodeInfo, nodeIndex, meshInfo, node, assign).then((mesh: BABYLON.Mesh) => {
                            // // Trail
                            // if (meshParticleConfig.trail) {
                            //     let indexMat = meshParticleConfig.trail.material;
                            //     let material = this.gltf.materials[indexMat];
                            //     let babylonMaterial = this._loadMaterialAsync("/materials/" + indexMat, indexMat, 0, material, null, 0);
                            //     meshParticle.psTool.trail.mesh.material = babylonMaterial;
                            //     meshParticle.psTool.trail.mesh.alphaIndex = material.extensions.PI_material.renderQueue;
                            // }
                            meshParticle.start();
                            // this.loaderRecord.particles.push(meshParticle);
                        });
                        meshParticle.metadata.loadPromise = loader;
                        this.loaderRecord.meshLoaders.push(loader);
                    } else {
                        let loader = meshParticle.metadata.loadPromise;
                        let instance = meshParticle.createInstance(nodeInfo.name);
                        instance.parent = node;
                        this.loaderRecord.meshLoaders.push(loader);
                    }
                } else {
                    let key = this.rootUrl + "#" + nodeInfo.name + "#" + nodeIndex;
                    let meshParticle: MeshParticleSystem = new MeshParticleSystem(nodeInfo.name, <Scene>this._scene, key);
                    let assign = (mesh: BABYLON.AbstractMesh) => {
                        mesh.setEnabled(false);
                        formatMeshParticle(meshParticleConfig, meshParticle);
                        meshParticle.parent = node;
                        meshParticle.setSourceMesh(<BABYLON.Mesh>mesh);
                    }
                    let loader = this._loadMeshAsync(``, nodeInfo, nodeIndex, meshInfo, node, assign).then((mesh: BABYLON.Mesh) => {
                        if (meshParticle.isDisposed()) return;

                        // Trail
                        if (meshParticleConfig.trail) {
                            let indexMat = meshParticleConfig.trail.material;
                            let material = this.gltf.materials[indexMat];

                            let forParticleSys = true;
                            let babylonMaterial = this._loadMaterialAsync("/materials/" + indexMat, indexMat, 0, material, <any>meshParticle.psTool.trail.mesh, 0, !forParticleSys);
                            if (!forParticleSys && ExtensionTools.Option.optimizeMaterialWithResTab) {
                                let mesh = <StandardMaterialMesh>meshParticle.psTool.trail.mesh;
                                mesh.materialData = new StandardMaterialData(this._scene);
                                mesh.materialData.fromMaterial(<any>babylonMaterial);
                                (<BABYLON.Mesh>mesh).onDisposeObservable.addOnce(function() {
                                    (<StandardMaterialData>mesh.materialData)?.dispose();
                                    mesh.materialData = undefined;
                                });
                            }

                            meshParticle.psTool.trail.mesh.material = babylonMaterial;
                            meshParticle.psTool.trail.mesh.alphaIndex = material.extensions.PI_material.renderQueue;
                        }
                        this.loaderRecord.particles.push(meshParticle);
                    });
                    this.loaderRecord.meshLoaders.push(loader);
                }
            } else {
                let loader = this._loadMeshAsync(``, nodeInfo, nodeIndex, meshInfo, node, () => {});
                this.loaderRecord.meshLoaders.push(loader);
            }
            node.parent = this.rootBabylonMesh;
            animationAbout[ETempNodeAnimationDataKey.Node] = node;
        } else {
            let node = new BABYLON.TransformNode(nodeInfo.name, this._scene);
            node.parent = this.rootBabylonMesh;
            animationAbout[ETempNodeAnimationDataKey.Node] = node;
        }
    };

    BABYLON.GLTF2.GLTFLoader.prototype.loadNodeAsync = function(this: any, context: string, node: any, assign: (babylonTransformNode: BABYLON.TransformNode) => void = () => { }): Promise<BABYLON.TransformNode> {
        // PI_BEGIN jzy
        let p = this.meshPromiseMap.get(node.index);
        if (p) {

            this.RecordPromise(`loadNodeAsync - jzy - ${this._fileName} - ${context}`, p);

            p.then((mesh: any) => {
                assign(mesh);
            });
            return p;
        }
        // PI_END

        const extensionPromise = this._extensionsLoadNodeAsync(context, node, assign);
        if (extensionPromise) {

            this.RecordPromise(`loadNodeAsync - extensionPromise - ${this._fileName} - ${context}`, extensionPromise);

            return extensionPromise;
        }

        if (node._babylonTransformNode) {
            throw new Error(`${context}: Invalid recursive node hierarchy`);
        }

        const promises = new Array<Promise<any>>();

        this.logOpen(`${context} ${node.name || ""}`);

        const loadNode = (babylonTransformNode: BABYLON.TransformNode) => {
            BABYLON.GLTF2.GLTFLoader.AddPointerMetadata(babylonTransformNode, context);
            (<any>BABYLON.GLTF2.GLTFLoader)._LoadTransform(node, babylonTransformNode);

            if (node.camera != undefined) {
                const camera = BABYLON.GLTF2.ArrayItem.Get(`${context}/camera`, this._gltf.cameras, node.camera);
                promises.push(this.loadCameraAsync(`/cameras/${(<any>camera).index}`, camera, (babylonCamera) => {
                    babylonCamera.parent = babylonTransformNode;
                }));
            }

            if (node.children) {
                for (const index of node.children) {
                    const childNode = BABYLON.GLTF2.ArrayItem.Get(`${context}/children/${index}`, this._gltf.nodes, index);
                    promises.push(this.loadNodeAsync(`/nodes/${(<any>childNode).index}`, childNode, (childBabylonMesh) => {
                        childBabylonMesh.parent = babylonTransformNode;
                    }));
                }
            }

            assign(babylonTransformNode);
        };

        if (node.mesh == undefined) {
            const nodeName = node.name || `node${node.index}`;
            this._babylonScene._blockEntityCollection = this._forAssetContainer;
            node._babylonTransformNode = new BABYLON.TransformNode(nodeName, this._babylonScene);
            this._babylonScene._blockEntityCollection = false;
            // PI_BEGIN
            if (node._boundingbox) {
                this._addBoundingBox(node);
            }
            // PI_END
            loadNode(node._babylonTransformNode);

            ////////////////////////
            let meshParticle: MeshParticleSystem;
            let meshParticleConfig = <IParticleSystemConfig>(<any>node).meshParticle;
            if (!meshParticleConfig) {
                meshParticleConfig = node.extras?.meshParticle;
            }
            if (meshParticleConfig) {
                let key = this.rootUrl + "#" + nodeName + "#" + node.index;
                (<any>this)._babylonScene._blockEntityCollection = this._forAssetContainer;
                meshParticle = new MeshParticleSystem(nodeName, (<any>this)._babylonScene, key);
                (<any>this)._babylonScene._blockEntityCollection = false;

                formatMeshParticle(meshParticleConfig, meshParticle);
                (<any>this)._particles.push(<any>meshParticle);

                meshParticle.parent = node._babylonTransformNode!;

                // Trail
                if (meshParticleConfig.trail) {
                    let material = (<any>this)._gltf.materials[meshParticleConfig.trail.material];
                    let forParticleSys = true;
                    let trailPromisre = this._loadMaterialAsync("/materials/" + meshParticleConfig.trail.material, material, meshParticle.psTool.trail.mesh, 0, function(babylonMaterial) {
                        if (!forParticleSys && ExtensionTools.Option.optimizeMaterialWithResTab) {
                            let mesh = <StandardMaterialMesh>meshParticle.psTool.trail.mesh;
                            mesh.materialData = new StandardMaterialData(this._scene);
                            mesh.materialData.fromMaterial(<any>babylonMaterial);
                            (<BABYLON.Mesh>mesh).onDisposeObservable.addOnce(function() {
                                (<StandardMaterialData>mesh.materialData)?.dispose();
                                mesh.materialData = undefined;
                            });
                        }

                        meshParticle.psTool.trail.mesh.material = babylonMaterial;
                        meshParticle.psTool.trail.mesh.alphaIndex = material.extensions.PI_material.renderQueue;
                    }, !forParticleSys);
                    promises.push(trailPromisre);
                }
                meshParticle.setSourceMesh(null);
            }
            ///////////////////////
        }
        else {
            const mesh = BABYLON.GLTF2.ArrayItem.Get(`${context}/mesh`, this._gltf.meshes, node.mesh);
            promises.push(this._loadMeshAsync(`/meshes/${(<any>mesh).index}`, node, mesh, loadNode));
        }

        this.logClose();

        // PI_BEGIN
        // return Promise.all(promises).then(() => {
        //  this._forEachPrimitive(node, (babylonMesh) => {
        //      if ((babylonMesh as Mesh).geometry && (babylonMesh as Mesh).geometry!.useBoundingInfoFromGeometry) {
        //          // simply apply the world matrices to the bounding info - the extends are already ok
        //          babylonMesh._updateBoundingInfo();
        //      } else {
        //          babylonMesh.refreshBoundingInfo(true);
        //      }
        //  });
        //
        //  return node._babylonTransformNode!;
        // });
        // //////////////////////////////
        const promise =  Promise.all(promises).then(() => {
            this._forEachPrimitive(node, (babylonMesh) => {
                let geometry = (babylonMesh as BABYLON.Mesh).geometry;
                if ((babylonMesh as BABYLON.Mesh)._boundingInfo && (babylonMesh as BABYLON.Mesh)._boundingInfo?.isImport) {
                    babylonMesh._updateBoundingInfo();
                }
                else if (geometry && (geometry!.useBoundingInfoFromGeometry)) {
                    // simply apply the world matrices to the bounding info - the extends are already ok
                    babylonMesh._updateBoundingInfo();
                } else {
                    babylonMesh.refreshBoundingInfo(true);
                }
            });

            return node._babylonTransformNode!;
        });

        this.meshPromiseMap.set(node.index, promise);

        return promise;
        // PI_END
    };

    (<any>BABYLON.GLTF2.GLTFLoader.prototype)._loadMeshAsync = function(this: any, context: string, node: any, mesh: any, assign: (babylonTransformNode: BABYLON.TransformNode) => void): Promise<BABYLON.TransformNode> {
        const primitives = mesh.primitives;
        if (!primitives || !primitives.length) {
            throw new Error(`${context}: Primitives are missing`);
        }

        if (primitives[0].index == undefined) {
            BABYLON.GLTF2.ArrayItem.Assign(primitives);
        }

        const promises = new Array<Promise<any>>();

        this.logOpen(`${context} ${mesh.name || ""}`);

        const name = node.name || `node${node.index}`;

        let meshParticle: MeshParticleSystem;
        let meshParticleConfig = <IParticleSystemConfig>(<any>node).meshParticle;
        if (!meshParticleConfig) {
            meshParticleConfig = node.extras?.meshParticle;
        }
        if (meshParticleConfig) {
            let key = (<any>this)._rootUrl + (<any>this)._fileName + "#" + name + "#" + node.index;
            (<any>this)._babylonScene._blockEntityCollection = this._forAssetContainer;
            meshParticle = new MeshParticleSystem(name, (<any>this)._babylonScene, key);
            (<any>this)._babylonScene._blockEntityCollection = false;

            formatMeshParticle(meshParticleConfig, meshParticle);
            (<any>this)._particles.push(<any>meshParticle);

            node._babylonTransformNode = meshParticle;

            // Trail
            if (meshParticleConfig.trail) {
                let forParticleSys = true;
                let material = (<any>this)._gltf.materials[meshParticleConfig.trail.material];
                let trailPromisre = this._loadMaterialAsync("/materials/" + meshParticleConfig.trail.material, material, meshParticle.psTool.trail.mesh, 0, function(babylonMaterial) {

                    if (!forParticleSys && ExtensionTools.Option.optimizeMaterialWithResTab) {
                        let mesh = <StandardMaterialMesh>meshParticle.psTool.trail.mesh;
                        mesh.materialData = new StandardMaterialData(this._scene);
                        mesh.materialData.fromMaterial(<any>babylonMaterial);
                        (<BABYLON.Mesh>mesh).onDisposeObservable.addOnce(function() {
                            (<StandardMaterialData>mesh.materialData)?.dispose();
                            mesh.materialData = undefined;
                        });
                    }

                    meshParticle.psTool.trail.mesh.material = babylonMaterial;
                    meshParticle.psTool.trail.mesh.alphaIndex = material.extensions.PI_material.renderQueue;
                }, !forParticleSys);
                promises.push(trailPromisre);
            }
        } else {
            (<any>this)._babylonScene._blockEntityCollection = this._forAssetContainer;
            node._babylonTransformNode = new BABYLON.TransformNode(name, (<any>this)._babylonScene);
            (<any>this)._babylonScene._blockEntityCollection = false;
        }
        node._primitiveBabylonMeshes = [];
            // let counter = 0;
            // for (const primitive of primitives) {
            //     let index = counter;
            //     const meshParticleConfig = <IParticleSystemConfig>(<any>node).meshParticle;
            //     if (meshParticleConfig) {
            //         (<any>this)._babylonScene._blockEntityCollection = this._forAssetContainer;
            //         const meshParticle = new MeshParticleSystem(name, (<any>this)._babylonScene);
            //         (<any>this)._babylonScene._blockEntityCollection = false;

            //         formatMeshParticle(meshParticleConfig, meshParticle);
            //         (<any>this)._particles.push(<any>meshParticle);

            //         (<any>this).geometryForMeshParticle = true;
            //         mesh.instanceID = undefined;
            //         let promise = this._loadMeshPrimitiveAsync(`${context}/primitives/${primitive.index}`, name, node, mesh, primitive, (babylonMesh: BABYLON.AbstractMesh) => {
            //             meshParticle.parent = node._babylonTransformNode!;
            //             node._primitiveBabylonMeshes![index] = babylonMesh;
            //             babylonMesh.setEnabled(false);
            //             //#region PI_BEGIN - SHADOW ABOUT
            //             if ((<BABYLON.InstancedMesh>babylonMesh).sourceMesh) {
            //                 babylonMesh = (<BABYLON.InstancedMesh>babylonMesh).sourceMesh;
            //             }
            //             babylonMesh.receiveShadows = !!mesh.receiveShadows;
            //             babylonMesh.castShadows = !!mesh.castShadows || !!mesh.castShadow;
            //             //#endregion
            //         }).then(() => {
            //             let babylonMesh = node._primitiveBabylonMeshes![index];
            //             meshParticle.setSourceMesh(<any>babylonMesh);
            //             babylonMesh.setEnabled(true);
            //         });
            //         promises.push(promise);

            //         // Trail
            //         if (meshParticleConfig.trail) {
            //             let material = (<any>this)._gltf.materials[meshParticleConfig.trail.material];
            //             let trailPromisre = this._loadMaterialAsync("/materials/" + meshParticleConfig.trail.material, material, null, 0, function(babylonMaterial) {
            //                 meshParticle.psTool.trail.mesh.material = babylonMaterial;
            //             });
            //             promises.push(trailPromisre);
            //         }
            //     }
            //     else {
            //         promises.push(this._loadMeshPrimitiveAsync(`${context}/primitives/${primitive.index}`, name, node, mesh, primitive, (babylonMesh: BABYLON.AbstractMesh) => {
            //             babylonMesh.parent = node._babylonTransformNode!;
            //             node._primitiveBabylonMeshes![index] = babylonMesh;
            //             //#region PI_BEGIN - SHADOW ABOUT
            //             if ((<BABYLON.InstancedMesh>babylonMesh).sourceMesh) {
            //                 babylonMesh = (<BABYLON.InstancedMesh>babylonMesh).sourceMesh;
            //             }
            //             babylonMesh.receiveShadows = !!mesh.receiveShadows;
            //             babylonMesh.castShadows = !!mesh.castShadows || !!mesh.castShadow;
            //             //#endregion
            //         }));
            //     }
            // }
            if (node.meshParticle || node.extras?.meshParticle) {
                (<any>this).geometryForMeshParticle = true;
                (<any>mesh).instanceID = undefined;
            }

            if (primitives.length === 1) {
                let primitive = primitives[0];
                promises.push(this._loadMeshPrimitiveAsync(`${context}/primitives/${primitive.index}`, `${name}`, node, mesh, primitive, (babylonMesh) => {
                    node._primitiveBabylonMeshes!.push(babylonMesh);
                    babylonMesh.geometryForMeshParticle = (<any>this).geometryForMeshParticle;

                    if (node.meshParticle || node.extras?.meshParticle) {
                        babylonMesh.setEnabled(false);
                    } else {
                        babylonMesh.parent = node._babylonTransformNode!;
                    }
                    if (node.billboard != undefined) {
                        babylonMesh.billboardMode = 7;
                    }

                    //#region PI_BEGIN - SHADOW ABOUT
                    if ((<BABYLON.InstancedMesh>babylonMesh).sourceMesh) {
                        babylonMesh = (<BABYLON.InstancedMesh>babylonMesh).sourceMesh;
                    }
                    babylonMesh.receiveShadows = !!mesh.receiveShadows;
                    babylonMesh.castShadows = !!mesh.castShadows || !!mesh.castShadow;
                    //#endregion

                }));
            } else {
                for (const primitive of primitives) {
                    promises.push(this._loadMeshPrimitiveAsync(`${context}/primitives/${primitive.index}`, `${name}#${primitive.index}`, node, mesh, primitive, (babylonMesh) => {
                        node._primitiveBabylonMeshes!.push(babylonMesh);
                        babylonMesh.geometryForMeshParticle = (<any>this).geometryForMeshParticle;

                        if (node.meshParticle || node.extras?.meshParticle) {
                            babylonMesh.setEnabled(false);
                        } else {
                            babylonMesh.parent = node._babylonTransformNode!;
                        }
                        if (node.billboard != undefined) {
                            babylonMesh.billboardMode = 7;
                        }

                        //#region PI_BEGIN - SHADOW ABOUT
                        if ((<BABYLON.InstancedMesh>babylonMesh).sourceMesh) {
                            babylonMesh = (<BABYLON.InstancedMesh>babylonMesh).sourceMesh;
                        }
                        babylonMesh.receiveShadows = !!mesh.receiveShadows;
                        babylonMesh.castShadows = !!mesh.castShadows || !!mesh.castShadow;
                        //#endregion
                    }));
                }
            }

        if (node.skin != undefined) {
            const skin = BABYLON.GLTF2.ArrayItem.Get(`${context}/skin`, <any[]>(<any>this)._gltf.skins, node.skin);
            promises.push((<any>this)._loadSkinAsync(`/skins/${skin.index}`, node, skin));
        }

        assign(node._babylonTransformNode);

        this.logClose();

        // let meshParticle: MeshParticleSystem;
        // let meshParticleConfig = <IParticleSystemConfig>(<any>node).meshParticle;
        // if (!meshParticleConfig) {
        //     meshParticleConfig = node.extras?.meshParticle;
        // }
        // if (meshParticleConfig) {
        //     let key = (<any>this)._rootUrl + (<any>this)._fileName + "#" + name + "#" + node.index;
        //     (<any>this)._babylonScene._blockEntityCollection = this._forAssetContainer;
        //     meshParticle = new MeshParticleSystem(name, (<any>this)._babylonScene, key);
        //     (<any>this)._babylonScene._blockEntityCollection = false;

        //     formatMeshParticle(meshParticleConfig, meshParticle);
        //     (<any>this)._particles.push(<any>meshParticle);

        //     meshParticle.parent = node._babylonTransformNode!;

        //     // Trail
        //     if (meshParticleConfig.trail) {
        //         let material = (<any>this)._gltf.materials[meshParticleConfig.trail.material];
        //         let trailPromisre = this._loadMaterialAsync("/materials/" + meshParticleConfig.trail.material, material, null, 0, function(babylonMaterial) {
        //             meshParticle.psTool.trail.mesh.material = babylonMaterial;
        //             meshParticle.psTool.trail.mesh.alphaIndex = material.extensions.PI_material.renderQueue;
        //         });
        //         promises.push(trailPromisre);
        //     }
        // }

        return Promise.all(promises).then(() => {
            node._babylonTransformNode.computeWorldMatrix();

            if (meshParticle) {
                meshParticle.setSourceMesh(<any>node._primitiveBabylonMeshes[0]);
                node._primitiveBabylonMeshes[0]!.setEnabled(true);
                // // 从Meshes列表移除
                // node._primitiveBabylonMeshes = [];
            }

            return node._babylonTransformNode!;
        });
    };

    
    BABYLON.MaterialHelper.PrepareDefinesForAttributes = function(mesh: BABYLON.AbstractMesh, defines: any, useVertexColor: boolean, useBones: boolean, useMorphTargets = false, useVertexAlpha = true): boolean {
        if (!defines._areAttributesDirty && defines._needNormals === defines._normals && defines._needUVs === defines._uvs) {
            return false;
        }

        defines._normals = defines._needNormals;
        defines._uvs = defines._needUVs;

        defines["NORMAL"] = (defines._needNormals && mesh.isVerticesDataPresent(BABYLON.VertexBuffer.NormalKind));

        if (defines._needNormals && mesh.isVerticesDataPresent(BABYLON.VertexBuffer.TangentKind)) {
            defines["TANGENT"] = true;
        }

        if (defines._needUVs) {
            defines["UV1"] = mesh.isVerticesDataPresent(BABYLON.VertexBuffer.UVKind);
            defines["UV2"] = mesh.isVerticesDataPresent(BABYLON.VertexBuffer.UV2Kind);
        } else {
            defines["UV1"] = false;
            defines["UV2"] = false;
        }

        if (useVertexColor) {
            var hasVertexColors = mesh.useVertexColors && mesh.isVerticesDataPresent(BABYLON.VertexBuffer.ColorKind);
            defines["VERTEXCOLOR"] = hasVertexColors;
            defines["VERTEXALPHA"] = mesh.hasVertexAlpha && hasVertexColors && useVertexAlpha;
        }

        // PI_BEGIN - MeshParticleSystem
        // if (mesh.hasThinInstances) {
            defines["PIMESHPS"] = mesh.renderAsMeshParticle;
        // }
        // PI_END

        if (useBones) {
            this.PrepareDefinesForBones(mesh, defines);
        }

        if (useMorphTargets) {
            this.PrepareDefinesForMorphTargets(mesh, defines);
        }

        return true;
    }
}
/**
 * 格式化粒子系统
 * @param config josn描述
 * @param mp 目标粒子系统
 */
export function formatMeshParticle(config: IParticleSystemConfig, mp: MeshParticleSystem) {
    const ps = mp.psTool;

    ps.looping = config.looping == 1;
    ps.duration = config.duration * 1000;
    ps.startDelay = config.startDelay * 1000;
    ps.maxParticles = config.maxParticles;
    ps.prewarm = config.prewarm == 1;
    ps.rateOverTime = config.emission[TEmissionIndexRate];
    ps.simulationSpace = config.simulationSpaceIsWorld ? EMeshParticleSpaceMode.World : EMeshParticleSpaceMode.Local;
    ps.scalingSpace = config.scalingMode;
    ps.renderAlignment = config.renderAlignment;
    ps.renderMode = config.renderMode;
    ps.stretchedLengthScale = config.stretchedLengthScale || 1;
    ps.stretchedVelocityScale = config.stretchedVelocityScale || 0;

    // bursts
    if (config.emission[TEmissionIndexBursts]) {
        const len = config.emission[TEmissionIndexBursts].length;
        for (let i = 0; i < len; i++) {
            ps.bursts[i] = config.emission[TEmissionIndexBursts][i];
        }
    }

    // emitter Shape
    ps.emitterShape = formatShape(config.shape);

    // startLifetime
    parseFloatInterpolation(<FloatInterpolation>ps.startLifetimeInterpolation, config.lifetime, TParamStartLifetime, 1000);

    // startSpeed
    parseFloatInterpolation(<FloatInterpolation>ps.startSpeedInterpolation, config.startSpeed, TParamStartSpeed, 1);

    // startSize
    StartSize.format(config.startSize, ps.startSizeInterpolation);

    // startRotation
    StartRotation.format(config.startRotation, ps.startRotationInterpolation);

    // startColor
    StartColor.format(config.startColor, ps.startColorInterpolation);

    // gravity
    parseFloatInterpolation(<FloatInterpolation>ps.gravityInterpolation.interpolation, config.gravity, TParamGravity, 1.0);

    // velocityOverLifetime
    if (config.velocityOverLifetime) {
        VelocityOverLifetime.format(config.velocityOverLifetime, ps.velocityOverLifetimeInterpolation);
        ps.velocityOverLifetimeInterpolation.isLocalSpace = (config.velocityOverLifetimeIsLocal !== 0);
        ps.enableVelocityOverLifeTime = true;
    }

    if (config.orbtialVelocity) {
        VelocityOverLifetime.format(config.orbtialVelocity, ps.localPositionModifier.orbitalRotateSpeed);
    }
    if (config.orbitalOffset) {
        VelocityOverLifetime.format(config.orbitalOffset, ps.localPositionModifier.orbitalOffset);
    }
    if (config.orbitalRadial) {
        parseFloatInterpolation(<FloatInterpolation>ps.localPositionModifier.radial, config.orbitalRadial, TParamStartSpeed, 1.0);
    }
    if (config.speedModifier) {
        parseFloatInterpolation(<FloatInterpolation>ps.localPositionModifier.speedModifier, config.speedModifier, TParamStartSpeed, 1.0);
    }

    // limitVelocityOverLifetime
    if (config.limitVelocityOverLifetime) {
        parseFloatInterpolation(<FloatInterpolation>ps.limitVelocityOverLifetimeInterpolation.interpolation, config.limitVelocityOverLifetime, TParamLimitVelocityOverLifetime, 1.0);
        ps.limitVelocityOverLifetimeInterpolation.dampen = config.limitVelocityOverLifetimeDampen || DefaultValue.limitVelocityOverLifetimeDampen;
        ps.enableLimitVelocityOverLifeTime = true;
    }

    // forceOverLifetime
    if (config.forceOverLifetime) {
        ForceOverLifetime.format(config.forceOverLifetime, ps.forceOverLifetimeInterpolation);
        ps.forceOverLifetimeInterpolation.isLocalSpace = (config.forceSpaceIsLocal !== 0);
        ps.enableForceOverLifeTime = true;
    }

    // colorOverLifetime
    if (config.colorOverLifetime) {
        Color4Interpolate.format(config.colorOverLifetime, ps.colorOverLifetimeInterpolation);
        ps.enableColorOverLifeTime = true;
    }
    if (config.colorBySpeed) {
        Color4Interpolate.format(config.colorBySpeed[0], ps.colorBySpeedInterpolation);
        ps.colorBySpeedInterpolation.rangeX = config.colorBySpeed[1];
        ps.colorBySpeedInterpolation.rangeY = config.colorBySpeed[2];
        ps.enableColorBySpeed = true;
    }

    // sizeOverLifetime
    if (config.sizeOverLifetime) {
        SizeOverLifetime.format(config.sizeOverLifetime, ps.sizeOverLifetimeInterpolation);
        ps.enableSizeOverLifeTime = true;
    }
    if (config.sizeBySpeed) {
        SizeOverLifetime.format(config.sizeBySpeed[0], ps.sizeBySpeedInterpolation);
        ps.sizeBySpeedInterpolation.rangeX = config.sizeBySpeed[1];
        ps.sizeBySpeedInterpolation.rangeY = config.sizeBySpeed[2];
        ps.enableSizeBySpeed = true;
    }

    // rotationOverLifetime
    if (config.rotationOverLifetime) {
        RotationOverLifetime.format(config.rotationOverLifetime, ps.rotationOverLifetimeInterpolation);
        ps.enableRotationOverLifeTime = true;
    }
    if (config.rotationBySpeed) {
        RotationOverLifetime.format(config.rotationBySpeed[0], ps.rotationBySpeedInterpolation);
        ps.rotationBySpeedInterpolation.rangeX = config.rotationBySpeed[1];
        ps.rotationBySpeedInterpolation.rangeY = config.rotationBySpeed[2];
        ps.enableRotationBySpeed = true;
    }

    if (config.custom1) {
        parseFloatInterpolation(ps.customDataForMainUV.uScale, config.custom1[0], TParamStartLifetime, 1.0);
        parseFloatInterpolation(ps.customDataForMainUV.vScale, config.custom1[1], TParamStartLifetime, 1.0);
        parseFloatInterpolation(ps.customDataForMainUV.uOffset, config.custom1[2], TParamStartLifetime, 1.0);
        parseFloatInterpolation(ps.customDataForMainUV.vOffset, config.custom1[3], TParamStartLifetime, 1.0);
        ps.enableCustomDataForMainUV = true;
    }

    /**
     * 导出时 有 textureSheet 必然对应 导出材质名称 为 SHADER_PI_SHADER_PS
     */
    if (config.textureSheet) {
        const tSheet = ps.textureSheetInterpolation;
        parseFloatInterpolation(tSheet.frameOverTime, config.textureSheet.frameOverTime, TParamTextureSheet, 1.0);
        tSheet.animMode = config.textureSheet.animMode;
        tSheet.customRow = config.textureSheet.customRow;
        tSheet.cycles = config.textureSheet.cycles;
        tSheet.rowMode = config.textureSheet.rowMode;
        parseFloatInterpolation(tSheet.startFrame, config.textureSheet.startFrame, TParamTextureSheet, 1.0);
        tSheet.tilesX = config.textureSheet.tilesX;
        tSheet.tilesY = config.textureSheet.tilesY;
        tSheet.timeMode = config.textureSheet.timeMode;
        ps.enableTextureSheet = true;
    }

    if (config.trail) {
        ps.trail = new TrailModifier(mp.getScene());
        (<any>ps.trail).mode = config.trail.mode;
        ps.trail.ratio = config.trail.ratio;
        parseFloatInterpolation(ps.trail.lifetime, config.trail.lifetime, null, 1.0);
        ps.trail.ribbonCount = config.trail.ribbonCount;
        ps.trail.attachRibbonsToTransfoem = config.trail.attachRTT == 1;
        ps.trail.minimunVertexDistance = config.trail.minDist;
        ps.trail.worldSpace = config.trail.worldSpace == 1;
        ps.trail.dieWithParticle = config.trail.dieWith == 1;
        ps.trail.textureMode = config.trail.texMode;
        ps.trail.sizeAffectsWidth = config.trail.sizeAWidth == 1;
        ps.trail.sizeAffectsLifetime = config.trail.sizeALifetime == 1;
        ps.trail.inheritParticleColor = config.trail.inheritColor == 1;
        parseColor4Gradient(ps.trail.colorOverLifetime.gradient, config.trail.colorOverLife, null);
        parseFloatInterpolation(ps.trail.widthOverTrail, config.trail.widthOverTrail, null, 1.0);
        parseColor4Gradient(ps.trail.colorOverTrail.gradient, config.trail.colorOverTrail, null);
        ps.enableTrail = true;
    }

    if (config.renderPivot) {
        ps.renderPivot.copyFromFloats(...config.renderPivot);
    }
}
/**
 *
 * @param shape 形状发射器json描述
 * @returns
 */
function formatShape(shape: IShape | undefined): IShapeEmitterType {
    if (!shape) {
        return new PointShapeEmitter();
    }
    else {
        let shapeEmitter: IShapeEmitterType;
        switch (shape.type) {
            case TShapeTypeCircle: {
                const temp = new CircleShapeEmitter(shape.radius, shape.radiusThickness);
                temp.arcMode = shape.arc.mode;
                temp.arcValue = shape.arc.value * Math.PI / 180;
                temp.arcSpread = shape.arc.spread;
                temp.arcSpeed = shape.arc.speed;
                shapeEmitter = <IShapeEmitterType>temp;
                break;
            }
            case TShapeTypeCone: {
                const temp = new ConeShapeEmitter(shape.radius, shape.angle * Math.PI / 180);
                temp.radiusRange = shape.radiusThickness;
                temp.height = shape.height;
                temp.heightRange = shape.emitAsVolume ? 1.0 : 0.0;
                temp.arcMode = shape.arc.mode;
                temp.arcValue = shape.arc.value * Math.PI / 180;
                temp.arcSpread = shape.arc.spread;
                temp.arcSpeed = shape.arc.speed;
                shapeEmitter = <IShapeEmitterType>temp;
                break;
            }
            case TShapeTypeHemisphere: {
                const temp = new HemisphereShapeEmitter(shape.radius, shape.radiusThickness);
                temp.arcMode = shape.arc.mode;
                temp.arcValue = shape.arc.value * Math.PI / 180;
                temp.arcSpread = shape.arc.spread;
                temp.arcSpeed = shape.arc.speed;
                shapeEmitter = <IShapeEmitterType>temp;
                break;
            }
            case TShapeTypeEdge: {
                const temp = new EdgeShapeEmitter();
                temp.size = shape.radius;
                temp.arcMode = shape.arc.mode;
                temp.arcValue = shape.arc.value * Math.PI / 180;
                temp.arcSpread = shape.arc.spread;
                temp.arcSpeed = shape.arc.speed;
                shapeEmitter = <IShapeEmitterType>temp;
                break;
            }
            case TShapeTypeSphere: {
                const temp = new SphereShapeEmitter(shape.radius, shape.radiusThickness);
                temp.arcMode = shape.arc.mode;
                temp.arcValue = shape.arc.value * Math.PI / 180;
                temp.arcSpread = shape.arc.spread;
                temp.arcSpeed = shape.arc.speed;
                shapeEmitter = <IShapeEmitterType>temp;
                break;
            }
            case TShapeTypeBox: {
                const temp = new BoxShapeEmitter();
                temp.emitMode = shape.boxEmitMode || EBoxShapeMode.Volume;
                shapeEmitter = <IShapeEmitterType>temp;
                break;
            }
            case TShapeTypeRectangle: {
                const temp = new RectangleShapeEmitter();
                shapeEmitter = <IShapeEmitterType>temp;
                break;
            }
        }

        if (shape.position && shape.rotation && shape.scale) {
            shapeEmitter.position.copyFromFloats(...shape.position);
            shapeEmitter.rotation.copyFromFloats(...shape.rotation);
            shapeEmitter.scaling.copyFromFloats(...shape.scale);
            // Unity Shape 旋转是绕轴 顺时针为正
            shapeEmitter.rotation.scaleInPlace(3.14159265359 / 180);
            BABYLON.Quaternion.FromEulerVectorToRef(shapeEmitter.rotation, BABYLON.TmpVectors.Quaternion[0]);
            BABYLON.Matrix.ComposeToRef(shapeEmitter.scaling, BABYLON.TmpVectors.Quaternion[0], shapeEmitter.position, shapeEmitter.localMatrix);
        }

        shapeEmitter.alignDirection = !!shape.alignDir;

        //
        if (shape.randomize) {
            shapeEmitter.randomizeDirection = shape.randomize[0];
            shapeEmitter.spherizeDirection = shape.randomize[1];
            shapeEmitter.randomizePosition = shape.randomize[2];
        }

        return shapeEmitter;
    }
}

const SHADER_PI_SHADER_PS = 'PI_SHADER_PS';
// CUSTOM_MATERIAL_SHADER_NAMES[SHADER_PI_SHADER_PS] = SHADER_PI_SHADER_PS;
export const SHADER_PI_SHADER_PS_NO_TEXSHEET = 'PI_SHADER_PS_NO_TEXSHEET';
// CUSTOM_MATERIAL_SHADER_NAMES[SHADER_PI_SHADER_PS_NO_TEXSHEET] = SHADER_PI_SHADER_PS_NO_TEXSHEET;
function _createMaterialPiShaderPS(context: string, babylonScene: BABYLON.Scene, materialName: string | undefined, index: number, materialInfo: any, babylonDrawMode: number): any {

    let name = materialName || "material" + index;
    let mat: BABYLON.StandardMaterial;

    mat = new BABYLON.StandardMaterial(name, babylonScene);

    // mat = <any>new BABYLON.CustomMaterial(name, babylonScene);
    // mat.name = SHADER_PI_SHADER_PS;

    // (<any>mat).Vertex_MainEnd(`
    // vec4 cell = floor(color / 10.0);
    // vMainUV1 = vMainUV1.xy / cell.xy + cell.zw / cell.xy;
    // vColor = color - cell * 10.0;
    // `);

    // (<any>mat).Fragment_Custom_Alpha(`
    // alpha *= vColor.a;
    // `);

    if (materialInfo.useLightmapAsShadowmap) {
        (<BABYLON.StandardMaterial>mat).useLightmapAsShadowmap = true;
    }
    if (materialInfo.useAlphaFromDiffuseTexture) {
        (<BABYLON.StandardMaterial>mat).useAlphaFromDiffuseTexture = true;
    }
    if (materialInfo.alphaMode) {
        (<BABYLON.StandardMaterial>mat).alphaMode = materialInfo.alphaMode;
    }

    mat.ambientColor = new BABYLON.Color3(1, 1, 1);

    mat.sideOrientation = babylonScene.useRightHandedSystem ? BABYLON.Material.CounterClockWiseSideOrientation : BABYLON.Material.ClockWiseSideOrientation;

    return mat;
}

function _createMaterialPiShaderPSNoTexSheet(context: string, babylonScene: BABYLON.Scene, materialName: string | undefined, index: number, materialInfo: any, babylonDrawMode: number): any {

    let name = materialName || "material" + index;
    let mat: BABYLON.StandardMaterial;

    mat = new BABYLON.StandardMaterial(name, babylonScene);

    // mat = new BABYLON.CustomMaterial(name, babylonScene);
    // (<any>mat).name = SHADER_PI_SHADER_PS_NO_TEXSHEET;

    // mat.Vertex_After_WorldPosComputed(`
    // vec4 cell = floor(color / 10.0);
    // `);

    // mat.Vertex_MainEnd(`
    // vColor = color - cell * 10.0;
    // `);

    // mat.Fragment_Custom_Alpha(`
    // alpha *= vColor.a;
    // `);

    if (materialInfo.useLightmapAsShadowmap) {
        (<any>mat).useLightmapAsShadowmap = true;
    }
    if (materialInfo.useAlphaFromDiffuseTexture) {
        (<any>mat).useAlphaFromDiffuseTexture = true;
    }
    if (materialInfo.alphaMode) {
        (<any>mat).alphaMode = materialInfo.alphaMode;
    }

    (<any>mat).ambientColor = new BABYLON.Color3(1, 1, 1);

    (<any>mat).sideOrientation = babylonScene.useRightHandedSystem ? BABYLON.Material.CounterClockWiseSideOrientation : BABYLON.Material.ClockWiseSideOrientation;

    return mat;
}

function _loadMaterialPropertiesAsyncPiShaderPS(context: string, material: any, mat: BABYLON.StandardMaterial, _loader: any): any {
    let m = material;

    var promises = new Array();

    mat.backFaceCulling = false;

    if (m.diffuseColor) {
        if (m.disableLighting !== undefined) {
            mat.disableLighting = m.disableLighting;
            (<BABYLON.StandardMaterial>mat).diffuseColor    = new BABYLON.Color3(m.diffuseColor[0], m.diffuseColor[1], m.diffuseColor[2]);
            mat.emissiveColor   = new BABYLON.Color3(m.diffuseColor[0], m.diffuseColor[1], m.diffuseColor[2]);
            // 不管是否受光照影响 - 默认都不需要高光
            (<BABYLON.StandardMaterial>mat).specularColor   = BABYLON.Color3.Black();
            // 不受光照时设置自发光颜色 - 白色 - 即表现为纹理自身颜色
            if (m.disableLighting) {
                (<BABYLON.StandardMaterial>mat).ambientColor    = BABYLON.Color3.Black();
            }
        // 原来的处理 - 没有文档说明为什么这样做
        } else {
            (<BABYLON.StandardMaterial>mat).diffuseColor    = new BABYLON.Color3(m.diffuseColor[0], m.diffuseColor[1], m.diffuseColor[2]);
            mat.emissiveColor   = new BABYLON.Color3(m.diffuseColor[0], m.diffuseColor[1], m.diffuseColor[2]);
            // 不管是否受光照影响 - 默认都不需要高光
            (<BABYLON.StandardMaterial>mat).specularColor   = BABYLON.Color3.Black();
            // 不管是否受光照影响 - 默认都不需要高光
            (<BABYLON.StandardMaterial>mat).ambientColor   = BABYLON.Color3.White();
            // (<BABYLON.StandardMaterial>mat).ambientColor = new BABYLON.Color3(m.diffuseColor[0], m.diffuseColor[1], m.diffuseColor[2]);
        }
    }
    if (m.diffuseTexture) {
        promises.push(_loader.loadTextureInfoAsync(context + "/PI_material/diffuseTexture", m.diffuseTexture, function(texture: BABYLON.Texture) {
            (<BABYLON.StandardMaterial>mat).diffuseTexture = texture;

            if (m.type.indexOf("Particle") >= 0) {
                (<BABYLON.Texture>(<BABYLON.StandardMaterial>mat).diffuseTexture).hasAlpha = true;
            }
            if (m.diffuseTexture.hasAlpha) {
                (<BABYLON.Texture>(<BABYLON.StandardMaterial>mat).diffuseTexture).hasAlpha = true;
            }
            if (m.diffuseTexture.offset) {
                (<any>texture).uOffset = m.diffuseTexture.offset[0];
                (<any>texture).vOffset = m.diffuseTexture.offset[1];
            }
            if (m.diffuseTexture.scale) {
                (<any>texture).uScale = m.diffuseTexture.scale[0];
                (<any>texture).vScale = m.diffuseTexture.scale[1];
            }
            if (m.diffusePreAlpha) {
                mat.diffuseTexturePremultiplyAlpha = true;
            }
        }));
    }
    if (m.bumpTexture) {
        promises.push(_loader.loadTextureInfoAsync(context + "/PI_material/bumpTexture", m.bumpTexture, function(texture) {
            (<BABYLON.StandardMaterial>mat).bumpTexture = texture;

            if (m.bumpTexture.offset) {
                (<any>texture).uOffset = m.bumpTexture.offset[0];
                (<any>texture).vOffset = m.bumpTexture.offset[1];
            }
            if (m.bumpTexture.scale) {
                (<any>texture).uScale = m.bumpTexture.scale[0];
                (<any>texture).vScale = m.bumpTexture.scale[1];
            }
            (<BABYLON.Texture>texture).level = material.bumpTexture.level || material.normalScale || 1.0;
        }));
    }
    if (m.opacityTexture) {
        promises.push(_loader.loadTextureInfoAsync(context + "/PI_material/opacityTexture", m.opacityTexture, function(texture) {
            (<BABYLON.StandardMaterial>mat).opacityTexture = texture;

            if (m.opacityTexture.offset) {
                (<any>texture).uOffset = m.opacityTexture.offset[0];
                (<any>texture).vOffset = m.opacityTexture.offset[1];
            }
            if (m.opacityTexture.scale) {
                (<any>texture).uScale = m.opacityTexture.scale[0];
                (<any>texture).vScale = m.opacityTexture.scale[1];
            }
            if (m.opacityApplyRGB) {
                mat.opacityTextureApplyRGB = true;
            }
        }));
    }

    return Promise.all(promises).then(function() { });
}

function _loadMaterialPropertiesAsyncPiShaderPSNew(context: string, material: any, mat: BABYLON.StandardMaterial, _loader: GLTFLoader): any {
    let m = material;

    var promises = new Array();

    mat.backFaceCulling = false;

    if (m.diffuseColor) {
        if (m.disableLighting !== undefined) {
            mat.disableLighting = m.disableLighting;
            (<BABYLON.StandardMaterial>mat).diffuseColor    = new BABYLON.Color3(m.diffuseColor[0], m.diffuseColor[1], m.diffuseColor[2]);
            mat.emissiveColor   = new BABYLON.Color3(m.diffuseColor[0], m.diffuseColor[1], m.diffuseColor[2]);
            // 不管是否受光照影响 - 默认都不需要高光
            (<BABYLON.StandardMaterial>mat).specularColor   = BABYLON.Color3.Black();
            // 不受光照时设置自发光颜色 - 白色 - 即表现为纹理自身颜色
            if (m.disableLighting) {
                (<BABYLON.StandardMaterial>mat).ambientColor    = BABYLON.Color3.Black();
            }
        // 原来的处理 - 没有文档说明为什么这样做
        } else {
            (<BABYLON.StandardMaterial>mat).diffuseColor    = new BABYLON.Color3(m.diffuseColor[0], m.diffuseColor[1], m.diffuseColor[2]);
            mat.emissiveColor   = new BABYLON.Color3(m.diffuseColor[0], m.diffuseColor[1], m.diffuseColor[2]);
            // 不管是否受光照影响 - 默认都不需要高光
            (<BABYLON.StandardMaterial>mat).specularColor   = BABYLON.Color3.Black();
            // 不管是否受光照影响 - 默认都不需要高光
            (<BABYLON.StandardMaterial>mat).ambientColor   = BABYLON.Color3.White();
            // (<BABYLON.StandardMaterial>mat).ambientColor = new BABYLON.Color3(m.diffuseColor[0], m.diffuseColor[1], m.diffuseColor[2]);
        }
    }
    if (m.diffuseTexture) {
        let texture = <BABYLON.Texture>_loader.loadTextureInfoAsync(context + "/PI_material/diffuseTexture", m.diffuseTexture);
            (<BABYLON.StandardMaterial>mat).diffuseTexture = texture;

            if (m.type.indexOf("Particle") >= 0) {
                (<BABYLON.Texture>(<BABYLON.StandardMaterial>mat).diffuseTexture).hasAlpha = true;
            }
            if (m.diffuseTexture.hasAlpha) {
                (<BABYLON.Texture>(<BABYLON.StandardMaterial>mat).diffuseTexture).hasAlpha = true;
            }
            if (m.diffuseTexture.offset) {
                (<any>texture).uOffset = m.diffuseTexture.offset[0];
                (<any>texture).vOffset = m.diffuseTexture.offset[1];
            }
            if (m.diffuseTexture.scale) {
                (<any>texture).uScale = m.diffuseTexture.scale[0];
                (<any>texture).vScale = m.diffuseTexture.scale[1];
            }
            if (m.diffusePreAlpha) {
                mat.diffuseTexturePremultiplyAlpha = true;
            }
    }
    if (m.bumpTexture) {
        let texture = <BABYLON.Texture>_loader.loadTextureInfoAsync(context + "/PI_material/bumpTexture", m.bumpTexture);
            (<BABYLON.StandardMaterial>mat).bumpTexture = texture;

            if (m.bumpTexture.offset) {
                (<any>texture).uOffset = m.bumpTexture.offset[0];
                (<any>texture).vOffset = m.bumpTexture.offset[1];
            }
            if (m.bumpTexture.scale) {
                (<any>texture).uScale = m.bumpTexture.scale[0];
                (<any>texture).vScale = m.bumpTexture.scale[1];
            }
            (<BABYLON.Texture>texture).level = material.bumpTexture.level || material.normalScale || 1.0;
    }
    if (m.opacityTexture) {
        let texture = <BABYLON.Texture>_loader.loadTextureInfoAsync(context + "/PI_material/opacityTexture", m.opacityTexture);
            (<BABYLON.StandardMaterial>mat).opacityTexture = texture;

            if (m.opacityTexture.offset) {
                (<any>texture).uOffset = m.opacityTexture.offset[0];
                (<any>texture).vOffset = m.opacityTexture.offset[1];
            }
            if (m.opacityTexture.scale) {
                (<any>texture).uScale = m.opacityTexture.scale[0];
                (<any>texture).vScale = m.opacityTexture.scale[1];
            }
            if (m.opacityApplyRGB) {
                mat.opacityTextureApplyRGB = true;
            }
    }

    return Promise.all(promises).then(function() { });
}

BABYLON.GLTF2.Loader.Extensions.PI_material.regist(
    SHADER_PI_SHADER_PS,
    _createMaterialPiShaderPS,
    _loadMaterialPropertiesAsyncPiShaderPS
);
LoadExtraMaterial.loadCalls.set(SHADER_PI_SHADER_PS, _createMaterialPiShaderPS);
LoadExtraMaterial.loadPropertyCalls.set(SHADER_PI_SHADER_PS, _loadMaterialPropertiesAsyncPiShaderPSNew);

BABYLON.GLTF2.Loader.Extensions.PI_material.regist(
    SHADER_PI_SHADER_PS_NO_TEXSHEET,
    _createMaterialPiShaderPSNoTexSheet,
    _loadMaterialPropertiesAsyncPiShaderPS
);
LoadExtraMaterial.loadCalls.set(SHADER_PI_SHADER_PS_NO_TEXSHEET, _createMaterialPiShaderPSNoTexSheet);
LoadExtraMaterial.loadPropertyCalls.set(SHADER_PI_SHADER_PS_NO_TEXSHEET, _loadMaterialPropertiesAsyncPiShaderPSNew);