import { BABYLON } from "../render3d/babylon";
// import * as BABYLON from "../pi_babylon/render3d/babylon";
import { ParticleSystemTool } from "./particle_system_tool";
import { TrailModifier } from "./modifier/trail";
import { IMeshParticleSystem, IMeshParticleSystemBatch, IMeshParticleSystemScene } from "../interface/particlesystem";
import { Scene } from "../render3d/scene";
import { IMeshMaterial } from "../interface/mesh";
import { ExtensionTools } from "../interface/index";

/**
 * 对接 Unity ParticleSystem Mesh Mode
 */
export class MeshParticleSystem extends BABYLON.TransformNode implements IMeshParticleSystem, IMeshMaterial {
    private static Counter = Number.MIN_SAFE_INTEGER;
    public static sqrt3 = Math.sqrt(3);
    public static registCompute = (call: () => void) => {
        ExtensionTools.setPermanentBefore(call);
    }
    public static registRecycle = (call: () => void) => {
        ExtensionTools.setPermanent(call);
    }
    public static unregistCompute = (call: () => void) => {
        ExtensionTools.clearPermanent(call);
    }
    public static unregistRecycle = (call: () => void) => {
        ExtensionTools.clearPermanent(call);
    }
    

    public readonly particleSysId: number;

    public psTool: ParticleSystemTool;

    alphaIndex: number = 3000;
    renderingGroupId: number = 0;

    public _sourceMesh: BABYLON.Mesh;
    // public _instanceMesh: BABYLON.Mesh;
    // public _instancePosition: Float32Array;
    // public _instanceUV: Float32Array;
    // public _instanceColor: Float32Array;
    // public _instanceIndices: Uint16Array;
    // private maxInstance: number = 100;

    public get material() {
        return this._sourceMesh.material;
    }

    private _alwaysSelectAsActive: boolean = true;
    public get alwaysSelectAsActive() {
        return false;
    }
    public set alwaysSelectAsActive(v: boolean) {
        this._alwaysSelectAsActive = v;
        // if (this._sourceMesh) {
        //     this._sourceMesh.alwaysSelectAsActiveMesh = v;
        // }
    }

    public readonly particleSytemBatch: IMeshParticleSystemBatch;
    constructor(name: string, scene: Scene, public readonly particleKey: string) {

        super(name, <Scene>scene);

        this.particleSysId = MeshParticleSystem.Counter++;

        this.particleSytemBatch = scene.getParticleSystemBatche(particleKey);

        if (this.getEngine()._caps.instancedArrays) {
            this.psTool = new ParticleSystemTool(<BABYLON.Scene>scene);
            this.psTool.getParentWorldMatrix = () => {
                return this.parent ? this.parent.computeWorldMatrix() : undefined;
            };

            this.psTool.getCameraRotationMatrixInvert = () => {
                (this._scene).activeCamera.computeWorldMatrix();
                return (this._scene).activeCamera.getViewMatrix().getRotationMatrix().invert();
            };
            this.psTool.getCameraPosition = () => {
                (this._scene).activeCamera.computeWorldMatrix();
                return (this._scene).activeCamera.globalPosition;
            };

            this.psTool.getCameraMatrix = () => {
                (this._scene).activeCamera.computeWorldMatrix();
                return (this._scene).activeCamera.getViewMatrix();
            };

            this.psTool.getWorldMatrix = () => {
                return this.worldMatrixFromCache;
            };
            this.psTool.getLocalMatrix = () => {
                return this._localMatrix;
            };
            this.psTool.getParentGlobalPosition = () => {
                return this.parent ? (<BABYLON.TransformNode>this.parent).absolutePosition : undefined;
            };
            this.psTool.getParentGlobalRotation = () => {
                return this.parent ? (<BABYLON.TransformNode>this.parent).absoluteRotationQuaternion : undefined;
            };
            this.psTool.getParentGlobalScale = () => {
                return this.parent ? (<BABYLON.TransformNode>this.parent).absoluteScaling : undefined;
            };
            this.psTool.getLocalPosition = () => {
                return this.position;
            };
            this.psTool.getLocalRotation = () => {
                return this.rotationQuaternion ? this.rotationQuaternion : BABYLON.Quaternion.FromEulerVector(this.rotation);
            };
            this.psTool.getLocalScale = () => {
                return this.scaling;
            };
        } else {
            this.psTool = new ParticleSystemTool(scene);
            this.psTool.getParentWorldMatrix = () => {
                return this.parent ? this.parent.computeWorldMatrix() : undefined;
            };
            this.psTool.getParentGlobalPosition = () => {
                return this.parent ? (<BABYLON.TransformNode>this.parent).absolutePosition : undefined;
            };
            this.psTool.getParentGlobalRotation = () => {
                return this.parent ? (<BABYLON.TransformNode>this.parent).absoluteRotationQuaternion : undefined;
            };
            this.psTool.getParentGlobalScale = () => {
                return this.parent ? (<BABYLON.TransformNode>this.parent).absoluteScaling : undefined;
            };
            this.psTool.getLocalPosition = () => {
                return this.position;
            };
            this.psTool.getLocalRotation = () => {
                return this.rotationQuaternion ? this.rotationQuaternion : BABYLON.Quaternion.FromEulerVector(this.rotation);
            };
            this.psTool.getLocalScale = () => {
                return this.scaling;
            };

            this.psTool.getCameraRotationMatrixInvert = () => {
                (this._scene).activeCamera.computeWorldMatrix();
                return (this._scene).activeCamera.getViewMatrix().getRotationMatrix().invert();
            };
            this.psTool.getCameraPosition = () => {
                (this._scene).activeCamera.computeWorldMatrix();
                return (this._scene).activeCamera.globalPosition;
            };

            this.psTool.getCameraMatrix = () => {
                (this._scene).activeCamera.computeWorldMatrix();
                return (this._scene).activeCamera.getViewMatrix();
            };

            this.psTool.getWorldMatrix = () => {
                return this.worldMatrixFromCache;
            };

            this.psTool.getLocalMatrix = () => {
                return this._localMatrix;
            };
            // this._instanceMesh = new BABYLON.Mesh(`${name}_disable_instance`, this._scene);
            // this._scene.onBeforeRenderObservable.add(this.alphaIndexForInstanceMesh);
            // console.warn(`Can not use instancedArrays`);
        }

    }

    // private alphaIndexForInstanceMesh = () => {
    //     if (this._sourceMesh) {
    //         this._sourceMesh.setEnabled(false);
    //         this._instanceMesh.alphaIndex = this._sourceMesh.alphaIndex;
    //     }
    // }

    public setSourceMesh(mesh: BABYLON.Mesh) {
        this._sourceMesh = mesh;
        if (mesh) {
            mesh.renderAsMeshParticle = true;
            this.particleSytemBatch.setSourceMesh(<any>mesh);
            mesh.doNotSyncBoundingInfo = true;
            if (!this.psTool) {
                mesh.setEnabled(false);
                return;
            }
            mesh.renderAsMeshParticle = true;
            mesh.useVertexColors = true;
            if (mesh.alphaIndex >= 3000) {
                mesh.hasVertexAlpha = true;
            }
            this.alphaIndex = mesh.alphaIndex;
            mesh.alwaysSelectAsActiveMesh = this._alwaysSelectAsActive;
            mesh.parent = null;
            mesh.freezeWorldMatrix(BABYLON.Matrix.Identity());
            mesh.setEnabled(false);
            // if (this._instanceMesh) {
            //     mesh.setEnabled(false);
            //     this._instanceMesh.alwaysSelectAsActiveMesh = this._alwaysSelectAsActive;
            //     this._instanceMesh.renderAsMeshParticle = true;
            //     this._instanceMesh.useVertexColors = true;
            //     this._instanceMesh.hasVertexAlpha = mesh.hasVertexAlpha;
            //     this._instanceMesh.parent = this;
            //     this._instanceMesh.freezeWorldMatrix(BABYLON.Matrix.Identity());
            //     this._instanceMesh.material = this._sourceMesh.material;
            // }
        }
        this.build();
    }

    public dispose(doNotRecurse?: boolean, disposeMaterialAndTextures?: boolean) {
        if (this.isDisposed()) {
            return;
        }
        this.stop();

        this.particleSytemBatch.disposeParticleSystem(this);

        this.psTool.dispose();

        this.psTool = undefined;

        // if (this._instanceMesh) {
        //     this._instanceMesh.dispose();
        // }
        if (this._sourceMesh) {
            this._sourceMesh.material?.dispose(true, true, true);
            this._sourceMesh.dispose(false, true);
        }

        // this._scene.onBeforeRenderObservable.removeCallback(this.alphaIndexForInstanceMesh);
        super.dispose(doNotRecurse, disposeMaterialAndTextures);
    }

    private _psReady: boolean = false;
    private build() {

        this.psTool.build();
        // if (this._sourceMesh) {
        //     if (this._instanceMesh) {
        //         // 
        //     }
        //     else {
        //         // this._psReady = true;
        //         this._sourceMesh.thinInstanceSetBuffer("matrix", this.psTool.mpMatrixList, 16, false);
        //         this._sourceMesh.thinInstanceSetBuffer("color", this.psTool.mpColorData, 4, false);
        //         // this.source.thinInstanceSetBuffer(ATTRIBUTE_PS_UV_SHEET, this.psTool.mpUVSheetData, 4, false);
        //         // var colorBuffer = new BABYLON.VertexBuffer(this._scene.getEngine(), this.psTool.mpColorData, BABYLON.VertexBuffer.ColorKind, true, false, 4, true);
        //         // this.source.setVerticesBuffer(colorBuffer);
        //     }
        // }
    }

    private _isMPPlaying: boolean = false;
    private _isOnStart: boolean = false;
    public start() {
        if (!this._isMPPlaying && this.psTool) {
            this.psTool.start();

            // MeshParticleSystem.registCompute(this._computeCall);
            // MeshParticleSystem.registRecycle(this._recycleCall);
            // this._scene.onBeforeRenderObservable.add(this._updateCall);

            this._isMPPlaying = true;
            this._isOnStart = true;
            this.particleSytemBatch.active(this);
        }
    }

    public _computeCall = () => {
        if (this.isEnabled()) {
            this.computeParents();
            let deltaTime = this._scene.deltaTime * this.speedRatio || 0;
            if (this._isOnStart) {
                deltaTime = 0;
            }
            this.psTool.checkDelay(deltaTime);
            this.psTool.varCompute(deltaTime);
            this._mpDirty = true;
            this._isOnStart = false;

            this._recycleCall();
        }
    }

    public speedRatio: number = 1;

    private _mpDirty = true;
    public _updateCall = () => {
        if (this._mpDirty) {
            this.computeParents();
            const count = this.psTool.mpUpdate();

            // if (this._instanceMesh) {
            //     if (this._instanceIndices == undefined) {
            //         let positions = this._sourceMesh.geometry.getVerticesData("position");
            //         let vertexCount = positions.length / 3;
            //         this.maxInstance = Math.min(this.maxInstance, Math.floor(65536 / Math.ceil(vertexCount)))
            //         let count = Math.min(this.maxInstance, this.psTool.maxParticles);

            //         this._instancePosition = new Float32Array(vertexCount * 3 * count);
            //         this._instanceColor = new Float32Array(vertexCount * 4 * count);
            //         this._instanceMesh.setVerticesData("position", this._instancePosition, true);
            //         this._instanceMesh.setVerticesData(BABYLON.VertexBuffer.ColorKind, this._instanceColor, true);

            //         let uvs = this._sourceMesh.geometry.getVerticesData(BABYLON.VertexBuffer.UVKind);
            //         this._instanceUV = new Float32Array(vertexCount * 2 * count);
            //         for (let i = 0; i < count; i++) {
            //             for (let j = 0; j < vertexCount; j++) {
            //                 this._instanceUV[(i * vertexCount + j) * 2 + 0] = uvs[j * 2 + 0];
            //                 this._instanceUV[(i * vertexCount + j) * 2 + 1] = uvs[j * 2 + 1];
            //             }
            //         }
            //         this._instanceMesh.setVerticesData(BABYLON.VertexBuffer.UVKind, this._instanceUV, true);

            //         let indices = this._sourceMesh.geometry.getIndices();
            //         let indicesCount = indices.length;
            //         let instanceIndices = new Uint16Array(indicesCount * count);
            //         let indicesTotal = indicesCount * count;
            //         for (let i = 0; i < count; i++) {
            //             for (let j = 0; j < indicesCount; j++) {
            //                 instanceIndices[i * indicesCount + j] = indices[j] + i * vertexCount;
            //             }
            //         }
            //         this._instanceMesh.setIndices(instanceIndices, undefined, false);

            //     }
            //     if (count > 0) {
            //         let matrix = BABYLON.Matrix.Identity();
            //         let position = BABYLON.Vector3.Zero();
            //         let sourcePosition = this._sourceMesh.geometry.getVerticesData("position");
            //         let vertexCount = Math.floor(sourcePosition.length / 3);
            //         for (let i = 0; i < count; i++) {
            //             BABYLON.Matrix.FromArrayToRef(this.psTool.mpMatrixList, i * 16, matrix);
            //             for (let j = 0; j < vertexCount; j++) {
            //                 BABYLON.Vector3.FromArrayToRef(sourcePosition, j * 3, position);
            //                 BABYLON.Vector3.TransformCoordinatesToRef(position, matrix, position);
            //                 this._instancePosition[(i * vertexCount + j) * 3 + 0] = position.x;
            //                 this._instancePosition[(i * vertexCount + j) * 3 + 1] = position.y;
            //                 this._instancePosition[(i * vertexCount + j) * 3 + 2] = position.z;
            //             }
            //         }
            //         this._instanceMesh.updateVerticesData("position", this._instancePosition);

            //         for (let i = 0; i < count; i++) {
            //             let r = this.psTool.mpColorData[i * 4 + 0];
            //             let g = this.psTool.mpColorData[i * 4 + 1];
            //             let b = this.psTool.mpColorData[i * 4 + 2];
            //             let a = this.psTool.mpColorData[i * 4 + 3];
            //             for (let j = 0; j < vertexCount; j++) {
            //                 this._instanceColor[(i * vertexCount + j) * 4 + 0] = r;
            //                 this._instanceColor[(i * vertexCount + j) * 4 + 1] = g;
            //                 this._instanceColor[(i * vertexCount + j) * 4 + 2] = b;
            //                 this._instanceColor[(i * vertexCount + j) * 4 + 3] = a;
            //             }
            //         }
            //         this._instanceMesh.updateVerticesData("color", this._instanceColor);
            //         this._instanceMesh.updateVerticesData("uv", this._instanceUV);
            //         this._instanceMesh.setEnabled(true);
            //         if (this._instanceMesh.subMeshes[0] != undefined) {
            //             this._instanceMesh.subMeshes[0].indexCount = count * this._sourceMesh.geometry.getIndices().length;
            //         }
            //     } else {
            //         this._instanceMesh.setEnabled(false);
            //     }
            // }
            // else
            //     if (this._sourceMesh) {
            //         this._sourceMesh.thinInstanceBufferUpdated("matrix");
            //         // this.source.updateVerticesData(BABYLON.VertexBuffer.ColorKind, this.psTool.mpColorData);
            //         this._sourceMesh.thinInstanceBufferUpdated("color");
            //         // this.source.thinInstanceBufferUpdated(ATTRIBUTE_PS_UV_SHEET);
            //         this._sourceMesh.thinInstanceCount = count;

            //         this._sourceMesh.setEnabled(count > 0 && this.isEnabled());
            //     }

            this._mpDirty = false;
            return count;
        }

        return 0;
    }

    private computeParents() {
        let max = 32;
        let list: BABYLON.TransformNode[] = [this];
        let node = <BABYLON.TransformNode>this;
        while (node.parent != null) {
            list.push(<BABYLON.TransformNode>node.parent);
            node = <any>node.parent;
            if (max <= 0) {
                break;
            }
        }
        let len = list.length;
        for (let i = len - 1; i >= 0; i--) {
            list[i].computeWorldMatrix();
        }
    }

    private _recycleCall = () => {
        this.psTool.recycle();
    }

    public stop() {
        if (this._isMPPlaying && this.psTool) {
            this.psTool.stop();

            // MeshParticleSystem.unregistCompute(this._computeCall);
            // MeshParticleSystem.unregistRecycle(this._recycleCall);
            // this._scene.onBeforeRenderObservable.removeCallback(this._updateCall);

            this._isMPPlaying = false;
            this.particleSytemBatch.remove(this);
        }
    }

    public reset() {
        this.stop();
    }

    private _particleEnable: boolean = false;
    _syncParentEnabledState() {
        super._syncParentEnabledState();
        if (this._psReady && this._sourceMesh?.isReady()) {
        // if (this._psReady) {
            let flag = (<any>this)._isEnabled && (<any>this)._isParentEnabled && this.isDisposed() == false;
            if (flag) {
                if (this._particleEnable == false) {
                    this.start();
                }
            } else {
                if (this._particleEnable == true) {
                    this.stop();
                }
            }
            this._particleEnable = flag;
        }
    }
}