import { BABYLON } from "../render3d/babylon";
import { IMeshParticleSystem } from "./particle_base";
import { IBakedKeyframes } from "../interface/index";

export class MeshParticleSystemBaked extends BABYLON.Mesh implements IMeshParticleSystem {
    _sourceMesh: BABYLON.Mesh;
    set alwaysSelectAsActive(v: boolean) {
        this.alwaysSelectAsActiveMesh = v;
    }
    get alwaysSelectAsActive() {
        return this.alwaysSelectAsActiveMesh;
    }
    setSourceMesh(mesh: BABYLON.Mesh): void {
        // throw new Error("Method not implemented.");
        this._sourceMesh = this;
    }
    start(): void {
        this._runningTime = 0;
        this._scene.onAfterAnimationsObservable.add(this._updateBakedFrame);
    }
    stop(): void {
        this._particleEnable = false;
        this._scene.onAfterAnimationsObservable.removeCallback(this._updateBakedFrame);
    }
    reset(): void {
        this._runningTime = 0;
    }
    _updateCall: () => void;
    get maxLifetimeOverage() {
        return this.bakedFrames.duration - this._runningTime;
    }
    activeParticles: number = 1;

    constructor(name: string, scene: BABYLON.Scene, bakedFrames: IBakedKeyframes, looping: boolean) {
        super(name, scene);
        this.bakedFrames = bakedFrames;
        this.looping = looping;
        this._sourceMesh = this;
    }

    public readonly bakedFrames: IBakedKeyframes;
    public readonly looping: boolean;
    public speed: number = 1;
    private _runningTime: number = 0;
    private _updateBakedFrame = () => {
        if (this._particleEnable == false) return;

        if (this._runningTime > this.bakedFrames.duration) {
            if (this.looping == false) {
                this.stop();
                return;
            }
        }
        this._runningTime += this._scene.deltaTime * this.speed;
        let frame = this._runningTime % this.bakedFrames.duration;
        frame = Math.floor(frame / this.bakedFrames.duration * this.bakedFrames.keyframes.length);
        let info = this.bakedFrames.keyframes[frame];
        if (this.subMeshes && this.subMeshes[0]) {
            this.subMeshes[0].indexStart = info.data[0].start;
            this.subMeshes[0].indexCount = info.data[0].count;
        }
    }

    private _particleEnable: boolean = false;
    _syncParentEnabledState() {
        super._syncParentEnabledState();
        if (this._sourceMesh?.isReady()) {
        // if (this._psReady) {
            let flag = (<any>this)._isEnabled && (<any>this)._isParentEnabled && this.isDisposed() == false;
            if (flag) {
                if (this._particleEnable == false) {
                    this.start();
                }
            } else {
                if (this._particleEnable == true) {
                    this.stop();
                }
            }
            this._particleEnable = flag;
        }
    }
}