import { BABYLON } from "../render3d/babylon";
// import * as BABYLON from "../pi_babylon/render3d/babylon";
import { ParticleSystemTool } from "./particle_system_tool";
import { TrailModifier } from "./modifier/trail";
import { MeshParticleSystem } from "./mesh_particle_system";
import { ExtensionTools } from "../interface/index";

/**
 * 对接 Unity ParticleSystem Mesh Mode
 */
export class MeshParticleSystemGeometry extends BABYLON.Mesh {
    protected baseVertexGeometry: BABYLON.Geometry;
    public static sqrt3 = Math.sqrt(3);
    public static registCompute = (call: () => void) => {
        ExtensionTools.setPermanentBefore(call);
    }
    public static registRecycle = (call: () => void) => {
        ExtensionTools.setPermanent(call);
    }
    public static unregistCompute = (call: () => void) => {
        ExtensionTools.clearPermanent(call);
    }
    public static unregistRecycle = (call: () => void) => {
        ExtensionTools.clearPermanent(call);
    }

    public psTool: ParticleSystemTool;


    // protected _alwaysSelectAsActive: boolean = true;
    // public get alwaysSelectAsActive() {
    //     return false;
    // }
    // public set alwaysSelectAsActive(v: boolean) {
    //     this._alwaysSelectAsActive = v;
    //     if (this.source) {
    //         this.source.alwaysSelectAsActiveMesh = v;
    //     }
    // }

    constructor(name: string, scene?: BABYLON.Scene, isPure?: boolean) {

        super(name, scene);

        if (this._scene.getEngine().getCaps().instancedArrays) {
            this.psTool = new ParticleSystemTool(scene);
            this.psTool.getParentWorldMatrix = () => {
                return undefined; // this.parent ? this.parent.computeWorldMatrix() : undefined;
            };

            this.psTool.getCameraRotationMatrixInvert = () => {
                (this._scene).activeCamera.computeWorldMatrix();
                return (this._scene).activeCamera.getViewMatrix().getRotationMatrix().invert();
            };
            this.psTool.getCameraPosition = () => {
                (this._scene).activeCamera.computeWorldMatrix();
                return (this._scene).activeCamera.globalPosition;
            };

            this.psTool.getCameraMatrix = () => {
                (this._scene).activeCamera.computeWorldMatrix();
                return (this._scene).activeCamera.getViewMatrix();
            };

            this.psTool.getWorldMatrix = () => {
                return BABYLON.Matrix.Identity(); this.worldMatrixFromCache;
            };

            this.psTool.getLocalMatrix = () => {
                return BABYLON.Matrix.Identity(); // this._localMatrix;
            };
        } else {
            console.error(`Can not use instancedArrays`);
        }
    }

    protected particleMesh: BABYLON.Mesh;
    protected particlePosition: Float32Array;
    protected particleColor: Float32Array;
    protected particleIndices: Uint16Array;
    protected particleUV: Float32Array;

    public setSourceMesh(mesh: BABYLON.Mesh) {
        this.particleMesh = mesh;
        if (mesh) {
            // mesh.doNotSyncBoundingInfo = true;
            // if (!this.psTool) {
            //     mesh.setEnabled(false);
            //     return;
            // }
            mesh.renderAsMeshParticle = true;
            mesh.useVertexColors = true;
            if (mesh.alphaIndex >= 3000) {
                mesh.hasVertexAlpha = true;
            }
            // mesh.alwaysSelectAsActiveMesh = this._alwaysSelectAsActive;
            mesh.parent = this;
            mesh.setEnabled(false);
            mesh.freezeWorldMatrix(BABYLON.Matrix.Identity());
        }
        this.renderAsMeshParticle = true;
        this.baseVertexGeometry = mesh.geometry;
        this.material = mesh.material;
        this.alphaIndex = mesh.alphaIndex;
        this.doNotSyncBoundingInfo = true;
        this.useVertexColors = mesh.useVertexColors;
        this.hasVertexAlpha = mesh.hasVertexAlpha;
        this.alwaysSelectAsActiveMesh = true;
        this._initGeometry(mesh);
        this.build();
        this.onBeforeRenderObservable.addOnce(() => {
            this.setEnabled(false);
        });
        mesh.material = null;
        Promise.resolve().then(() => { mesh.dispose(); });
    }

    public dispose(doNotRecurse?: boolean, disposeMaterialAndTextures?: boolean) {
        if (this.isDisposed()) {
            return;
        }
        this.stop();

        this.psTool?.dispose();

        this.psTool = undefined;

        // if (this.source) {
        //     this.source.material?.dispose(true, true, true);
        //     this.source.dispose(false, true);
        // }

        super.dispose(doNotRecurse, disposeMaterialAndTextures);
    }

    protected _psReady: boolean = false;
    protected build() {

        this.psTool.build();
        // if (this.source) {
        //     // this._psReady = true;
        //     this.source.thinInstanceSetBuffer("matrix", this.psTool.mpMatrixList, 16, false);
        //     this.source.thinInstanceSetBuffer("color", this.psTool.mpColorData, 4, false);
        //     // this.source.thinInstanceSetBuffer(ATTRIBUTE_PS_UV_SHEET, this.psTool.mpUVSheetData, 4, false);
        //     // var colorBuffer = new BABYLON.VertexBuffer(this._scene.getEngine(), this.psTool.mpColorData, BABYLON.VertexBuffer.ColorKind, true, false, 4, true);
        //     // this.source.setVerticesBuffer(colorBuffer);
        // }
    }

    protected _isMPPlaying: boolean = false;
    public start() {
        if (!this._isMPPlaying && this.psTool) {
            this.psTool.start();

            MeshParticleSystem.registCompute(this._computeCall);
            // MeshParticleSystem.registRecycle(this._recycleCall);
            this._scene.onBeforeRenderObservable.add(this._updateCall);

            this._isMPPlaying = true;
        }
    }

    protected _computeCall = () => {
        if (this.geometry) {
            this.computeParents();
            this.psTool.varCompute(this._scene.deltaTime * this.speedRatio || 0);
            this._mpDirty = true;

            this._recycleCall();
        }
    }

    public speedRatio: number = 1;

    protected _mpDirty = true;
    protected _updateCall = () => {
        if (this._mpDirty && this.geometry) {
            this.computeParents();
            const count = this.psTool.mpUpdate();

            this._buildGeometry(this.psTool.mpMatrixList, this.psTool.mpColorData, count, this.geometry);

            // if (this.source) {
            //     this.source.thinInstanceBufferUpdated("matrix");
            //     // this.source.updateVerticesData(BABYLON.VertexBuffer.ColorKind, this.psTool.mpColorData);
            //     this.source.thinInstanceBufferUpdated("color");
            //     // this.source.thinInstanceBufferUpdated(ATTRIBUTE_PS_UV_SHEET);
            //     this.source.thinInstanceCount = count;

            //     this.source.setEnabled(count > 0 && this.isEnabled());
            // }

            this._mpDirty = false;
        }
    }

    protected computeParents() {
        let max = 32;
        let list: BABYLON.TransformNode[] = [this];
        let node = <BABYLON.TransformNode>this;
        while (node.parent != null) {
            list.push(<BABYLON.TransformNode>node.parent);
            node = <any>node.parent;
            if (max <= 0) {
                break;
            }
        }
        let len = list.length;
        for (let i = len - 1; i >= 0; i--) {
            list[i].computeWorldMatrix();
        }
    }

    protected _recycleCall = () => {
        this.psTool.recycle();
    }

    public stop() {
        if (this._isMPPlaying && this.psTool) {
            this.psTool.stop();

            MeshParticleSystem.unregistCompute(this._computeCall);
            // MeshParticleSystem.unregistRecycle(this._recycleCall);
            this._scene.onBeforeRenderObservable.removeCallback(this._updateCall);

            this._isMPPlaying = false;
        }
    }

    public reset() {
        this.stop();
    }

    // protected _particleEnable: boolean = false;
    // _syncParentEnabledState() {
    //     super._syncParentEnabledState();
    //     if (this._psReady && this.particleMesh?.isReady()) {
    //         let flag = (<any>this)._isEnabled && (<any>this)._isParentEnabled && this.isDisposed() == false;
    //         if (flag) {
    //             if (this._particleEnable == false) {
    //                 this.start();
    //             }
    //         } else {
    //             if (this._particleEnable == true) {
    //                 this.stop();
    //             }
    //         }
    //         this._particleEnable = flag;
    //     }
    // }

    public static SupportVertexKinds = [BABYLON.VertexBuffer.PositionKind, BABYLON.VertexBuffer.ColorKind, BABYLON.VertexBuffer.UVKind];
    protected _baseMeshPosition: BABYLON.FloatArray;
    // protected _baseMeshHasColor: BABYLON.FloatArray ;
    protected _baseMeshHasUV: BABYLON.FloatArray;
    protected _baseMeshIndices: BABYLON.IndicesArray;
    protected _particleCount: number = 0;
    protected _initGeometry(particleMesh: BABYLON.Mesh) {
        this._particleCount = Math.min(this.psTool.maxParticles, 32);
        this._baseMeshPosition = particleMesh.geometry.getVerticesData(BABYLON.VertexBuffer.PositionKind);
        let pointCount = Math.floor(this._baseMeshPosition.length / 3);
        // this._baseMeshHasColor      = particleMesh.geometry.getVerticesData(BABYLON.VertexBuffer.ColorKind);
        this._baseMeshHasUV = particleMesh.geometry.getVerticesData(BABYLON.VertexBuffer.UVKind);
        this._baseMeshIndices = particleMesh.geometry.getIndices();
        let geometry = new BABYLON.Geometry(this.name, this._scene, undefined, true);
        this.particlePosition = new Float32Array(this._baseMeshPosition.length * this.psTool.maxParticles);
        this.particleColor = new Float32Array(pointCount * 4 * this.psTool.maxParticles);
        this.particleUV = new Float32Array(this._baseMeshHasUV.length * this.psTool.maxParticles);
        this.particleIndices = new Uint16Array(this._baseMeshIndices.length * this.psTool.maxParticles);
        let idx = 0; let idx2 = 0;
        for (let i = 0; i < this.psTool.maxParticles; i++) {
            let indicesIdx = pointCount * i;
            this._baseMeshIndices.forEach(v => {
                this.particleIndices[idx++] = v + indicesIdx;
            });

            // this._baseMeshHasUV.forEach((v) => {
            //     this.particleUV[idx2++] = v;
            // });
        }
        geometry.setVerticesData(BABYLON.VertexBuffer.PositionKind, this.particlePosition, true);
        geometry.setVerticesData(BABYLON.VertexBuffer.ColorKind, this.particleColor, true);
        geometry.setVerticesData(BABYLON.VertexBuffer.UVKind, this.particleUV, true);
        geometry.setIndices(this.particleIndices);
        geometry.applyToMesh(this);
        let box = this.getBoundingInfo();
        box.reConstruct(particleMesh.getBoundingInfo().minimum.clone(), particleMesh.getBoundingInfo().maximum.clone());
    }
    protected _buildGeometry(matrixList: Float32Array, colors: Float32Array, particleCount: number, geometry: BABYLON.Geometry) {
        let matrix = BABYLON.Matrix.Identity();
        let position = BABYLON.Vector3.Zero();
        for (let i = 0; i < particleCount; i++) {
            BABYLON.Matrix.FromArrayToRef(matrixList, i * 16, matrix);
            let r = colors[i * 4 + 0]; let g = colors[i * 4 + 1]; let b = colors[i * 4 + 2]; let a = colors[i * 4 + 3];
            let us = Math.floor(r * 0.1) / 1000; let vs = Math.floor(g * 0.1) / 1000; let uo = Math.floor(b * 0.1) / 1000; let vo = Math.floor(a * 0.1) / 1000;
            // r -= us * 10000; g -= vs * 10000; b -= uo * 10000; a -= vo * 10000;

            let pointCount = Math.floor(this._baseMeshPosition.length / 3);
            for (let j = 0; j < pointCount; j++) {
                let pointIndex = (i * pointCount + j);
                BABYLON.Vector3.TransformCoordinatesFromFloatsToRef(this._baseMeshPosition[j * 3 + 0], this._baseMeshPosition[j * 3 + 1], this._baseMeshPosition[j * 3 + 2], matrix, position);
                this.particlePosition[pointIndex * 3 + 0] = position.x;
                this.particlePosition[pointIndex * 3 + 1] = position.y;
                this.particlePosition[pointIndex * 3 + 2] = position.z;

                this.particleColor[pointIndex * 4 + 0] = r;
                this.particleColor[pointIndex * 4 + 1] = g;
                this.particleColor[pointIndex * 4 + 2] = b;
                this.particleColor[pointIndex * 4 + 3] = a;

                let u = this._baseMeshHasUV[j * 2 + 0]; let v = this._baseMeshHasUV[j * 2 + 1];
                this.particleUV[pointIndex * 2 + 0] = u * us + uo;
                this.particleUV[pointIndex * 2 + 1] = v * vs + vo;
            }
            geometry.updateVerticesData(BABYLON.VertexBuffer.PositionKind, this.particlePosition);
            geometry.updateVerticesData(BABYLON.VertexBuffer.ColorKind, this.particleColor);
            geometry.updateVerticesData(BABYLON.VertexBuffer.UVKind, this.particleUV);
        }
        geometry.updateIndices(this.particleIndices.slice(0, particleCount * this._baseMeshIndices.length));
    }
}