import { BABYLON } from "../../render3d/babylon";
// import * as BABYLON from "../../pi_babylon/render3d/babylon";
import { EInterpolationCurveMode } from "../iparticle_system_config";
import { Color4Gradient, FloatInterpolation, parseFloatInterpolation } from "../interpolation";
import { IParticle } from "../particle_base";
import { ColorOverLifetime } from "./color_over_lifetime";
import { TrailGeometryModifier } from "./trial_geometry";
import { MathDataPool } from "../../render3d/extends/math";

export enum ETrailMode {
    Particles = 0
}

export enum ETrailTextureMode {
    Stretch = 0,
    Tiled = 1,
    DistributePerSegment = 2,
    RepeatPerSegment = 3
}

export class PathRecord {
    private list: [number, number, number][] = [];
    public record() {

    }
}

export class TrailModifier {
    private static tempOne: BABYLON.Vector3 = new BABYLON.Vector3(1, 1, 1);
    private static tempPoint1: BABYLON.Vector3 = new BABYLON.Vector3();
    private static tempPoint2: BABYLON.Vector3 = new BABYLON.Vector3();
    private static tempPoint: BABYLON.Vector3 = new BABYLON.Vector3();
    private static tempDirection: BABYLON.Vector3 = new BABYLON.Vector3();
    private static tempNormal: BABYLON.Vector3 = new BABYLON.Vector3();
    private static tempExtend: BABYLON.Vector3 = new BABYLON.Vector3();
    private static tempView: BABYLON.Vector3 = new BABYLON.Vector3();
    private static tempColor4A: [number, number, number, number] = [1, 1, 1, 1];
    private static tempColor4B: [number, number, number, number] = [1, 1, 1, 1];
    public static counter = 0;
    public static _Scale = 1 / Math.sqrt(3);

    private _mode: ETrailMode = ETrailMode.Particles;
    /**
     *
     *  * 渲染模式
     *      * Particles
     *      * 渲染为粒子的路径线
     */
    public set mode(value: ETrailMode) {
        this._mode = value;
        this.updateModify();
    }
    public get mode() {
        return this._mode;
    }

    /**
     * 可生成路径轨迹的粒子的百分比
     */
    public ratio: number = 1;

    /**
     * 使用粒子多少生命百分比内的路径点
     */
    public lifetime: FloatInterpolation = new FloatInterpolation();

    /**
     * ribbon 模式 间隔多少个粒子进行连接
     */
    public ribbonCount: number = 1;

    /**
     * ribbon 末尾是否连接到粒子系统的位置
     */
    public attachRibbonsToTransfoem: boolean = false;

    /**
     * 新路径点与作为轨迹用的最后一个点之间距离超过多少，才将新作为轨迹用的点
     */
    public minimunVertexDistance: number = 1.0;

    private _useWorldSpace: boolean = false;
    /**
     * 生成轨迹点的坐标是否使用其当时的世界坐标
     */
    public set worldSpace(value: boolean) {
        this._useWorldSpace = value;
        this.updateModify();
    }
    public get worldSpace() {
        return this._useWorldSpace;
    }

    /**
     * 拖尾是否在粒子死亡时即死亡
     */
    public dieWithParticle: boolean = true;

    private _textureMode: ETrailTextureMode = ETrailTextureMode.Stretch;
    /**
     * 拖尾的纹理平铺方式
     */
    public set textureMode(value: ETrailTextureMode) {
        this._textureMode = value;
    }
    public get TextureMode() {
        return this._textureMode;
    }

    /**
     * 粒子的Size是否实时影响拖尾宽度
     */
    public sizeAffectsWidth: boolean = false;
    /**
     * 粒子的Size是否影响拖尾的 Lifetime 参数
     */
    public sizeAffectsLifetime: boolean = false;
    /**
     * 粒子的实时颜色是否影响拖尾颜色的整体变化
     */
    public inheritParticleColor: boolean;
    /**
     * 拖尾的颜色随生命增长的整体变化
     */
    public colorOverLifetime: ColorOverLifetime = new ColorOverLifetime(new Color4Gradient());
    /**
     * 拖尾的初始宽度
     */
    public widthOverTrail: FloatInterpolation = new FloatInterpolation();
    /**
     * 拖尾颜色在路径线上的分布情况
     */
    public colorOverTrail: ColorOverLifetime = new ColorOverLifetime(new Color4Gradient());

    private static positions: number[]  ;
    private static colors: number[]     ;
    private static uvs: number[]        ;
    private static indices: number[]    = [];
    private static maxKeyPoint = 400;

    private geometry: BABYLON.Geometry;
    public mesh: BABYLON.Mesh;

    constructor(scene: BABYLON.Scene) {
        if (!TrailModifier.positions) {
            TrailModifier.positions = [];
            TrailModifier.colors = [];
            TrailModifier.uvs = [];
            for (let i = 0; i < TrailModifier.maxKeyPoint; i++) {
                TrailModifier.positions.push(0);
                TrailModifier.positions.push(0);
                TrailModifier.positions.push(0);
                TrailModifier.positions.push(0);
                TrailModifier.positions.push(0);
                TrailModifier.positions.push(0);

                TrailModifier.colors.push(0);
                TrailModifier.colors.push(0);
                TrailModifier.colors.push(0);
                TrailModifier.colors.push(0);
                TrailModifier.colors.push(0);
                TrailModifier.colors.push(0);
                TrailModifier.colors.push(0);
                TrailModifier.colors.push(0);

                TrailModifier.uvs.push(0);
                TrailModifier.uvs.push(0);
                TrailModifier.uvs.push(0);
                TrailModifier.uvs.push(0);
            }
        }
        let id = TrailModifier.counter++;
        let vertexData = new BABYLON.VertexData();
        this.geometry = new BABYLON.Geometry(`Trail${id}`, scene, vertexData, true);
        this.geometry.setVerticesData("position", TrailModifier.positions, true);
        this.geometry.setVerticesData("uv", TrailModifier.uvs, true);
        this.geometry.setVerticesData("color", TrailModifier.colors, true);
        this.geometry.setIndices(TrailModifier.indices);

        let root = scene.metadata.trailRoot;
        if (!root) {
            root = new BABYLON.TransformNode(`Trails`, scene);
            scene.metadata.trailRoot = root;
        }
        this.mesh = new BABYLON.Mesh(`Trail${id}`, scene);
        this.mesh.doNotSyncBoundingInfo = true;
        this.geometry.applyToMesh(this.mesh);
        this.mesh.setEnabled(false);
        this.mesh.parent = root;
        this.mesh.alwaysSelectAsActiveMesh = true;

        parseFloatInterpolation(this.widthOverTrail, [1, 2, 2, 0.5], null);
        parseFloatInterpolation(this.lifetime, [1, 2, 1, 0.5], null);
    }

    private _enabled: boolean = false;
    public set enable(value: boolean) {
        this._enabled = value;
        if (!value) {
            this.mesh.setEnabled(value);
        }
    }

    public dispose() {
        this.mesh.material?.dispose(true, true, true);
        this.mesh.dispose(true, true);
        this.geometry.dispose();
    }

    public recordTrailPath(viewDirection: BABYLON.Vector3, parentMatrix: BABYLON.Matrix, globalPosition: BABYLON.Vector3, localPosition: BABYLON.Vector3, particle: IParticle) {
        if (particle.trialLifetime == 0) {
            return;
        }
        this._recordTrailPath(this, viewDirection, parentMatrix, globalPosition, localPosition, particle);
    }
    public _recordTrailPath: (trail: TrailModifier, viewDirection: BABYLON.Vector3, parentMatrix: BABYLON.Matrix, globalPosition: BABYLON.Vector3, localPosition: BABYLON.Vector3, particle: IParticle) => void = TrailModifier._recordTrailPathWorld;
    private static _recordTrailPathWorld(trail: TrailModifier, viewDirection: BABYLON.Vector3, parentMatrix: BABYLON.Matrix, globalPosition: BABYLON.Vector3, localPosition: BABYLON.Vector3, particle: IParticle) {
        parentMatrix = particle.startWorldMatrix;
        let positionList = particle.globalPositionList;
        let lifeRecordList = particle.lifeRecordList;
        let maxAgeAmount = particle.age;
        let oldLength = positionList.length;
        let lastPop = false;
        if (maxAgeAmount != lifeRecordList[oldLength - 1]) {
            positionList.push(MathDataPool.Vector3(globalPosition.x, globalPosition.y, globalPosition.z));
            lifeRecordList.push(maxAgeAmount);
            if (oldLength > 1) {
                let lastPosition = positionList[oldLength - 2];
                lastPosition.subtractToRef(globalPosition, TrailModifier.tempDirection);
                if (TrailModifier.tempDirection.length() < trail.minimunVertexDistance) {
                    lastPop = true;
                }
            }
        }

        let pointCount = 0;
        let newLength = positionList.length;
        if (newLength >= 2) {
            let last = positionList[0];
            let curr = positionList[1];
            curr.subtractToRef(last, TrailModifier.tempDirection);

            BABYLON.Vector3.TransformNormalToRef(TrailModifier.tempOne, parentMatrix, TrailModifier.tempNormal);
            let tempSize = TrailModifier.tempNormal.length();

            let trailAge = 0;
            let checkAge = Math.min(particle.age, particle.lifetime);
            trail.colorOverLifetime.gradient.interpolate(checkAge / particle.lifetime, TrailModifier.tempColor4A, particle.colorOverLifetimeAmount);
            for (let i = 0; i <= newLength; i++) {
                if (i < newLength) {
                    curr = positionList[i];
                }
                let delta = (particle.age - particle.lifeRecordList[i]);
                if (delta <= particle.trialLifetime) {
                    delta = checkAge - particle.lifeRecordList[i];
                }
                trailAge = Math.max(trailAge, delta);
                if (trail._count < TrailModifier.maxKeyPoint && delta <= particle.trialLifetime) {
                    if (i > 0 && i < newLength - 1) {
                        curr.subtractToRef(last, TrailModifier.tempDirection);
                    }

                    let index = trail._count;
                    let trailAmount = delta / (trailAge);

                    BABYLON.Vector3.CrossToRef(TrailModifier.tempDirection, viewDirection, TrailModifier.tempExtend);
                    const xSquareLength = TrailModifier.tempExtend.lengthSquared();
                    if (xSquareLength === 0) {
                        TrailModifier.tempExtend.y = 1.0;
                    } else {
                        TrailModifier.tempExtend.normalizeFromLength(Math.sqrt(xSquareLength));
                    }
                    let width = particle.trialWidth * 0.5 * tempSize * TrailModifier._Scale;
                    if (trail.sizeAffectsWidth) {
                        width *= particle.scaling.length() * TrailModifier._Scale;
                    }
                    if (trail.widthOverTrail.mode == EInterpolationCurveMode.Curve || trail.widthOverTrail.mode == EInterpolationCurveMode.TwoCurves) {
                        width *= trail.widthOverTrail.interpolate(trailAmount, particle.baseRandom[3]);
                    }
                    TrailModifier.tempExtend.scaleInPlace(width);

                    TrailGeometryModifier.modifyPosition(
                        last.x, last.y, last.z,
                        TrailModifier.tempExtend.x, TrailModifier.tempExtend.y, TrailModifier.tempExtend.z,
                        TrailModifier.positions, index
                    );
                    TrailGeometryModifier.modifyUV(
                        trailAmount,
                        TrailModifier.uvs, index
                    );

                    trail.colorOverTrail.gradient.interpolate(trailAmount, TrailModifier.tempColor4B, particle.colorOverLifetimeAmount);
                    TrailModifier.tempColor4B[0] *= TrailModifier.tempColor4A[0];
                    TrailModifier.tempColor4B[1] *= TrailModifier.tempColor4A[1];
                    TrailModifier.tempColor4B[2] *= TrailModifier.tempColor4A[2];
                    TrailModifier.tempColor4B[3] *= TrailModifier.tempColor4A[3];
                    if (trail.inheritParticleColor) {
                        TrailModifier.tempColor4B[0] *= particle.color.r;
                        TrailModifier.tempColor4B[1] *= particle.color.g;
                        TrailModifier.tempColor4B[2] *= particle.color.b;
                        TrailModifier.tempColor4B[3] *= particle.color.a;
                    }
                    TrailGeometryModifier.modifyColor(
                        TrailModifier.tempColor4B[0], TrailModifier.tempColor4B[1], TrailModifier.tempColor4B[2], TrailModifier.tempColor4B[3],
                        TrailModifier.colors, index
                    );

                    if (pointCount > 0) {
                        TrailGeometryModifier.modifyIndices(
                            TrailModifier.indices, index
                        );
                    }
                    pointCount += 1;
                    trail._count += 1;

                    if (delta < 0) {
                        break;
                    }
                } else {
                    trailAge = 0;
                }
                last = curr;
            }
        }
        if (lastPop) {
            MathDataPool.Recycle(positionList.pop());
            lifeRecordList.pop();
        }
    }
    private _count = 0;
    public build() {
        this.geometry.updateVerticesData("position", TrailModifier.positions, false);
        this.geometry.updateVerticesData("uv", TrailModifier.uvs, false);
        this.geometry.updateVerticesData("color", TrailModifier.colors, false);
        this.geometry.setIndices(TrailModifier.indices);

        this.mesh.setEnabled(this._enabled && this._count >= 2);

        // this.positions.length = 0;
        // this.colors.length = 0;
        // this.uvs.length = 0;
        TrailModifier.indices.length = 0;
        this._count = 0;
    }
    private static _recordTrailPathLocal(trail: TrailModifier, viewDirection: BABYLON.Vector3, parentMatrix: BABYLON.Matrix, globalPosition: BABYLON.Vector3, localPosition: BABYLON.Vector3, particle: IParticle) {
        parentMatrix = particle.startWorldMatrix;
        let positionList = particle.localPositionList;
        let lifeRecordList = particle.lifeRecordList;
        let maxAgeAmount = particle.age;
        let oldLength = positionList.length;
        let lastPop = false;
        if (maxAgeAmount != lifeRecordList[oldLength - 1]) {
            positionList.push(MathDataPool.Vector3(localPosition.x, localPosition.y, localPosition.z));
            lifeRecordList.push(maxAgeAmount);
            if (oldLength > 1) {
                let lastPosition = positionList[oldLength - 2];
                lastPosition.subtractToRef(localPosition, TrailModifier.tempDirection);
                if (TrailModifier.tempDirection.length() < trail.minimunVertexDistance) {
                    lastPop = true;
                }
            }
        }


        let pointCount = 0;
        let newLength = positionList.length;

        if (newLength >= 2) {
            let last = TrailModifier.tempPoint1;
            BABYLON.Vector3.TransformCoordinatesToRef(positionList[0], parentMatrix, last);
            let curr = TrailModifier.tempPoint2;
            BABYLON.Vector3.TransformCoordinatesToRef(positionList[1], parentMatrix, curr);
            curr.subtractToRef(last, TrailModifier.tempDirection);

            BABYLON.Vector3.TransformNormalToRef(TrailModifier.tempOne, parentMatrix, TrailModifier.tempNormal);
            let tempSize = TrailModifier.tempNormal.length();

            let trailAge = 0;
            let checkAge = Math.min(particle.age, particle.lifetime);
            trail.colorOverLifetime.gradient.interpolate(checkAge / particle.lifetime, TrailModifier.tempColor4A, particle.colorOverLifetimeAmount);
            for (let i = 0; i <= newLength; i++) {
                if (i < newLength) {
                    curr = TrailModifier.tempPoint2;
                    BABYLON.Vector3.TransformCoordinatesToRef(positionList[i], parentMatrix, curr);
                }
                let delta = (maxAgeAmount - particle.lifeRecordList[i]);
                if (delta <= particle.trialLifetime) {
                    delta = checkAge - particle.lifeRecordList[i];
                }
                trailAge = Math.max(trailAge, delta);
                if (trail._count < TrailModifier.maxKeyPoint && delta <= particle.trialLifetime) {
                    if (i > 0 && i < newLength - 1) {
                        curr.subtractToRef(last, TrailModifier.tempDirection);
                    }

                    let index = trail._count;
                    let trailAmount = delta / trailAge;

                    BABYLON.Vector3.CrossToRef(TrailModifier.tempDirection, viewDirection, TrailModifier.tempExtend);
                    const xSquareLength = TrailModifier.tempExtend.lengthSquared();
                    if (xSquareLength === 0) {
                        TrailModifier.tempExtend.y = 1.0;
                    } else {
                        TrailModifier.tempExtend.normalizeFromLength(Math.sqrt(xSquareLength));
                    }

                    let width = particle.trialWidth * 0.5 * tempSize * TrailModifier._Scale;
                    if (trail.sizeAffectsWidth) {
                        width *= particle.scaling.length() * TrailModifier._Scale;
                    }
                    if (trail.widthOverTrail.mode == EInterpolationCurveMode.Curve || trail.widthOverTrail.mode == EInterpolationCurveMode.TwoCurves) {
                        width *= trail.widthOverTrail.interpolate(trailAmount, particle.baseRandom[3]);
                    }
                    TrailModifier.tempExtend.scaleInPlace(width);

                    TrailGeometryModifier.modifyPosition(
                        last.x, last.y, last.z,
                        TrailModifier.tempExtend.x, TrailModifier.tempExtend.y, TrailModifier.tempExtend.z,
                        TrailModifier.positions, index
                    );
                    TrailGeometryModifier.modifyUV(
                        trailAmount,
                        TrailModifier.uvs, index
                    );

                    trail.colorOverTrail.gradient.interpolate(trailAmount, TrailModifier.tempColor4B, particle.colorOverLifetimeAmount);
                    TrailModifier.tempColor4B[0] *= TrailModifier.tempColor4A[0];
                    TrailModifier.tempColor4B[1] *= TrailModifier.tempColor4A[1];
                    TrailModifier.tempColor4B[2] *= TrailModifier.tempColor4A[2];
                    TrailModifier.tempColor4B[3] *= TrailModifier.tempColor4A[3];
                    if (trail.inheritParticleColor) {
                        TrailModifier.tempColor4B[0] *= particle.color.r;
                        TrailModifier.tempColor4B[1] *= particle.color.g;
                        TrailModifier.tempColor4B[2] *= particle.color.b;
                        TrailModifier.tempColor4B[3] *= particle.color.a;
                    }
                    TrailGeometryModifier.modifyColor(
                        TrailModifier.tempColor4B[0], TrailModifier.tempColor4B[1], TrailModifier.tempColor4B[2], TrailModifier.tempColor4B[3],
                        TrailModifier.colors, index
                    );

                    if (pointCount > 0) {
                        TrailGeometryModifier.modifyIndices(
                            TrailModifier.indices, index
                        );
                    }
                    pointCount += 1;
                    trail._count += 1;
                    if (delta < 0) {
                        break;
                    }
                } else {
                    trailAge = 0;
                }
                last.copyFrom(curr);
            }
        }
        if (lastPop) {
            MathDataPool.Recycle(positionList.pop());
            lifeRecordList.pop();
        }
    }

    private updateModify() {
        switch (this._mode) {
            case (ETrailMode.Particles): {
                if (this._useWorldSpace) {
                    this._recordTrailPath = TrailModifier._recordTrailPathWorld;
                } else {
                    this._recordTrailPath = TrailModifier._recordTrailPathLocal;
                }
                break;
            }
            default: {
                if (this._useWorldSpace) {
                    this._recordTrailPath = TrailModifier._recordTrailPathWorld;
                } else {
                    this._recordTrailPath = TrailModifier._recordTrailPathLocal;
                }
            }
        }
    }

    private static forRibbonWorldSpace(cameraGlobalPos: BABYLON.Vector3, particles: IParticle[], modifier: TrailModifier) {

    }
    private static forRibbonLocalSpace(cameraGlobalPos: BABYLON.Vector3, particles: IParticle[], modifier: TrailModifier) {

    }

    private static modifyGeometry(ax: number, ay: number, az: number, bx: number, by: number, bz: number, data: ArrayLike<number>, offset: number) {

    }
}