import { MathDataPool } from "../render3d/extends/math";
import { BABYLON } from "../render3d/babylon";
// import * as BABYLON from "../pi_babylon/render3d/babylon";
import { IParticle } from "./particle_base";

export class Particle implements IParticle {
    startWorldMatrix: BABYLON.Matrix;
    startWorldMatrixInvert: BABYLON.Matrix;
    emitWorldMatrix: BABYLON.Matrix;
    startScaling: BABYLON.Vector3;
    startColor: BABYLON.Color4;
    emitRotation: BABYLON.Vector3;
    position: BABYLON.Vector3;
    rotation: BABYLON.Vector3;
    scaling: BABYLON.Vector3;
    velocity: BABYLON.Vector3;
    direction: BABYLON.Vector3;
    readldirection: BABYLON.Vector3;
    directionScale: number = 1;
    directionLength: number = 0;
    color: BABYLON.Color4;
    textureStartFrame: number = 0;
    textureRow: number = 0;
    age: number = Number.MAX_SAFE_INTEGER;
    lifetime: number = 1000;
    trialLifetime: number = 0;
    trialWidth: number = 1;
    uv: [number, number, number, number] = [1, 1, 0, 0];
    lifeRecordList: number[] = [];
    globalPositionList: BABYLON.Vector3[] = [];
    localPositionList: BABYLON.Vector3[] = [];
    globalRecordList: [number, number, number, number][] = [];
    localRecordList: [number, number, number, number][] = [];
    colorOverLifetimeAmount: number = Math.random();
    baseRandom: [number, number, number, number];
    recycle(): void {
        MathDataPool.Recycle(this.startWorldMatrix);
        MathDataPool.Recycle(this.startWorldMatrixInvert);
        MathDataPool.Recycle(this.emitWorldMatrix);
        MathDataPool.Recycle(this.startScaling);
        MathDataPool.Recycle(this.startColor);
        MathDataPool.Recycle(this.emitRotation);
        MathDataPool.Recycle(this.position);
        MathDataPool.Recycle(this.rotation);
        MathDataPool.Recycle(this.scaling);
        MathDataPool.Recycle(this.velocity);
        MathDataPool.Recycle(this.direction);
        MathDataPool.Recycle(this.readldirection);
        MathDataPool.Recycle(this.color);
        this.globalPositionList.forEach(function(item) {
            MathDataPool.Recycle(item);
        });
        this.localPositionList.forEach(function(item) {
            MathDataPool.Recycle(item);
        });
        this.startWorldMatrix       = undefined;
        this.startWorldMatrixInvert = undefined;
        this.emitWorldMatrix        = undefined;
        this.startScaling           = undefined;
        this.startColor             = undefined;
        this.emitRotation           = undefined;
        this.position               = undefined;
        this.rotation               = undefined;
        this.scaling                = undefined;
        this.velocity               = undefined;
        this.direction              = undefined;
        this.readldirection         = undefined;
        this.color                  = undefined;

        this.age = Number.MAX_SAFE_INTEGER;
        this.lifetime = 1000;
        this.trialLifetime = 0;
        this.directionLength = 0;
        this.directionScale = 1;
        this.baseRandom = [Math.random(), Math.random(), Math.random(), Math.random()];
        this.uv[0] = 1, this.uv[1] = 1, this.uv[2] = 0, this.uv[3] = 0;
        this.globalPositionList.length = 0;
        this.localPositionList.length = 0;
        this.lifeRecordList.length = 0;
    }
    reset() {
        this.startWorldMatrix   = MathDataPool.Matrix();
        this.startWorldMatrixInvert = MathDataPool.Matrix();
        this.emitWorldMatrix    = MathDataPool.Matrix();
        this.startScaling       = MathDataPool.Vector3(1, 1, 1);
        this.startColor         = MathDataPool.Color4(1, 1, 1, 1);
        this.emitRotation       = MathDataPool.Vector3(0, 0, 0);
        this.position           = MathDataPool.Vector3(0, 0, 0);
        this.rotation           = MathDataPool.Vector3(0, 0, 0);
        this.scaling            = MathDataPool.Vector3(1, 1, 1);
        this.velocity           = MathDataPool.Vector3(0, 0, 0);
        this.direction          = MathDataPool.Vector3(0, 0, 0);
        this.readldirection     = MathDataPool.Vector3(0, 0, 0);
        this.color              = MathDataPool.Color4(1, 1, 1, 1);

        this.age = Number.MAX_SAFE_INTEGER;
        this.lifetime = 1000;
        this.trialLifetime = 0;
        this.directionLength = 0;
        this.directionScale = 1;
        this.baseRandom = [Math.random(), Math.random(), Math.random(), Math.random()];
        this.uv[0] = 1, this.uv[1] = 1, this.uv[2] = 0, this.uv[3] = 0;
        this.globalPositionList.length = 0;
        this.localPositionList.length = 0;
        this.lifeRecordList.length = 0;
    }
}