
import { BABYLON } from "../render3d/babylon";
import { SceneStruct } from "../scene_struct";
import { ShaderMaterial } from "../render3d/extends/shader_material";

export interface IMixGroundTextureTilloff {
    uScale: number;
    vScale: number;
    uOffset: number;
    vOffset: number;
}

export class GridInfo {
    static TOP = 1;
    static RIGHT = 2;
    static BOTTOM = 4;
    static LEFT = 8;
    static RIGHT_TOP = 16;
    static RIGHT_BOTTOM = 32;
    static LEFT_BOTTOM = 64;
    static LEFT_TOP = 128;
}

export class MixGroungMaterial extends ShaderMaterial {
    static KEY = "MixGroungMaterial";
    static POSITION = "position";
    static UV = "uv";
    static ASSET_TEX0 = "ASSET_TEX0";
    static ASSET_TEX1 = "ASSET_TEX1";
    static ASSET_TEX2 = "ASSET_TEX2";
    static ASSET_TEX3 = "ASSET_TEX3";
    static ASSET_TEX4 = "ASSET_TEX4";
    static ASSET_TILLOFF_0 = "ASSET_TILLOFF_0";
    static ASSET_TILLOFF_1 = "ASSET_TILLOFF_1";
    static ASSET_TILLOFF_2 = "ASSET_TILLOFF_2";
    static ASSET_TILLOFF_3 = "ASSET_TILLOFF_3";
    static ASSET_TILLOFF_4 = "ASSET_TILLOFF_4";
    static MIX_TEX = "CHANNEL_MIX_TEX";
    static MIX_TILLOFF = "MIX_TILLOFF";
    static GRID_TEX = "GRID_TEX";
    static WORLD_VIEW_PROJECTION = "worldViewProjection";
    static VIEW_PROJECTION = "viewProjection";
    static VIEW = "view";
    static WORLD = "world";
    static LIGHT_TEX = "LIGHT_TEX";
    static SCENE_SIZE = "SCENE_SIZE";
    static GRID_PARAM = "GRID_PARAM";
    static LIGHT_PARAM = "LIGHT_PARAM";

    private _asset_tex0: BABYLON.Texture;
    private _asset_tex1: BABYLON.Texture;
    private _asset_tex2: BABYLON.Texture;
    private _asset_tex3: BABYLON.Texture;
    private _asset_tex4: BABYLON.Texture;
    private _mix_tex: BABYLON.Texture;
    private _light_tex: BABYLON.Texture;
    private _grid_tex: BABYLON.Texture;

    private _asset_tilloff_0: BABYLON.Vector4 = new BABYLON.Vector4(0, 0, 1, 1);
    private _asset_tilloff_1: BABYLON.Vector4 = new BABYLON.Vector4(0, 0, 1, 1);
    private _asset_tilloff_2: BABYLON.Vector4 = new BABYLON.Vector4(0, 0, 1, 1);
    private _asset_tilloff_3: BABYLON.Vector4 = new BABYLON.Vector4(0, 0, 1, 1);
    private _asset_tilloff_4: BABYLON.Vector4 = new BABYLON.Vector4(0, 0, 1, 1);
    private _mix_tilloff: BABYLON.Vector4 = new BABYLON.Vector4(0, 0, 1, 1);
    private _scene_param: BABYLON.Vector4 = new BABYLON.Vector4(8, 8, 0, 0.05);// 调整格子间的渐变宽度
    private _grid_param: BABYLON.Vector4 = new BABYLON.Vector4(0.3, 0.05, 0.8, 0.8);// 调整格子间的对比度
    private _light_param: BABYLON.Vector4 = new BABYLON.Vector4(1, 1, 1, 1);// 调整格子间的对比度
    private _scene_width: number = 9;
    private _scene_height: number = 9;
    private _time: number = Date.now();

    public set assetTex0(val: BABYLON.Texture) {
        if (this._asset_tex0) {
            this._asset_tex0.dispose();
        }
        this._asset_tex0 = val;
        this.setTexture(MixGroungMaterial.ASSET_TEX0, this._asset_tex0);
    }
    public get assetTex0(): BABYLON.Texture {
        return this._asset_tex0;
    }

    public set assetTex1(val: BABYLON.Texture) {
        if (this._asset_tex1) {
            this._asset_tex1.dispose();
        }
        this._asset_tex1 = val;
        this.setTexture(MixGroungMaterial.ASSET_TEX1, this._asset_tex1);
    }
    public get assetTex1(): BABYLON.Texture {
        return this._asset_tex1;
    }

    public set assetTex2(val: BABYLON.Texture) {
        if (this._asset_tex2) {
            this._asset_tex2.dispose();
        }
        this._asset_tex2 = val;
        this.setTexture(MixGroungMaterial.ASSET_TEX2, this._asset_tex2);
    }
    public get assetTex2(): BABYLON.Texture {
        return this._asset_tex2;
    }

    public set assetTex3(val: BABYLON.Texture) {
        if (this._asset_tex3) {
            this._asset_tex3.dispose();
        }
        this._asset_tex3 = val;
        this.setTexture(MixGroungMaterial.ASSET_TEX3, this._asset_tex3);
    }
    public get assetTex3(): BABYLON.Texture {
        return this._asset_tex3;
    }

    public set assetTex4(val: BABYLON.Texture) {
        if (this._asset_tex4) {
            this._asset_tex4.dispose();
        }
        this._asset_tex4 = val;
        this.setTexture(MixGroungMaterial.ASSET_TEX4, this._asset_tex4);
    }
    public get assetTex4(): BABYLON.Texture {
        return this._asset_tex4;
    }

    public set mixTex(val: BABYLON.Texture) {
        if (this._mix_tex) {
            this._mix_tex.dispose();
        }
        this._mix_tex = val;
        this.setTexture(MixGroungMaterial.MIX_TEX, this._mix_tex);
    }
    public get mixTex(): BABYLON.Texture {
        return this._mix_tex;
    }

    public set lightTex(val: BABYLON.Texture) {
        if (this._light_tex) {
            this._light_tex.dispose();
        }
        this._light_tex = val;
        this.setTexture(MixGroungMaterial.LIGHT_TEX, this._light_tex);
    }
    public get lightTex(): BABYLON.Texture {
        return this._light_tex;
    }

    public set gridTex(val: BABYLON.Texture) {
        if (this._grid_tex) {
            this._grid_tex.dispose();
        }
        this._grid_tex = val;
        this.setTexture(MixGroungMaterial.GRID_TEX, this._grid_tex);
    }
    public set assetTilloff0(param: IMixGroundTextureTilloff) {
        this._asset_tilloff_0.copyFromFloats(param.uOffset, param.vOffset, param.uScale, param.vScale);
        // this.setVector4(MixGroungMaterial.ASSET_TILLOFF_1, this._asset_tilloff_0);
    }
    public set assetTilloff1(param: IMixGroundTextureTilloff) {
        this._asset_tilloff_1.copyFromFloats(param.uOffset, param.vOffset, param.uScale, param.vScale);
        // this.setVector4(MixGroungMaterial.ASSET_TILLOFF_1, this._asset_tilloff_1);
    }
    public set assetTilloff2(param: IMixGroundTextureTilloff) {
        this._asset_tilloff_2.copyFromFloats(param.uOffset, param.vOffset, param.uScale, param.vScale);
        // this.setVector4(MixGroungMaterial.ASSET_TILLOFF_2, this._asset_tilloff_2);
    }
    public set assetTilloff3(param: IMixGroundTextureTilloff) {
        this._asset_tilloff_3.copyFromFloats(param.uOffset, param.vOffset, param.uScale, param.vScale);
        // this.setVector4(MixGroungMaterial.ASSET_TILLOFF_3, this._asset_tilloff_3);
    }
    public set assetTilloff4(param: IMixGroundTextureTilloff) {
        this._asset_tilloff_4.copyFromFloats(param.uOffset, param.vOffset, param.uScale, param.vScale);
        // this.setVector4(MixGroungMaterial.ASSET_TILLOFF_4, this._asset_tilloff_4);
    }
    public set mixTilloff(param: IMixGroundTextureTilloff) {
        this._mix_tilloff.copyFromFloats(param.uOffset, param.vOffset, param.uScale, param.vScale);
        // this.setVector4(MixGroungMaterial.MIX_TILLOFF, this._mix_tilloff);
    }

    public set sceneHSize(val: number) {
        this._scene_param.x = val;
    }

    public set sceneVSize(val: number) {
        this._scene_param.y = val;
    }

    public set gridEdgeLineBaseSize(val: number) {
        this._grid_param.x = val;
    }

    public set gridEdgeLineDeltaSize(val: number) {
        this._grid_param.y = val;
    }

    public set gridEdgeLineStrength(val: number) {
        this._grid_param.z = val;
    }

    public set gridEdgeLightStrength(val: number) {
        this._grid_param.w = val;
    }

    public set gridEdgeLightRangeSize(val: number) {
        this._scene_param.w = val;
    }

    constructor(
        scene: SceneStruct,
        materialName: string,
    ) {
        super(
            materialName,
            scene.impl,
            `MixGroungMaterial`,
            {
                attributes: [MixGroungMaterial.POSITION, MixGroungMaterial.UV],
                uniforms: [
                    MixGroungMaterial.WORLD_VIEW_PROJECTION,
                    MixGroungMaterial.ASSET_TEX0,
                    MixGroungMaterial.ASSET_TEX1,
                    MixGroungMaterial.ASSET_TEX2,
                    MixGroungMaterial.ASSET_TEX3,
                    MixGroungMaterial.ASSET_TEX4,
                    MixGroungMaterial.ASSET_TILLOFF_0,
                    MixGroungMaterial.ASSET_TILLOFF_1,
                    MixGroungMaterial.ASSET_TILLOFF_2,
                    MixGroungMaterial.ASSET_TILLOFF_3,
                    MixGroungMaterial.ASSET_TILLOFF_4,
                    MixGroungMaterial.MIX_TEX,
                    MixGroungMaterial.MIX_TILLOFF,
                    MixGroungMaterial.LIGHT_TEX,
                    MixGroungMaterial.GRID_TEX,
                    MixGroungMaterial.SCENE_SIZE,
                    MixGroungMaterial.GRID_PARAM,
                    MixGroungMaterial.LIGHT_PARAM,
                ],
                needAlphaBlending: false
            }
        );

        this.setVector4(MixGroungMaterial.ASSET_TILLOFF_0, this._asset_tilloff_0);
        this.setVector4(MixGroungMaterial.ASSET_TILLOFF_1, this._asset_tilloff_1);
        this.setVector4(MixGroungMaterial.ASSET_TILLOFF_2, this._asset_tilloff_2);
        this.setVector4(MixGroungMaterial.ASSET_TILLOFF_3, this._asset_tilloff_3);
        this.setVector4(MixGroungMaterial.ASSET_TILLOFF_4, this._asset_tilloff_4);
        this.setVector4(MixGroungMaterial.MIX_TILLOFF, this._mix_tilloff);
        this.setVector4(MixGroungMaterial.SCENE_SIZE, this._scene_param);
        this.setVector4(MixGroungMaterial.GRID_PARAM, this._grid_param);
        this.setVector4(MixGroungMaterial.LIGHT_PARAM, this._light_param);

        scene.impl.onBeforeRenderObservable.add(() => {
            this._scene_param.z = (Date.now() - this._time) / 1000;
            this._light_param.w = this._light_tex ? this._light_tex.level : 1.;
        });
    }
    isReady(mesh?: BABYLON.AbstractMesh, useInstances?: boolean): boolean {
        return (this._asset_tex0 && this._asset_tex0.isReady())
        && (this._asset_tex1 && this._asset_tex1.isReady())
        && (this._asset_tex2 && this._asset_tex2.isReady())
        && (this._asset_tex3 && this._asset_tex3.isReady())
        && (this._asset_tex4 && this._asset_tex4.isReady())
        && (this._mix_tex && this._mix_tex.isReady())
        && (this._light_tex && this._light_tex.isReady())
        && super.isReady(mesh, useInstances);
    }

    bind(world: BABYLON.Matrix, mesh?: BABYLON.Mesh, effectOverride?: BABYLON.Effect): void {
        // this._scene_param.copyFromFloats(this._scene_width, this._scene_height, (Date.now() - this._time) / 1000, 0);
        super.bind(world, mesh, effectOverride);
    }
    dispose(forceDisposeEffect?: boolean, forceDisposeTextures?: boolean, notBoundToMesh?: boolean): void {
        this.assetTex0?.dispose();
        this.assetTex1?.dispose();
        this.assetTex2?.dispose();
        this.assetTex3?.dispose();
        this.assetTex4?.dispose();
        this.mixTex?.dispose();
        this._light_tex?.dispose();
        this._grid_tex?.dispose();
        super.dispose(forceDisposeEffect, forceDisposeTextures, notBoundToMesh);
    }
}

BABYLON.Effect.ShadersStore[`${MixGroungMaterial.KEY}VertexShader`] = `
attribute vec3 ${MixGroungMaterial.POSITION};
attribute vec2 ${MixGroungMaterial.UV};

uniform mat4 ${MixGroungMaterial.WORLD_VIEW_PROJECTION};

varying vec2 vUV;

void main() {
    vUV = ${MixGroungMaterial.UV};
    gl_Position         = ${MixGroungMaterial.WORLD_VIEW_PROJECTION} * vec4(${MixGroungMaterial.POSITION}, 1.0);
}
`;
BABYLON.Effect.ShadersStore[`${MixGroungMaterial.KEY}FragmentShader`] = `
uniform sampler2D ${MixGroungMaterial.ASSET_TEX0};
uniform sampler2D ${MixGroungMaterial.ASSET_TEX1};
uniform sampler2D ${MixGroungMaterial.ASSET_TEX2};
uniform sampler2D ${MixGroungMaterial.ASSET_TEX3};
uniform sampler2D ${MixGroungMaterial.ASSET_TEX4};
uniform sampler2D ${MixGroungMaterial.MIX_TEX};
uniform sampler2D ${MixGroungMaterial.LIGHT_TEX};
uniform sampler2D ${MixGroungMaterial.GRID_TEX};

uniform vec4 ${MixGroungMaterial.ASSET_TILLOFF_0};
uniform vec4 ${MixGroungMaterial.ASSET_TILLOFF_1};
uniform vec4 ${MixGroungMaterial.ASSET_TILLOFF_2};
uniform vec4 ${MixGroungMaterial.ASSET_TILLOFF_3};
uniform vec4 ${MixGroungMaterial.ASSET_TILLOFF_4};
uniform vec4 ${MixGroungMaterial.MIX_TILLOFF};
uniform vec4 ${MixGroungMaterial.SCENE_SIZE};
uniform vec4 ${MixGroungMaterial.GRID_PARAM};
uniform vec4 ${MixGroungMaterial.LIGHT_PARAM};

varying vec2 vUV;

const float PI = 3.1415926;
const vec3 LuminanceEncodeApprox = vec3(0.2126, 0.7152, 0.0722);
// BABYLON
float GetLuminance(vec3 color) {
    return clamp(dot(color, LuminanceEncodeApprox), 0., 1.);
}
mat2 rotate2d(float _angle){
    return mat2(cos(_angle),-sin(_angle),
                sin(_angle),cos(_angle));
}

void main() {
    vec2 mainUV = vUV;
    // // vec2 mainUV = vec2(vUV.x - 0.7071067811 * vUV.y, vUV.y / 0.7071067811);
    // vec2 mainUV = vUV - 0.5;
    // mainUV = rotate2d(PI * 0.25) * mainUV;
    // mainUV = mainUV + 0.5;
    vec4 diffuseColor = vec4(1., 1., 1., 1.);

    vec4 mixColor = texture2D(${MixGroungMaterial.MIX_TEX}, mainUV * ${MixGroungMaterial.MIX_TILLOFF}.zw + ${MixGroungMaterial.MIX_TILLOFF}.xy);

    vec4 asset0 = texture2D(${MixGroungMaterial.ASSET_TEX0}, mainUV * (${MixGroungMaterial.ASSET_TILLOFF_0}.zw ) + ${MixGroungMaterial.ASSET_TILLOFF_0}.xy);
    vec4 asset1 = texture2D(${MixGroungMaterial.ASSET_TEX1}, mainUV * (${MixGroungMaterial.ASSET_TILLOFF_1}.zw ) + ${MixGroungMaterial.ASSET_TILLOFF_1}.xy);
    vec4 asset2 = texture2D(${MixGroungMaterial.ASSET_TEX2}, mainUV * (${MixGroungMaterial.ASSET_TILLOFF_2}.zw ) + ${MixGroungMaterial.ASSET_TILLOFF_2}.xy);
    vec4 asset3 = texture2D(${MixGroungMaterial.ASSET_TEX3}, mainUV * (${MixGroungMaterial.ASSET_TILLOFF_3}.zw ) + ${MixGroungMaterial.ASSET_TILLOFF_3}.xy);
    vec4 asset4 = texture2D(${MixGroungMaterial.ASSET_TEX4}, mainUV * (${MixGroungMaterial.ASSET_TILLOFF_4}.zw ) + ${MixGroungMaterial.ASSET_TILLOFF_4}.xy);

    vec4 lightColor = texture2D(${MixGroungMaterial.LIGHT_TEX}, mainUV) * ${MixGroungMaterial.LIGHT_PARAM}.w;

    diffuseColor.rgb = asset0.rgb;
    diffuseColor.rgb = mix(diffuseColor.rgb, asset1.rgb, mixColor.r);
    diffuseColor.rgb = mix(diffuseColor.rgb, asset2.rgb, mixColor.g);
    diffuseColor.rgb = mix(diffuseColor.rgb, asset3.rgb, mixColor.b);
    diffuseColor.rgb = mix(diffuseColor.rgb, asset4.rgb, 1.0 - mixColor.a);

    float lightStrength = 1.; //GetLuminance(diffuseColor.rgb);
    // lightStrength = pow(lightStrength, 0.5);
    // grid
    vec2 grid = fract(mainUV * ${MixGroungMaterial.SCENE_SIZE}.xy);
    vec2 gridRowCol = floor(mainUV * ${MixGroungMaterial.SCENE_SIZE}.xy);

    float random = GetLuminance(diffuseColor.rgb) + cos(54981. * gridRowCol.x * gridRowCol.y);

    float lineDiff = ${MixGroungMaterial.SCENE_SIZE}.w;
    // 网格线
    float gridLine = smoothstep(0.5 - lineDiff, 0.5, abs(grid.x - 0.5)) + smoothstep(0.5 - lineDiff, 0.5, abs(grid.y - 0.5));
    // 网格线 由内向外增强
    gridLine = smoothstep(0.5, 1., gridLine);

    vec4 gridData = texture2D(${MixGroungMaterial.GRID_TEX}, mainUV);
    float gridLum = gridData.r + gridData.g + gridData.b;
    // 为不同颜色实现时间差异,以实现呼吸差异
    float gridFlag = gridData.a * 255.;

    float deltaLightStrength = 0.4 + smoothstep(-0.5, 1.5, random) * 0.6;
    // 网格线 左&下 变暗, 右&上 变亮 模拟右上向左下的光照下的瓷砖边界凹槽
    float gridEdgeLightStrength = 1.0 - gridLine * (mix(${MixGroungMaterial.GRID_PARAM}.w * deltaLightStrength, 0.-${MixGroungMaterial.GRID_PARAM}.w * deltaLightStrength, smoothstep(0.9, 1.1, grid.x + grid.y)));
    lightStrength = lightStrength * gridEdgeLightStrength;

    // 这里控制网格线是否显示
    // diffuseColor.rgb = diffuseColor.rgb * lightStrength;

    float randomEdge = gridData.r * 1. + gridData.g * 2. + gridData.b * 4.;
    float edge = ${MixGroungMaterial.GRID_PARAM}.x + ${MixGroungMaterial.GRID_PARAM}.y * cos((${MixGroungMaterial.SCENE_SIZE}.z + randomEdge) * PI * 2. * 0.2);
    float leftV = (1.0 - smoothstep(0.05, edge, grid.x));
    float bottomV = (1.0 - smoothstep(0.05, edge, grid.y));
    float rightV = smoothstep(1.0 - edge, 1.0 - 0.05, grid.x);
    float topV = smoothstep(1.0 - edge, 1.0 - 0.05, grid.y);

    float gridLeftTop = step(${GridInfo.LEFT_TOP}., gridFlag);
    gridFlag = gridFlag - ${GridInfo.LEFT_TOP}. * gridLeftTop;
    float gridLeftBottom = step(${GridInfo.LEFT_BOTTOM}., gridFlag);
    gridFlag = gridFlag - ${GridInfo.LEFT_BOTTOM}. * gridLeftBottom;
    float gridRightBottom = step(${GridInfo.RIGHT_BOTTOM}., gridFlag);
    gridFlag = gridFlag - ${GridInfo.RIGHT_BOTTOM}. * gridRightBottom;
    float gridRightTop = step(${GridInfo.RIGHT_TOP}., gridFlag);
    gridFlag = gridFlag - ${GridInfo.RIGHT_TOP}. * gridRightTop;
    float gridLeft = step(${GridInfo.LEFT}., gridFlag);
    gridFlag = gridFlag - ${GridInfo.LEFT}. * gridLeft;
    float gridBottom = step(${GridInfo.BOTTOM}., gridFlag);
    gridFlag = gridFlag - ${GridInfo.BOTTOM}. * gridBottom;
    float gridRight = step(${GridInfo.RIGHT}., gridFlag);
    gridFlag = gridFlag - ${GridInfo.RIGHT}. * gridRight;
    float gridTop = step(${GridInfo.TOP}., gridFlag);
    float gridEdge = max( mix(0., bottomV, gridTop),
        max( mix(0., rightV, gridRight),
            max( mix(0., leftV, gridLeft),
                max( mix(0., topV, gridBottom),
                    max( mix(0., leftV * bottomV, gridLeftTop),
                        max( mix(0., leftV * topV, gridLeftBottom),
                            max( mix(0., rightV * topV, gridRightBottom),
                                mix(0., rightV * bottomV, gridRightTop)
                            )
                        )
                    )
                )
            )
        )
    );
    gridEdge = gridEdge * step(0.01, gridLum);
    // gridEdge = smoothstep(0.0, 0.8, gridEdge) - step(0.95, gridEdge);
    gridEdge = gridEdge * ${MixGroungMaterial.GRID_PARAM}.z;
    // gridEdge = smoothstep(0.0, ${MixGroungMaterial.GRID_PARAM}.z, gridEdge) - step(${MixGroungMaterial.GRID_PARAM}.z, gridEdge);

    // diffuseColor.rgb = gridData.rgb; //
    diffuseColor.rgb = mix(diffuseColor.rgb, gridData.rgb, gridEdge);

    // diffuseColor.rgb = diffuseColor.rgb * lightColor.rgb * smoothstep(0., 1.0 + deltaLightStrength, 1.0 + deltaLightStrength); // * (1.0 + gridEdge * 2.);
    diffuseColor.rgb = diffuseColor.rgb * lightColor.rgb; // * (1.0 + gridEdge * 2.);

    gl_FragColor = diffuseColor;
}
`;

export class MixGround {
    private plane: BABYLON.Mesh;
    private root: BABYLON.TransformNode;
    public material: MixGroungMaterial;
    constructor(
        public scene: SceneStruct,
        hsize: number,
        vsize: number,
    ) {
        this.material = new MixGroungMaterial(scene, "MixGround");

        let geo = new BABYLON.Geometry("mixgroundgeo", scene.impl);
        let mesh = new BABYLON.Mesh("MixGround", scene.impl);
        let position = [];
        let temp: [number, number] = [0, 0];
        MixGround.rowcolToPosition(-vsize / 2, -hsize / 2, temp);
        // position.push(temp[0]); position.push(temp[1]);position.push(0); 
        position.push(temp[0]); position.push(temp[1]);position.push(0); 
        MixGround.rowcolToPosition(-vsize / 2, hsize / 2, temp);
        // position.push(temp[0]); position.push(temp[1]);position.push(0); 
        position.push(temp[0]); position.push(temp[1]);position.push(0); 
        MixGround.rowcolToPosition(vsize / 2, hsize / 2, temp);
        // position.push(temp[0]); position.push(temp[1]);position.push(0); 
        position.push(temp[0]); position.push(temp[1]);position.push(0); 
        MixGround.rowcolToPosition(vsize / 2, -hsize / 2, temp);
        // position.push(temp[0]); position.push(temp[1]);position.push(0); 
        position.push(temp[0]); position.push(temp[1]);position.push(0); 

        geo.setVerticesData( BABYLON.VertexBuffer.PositionKind, position );
        geo.setVerticesData( BABYLON.VertexBuffer.UVKind, [0, 0, 1, 0, 1, 1, 0, 1] );
        geo.setVerticesData( BABYLON.VertexBuffer.NormalKind, [0, 0, 1, 0, 0, 1, 0, 0, 1, 0, 0, 1] );
        geo.setIndices( [0, 1, 2, 0, 2, 3] );
        
        geo.applyToMesh(mesh);
        this.plane = mesh;
        // this.plane = BABYLON.PlaneBuilder.CreatePlane("MixGround", { size: 1 }, scene.impl);

        this.plane.material = this.material;
        this.plane.material.needAlphaBlending = () => true
        // debugger;

        this.root = new BABYLON.TransformNode("MixGround", this.scene.impl);
        this.root.rotation.copyFromFloats(Math.PI / 2, 0, 0);
        this.plane.parent = this.root;
        window["rr"] = this.root;
    }

    public set position(val: BABYLON.Vector3) {
        this.root.position.copyFrom(val);
    }
    public set scaling(val: BABYLON.Vector3) {
        this.plane.scaling.copyFrom(val);
    }
    public get position() {
        return this.root.position;
    }
    public get scaling() {
        return this.plane.scaling;
    }
    public set alphaIndex(val:number) {
        this.plane.alphaIndex = val;
    }

    public dispose() {
        this.material?.dispose(true, true);
        this.plane?.dispose(true, true);
    }
    
    public static rotateSin = Math.sin(0 / 180 * Math.PI);
    public static rotateCos = Math.cos(0 / 180 * Math.PI);
    /**
     * 行列坐标转换为场景坐标
     * 以 Map 中心 为旋转中心 顺时针旋转 45 度, 再纵向缩放为 0.8 
     * @param row 
     * @param col 
     * @param result 
     */
    public static rowcolToPosition(row: number, col: number, result: [number, number]) {
        let x = col;
        let y = row;

        result[0] = MixGround.rotateCos * x - MixGround.rotateSin * y;
        result[1] = MixGround.rotateSin * x + MixGround.rotateCos * y;

        // result[0] += 1 / 2;
        // result[1] += 1 / 2;

        // result[0] += this.tileMap.width / 2;
        result[1] *= MixGround.scaleY;
    }
    public static scaleY = 0.999;
}