
import { NodeTools, LoaderTool } from "./scene_tool";
import { IModelChildOpt, IModelAttachOpt, IModelOpt, IImageSolt, IMaterialSolt, IMaterialReplace, IModelObj, IPrefabAnime, IModel } from "./interface/modelobj";
import { MeshParticleSystem } from "./mesh_particle/mesh_particle_system";
import { BABYLON } from "./render3d/babylon";
import { IGLTFNew } from "./gltf/interface";
import { IMeshMaterial, IMeshRenderQueue } from "./interface/mesh";
import { INodeCollider } from "./interface/collider";
import { EventState, ExtensionTools, IOfflineAnimationCfg, Observable, Res, ResTab } from './interface/index';
import { PointerInfo } from "./render/events/pointerEvents";
import { PiCollider } from "./scene_pick";
import { CustomScene } from "./index";
import { IColliderItem, IPiCollider } from "./interface/extends";
import { GLTFLoader } from "./gltf/glTFLoader";

declare let module;
let logger;
ExtensionTools.RegistStartUp(() => { logger = ExtensionTools.Logger(module.name, "ModelObj"); });

export class Model implements IModel, IColliderItem {
    private _extendData: any;
    set extendData(val: any) {
        this._extendData = val;
    }
    get extendData() {
        return this._extendData;
    }
    rootImpl: BABYLON.TransformNode;
    isDisposed: boolean;
    protected _isDisposeOver: boolean = false;
    public isLoaded: boolean;
    public position: [number, number, number];
    public scale: [number, number, number];
    public rotate: [number, number, number];
    public lookAt: [number, number, number];
    public color: [number, number, number];
    public alpha: number;
    public animOpt: IPrefabAnime;

    protected _animationMap?: Map<string, BABYLON.AnimationGroup> = new Map();
    public get animationMap() {
        return this._animationMap;
    }

    constructor(
        protected _name: string,
        public readonly scene: CustomScene,
    ) {

    }
    setAnimeFrame(animName: string, progress: number): IModel {
        // throw new Error("Method not implemented.");
        return this;
    }
    getChild(index?: number): BABYLON.TransformNode {
        return this.rootImpl;
    }
    _colliderBoundle?: IPiCollider;
    _boundingInfo?: BABYLON.BoundingInfo;
    get worldMatrixFromCache(): BABYLON.Matrix {
        return this.rootImpl?.worldMatrixFromCache;
    }
    protected _loadPromise?: Promise<any>

    public get name() {
        return this._name;
    }

    public get loadPromise() {
        return this._loadPromise;
    }

    protected _isEnabled: boolean;
    public isEnabled(): boolean {
        if (this._isEnabled === undefined) {
            return true;
        } else {
            return this._isEnabled;
        };
    }
    protected _rayID: number;
    public get rayID() {
        return this._rayID;
    }
    public set rayID(val: any) {
        this._rayID = val;
    }
    protected _state: any;
    public modifyState(state: any) {
        this._state = state;
        this._modifyState();
    }
    protected _modifyState() {
        if (this.isLoaded && this.isDisposed == false) {
            //
        }
    }

    public setPostion(data: number[]) {
        this.position = [data[0], data[1], data[2]];
    }
    public setPosition(x: number | number[], y?: number, z?: number) {
        if (x instanceof Array) {
            this.position = [x[0], x[1], x[2]];
        } else {
            this.position = [x, y, z];
        }
    }
    public setScale(x: number | number[], y?: number, z?: number) {
        this.setScaling(x, y, z);
    }
    public setScaling(x: number | number[], y?: number, z?: number) {
        if (x instanceof Array) {
            this.scale = [x[0], x[1], x[2]];
        } else {
            this.scale = [x, y, z];
        }
    }
    public setRotation(x: number | number[], y?: number, z?: number) {
        if (x instanceof Array) {
            this.rotate = [x[0], x[1], x[2]];
        } else {
            this.rotate = [x, y, z];
        }
    }
    public setRotate(x: number | number[], y?: number, z?: number) {
        this.setRotation(x, y, z);
    }
    public setLookAt(x: number | number[], y?: number, z?: number) {
        if (x instanceof Array) {
            this.lookAt = [x[0], x[1], x[2]];
        } else {
            this.lookAt = [x, y, z];
        }
    }
    public setEnabled(isEnabled: boolean) {
        this._isEnabled = isEnabled;
        this.rootImpl?.setEnabled(isEnabled);
    }
    setColor(r: number, g: number, b: number): void {
        this.color = [r, g, b];
    }
    setAlpha(val: number): void {
        this.alpha = val;
    }
    getDistanceToCamera(camera: BABYLON.Camera) {
        if (this.rootImpl) {
            return this.rootImpl.getAbsolutePosition().subtract(camera.globalPosition).length();
        } else {
            return Number.MAX_SAFE_INTEGER;
        }
    }
    dispose(disposeMaterialAndTextures?: boolean): void {
        this.onDisposeObservable.notifyObservers(this);
        this.onDisposeObservable.clear();
        this.onPointerObservable.clear();
        this.rootImpl?.dispose(false, disposeMaterialAndTextures);
    }
    addAnimationEventListen(key: string, func: () => boolean): void {
        // throw new Error("Method not implemented.");
    }
    pickAboutItems(): INodeCollider[] {
        // throw new Error("Method not implemented.");
        return [];
    }
    setAnim(animOpt: IPrefabAnime): void {
        // throw new Error("Method not implemented.");
    }
    setAnimeSpeed(param: number): void {
        // throw new Error("Method not implemented.");
    }
    setParent(parent: IModel, indexOfParent?: number): void {
        this.rootImpl.parent = <BABYLON.TransformNode>parent?.rootImpl;
    }
    reuse() {

    }

    public onPointerObservable = new Observable<PointerInfo>();
    public onDisposeObservable = new Observable<Model>();
    protected onPointer = (info: PointerInfo, eventState: EventState) => {
        this.onPointerObservable.notifyObservers(info);
    }

    protected _animations?: BABYLON.AnimationGroup[];
    public get animations() {
        return this._animations;
    }

    protected _renderingGroupId: number;
    protected _alphaIndex: number;
    protected _lastuseAlphaIndex: number;
    set alphaIndex(param: number) {
        this._alphaIndex = param;
    }
    public get alphaIndex() {
        // debugger;
        return this._alphaIndex;
    }
    public get renderingGroupId() {
        return this._renderingGroupId;
    }
    public set renderingGroupId(param: number) {
        // debugger;
        this._renderingGroupId = param;
    }

    protected __cached: boolean;
    public get _cached() {
        return this.__cached;
    }
    public set _cached(val: boolean) {
        if (this.__cached == undefined) {
            this.__cached = val;
        }
    }

    public unuse(): void {
        this.onPointerObservable.clear();
        this._pointerUps.length = 0;
        this._pointerDowns.length = 0;
        this.setEnabled(false);
        this._modifyUseState();
    }
    protected _modifyUseState(): void {
        if (this._isEnabled == false) {
            this.animationMap?.forEach((group) => {
                group.onAnimationGroupEndObservable.clear();
                group.onAnimationGroupPlayObservable.clear();
                group.onAnimationGroupLoopObservable.clear();
                // group.onAnimationGroupFrameEventObservable.clear();
                group.stop();
            });
        }
    }

    public renderQueueItems(): IMeshRenderQueue[] {
        let result: IMeshRenderQueue[] = [];
        return result;
    }

    public renderMaterialItems(): IMeshMaterial[] {
        let result: IMeshMaterial[] = [];
        return result;
    }

    /**
     * 停止当前的动画 - （模型）一次只播一个动画
     */
    public stopAnim() {

    }
    /**
     * 暂停当前的动画 - （模型）一次只播一个动画
     */
    public pauseAnim() {

    }
    protected _pointerUps: Function[] = [];
    protected _pointerDowns: Function[] = [];
    /**
     * 添加事件监听
     * * 如果不是SpriteModel需要手动调用 collider 设置命中盒子
     * @param etype 
     * @param func 
     */
    addEventListener(etype: "pointerup" | "pointerdown", func: Function) {
        if (this.isDisposed) return;
        this._isPickable = true;
        if (etype == "pointerup") {
            let idx = this._pointerUps.indexOf(func);
            if (idx < 0) {
                this._pointerUps.push(func);
            }
        } else {
            let idx = this._pointerDowns.indexOf(func);
            if (idx < 0) {
                this._pointerDowns.push(func);
            }
        }
    }
    removeEventListener(etype: "pointerup" | "pointerdown", func: Function) {
        if (this.isDisposed) return;
        this._isPickable = true;
        if (etype == "pointerup") {
            let idx = this._pointerUps.indexOf(func);
            if (idx >= 0) {
                this._pointerUps.splice(idx, 1);
            }
        } else {
            let idx = this._pointerDowns.indexOf(func);
            if (idx < 0) {
                this._pointerDowns.splice(idx, 1);
            }
        }
    }
    /**
     * 自定义的盒子
     */
    protected _colliderCustom: PiCollider;
    protected _isPickable: boolean;
    /**
     * 默认值为 (-0.5, -0.5, -0.5) 到 (0.5, 0.5, 0.5)
     * @param min 
     * @param max 
     */
    collider(min?: BABYLON.Vector3, max?: BABYLON.Vector3) {
        if (this.isDisposed) return;
        this._isPickable = true;
        if (this._colliderCustom == undefined) {
            // 如果有精灵自动创建的内容则进行覆盖
            this._colliderBoundle = undefined;
            this._colliderCustom = new PiCollider(min, max, this.rootImpl?.worldMatrixFromCache);
        } else {
            this._colliderCustom.reConstruct(min, max, this.rootImpl?.worldMatrixFromCache);
        }
        this._updateCollider();
    }
    protected _colliderSphereTreshold: number = 0;
    colliderSphereTreshold(val: number) {
        this._colliderSphereTreshold = val;
        this._updateCollider();
    }
    protected _listenAdded: boolean = false;
    protected _updateCollider() {
        if (this.isDisposed) return;
        if (this._colliderBoundle instanceof PiCollider) {
            this._colliderBoundle.intersectionTreshold = this._colliderSphereTreshold;
        }
        if (this._isPickable && !this._listenAdded) {
            this.scene.addEventListener(this, "pointerup", this._onPointerUp);
            this.scene.addEventListener(this, "pointerdown", this._onPointerDown);
            this._listenAdded = true;
        }
    }
    protected _onPointerUp = () => {
        if (this.isDisposed) return;
        this._pointerUps.forEach(function(func) {
            try {
                func();
            } catch (error) {
                
            }
        })
    }
    protected _onPointerDown = () => {
        if (this.isDisposed) return;
        this._pointerDowns.forEach(function(func) {
            try {
                func();
            } catch (error) {
                
            }
        })
    }
}

const path = [
    "translation",  "rotation",     "scale",            "uScale",       "vScale",
    "uOffset",      "vOffset",      "alpha",            "color",        "ort",
    "per",          "piIsActive",   "rotation_eular",   "intensity",    "lightdiffuse",
    "alphaCutOff",  "cellId",       "uScaleOp",         "vScaleOp",     "uOffsetOp",
    "vOffsetOp",    "maskCutoff",   "uScaleM",      "vScaleM",           "uOffsetM",
    "vOffsetM"
];

const propertype = [
    "position",  "rotationQuaternion",     "scaling",            "uScale",       "vScale",
    "uOffset",      "vOffset",      "alpha",            "color",        "",
    "",          "piIsActive",   "rotation",   "",    "",
    "",  "",       "",         "",     "",
    "",    "",   "",      "",           "",
    ""
];

export class ModelObj extends Model implements IModelObj {
    public static appendAnimation(root: BABYLON.TransformNode, appenAnimation: IOfflineAnimationCfg) {
        let childs = root.getChildTransformNodes();
        let childNames = [];
        childs.forEach((child) => {
            childNames.push(child.name);
        });
        let result = [];
        if (appenAnimation) {
            let groupCount = appenAnimation[0].length;
            let buffer: Uint8Array;
            if (appenAnimation[2] instanceof Uint8Array) {
                buffer = appenAnimation[2];
            } else {
                buffer = new Uint8Array(appenAnimation[2]);
            }
            for (let groupidx = 0; groupidx < groupCount; groupidx++) {
                let animations = appenAnimation[1][groupidx];
                let groupName = appenAnimation[0][groupidx];
                let animationGroup = new BABYLON.AnimationGroup(groupName, root._scene);
                result.push(animationGroup);

                animations.forEach((info) => {
                    
                    let [property, name, inoffset, incount, inbufferoff, inbufferlen, outoffset, outcount, outbufferoff, outbufferlen, interpolation, keys] = info;
                    let propertylen = 1;
                    let animationType = BABYLON.Animation.ANIMATIONTYPE_FLOAT;
                    if (property == 0 || property == 2 || property == 12) {
                        propertylen = 3;
                        animationType = BABYLON.Animation.ANIMATIONTYPE_VECTOR3;
                    } else if (property == 8 || property == 14) {
                        propertylen = 3;
                        animationType = BABYLON.Animation.ANIMATIONTYPE_COLOR3;
                    } else if (property == 1) {
                        propertylen = 4;
                        animationType = BABYLON.Animation.ANIMATIONTYPE_QUATERNION;
                    }
                    let idx = childNames.indexOf(name);
                    if (idx >= 0) {
                        let target = childs[idx];
                        let input = new Float32Array(buffer.buffer, inoffset + inbufferoff, incount);
                        let output = new Float32Array(buffer.buffer, outoffset + outbufferoff, propertylen * outcount);
                        let data = {
                            input,
                            output,
                            interpolation,
                        };
                        let targetPath = path[property];
                        let propertypePath = propertype[property];
                        if (targetPath && propertypePath) {
                            if (keys == undefined) {
                                let outputBufferOffset = 0;
                                let getNextOutputValue: () => BABYLON.Vector3 | BABYLON.Quaternion | number[] | number | BABYLON.Color3;
                                switch (propertypePath) {
                                    case 'position': {
                                        getNextOutputValue = () => {
                                            const value = BABYLON.Vector3.FromArray(data.output, outputBufferOffset);
                                            outputBufferOffset += 3;
                                            return value;
                                        };
                                        break;
                                    }
                                    case 'rotationQuaternion': {
                                        getNextOutputValue = () => {
                                            const value = BABYLON.Quaternion.FromArray(data.output, outputBufferOffset);
                                            outputBufferOffset += 4;
                                            return value;
                                        };
                                        break;
                                    }
                                    case 'color': {
                                        getNextOutputValue = () => {
                                            const value = BABYLON.Color3.FromArray(data.output, outputBufferOffset);
                                            outputBufferOffset += 3;
                                            return value;
                                        };
                                        break;
                                    }
                                    case 'rotation': {
                                        getNextOutputValue = () => {
                                            const value = BABYLON.Vector3.FromArray(data.output, outputBufferOffset);
                                            outputBufferOffset += 3;
                                            return value;
                                        };
                                        break;
                                    }
                                    case 'scaling': {
                                        getNextOutputValue = () => {
                                            const value = BABYLON.Vector3.FromArray(data.output, outputBufferOffset);
                                            outputBufferOffset += 3;
                                            return value;
                                        };
                                        break;
                                    }
                                    case 'diffuse': {
                                        getNextOutputValue = () => {
                                            const value = BABYLON.Color3.FromArray(data.output, outputBufferOffset);
                                            outputBufferOffset += 3;
                                            return value;
                                        };
                                        break;
                                    }
                                    case 'influence': {
                                        // getNextOutputValue = () => {
                                        //     // tslint:disable-next-line:prefer-array-literal
                                        //     const value = new Array<number>(targetNode._numMorphTargets!);
                                        //     // tslint:disable-next-line:no-non-null-assertion
                                        //     for (let i = 0; i < targetNode._numMorphTargets!; i++) {
                                        //         value[i] = data.output[outputBufferOffset++];
                                        //     }
                    
                                        //     return value;
                                        // };
                                        break;
                                    }
                                    default: {
                                        getNextOutputValue = () => {
                                            return data.output[outputBufferOffset++];
                                        };
                                    }
                                }
                    
                                let getNextKey: (frameIndex: number) => BABYLON.IAnimationKey;
                                // tslint:disable-next-line:switch-default
                                switch (data.interpolation) {
                                    case 0: {
                                        getNextKey = (frameIndex) => ({
                                            frame: data.input[frameIndex],
                                            value: getNextOutputValue(),
                                            interpolation: BABYLON.AnimationKeyInterpolation.STEP
                                        });
                                        break;
                                    }
                                    case 1: {
                                        getNextKey = (frameIndex) => ({
                                            frame: data.input[frameIndex],
                                            value: getNextOutputValue()
                                        });
                                        break;
                                    }
                                    case 3: {
                                        getNextKey = (frameIndex) => {
                                            const inTangent = getNextOutputValue();
                                            const value = getNextOutputValue();
                                            const outTangent = getNextOutputValue();
                    
                                            return {
                                                frame: data.input[frameIndex],
                                                inTangent: inTangent,
                                                value: value,
                                                outTangent: outTangent
                                            };
                                        };
                                        break;
                                    }
                                    case 2 /* CUBICSPLINE */: {
                                        getNextKey = (frameIndex) => {
                                            const inAndOut = getNextOutputValue();
                    
                                            return ({
                                                frame: data.input[frameIndex],
                                                inTangent: inAndOut,
                                                value: getNextOutputValue(),
                                                outTangent: inAndOut
                                            });
                                        };
                                    }
                                }
        
                                // tslint:disable-next-line:prefer-array-literal
                                keys = [];
                                let frame = data.input[0];
                                for (let frameIndex = 0; frameIndex < data.input.length; frameIndex++) {
                                    // tslint:disable-next-line:no-non-null-assertion
                                    if (frame <= data.input[frameIndex]) {
                                        keys.push(getNextKey!(frameIndex));
                                        frame = data.input[frameIndex];
                                    }
                                }
    
                                info[11] = keys;
                            }
    
                            let babylonAnimation = new BABYLON.Animation(groupName + "#" + name, propertypePath, 1, animationType);
                            babylonAnimation.setKeys(keys);
                            target.animations.push(babylonAnimation);
                            animationGroup.addTargetedAnimation(babylonAnimation, target);
                        }
                    }
                });
            }
        }
        return result;
    }
    public isDisposed: boolean = false;
    protected _isDisposeOver: boolean = false;
    /**
     * 导入的动画是否默认播放
     */
    public animDefault: boolean;
    public particleAutoStart: boolean;
    public gltfModifyConfig: {
        imageSolts?: string[] | IImageSolt[];
        materialSolts?: string[] | IMaterialSolt[];
        materialReplace?: IMaterialReplace[];
        resTab?: ResTab;
    } = {};
    /**
     * 切换动画时是否平滑过渡
     */
    public changeAnimEnableBlending: boolean = false;

    /**
     * 切换动画时过渡 需要的时间 单位:秒
     */
    public changeAnimBlendUseSeconds: number = (1 / 0.01) * (1000 / 60) / 1000;

    /**
     * 切换动画时过渡速度 默认 0.01 - 每帧进度
     * @deprecated
     */
    public set changeAnimBlendSpeed(val: number) {
        this.changeAnimBlendUseSeconds = (1 / val) * (1000 / 60) / 1000;
    }
    /**
     * @deprecated
     */
    public workWithAnimationGroupChange: boolean = false;
    public needDepthPrePass: boolean = false;
    /**
     * 是否具有骨骼动画优化
     */
    public skeletonAnimOptimize: boolean = false;
    /**
     * 动画切换前是否重置各属性为无动画状态
     */
    public initialWhileAnimChange: boolean = true;
    public resTab: ResTab;

    /**
     * 父节点
     */
    // public parent: ModelObj;
    // protected _parent: ModelObj;
    // public get parent() {
    //     return this._parent;
    // }
    // public readonly name: string;
    // public readonly path: string;
    protected _path: string;
    public get path() {
        return this._path;
    }
    // public readonly fileName: string;
    protected _fileName: string;
    public get fileName() {
        return this._fileName;
    }
    // public readonly modelName: string | string[];
    protected _modelName: string | string[];
    public get modelName() {
        return this._modelName;
    }
    public insertedCall: (model: ModelObj) => any;
    /**
     * 可拓展的实例化物件创建方法
     */
    // public createInstancedMesh: (instanceID: any, scene: SceneStruct) => BABYLON.InstancedMesh;

    protected _meshMap: Map<string, BABYLON.Mesh> = new Map();
    public get meshMap() {
        // throw new Error("Method not implemented.");
        // console.error("API deprecated !!");
        return this._meshMap;
    }
    protected _skeletonMap: Map<string, BABYLON.Skeleton> = new Map();
    protected _skeletons: BABYLON.Skeleton[] = [];
    public get skeletonMap() {
        // throw new Error("Method not implemented.");
        // console.error("API deprecated !!");
        return this._skeletonMap;
    }
    // public readonly particleSysMap: Map<string, MeshParticleSystem> = new Map();
    public get particleSysList() {
        return this._particleSystems;
    }
    public get attachOptList() {
        // throw new Error("Method not implemented.");
        console.error("API deprecated !!");
        return;
    }
    public get childOptList() {
        // throw new Error("Method not implemented.");
        console.error("API deprecated !!");
        return;
    }
    public get particleSysMap() {
        return this._particleSystems;
    }
    /**
     * 模型当前动画
     */
    protected _animation: BABYLON.AnimationGroup;
    public get animation() {
        return this._animation;
    }
    /**
     * 模型当前动画 对应 骨头动画
     */
    // protected animationForBone: BABYLON.AnimationGroup;
    /**
     * 当前动画是否循环播放
     */
    protected isAnimationLoop: boolean = false;
    protected _animEventMap: Map<string, { event: BABYLON.AnimationEvent, group: BABYLON.AnimationGroup, idxs: number[] }>;
    protected get animEventMap() {
        // throw new Error("Method not implemented.");
        // console.error("API deprecated !!");
        return this._animEventMap;
    }
    /**
     * 加载时准备的克隆数据
     */
    // protected cloneData: ICloneData;
    protected get cloneData() {
        // throw new Error("Method not implemented.");
        console.error("API deprecated !!");
        return;
    }
    /**
     * 骨骼缓存文件路径
     */
    protected _sklaiPath: string;
    protected get sklaiPath() {
        // throw new Error("Method not implemented.");
        // console.error("API deprecated !!");
        return this._sklaiPath;
    }
    protected _sklaiRes: Res<any, ArrayBuffer>;
    protected get sklaiRes() {
        // throw new Error("Method not implemented.");
        // console.error("API deprecated !!");
        return this._sklaiRes;
    }
    /**
     * 克隆标识
     */
    // protected isClone: boolean = false;
    protected get isClone() {
        // throw new Error("Method not implemented.");
        console.error("API deprecated !!");
        return;
    }
    /**
     * 模型创建时的配置
     */
    protected option: IModelOpt;
    protected cloneInfo: { ref: number } = { ref: 0 };
    public get isReady() {
        // // throw new Error("Method not implemented.");
        console.error("API deprecated !!");
        return this.rootImpl; // && this.rootImpl.isReady;
    }
    _loader: GLTFLoader;
    constructor(meshName: string, scene: CustomScene, opt: IModelOpt = {}) {
        super(meshName, scene);

        this.scene._modelMap.set(this, this);

        this._name = meshName;
        // this._parent = opt.parent;
        this._rayID = opt.rayID;
        this._path = opt.path || '';
        this._fileName = opt.fileName;
        this._modelName = opt.modelName;
        this.insertedCall = opt.insertedCall;
        this.animDefault = !!opt.animDefault;
        this.particleAutoStart = !!opt.particleAutoStart;
        // this._isEnabled = true;

        this.resTab = ExtensionTools.CreateResTab();

        this.gltfModifyConfig.imageSolts = opt.imageSolts;
        this.gltfModifyConfig.materialSolts = opt.materialSolts;
        this.gltfModifyConfig.materialReplace = opt.materialReplace;
        this.gltfModifyConfig.resTab = this.resTab;

        this.needDepthPrePass = !!opt.needDepthPrePass;
        this.workWithAnimationGroupChange = opt.workWithAnimationGroupChange;
        this.skeletonAnimOptimize = !!opt.skeletonAnimOptimize;
        // this.initialWhileAnimChange = !!opt.initialWhileAnimChange;
        this.option = opt;

        if (this.fileName) {
            const arr = [];
            arr.push(
                LoaderTool.insertMeshAsync(this).then((res) => {
                    this._gltf = res.gltf;
                    res.particles?.forEach((item) => {
                        item._psReady = true;
                    })
                    this.loaded(res._root_, res.nodes, res.meshes, res.particles, res.materials, res.animationGroups, (<any>res).skeletons);
                    return this;
                })
            );
            if (this.skeletonAnimOptimize) {
                arr.push(this.loadSklai());
            }

            this._loadPromise = Promise.all(arr).then((value: [void, [string, Res<any, ArrayBuffer>] | undefined]) => {
                if (this.isDisposed) {
                    this.dispose();
                } else {
                    this.analySkin();
                    if (this.insertedCall) {
                        this.insertedCall(this);
                    }
                }
                return this;
            }).catch((e) => {
                console.error(e);
                this.dispose();
            });
        }
    }
    public static Creation(meshName: string, scene: CustomScene, opt: IModelOpt = {}) {
        return new ModelObj(meshName, scene, opt);
    }
    protected _transformNodes?: BABYLON.TransformNode[];
    protected _meshes?: BABYLON.AbstractMesh[];
    protected _materials?: BABYLON.Material[];
    protected _particleSystems?: MeshParticleSystem[];
    protected _gltf?: Res<any, IGLTFNew>;
    public get transformNodes() {
        return this._transformNodes;
    }
    public get meshes() {
        return this._meshes;
    }
    public get materials() {
        return this._materials;
    }
    public get particleSystems() {
        return this._particleSystems;
    }
    // protected frameListen: (data: BABYLON.IAnimationEventDataFrame) => void;
    /**
     * 添加动画事件监听
     * @deprecated use animationGroup.onAnimationGroupFrameEventObserva
     * @param key 动画事件关键字
     * @param func 监听 - 返回值为 true 表示该监听仅生效一次
     */
    public addAnimationEventListen(key: string, func: () => boolean) {
        // throw new Error("Method not implemented.");
        // console.error("API deprecated !!");
        return this.loadPromise.then(() => {

            this.initAnimationEventMap();

            if (this.animEventMap) {
                const eventInfo = this.animEventMap.get(key);
                if (eventInfo) {
                    if ((<any>eventInfo).event._pi_listenList?.includes(func)) {
                        console.warn(`再次相同监听 AnimationEvent - ${key}, 被忽略`);
                    }
                    else {
                        let index = eventInfo.event._pi_listenList.length;
                        eventInfo.event._pi_addListen(func);
                        eventInfo.idxs.push(index);
                    }
                    return;
                }
            }
            console.warn(`AnimationEvent Not Found - ${key}`);
        });
    }
    /**
     * 返回 Promise
     * @param opt
     */
    public loadAsync(opt: IModelOpt) {
        if (!this.fileName) {
            // this._parent = opt.parent;
            this._rayID = opt.rayID;
            this._path = opt.path || '';
            this._fileName = opt.fileName;
            this._modelName = opt.modelName;
            // this.insertedCall = opt.insertedCall;
            this.animDefault = !!opt.animDefault;
            this.particleAutoStart = !!opt.particleAutoStart;
            // this._isEnabled = true;

            this.gltfModifyConfig.imageSolts = opt.imageSolts;
            this.gltfModifyConfig.materialSolts = opt.materialSolts;
            this.gltfModifyConfig.materialReplace = opt.materialReplace;

            this.needDepthPrePass = !!opt.needDepthPrePass;
            this.workWithAnimationGroupChange = opt.workWithAnimationGroupChange;
            this.skeletonAnimOptimize = !!opt.skeletonAnimOptimize;
            // this.initialWhileAnimChange = !!opt.initialWhileAnimChange;
            this.option = opt;

            const arr = [];
            arr.push(
                LoaderTool.insertMeshAsync(this).then((res) => {
                    res.particles?.forEach((item) => {
                        item._psReady = true;
                    })
                    this.loaded(res._root_, res.nodes, res.meshes, res.particles, res.materials, res.animationGroups, (<any>res).skeletons);
                    return this;
                })
            );
            if (this.skeletonAnimOptimize) {
                arr.push(this.loadSklai());
            }

            this._loadPromise = Promise.all(arr).then((value: [void, [string, Res<any, ArrayBuffer>] | undefined]) => {
                if (this.isDisposed) {
                    this.dispose();
                } else {
                    this.analySkin();
                    if (this.insertedCall) {
                        this.insertedCall(this);
                    }
                }
                return this;
            }).catch((e) => {
                console.error(e);
                this.dispose();
            });
        }
        else {
            this._loadPromise = Promise.reject(`已在创建时进行加载`);
        }

        return this._loadPromise;
    }
    protected loadSklai(): Promise<void> {
        return LoaderTool.loadSklai(this, this.option.sklaiPath, this.option.sklaiName);
    }
    protected analySkin() {
        // return;
        const filePath = this.sklaiPath;
        let ab = this.sklaiRes?.link;

        if (!ab) {
            return;
        }
        const engine = this.scene.impl.getEngine();
        this.skeletonMap.forEach((skl) => {
            let sklTable = engine.skeletonCacheMatrixList.find((v) => {
                return v.skeletonName == skl.name;
            });

            if (!sklTable) {
                sklTable = new BABYLON.SkeletonMatrixCacheTable(filePath, skl.name, engine);
            }
            sklTable.useForSkeleton(skl);
            skl.pi_cacheMatrixTable = sklTable;
            const map = skl.pi_cacheMatrixInfoMap;
            if (map) {
                skl.pi_useCacheSkinMaterix = true;

                map.forEach((data) => {
                    const [dataLength, precision, startSecond, endSecond, totalFrameCount, dataStart, dataEnd, skinIndex, skinName, actionName] = data;
                    sklTable.cache(actionName, {
                        dataLength, precision, startSecond, endSecond, totalFrameCount, dataStart, dataEnd
                    }, <Uint8Array>ab);
                });
            }
        });
    }
    public dispose = (disposeMaterialAndTextures: boolean = true) => {
        this._dispose(disposeMaterialAndTextures);

        super.dispose();

        this.isDisposed = true;
    }
    protected _dispose(disposeMaterialAndTextures?: boolean) {

        this.scene?._modelMap?.delete(this);

        if (this._loader) {
            this._loader.dispose();
        }

        // this.resTab.clear();
        let resTab = this.resTab; // this.resTab = undefined;
        Promise.resolve().then(() => { resTab.clear() });

        if (this._gltf) {
            this._gltf?.free();
            // console.error("Dispose GLTF");
        }

        this.animEventMap?.forEach((item) => {
            item.idxs.forEach((idx) => {
                item.event._pi_listenList[idx] = undefined;
            });
        });
        this.animEventMap?.clear();

        if (this._animations) {
            this._animations.forEach((item) => {
                item.stop();
                item.dispose();
            });
            this._animations.length = 0;
        }

        this._meshMap?.forEach((mesh) => {
            if ((<any>mesh).materialData) {
                mesh.material = undefined;
            }
        });
        this._meshMap?.clear();

        // this._materials?.forEach((item) => {
        //     item.dispose();
        // });
        (<BABYLON.AbstractMesh>this.rootImpl)?.dispose(false, disposeMaterialAndTextures);

        this._gltf = undefined;
        this._transformNodes = undefined;
        this._meshes = undefined;
        this._particleSystems = undefined;
        this._materials = undefined;
        this._animations = undefined;
        this.rootImpl = undefined;

        this.animOpt = undefined;
        this._animation = undefined;
        // this._parent = undefined;
        this.lookAt = undefined;
        this.rotate = undefined;
        this.position = undefined;
        this.scale = undefined;
        this.insertedCall = undefined;
        this.gltfModifyConfig = undefined;
        // this.option = undefined;
    }
    public reuse() {
        if (this.isDisposed) return;
        if (this.animOpt == undefined && this.animDefault && this._animations && this._animations.length > 0) {
            this.animOpt = {
                animName: this.animations[0].name,
                isLoop: true,
            }
            this._changeAnim();
        }
    }
    protected _modifyUseState(): void {
        if (this._isEnabled == false) {
            this.animationMap?.forEach((group) => {
                group.onAnimationGroupEndObservable.clear();
                group.onAnimationGroupPlayObservable.clear();
                group.onAnimationGroupLoopObservable.clear();
                // group.onAnimationGroupFrameEventObservable.clear();
                group.stop();
            });
        }
    }
    protected extendLoaded() {

    }
    protected appendAnimationEnd(animationGroup: BABYLON.AnimationGroup) {

    }
    // tslint:disable-next-line:max-line-length
    protected loaded = (root: BABYLON.TransformNode, nodes: BABYLON.TransformNode[], meshes: BABYLON.AbstractMesh[], particleSystems: MeshParticleSystem[], materials: BABYLON.Material[], animationGroups: BABYLON.AnimationGroup[], skeletons: BABYLON.Skeleton[]) => {
        this.scene.isDirty = true;

        root.name = this._name;
        this.rootImpl = root;
        this._particleSystems = particleSystems;
        this._meshes = meshes;
        this._animations = animationGroups;
        this._materials = materials;
        this._transformNodes = nodes;
        this._skeletons = skeletons;

        if (skeletons !== undefined) {
            skeletons.forEach((skeleton) => {
                this.skeletonMap.set(skeleton.id, skeleton);
            });
        }

        if (meshes !== undefined) {
            meshes.forEach((item) => {
                let mesh = <BABYLON.Mesh>item;
                this._meshMap.set(mesh.id, mesh);
                // (<any>mesh).rayID = this.rayID;
                if (mesh.material && this.needDepthPrePass != undefined) {
                    // mesh.material.needDepthPrePass = this.needDepthPrePass;
                }

            });
        }

        this.rootImpl.getChildTransformNodes().forEach((item) => {
            if ((<any>item).onPointerObservable == undefined) {
                (<any>item).onPointerObservable = new Observable();
                item.onDisposeObservable.addOnce(() => {
                    (<any>item).onPointerObservable?.clear();
                })
            }
        });
        if ((<any>this.rootImpl).onPointerObservable == undefined) {
            (<any>this.rootImpl).onPointerObservable = new Observable();
            this.rootImpl.onDisposeObservable.addOnce(() => {
                (<any>this.rootImpl).onPointerObservable?.clear();
            })
        }
        (<any>this.rootImpl).onPointerObservable.add(this.onPointer);

        if (this.isDisposed == true) {
            return;
        }

        this.extendLoaded();

        ModelObj.appendAnimation(root, this.option?.appendAnimation)?.forEach((anim) => {
            if (animationGroups == undefined) {
                animationGroups = [];
            }
            this.appendAnimationEnd(anim);
            animationGroups.push(anim);
        });
        if (animationGroups !== undefined) {
            this._animation = animationGroups[0];
            animationGroups.forEach((animGroup) => {
                this._animationMap.set(animGroup.name, animGroup);
            });
            if (this.initialWhileAnimChange) {
                this.animationsInitRecorde();
            }

        }

        if (this.animOpt == undefined && this.animDefault && animationGroups[0]) {
            this.animOpt = {
                animName: animationGroups[0].name,
                isLoop: true,
            };
        }

        this.isLoaded = true;
        this._changePosition();
        this._changeScale();
        this._changeRotate();
        this._changeLookAt();
        this._changeAlpha();
        this._changeEnable();
        this._changeColor();

        this._updateCollider();
        this._updateAttach();
        this._updateAttachOpt();
        this._updateChildOpt();
        this._updateParent();
        // this._changeParentNode();
        this._changeAlphaIndex();
        this._changeRenderingGroupId();
        this._changeAnim();
        this._changeAnimeSpeed();

        this._modifyUseState();
        this._modifyState();

        // if (skeletons !== undefined) {
        //     skeletons.forEach((skeleton) => {
        //         this.skeletonMap.set(skeleton.id, skeleton);
        //     });
        // }

        return;
    }
    protected animEndCall = () => {
        // throw new Error("Method not implemented.");
        console.error("API deprecated !!");
    }
    /**
     * 动画执行结束的回调
     */
    protected animEndCB() {
        // throw new Error("Method not implemented.");
        console.error("API deprecated !!");
    }
    /**
     * 停止模型所有动画 - (场景模型) 同时存在多个动画，不存在同一个模型的动画切换
     * @description 与 start 配对
     * @deprecated
     */
    public stopAllAnim() {
        // throw new Error("Method not implemented.");
        // console.error("API deprecated !!");
    }
    /**
     * 暂停模型所有动画 - (场景模型) 同时存在多个动画，不存在同一个模型的动画切换
     * @description 与 play 配对
     * @deprecated
     */
    public pauseAllAnim() {
        // throw new Error("Method not implemented.");
        // console.error("API deprecated !!");
    }
    /**
     * 继续播放模型所有动画 - (场景模型) 同时存在多个动画，不存在同一个模型的动画切换
     * @description 与 pause 配对
     * @param loop 是否循环
     * @deprecated
     */
    public playAllAnim(loop: boolean = true) {
        // throw new Error("Method not implemented.");
        // console.error("API deprecated !!");
    }
    /**
     * 重新播放模型所有动画 - (场景模型) 同时存在多个动画，不存在同一个模型的动画切换
     * @description 与 pause 配对,从第0帧开始
     * @deprecated
     */
    public restartAllAnim() {
        // throw new Error("Method not implemented.");
        // console.error("API deprecated !!");
    }
    /**
     * 重新播放模型所有粒子
     * @description 与 pause 配对,从第0帧开始
     * @deprecated
     */
    public restartAllParticle() {
        // throw new Error("Method not implemented.");
        // console.error("API deprecated !!");
    }
    /**
     * 启动模型所有动画 - (场景模型) 同时存在多个动画，不存在同一个模型的动画切换
     * @description 与 stop 配对
     * @param loop 是否循环
     * @param speed 播放速度
     * @deprecated
     */
    public startAllAnim(loop: boolean = true, speed?: number) {
        // throw new Error("Method not implemented.");
        // console.error("API deprecated !!");
        if (this.animOpt == undefined) {
            this.animDefault = true;
            if (this.animations?.length > 0) {
                this.animOpt = {
                    animName: this.animations[0].name,
                    isLoop: loop,
                }
            }
        }
        if (speed) {
            this.setAnimeSpeed(speed);
        }
        this._changeAnim();
    }
    /**
     * 停止模型所有粒子
     * @deprecated
     */
    public stopAllParticle() {
        // throw new Error("Method not implemented.");
        // console.error("API deprecated !!");
    }
    /**
     * 启动模型所有粒子
     * @param delay 延时启动时间
     * @deprecated
     */
    public startAllParticle(delay: number = 0) {
        // throw new Error("Method not implemented.");
        // console.error("API deprecated !!");
    }
    /**
     * 附加到指定骨骼
     * @param opt
     */
    public setAttach(opt: IModelAttachOpt) {
        // throw new Error("Method not implemented.");
        // console.error("API deprecated !!");
    }
    /**
     * 添加子模型
     * @param opt
     */
    public setChildModelOpt(opt: IModelChildOpt) {
        // throw new Error("Method not implemented.");
        // console.error("API deprecated !!");
    }

    public setAlpha(data: number) {
        this.alpha = data;
        this._changeAlpha();
    }
    public setPostion(data: number[]) {
        this.position = [data[0], data[1], data[2]];
        this._changePosition();
    }
    public setPosition(x: number | number[], y?: number, z?: number) {
        if (x instanceof Array) {
            this.position = [x[0], x[1], x[2]];
        } else {
            this.position = [x, y, z];
        }
        // this.position = [x, y, z];
        this._changePosition();
    }
    public setScaling(x: number | number[], y?: number, z?: number) {
        if (x instanceof Array) {
            this.scale = [x[0], x[1], x[2]];
        } else {
            this.scale = [x, y, z];
        }
        // this.scale = [x, y, z];
        this._changeScale();
    }
    public setScale(x: number | number[], y?: number, z?: number) {
        this.setScaling(x, y, z);
    }
    public setRotation(x: number | number[], y?: number, z?: number) {
        if (x instanceof Array) {
            this.rotate = [x[0], x[1], x[2]];
        } else {
            this.rotate = [x, y, z];
        }
        // this.rotate = [x, y, z];
        this._changeRotate();
    }
    public setRotate(x: number | number[], y?: number, z?: number) {
        this.setRotation(x, y, z);
    }
    public setLookAt(x: number | number[], y?: number, z?: number) {
        if (x instanceof Array) {
            this.lookAt = [x[0], x[1], x[2]];
        } else {
            this.lookAt = [x, y, z];
        }
        this._changeLookAt();
    }
    public setColor(r: number, g: number, b: number) {
        this.color = [r, g, b];
        this._changeColor();
    }
    // protected _parent: IPrefabRes;
    // protected _indexOfParent: number;
    // public setParent(parent: IPrefabRes, indexOfParent?: number): void {
    //     this._parent = parent;
    //     this._indexOfParent = indexOfParent;
    //     this._changeParentNode();
    // }
    /**
     * @deprecated
     * @deprecated use setEnabled
     * @param data 
     */
    public setVisible(data: boolean) {
        throw new Error("Method not implemented. Please Use `setEnabled`.");
    }

    public setEnabled(isEnabled: boolean) {
        this._isEnabled = isEnabled;
        this._changeEnable();
    }
    /**
     * 启动指定某个动画 - （模型）一次只播一个动画
     * @param animOpt
     */
    public setAnim(animOpt: IPrefabAnime) {
        this.animOpt = animOpt;
        if (animOpt && animOpt.speed) {
            this.setAnimeSpeed(animOpt.speed);
        }
        this._changeAnim();
    }
    set alphaIndex(param: number) {
        this._alphaIndex = param;
        this._changeAlphaIndex();
    }
    public get alphaIndex() {
        // debugger;
        return this._alphaIndex;
    }
    public get renderingGroupId() {
        return this._renderingGroupId;
    }
    public set renderingGroupId(param: number) {
        // debugger;
        this._renderingGroupId = param;
        this._changeRenderingGroupId();
    }

    public renderQueueItems(): IMeshRenderQueue[] {
        let result: IMeshRenderQueue[] = [];
        this._meshes?.forEach((item) => {
            result.push(item);
        });
        this._particleSystems?.forEach((item) => {
            result.push(item);
            if (item.psTool?.trail?.mesh) {
                result.push(item.psTool.trail.mesh);
            }
        });

        return result;
    }

    public renderMaterialItems(): IMeshMaterial[] {
        let result: IMeshMaterial[] = [];
        this._meshes?.forEach((item) => {
            result.push(<IMeshMaterial>item);
        });
        this._particleSystems?.forEach((item) => {
            result.push(<IMeshMaterial>item);
        });

        return result;
    }

    public pickAboutItems(): INodeCollider[] {
        let result: INodeCollider[] = [];
        this._meshes?.forEach((item) => {
            result.push(<any>item);
        });

        return result;
    }

    protected _animeSpeed = 1.;
    setAnimeSpeed(param: number): void {
        this._animeSpeed = param;
        this._changeAnimeSpeed();
    }
    /**
     * 停止当前的动画 - （模型）一次只播一个动画
     */
    public stopAnim() {
        if (this.workWithAnimationGroupChange) {
            return;
        }
        if (this.isLoaded && !this.isDisposed) {
            this.stopAnimation();
        } else {
            this.animOpt = undefined;
        }
    }
    /**
     * 暂停当前的动画 - （模型）一次只播一个动画
     */
    public pauseAnim() {
        if (this.isLoaded && !this.isDisposed) {
            this.pauseAnimation();
        } else {
            this.animOpt = undefined;
        }
    }
    // protected _changeParentNode() {
    //     if (!this.isLoaded || this.isDisposed) { return }
    //     this.rootImpl.parent = this._parent?.getChild(this._indexOfParent);
    // }
    protected _changeAlpha() {
        if (!this.isLoaded || this.isDisposed) { return }
        if (this.alpha !== undefined) {
            // this.materials?.forEach((material) => {
            //     material.alpha = this.alpha;
            // });
            this.meshMap.forEach((mesh) => {
                mesh.visibility = this.alpha;
            });
        }
    }
    protected _changeColor(): void {
        if (!this.isLoaded || this.isDisposed) { return }
        if (this.color) {
            // this.materials?.forEach((material) => {
            //     material.mainColor(new BABYLON.Color3(...this.color));
            // });
        }
    }
    protected _changePosition() {
        if (!this.isLoaded || this.isDisposed) { return }
        if (this.position) {
            NodeTools.positNode(this.rootImpl, this.position);
        }
    }
    protected _changeScale() {
        if (!this.isLoaded || this.isDisposed) { return }
        if (this.scale) {
            NodeTools.scaleNode(this.rootImpl, this.scale);
        }
    }
    protected _changeRotate() {
        if (!this.isLoaded || this.isDisposed) { return }
        if (this.rotate) {
            NodeTools.rotateNode(this.rootImpl, this.rotate);
        }
    }
    protected _changeLookAt() {
        if (!this.isLoaded || this.isDisposed) { return }
        if (this.lookAt) {
            NodeTools.nodeLookAt(<BABYLON.TransformNode>this.rootImpl, this.lookAt);
        }
    }
    /**
     * @deprecated
     */
    protected _changeVisible() {
        this._changeEnable();
    }

    protected _changeEnable() {
        if (!this.isLoaded || this.isDisposed) { return }
        if (this._isEnabled !== undefined) {
            if (this._isEnabled == true && this.rootImpl.isEnabled() != this._isEnabled) {
                this._changeAnim();
            }
        }
        this.rootImpl.setEnabled(this.isEnabled());
    }
    protected _updateAttachOpt() {
        // let useCount = 0;
        // let tempIndex = 0;
        // let newListLen = 0;
        // const listLen = this.attachOptList.length;

        // for (let len = listLen - 1; len >= 0; len--) {
        //     const opt = this.attachOptList[len];

        //     if (opt.mesh !== undefined) {
        //         const skeleton = this.skeletonMap.get(opt.skeletonName);

        //         if (skeleton !== undefined) {
        //             const boneIndex = skeleton.getBoneIndexByName(opt.boneName);
        //             if (boneIndex >= 0) {
        //                 // opt.mesh.attachToBone(skeleton.bones[boneIndex], this.rootImpl);
        //                 NodeTools.nodeTransform(opt.mesh, opt.transform);
        //             }
        //         }

        //         this.attachOptList[len] = undefined;
        //         useCount++;
        //     }
        // }

        // newListLen = listLen - useCount;
        // for (let i = 0, k = listLen - useCount; i < k; i++) {
        //     while (this.attachOptList[tempIndex] === undefined && tempIndex < listLen) {
        //         tempIndex++;
        //     }

        //     if (this.attachOptList[tempIndex] !== undefined) {
        //         this.attachOptList[i] = this.attachOptList[tempIndex];
        //     }
        // }

        // this.attachOptList.length = newListLen;
    }
    protected _updateChildOpt() {
        // this.childOptList.forEach((opt) => {
        //     if (opt.isFinished !== true) {
        //         if (opt.model !== undefined) {
        //             if (opt.model.rootImpl !== undefined) {
        //                 opt.isFinished = true;
        //                 opt.model.rootImpl.parent = this.rootImpl;
        //                 NodeTools.nodeTransform(opt.model.rootImpl, opt.transform);
        //             }
        //         } else if (opt.mesh !== undefined) {
        //             opt.isFinished = true;
        //             opt.mesh.parent = this.rootImpl;
        //             NodeTools.nodeTransform(opt.mesh, opt.transform);
        //         }
        //     }
        // });
    }
    protected _updateAttach() {
        //
    }
    protected _changeAnim() {
        if (!this.isLoaded || this.isDisposed) { return }
        if (this.animOpt !== undefined) {
            if (this._animation) {
                this._stopAnimation(this._animation);
                this._animation = undefined;
            }

            let anime = this.animationMap.get(this.animOpt.animName);
            if (anime) {
                this.animationsInitUse();

                if (this.animOpt.startCall) {
                    anime.onAnimationGroupPlayObservable.addOnce(this.animOpt.startCall);
                }
                if (this.animOpt.endCall) {
                    anime.onAnimationGroupEndObservable.addOnce(this.animOpt.endCall);
                }
                if (this.animOpt.loopCall) {
                    anime.onAnimationGroupLoopObservable.addOnce(this.animOpt.loopCall);
                }
                // if (this.animOpt.frameCall) {
                //     anime.onAnimationGroupFrameEventObservable.addOnce(this.animOpt.frameCall);
                // }
                if (this.animOpt.from || this.animOpt.to) {
                    anime.start(this.animOpt.isLoop, undefined, this.animOpt.from, this.animOpt.to);
                } else {
                    anime.play(this.animOpt.isLoop);
                }
                anime.speedRatio = this._animeSpeed;
            }
            this._animation = anime;
        } else {
            if (this._animation) {
                this._stopAnimation(this._animation);
                this._animation = undefined;
            }
        }
    }
    protected _changeAlphaIndex() {
        if (this.isLoaded && this._meshes && this._alphaIndex != undefined) {
            let min = Number.MAX_SAFE_INTEGER;
            this._meshes.forEach((item) => {
                min = Math.min(min, item.alphaIndex);
            });
            this._particleSystems.forEach((item) => {
                min = Math.min(min, item.alphaIndex);
                if (item.psTool?.trail?.mesh) {
                    min = Math.min(min, item.psTool.trail.mesh.alphaIndex);
                }
            });
            this._meshes.forEach((item) => {
                item.alphaIndex = item.alphaIndex - min + this._alphaIndex;
            });
            this._particleSystems.forEach((item) => {
                item.alphaIndex = item.alphaIndex - min + this._alphaIndex;
                if (item.psTool?.trail?.mesh) {
                    item.psTool.trail.mesh.alphaIndex = item.alphaIndex - min + this._alphaIndex;
                }
            });
        }
    }
    protected _changeRenderingGroupId() {
        if (this.isLoaded && this._meshes && this._renderingGroupId != undefined) {
            this._meshes.forEach((item) => {
                item.renderingGroupId = this._renderingGroupId;
            });
            this._particleSystems.forEach((item) => {
                item.renderingGroupId = this._renderingGroupId;
                if (item.psTool?.trail?.mesh) {
                    item.psTool.trail.mesh.renderingGroupId = this._renderingGroupId;
                }
            });
        }
    }
    protected _changeAnimeSpeed() {
        if (!this.isLoaded || this.isDisposed) { return }
        if (this.animation) {
            this.animation.speedRatio = this._animeSpeed;
        }
        this._particleSystems?.forEach((item) => {
            item.speedRatio = this._animeSpeed;
        });
    }
    protected parent: Model = undefined;
    protected parentIndex: number;
    setParent(parent: IModel, parentIndex?: number): void {
        this.parent = <Model>parent;
        this.parentIndex = parentIndex;
        this._updateParent();
    }
    getChild(index?: number): BABYLON.TransformNode {
        if (!this.isLoaded || this.isDisposed) { return }
        if (index != undefined) {
            return <BABYLON.TransformNode>this.transformNodes[index];
        } else {
            return this.rootImpl;
        }
    }
    protected _updateParent() {
        if (this.isLoaded && !this.isDisposed && this.rootImpl) {
            if (this.parent && !this.isDisposed) {
                if (this.parent.isLoaded) {
                    this.rootImpl.parent = this.parent.getChild(this.parentIndex);
                } else {
                    let parent = this.parent;
                    this.parent.loadPromise.then((model) =>{
                        if (parent == model) {
                            this.rootImpl.parent = parent?.getChild(this.parentIndex);
                        }
                    });
                }
            } else {
                this.rootImpl.parent = null;
            }
        }
    }
    /**
     * 更改模型动画
     * @param animName 目标动画名称
     * @param isLoop 是否循环播放
     * @param stopFlag 动画停止配置
     * @param endCall 动画结束回调
     */
    // protected changeAnimation = () => {
    //     this._animation = this._changeAnimation(this.animOpt, false);
    //     this.isAnimationLoop = this.animOpt.isLoop;
    // }
    // protected _changeAnimation(animOpt: IModelAnimOpt, forBoneAttach: boolean = false) {

    //     const speed = animOpt.speed === undefined ? 1 : animOpt.speed;
    //     const animName = animOpt.animName + (forBoneAttach ? BABYLON.Skeleton.ANIME_NAME_SUFFIX_WHILE_BAKED : "");
    //     const isLoop = !!animOpt.isLoop;

    //     const animGroup = this.animationMap.get(animName);


    //     if (animGroup !== undefined) {
    //         if (this.changeAnimEnableBlending) {
    //             animGroup.enableBlend(false, this.changeAnimBlendUseSeconds);
    //         }

    //         if (animGroup.isStarted && isLoop) {
    //             logger.warn(`${animName} - Animation is played!`);
    //             animGroup.onAnimationGroupEndObservable.clear();
    //             animGroup.stop();
    //         }

    //         animGroup.speed = speed;
    //         animGroup.start(isLoop, null);
    //     } else {
    //         if (!forBoneAttach) {
    //             logger.warn(`${animName} - Animation Do Not Exist!`);
    //         }
    //     }

    //     return animGroup;
    // }
    protected stopAnimation = () => {
        this._stopAnimation(this._animation);
    }
    protected _stopAnimation = (animation: BABYLON.AnimationGroup) => {
        if (animation && this.changeAnimEnableBlending && animation.isStarted) {
            // animation.enableBlend(true, this.changeAnimBlendUseSeconds);
        } else {
            if (animation) {
                animation.onAnimationGroupEndObservable.clear();
                // animation.onAnimationGroupPlayObservable.clear();
                animation.stop();
            }
        }
    }
    protected pauseAnimation = () => {
        this._animation && this._animation.pause();
    }
    protected initAnimationEventMap() {
        // throw new Error("Method not implemented.");
        // console.error("API deprecated !!");
        // 动画事件
        this.animationMap.forEach((ag) => {
            if (ag.animatables) {
                ag.animatables.forEach((at) => {
                    const list = at.getAnimations();
                    list && list.forEach((a) => {
                        const events = (<BABYLON.AnimationEvent[]>(<any>a)._events);
                        events.forEach((event) => {
                            if (!this.animEventMap) {
                                this._animEventMap = new Map();
                            }
                            this.animEventMap.set(event._pi_eventKey, { event, group: ag, idxs: [] });
                        });
                    });
                });
            }
        });
    }
    // -------------------------------------------------------------------- 克隆
    /**
     * 以当前模型 进行克隆
     * @deprecated
     * @param newName 新的名称
     * @param rayID rayID
     */
    public clone(newName: string, rayID: number = -1) {
        // throw new Error("Method not implemented.");
        console.error("API deprecated !!");
    }
    /**
     * @deprecated
     * @param newName 
     * @param rayID 
     */
    protected _clone(newName: string, rayID: number = -1) {
        // throw new Error("Method not implemented.");
        console.error("API deprecated !!");
    }
    /**
     * @deprecated
     * @param oldParent 
     * @param newParent 
     */
    protected _instanceChild(oldParent: BABYLON.TransformNode, newParent: BABYLON.TransformNode) {
        // throw new Error("Method not implemented.");
        console.error("API deprecated !!");
    }
    /**
     * 克隆动画 - 动画需要作用的目标对象
     * @deprecated
     * @param newNodes 节点
     * @param newSkeletons 骨骼
     * @param newMaterials 材质
     * @param newTextures 纹理
     */
    protected _cloneAnimation(newNodes: { metadata: any }[], newSkeletons: { metadata: any }[], newMaterials: { metadata: any }[], newTextures: { metadata: any }[], bones: { metadata: any }[]) {

        // throw new Error("Method not implemented.");
        console.error("API deprecated !!");
    }
    /**
     * @deprecated
     * @param newMeshes 
     */
    protected _cloneSkeleton(newMeshes: BABYLON.AbstractMesh[]) {
        // throw new Error("Method not implemented.");
        console.error("API deprecated !!");
    }
    /**
     * @deprecated
     * @param newNodes 
     */
    protected _cloneParicle(newNodes: BABYLON.TransformNode[]) {
        // throw new Error("Method not implemented.");
        console.error("API deprecated !!");
    }
    /**
     * @deprecated
     */
    protected animationsInitRecordMap: Map<any, Map<string, any>> = new Map();
    /**
     * @deprecated
     */
    protected animationsInitRecorde() {
        this.animationMap.forEach((animGroup) => {
            animGroup.targetedAnimations.forEach((targetAnim) => {
                const target = targetAnim.target;
                if (targetAnim.animation.targetProperty) {
                    const targetPath = targetAnim.animation.targetProperty;
                    this.animationInitRecorde(target, targetPath);
                }
                else if (targetAnim.animation.targetPropertyPath) {
                    const paths = targetAnim.animation.targetPropertyPath;
                    const len = paths.length;
                    for (let i = 0; i < len; i++) {
                        this.animationInitRecorde(target, paths[i]);
                    }
                }

            });
        });

    }
    /**
     * @deprecated
     * @param target 
     * @param path 
     */
    protected animationInitRecorde(target: any, path: string) {
        const paths = path.split('.');
        let data: any = target;

        const len = paths.length;
        if (len > 0) {
            if (data instanceof Array) {
                data.forEach((target) => {
                    let targetProperty = target;
                    for (let i = 0; i < len; i++) {
                        targetProperty = targetProperty[paths[i]];
                    }

                    let map = this.animationsInitRecordMap.get(target);
                    if (!map) {
                        map = new Map();
                        this.animationsInitRecordMap.set(target, map);
                    }

                    map.set(path, targetProperty);
                });
            } else {
                let targetProperty = data;
                for (let i = 0; i < len; i++) {
                    targetProperty = targetProperty[paths[i]];
                }

                let map = this.animationsInitRecordMap.get(target);
                if (!map) {
                    map = new Map();
                    this.animationsInitRecordMap.set(target, map);
                }

                map.set(path, targetProperty);
            }
        }

    }
    /**
     * @deprecated
     */
    protected animationsInitUse() {
        this.animationsInitRecordMap.forEach((map, target) => {
            if (map) {
                map.forEach((data, path) => {
                    this.animationInitUse(target, path, data);
                });
            }
        });

    }
    /**
     * @deprecated
     * @param target 
     * @param path 
     * @param data 
     */
    protected animationInitUse(target: any, path: string, data: any) {
        const paths = path.split('.');
        const len = paths.length;
        if (len > 0) {
            for (let i = 0; i < len - 1; i++) {
                target = target[paths[i]];
            }

            target[paths[len - 1]] = data;
        }

    }
    
    /**
     * loaded逻辑中需要在父节点设置之前
     * @returns 
     */
    protected _updateCollider() {
        if (this.isDisposed) return;
        if (this.isLoaded && this._isPickable) {
            // 不是精灵内部自动创建的盒子
            if (this._colliderBoundle == undefined) {
                if (this._colliderCustom) {
                    this._colliderBoundle = this._colliderCustom;
                    this.rootImpl.onAfterWorldMatrixUpdateObservable.add(() => {
                        this._colliderBoundle.update(this.worldMatrixFromCache);
                    });
                } else {
                    let min = new BABYLON.Vector3(-0.5, -0.5, -0.5);
                    let max = new BABYLON.Vector3( 0.5,  0.5,  0.5);
                    // this.meshes.forEach((mesh) => {
                    //     if (mesh._boundingInfo) {
                            
                    //     }
                    // });
                    this._colliderCustom = new PiCollider(min, max);
                    this._colliderBoundle = this._colliderCustom;
                }
            }
            this.scene.addCollider(this);
            super._updateCollider();
        }
    }
    public setAnimeFrame(animName: string, progress: number): IModel {
        if (this.isLoaded && !this.isDisposed) {
            let anime = this.animationMap.get(animName);
            if (anime != this.animation) {
                this.stopAnimation();
                this._animation = anime;
            }
            if (anime) {
                if (!anime.isStarted) {
                    anime.play(true);
                }
                anime.pause();
                anime.goto(progress);
            }
        }
        return this;
    }
}
