
import { Model, ModelObj } from "./model_obj";
import { BABYLON } from "./render3d/babylon";
import { SceneStruct } from "./scene_struct";
import { IPrefab, IPrefabCreation, IPrefabPool, IPrefabType, IPrefabID } from "./interface/modelprefab";
import { IModel, IModelOpt, IPrefabAnime } from "./interface/modelobj";
import { PointerInfo } from "./render/events/pointerEvents";
import { EventState, Observable } from "./interface/observable";
import { CustomScene } from "./index";

export interface IPrefabRes {
    /**
     * 创建时的拓展数据 - 自定义作用
     */
    extendData: any;
    _cached: boolean;
    unuse(): void;
    reuse(): void;
    dispose(): void;
}

export class NodeObj implements IPrefab {
    public static PrefabInfo: IPrefabType = "Transform";
    public static PrefabCreation: IPrefabCreation = { single: false, cacheCount: 128 };
    public static CheckIsNode(info: IPrefabType) {
        return info == NodeObj.PrefabInfo;
    }
    public static baseRate: number = 100;
    protected _info: IPrefabType;
    public get info() {
        return this._info;
    }
    protected _id: IPrefabID;
    protected _data: Model;
    protected _isDisposed: boolean = false;
    public get isDisposed() {
        return this._isDisposed;
    }
    protected childs: IPrefab[] = [];
    
    extendData: any;
    pool: PrefabModelObjPool;
    protected _loadPromise: Promise<NodeObj>;
    get scene() {
        return this._data?.scene;
    };
    get animationMap() {
        return this._data?.animationMap;
    }
    get loadPromise() {
        if (this._loadPromise == undefined) {
            if (this._isDisposed) {
                this._loadPromise = Promise.reject("Disposed!!");
            } else {
                // if (this._data.isLoaded) {
                //     this._loadPromise = Promise.resolve().then(() => {
                //         this._loaded();
                //         return (this);
                //     });
                // } else 
                if (this._data.loadPromise) {
                    this._loadPromise = this._data.loadPromise.then(() => {
                        this._loaded();
                        return (this);
                    });
                } else {
                    //
                }
            }
        }
        return this._loadPromise;
    }
    get name(): string {
        return this._data?.name;
    }
    get animations(): { name: string; }[] {
        return this._data?.animations;
    }
    constructor(scene: SceneStruct, info: IPrefabType, pool: PrefabModelObjPool, data: Model) {
        this._id = scene.prefabID();
        this._data = data;
        this.pool = pool;
        this._info = info;
        pool.Prefabs.set(this, this);
    }
    get alphaIndex(): number {
        return this._data.alphaIndex;
    }
    set alphaIndex(val: number) {
        this._data.alphaIndex = val;
    }
    setAlphaIndex(val: number): NodeObj {
        this._data.alphaIndex = val;
        return this;
    }
    get renderingGroupId(): number {
        return this._data.renderingGroupId;
    }
    set renderingGroupId(val: number) {
        this._data.renderingGroupId = val;
    }
    setRenderingGroupId(val: number): NodeObj {
        this._data.renderingGroupId = val;
        return this;
    }
    get resource() {
        return this._data;
    }
    get meshMap() {
        return undefined;
    }
    get rootImpl() {
        return this._data?.rootImpl;
    }
    get position() {
        return this._data?.position || [0, 0, 0];
    }
    get scale() {
        return this._data?.scale || [1,1,1];
    }
    get rotate() {
        return this._data?.rotate || [0, 0, 0];
    }
    get animOpt() {
        return this._data?.animOpt;
    }
    protected _rayID: any;
    get rayID() {
        return this._rayID;
    }
    set rayID(val: any) {
        if (this._data) {
            this._data.rayID = val;
        }
        this._rayID = val;
    }
    setRayID(val: number): NodeObj {
        this.rayID = val;
        return this;
    }
    
    protected _parent: IPrefab = undefined;
    public get parent() {
        return this._parent;
    }
    protected _parentIndex: number;

    public onDisposeObservable = new Observable<NodeObj>();
    isEnabled(): boolean {
        return this._data.isEnabled();
    }
    loadAsync() {

    }
    pickAboutItems() {
        return this._data?.pickAboutItems();
    }
    public static Creation(scene: SceneStruct, info: IPrefabType, pool: PrefabModelObjPool, model: ModelObj) {
        let result = new NodeObj(scene, info, pool, model);
        let p = result.loadPromise;
        return result;
    }
    public static CreationResource(name: string, scene: CustomScene) {
        let model = new ModelObj(name, scene);
        let rootImpl = new BABYLON.TransformNode(name, scene.impl);
        model.rootImpl = rootImpl;
        (<any>model)._loadPromise = Promise.resolve().then(() => {
            (<any>model).loaded(rootImpl, [rootImpl], [], [], [], [], [], []);
            return model;
        });

        return model;
    }
    protected _loaded() {
        this._data?.setEnabled(this._isEnabled);
    }
    pauseAnim(): NodeObj {
        // this.resource?.pauseAnim();
        return this;
    }
    addAnimationEventListen(key: string, func: () => boolean): NodeObj {
        this.resource?.addAnimationEventListen(key, func);
        return this;
    }
    getChild(index?: number): BABYLON.TransformNode {
        return this._data?.getChild(index);
    }
    setParent(parent: IPrefab, indexOfParent?: number): NodeObj {
        this._parent = parent;
        this._parentIndex = indexOfParent;
        this._data?.setParent(parent?.resource, indexOfParent);
        return this;
    }
    setPosition(x: number, y: number, z: number): NodeObj {
        this._data?.setPosition(x, y, z);
        return this;
    }
    getPosition(): number[] {
        return this._data?.position || [0, 0, 0]
    }
    setScaling(x: number, y: number, z: number): NodeObj {
        this._data?.setScaling(x, y, z);
        return this;
    }
    setScale(x: number, y: number, z: number): NodeObj {
        this._data?.setScale(x, y, z);
        return this;
    }
    setRotation(x: number, y: number, z: number): NodeObj {
        this._data?.setRotation(x, y, z);
        return this;
    }
    setRotate(x: number, y: number, z: number): NodeObj {
        this._data?.setRotate(x, y, z);
        return this;
    }
    setAlpha(param: number): NodeObj {
        this._data?.setAlpha(param);
        return this;
    }
    setColor(r: number, g: number, b: number): NodeObj {
        this._data?.setColor(r, g, b);
        return this;
    }
    setAnim(param: IPrefabAnime): NodeObj {
        this._data?.setAnim(param);
        return this;
    }
    protected _isEnabled: boolean = false;
    setEnabled(param: boolean): NodeObj {
        this._isEnabled = param;
        if (this._data?.isLoaded) {
            this._data?.setEnabled(this._isEnabled);
        }
        // if (param) {
        //     this._data?.startAllAnim(false);
        // }
        return this;
    }
    setLookAt(x: number, y: number, z: number): NodeObj {
        this._data?.setLookAt(x, y, z);
        return this;
    }
    // setRenderQueue(param: number): void {
    //     this._data?.setRenderQueue(param);
    // }
    setAnimeSpeed(param: number): NodeObj {
        this._data?.setAnimeSpeed(param);
        return this;
    }
    resourceUnuse(): NodeObj {
        if (this._data) {
            this._data.setParent(null);
            this._data.setPosition(0, 0, 0);
            this._data.setRotate(0, 0, 0);
            this._data.setScaling(1, 1, 1);
            this._data.setAnim(undefined);
            this._data.alphaIndex = 0;
            this._data.setAnimeSpeed(1);
        }
        return this;
    }
    getDistanceToCamera(camera: BABYLON.Camera) {
        if (this.rootImpl) {
            return this.rootImpl.getAbsolutePosition().subtract(camera.globalPosition).length();
        } else {
            return Number.MAX_SAFE_INTEGER;
        }
    }
    dispose(): void {
        if (this._isDisposed) { return; }
        this.onDisposeObservable.notifyObservers(this);
        this.onDisposeObservable.clear();
        this.childs.forEach((item) => {
            item.dispose();
        });

        this.resourceUnuse();
        this.pool.Prefabs.delete(this);
        this.pool.recyclePrefab(this._info, this._data);

        this.childs = undefined;
        this._id = undefined;
        this._info = undefined;
        this._data = undefined;
        this.pool = undefined;
        this._id = undefined;
        this._isDisposed = true;
        // throw new Error("Method not implemented.");
    }
    id(): number {
        return this._id;
    }
    renderQueueItems() {
        return this._data?.renderQueueItems();
    }
    renderMaterialItems() {
        return this._data?.renderMaterialItems();
    }
    stopAnim(): NodeObj {
        this._data?.stopAnim();
        return this
    }
    isDispose(): boolean {
        return this._isDisposed;
    }
    isLoadDone(): boolean {
        return this._data.isLoaded;
    }
    /**
     * @deprecated
     */
    startAllAnim(): void {
        // this._data?.startAllAnim();
    }
    /**
     * 停止模型所有动画 - (场景模型) 同时存在多个动画，不存在同一个模型的动画切换
     * @description 与 start 配对
     * @deprecated
     */
    stopAllAnim() {
        // throw new Error("Method not implemented.");
        // console.error("API deprecated !!");
    }
    /**
     * 停止模型所有粒子
     * @deprecated
     */
    stopAllParticle() {
        // throw new Error("Method not implemented.");
        // console.error("API deprecated !!");
    }
    /**
     * 暂停模型所有动画 - (场景模型) 同时存在多个动画，不存在同一个模型的动画切换
     * @description 与 play 配对
     * @deprecated
     */
    public pauseAllAnim() {
        // throw new Error("Method not implemented.");
        // console.error("API deprecated !!");
    }
    /**
     * 继续播放模型所有动画 - (场景模型) 同时存在多个动画，不存在同一个模型的动画切换
     * @description 与 pause 配对
     * @param loop 是否循环
     * @deprecated
     */
    public playAllAnim(loop: boolean = true) {
        // throw new Error("Method not implemented.");
        // console.error("API deprecated !!");
    }
    /**
     * 重新播放模型所有动画 - (场景模型) 同时存在多个动画，不存在同一个模型的动画切换
     * @description 与 pause 配对,从第0帧开始
     * @deprecated
     */
    public restartAllAnim() {
        // throw new Error("Method not implemented.");
        // console.error("API deprecated !!");
    }
    /**
     * 重新播放模型所有粒子
     * @description 与 pause 配对,从第0帧开始
     * @deprecated
     */
    public restartAllParticle() {
        // throw new Error("Method not implemented.");
        // console.error("API deprecated !!");
    }
    /**
     * 重新播放模型所有粒子
     * @description 与 pause 配对,从第0帧开始
     * @deprecated
     */
    public startAllParticle() {
        // throw new Error("Method not implemented.");
        // console.error("API deprecated !!");
    }
    /**
     * 添加事件监听
     * * 如果不是SpriteModel需要手动调用 collider 设置命中盒子
     * @param etype 
     * @param func 
     */
    addEventListener(etype: "pointerup" | "pointerdown", func: Function) {
        this._data?.addEventListener(etype, func);
    }
    removeEventListener(etype: "pointerup" | "pointerdown", func: Function) {
        this._data?.removeEventListener(etype, func);
    }
    /**
     * 默认值为 (-0.5, -0.5, -0.5) 到 (0.5, 0.5, 0.5)
     * @param min 
     * @param max 
     */
    collider(min?: BABYLON.Vector3, max?: BABYLON.Vector3) {
        this._data?.collider(min, max);
    }
    colliderSphereTreshold(val: number) {
        this._data?.colliderSphereTreshold(val);
    }
    modifyState(state: any) {
        this._data?.modifyState(state);
    }
    setAnimeFrame(animName: string, progress: number): IModel {
        this._data?.setAnimeFrame(animName, progress);
        return this;
    }
}

export class ModelPrefab extends NodeObj {
    protected static DIRTY_POSITION = 0b0000_0000_0000_0001;
    protected static DIRTY_SCALE = 0b0000_0000_0000_0010;
    protected static DIRTY_ROTATION = 0b0000_0000_0000_0100;
    protected static DIRTY_ANIME = 0b0000_0000_0000_1000;
    protected static DIRTY_ENABLE = 0b0000_0000_0001_0000;
    protected static DIRTY_COLOR = 0b0000_0000_0010_0000;
    protected static DIRTY_ALPHA = 0b0000_0000_0100_0000;
    protected static DIRTY_RENDERQUEUE = 0b0000_0000_1000_0000;
    protected static DIRTY_ANIMESPEED = 0b0000_0001_0000_0000;
    protected static DIRTY_PARENT = 0b0000_0010_0000_0000;
    protected _isLoaded: boolean;
    constructor(scene: SceneStruct, info: IPrefabType, pool: PrefabModelObjPool, model: ModelObj) {
        super(scene, info, pool, model);
        this._loadPromise = new Promise(
            (resolve, reject) => {
                if (this._isDisposed) {
                    reject("Disposed!!");
                } else {
                    // if (this._data.isLoaded) {
                    //     this._loaded();
                    //     resolve(this);
                    // } else {
                        this._data.loadPromise.then(() => {
                            this._loaded();
                            resolve(this);
                        })
                        .catch(e => {
                            console.log(this.name);
                            reject(e);
                        });
                    }
                // }
            }
        );
    }

    public static Creation(scene: SceneStruct, info: IPrefabType, pool: PrefabModelObjPool, model: ModelObj) {
        return new ModelPrefab(scene, info, pool, model);
    }
    public static CreationResource(name: string, scene: SceneStruct, param: IModelOpt = {}) {
        return ModelObj.Creation(name, <any>scene, param);
    }
    protected _loaded = () => {
        if (this._isDisposed) {
            return;
        }
        if (this._data) {
            this._isLoaded = true;
            this._data.setEnabled(this._isEnabled);
        }
        return this;
    }
    protected px: number = 0;
    protected py: number = 0;
    protected pz: number = 0;
    protected sx: number = 1;
    protected sy: number = 1;
    protected sz: number = 1;
    protected rx: number = 0;
    protected ry: number = 0;
    protected rz: number = 0;
    protected r: number = 1;
    protected g: number = 1;
    protected b: number = 1;
    protected alpha: number = 1;
    protected enable: boolean = false;
    protected anime: IPrefabAnime = undefined;
    protected animeSpeed: number = 1.0;
    protected lastusequeue: number = 0;
    protected renderqueue: number = 0;
    declare protected _data: ModelObj;
    dispose(): void {
        if (this._isDisposed) { return; }
        this.setPosition(0, 0, 0);
        this.setScaling(1, 1, 1);
        this.setRotate(0, 0, 0);
        this.setAnim(undefined);
        this.alphaIndex = 0;
        this.setAnimeSpeed(1);
        super.dispose();
        // this._isLoaded = false;
    }

    get fileName(): string {
        return this._data?.fileName;
    }
    get path(): string {
        return this._data?.path;
    }
    renderQueueItems() {
        return this._data?.renderQueueItems();
    }
    renderMaterialItems() {
        return this._data?.renderMaterialItems();
    }
    stopAnim(): ModelPrefab {
        this._data?.stopAnim();
        return this
    }
    pauseAnim(): ModelPrefab {
        this.resource?.pauseAnim();
        return this;
    }
    get resource() {
        return this._data;
    }
    get meshMap() {
        return this.resource?.meshMap;
    }
    setAlpha(param: number): ModelPrefab {
        this._data?.setAlpha(param);
        // this._data?.meshMap.forEach((mesh) => {
        //     if (mesh) {
        //         // // 自定义材质的透明度变化
        //         // if (mesh instanceof BABYLON.InstancedMesh) {
        //         //     mesh.material.alpha = param;
        //         // } else if (mesh.material) {
        //         //     mesh.material.alpha = param;
        //         // } else {
        //         //     // 非自定义材质的透明度变化
        //         //     // mesh.visibility = param;
        //         // }
        //         ((<StandardMaterialMesh>mesh).materialData || <BABYLON.StandardMaterial>mesh.material).alpha = param;
        //     }
        // });
        return this;
    }
}

export class PrefabModelObjPool implements IPrefabPool {
    public Prefabs: Map<any, any> = new Map();
    public get cfgmap(): Map<IPrefabType, IPrefabCreation> {
        return PrefabModelObjPool.cfgMap;
    }
    private static SingleCounter: number = Number.MIN_SAFE_INTEGER;
    public static SpecialPrefabID(): string {
        return `${this.SingleCounter++}`;
    }
    public static readonly cfgMap: Map<IPrefabType, IPrefabCreation> = new Map();
    public static factoryMap: Map<IPrefabType, (scene: SceneStruct, info: IPrefabType, pool: PrefabModelObjPool, model: IPrefabRes) => IPrefab> = new Map();
    public static resFactoryMap: Map<IPrefabType, (name: any, scene: SceneStruct, param?: any) => IPrefabRes> = new Map();
    public static registPrefabFactory<T extends IPrefab>(key: IPrefabType, func: (scene: SceneStruct, info: IPrefabType, pool: PrefabModelObjPool, model: IPrefabRes) => T) {
        this.factoryMap.set(key, func);
    }
    public static registPrefabResFactory<T extends IPrefabRes, I>(key: IPrefabType, creation: (name: I, scene: SceneStruct, param?: any) => T) {
        this.resFactoryMap.set(key, creation);
    }
    public unuse_pools: Map<IPrefabType, IPrefabRes[]> = new Map();
    public pools: Map<IPrefabType, IPrefabRes[]> = new Map();
    public single_pools: Map<IPrefabType, IPrefabRes> = new Map();
    public others: Map<IPrefabRes, IPrefabRes> = new Map();
    public scene: SceneStruct;
    constructor(scene: SceneStruct) {
        this.scene = scene;
        this.scene.onBeforeDispose(this._dispose);
        this.registPrefabCfg(NodeObj.PrefabInfo, NodeObj.PrefabCreation);
        PrefabModelObjPool.registPrefabFactory(NodeObj.PrefabInfo, NodeObj.Creation);
        PrefabModelObjPool.registPrefabResFactory(NodeObj.PrefabInfo, NodeObj.CreationResource);
    }
    // defaultPrefabResFactory(name: any, scene: SceneStruct, param?: any): IPrefabRes {
    //     return undefined;
    // }
    recyclePrefab(info: IPrefabType, val: IPrefabRes): void {
        let key = info;
        if (val._cached) {
            let unuselist = this.unuse_pools.get(key);
            if (unuselist) {
                unuselist.push(val);
            }
            val.unuse();
        } else {
            val.dispose();
            this.others.delete(val);
        }
    }
    registPrefabCfg<IC extends IPrefabCreation>(info: IPrefabType, param: IC): void {
        this.cfgmap.set(info, param);
    }
    createPrefab<T extends IPrefab, IC extends IPrefabCreation, I>(id: I, info: IPrefabType, param?: IC, extendData?: any): T {
        let key = info;
        let cfg = this.cfgmap.get(key);
        if (cfg == undefined) {
            this.cfgmap.set(key, param);
            cfg = param;
        }
        if (cfg == undefined) {
            throw new Error(`Not Get [${key}] Prefab Creation Param !!`);
        }

        let create = PrefabModelObjPool.factoryMap.get(key);
        if (create == undefined) {
            throw new Error(`Not Get [${key}] Prefab Creation Function !!`);
        }
        let creation = PrefabModelObjPool.resFactoryMap.get(key);
        if (creation == undefined) {
            throw new Error(`Not Get [${key}] PrefabRes Creation Function !!`);
        }

        if (cfg.single !== false) {
            let prefabres = this.single_pools.get(key);
            if (prefabres == undefined && cfg.path && cfg.fileName) {
                prefabres = creation(id, this.scene, cfg);
                prefabres._cached = true;
                this.single_pools.set(key, prefabres);
            } else {
                prefabres.reuse();
            }
            let result = create(this.scene, key, this, prefabres) as T;
            prefabres.extendData = extendData;
            result.rayID = id;
            return result;
        } else {
            let unuselist = this.unuse_pools.get(key);
            let poollist = this.pools.get(key);
            if (unuselist == undefined) {
                this.unuse_pools.set(key, []);
                unuselist = this.unuse_pools.get(key);
            }
            if (poollist == undefined) {
                this.pools.set(key, []);
                poollist = this.pools.get(key);
            }
            let prefabres = unuselist.pop();
            if (prefabres == undefined) {
                prefabres = creation(id, this.scene, cfg);
                if (poollist.length < (cfg.cacheCount || 0)) {
                    prefabres._cached = true;
                    poollist.push(prefabres);
                } else {
                    prefabres._cached = false;
                    this.others.set(prefabres, prefabres);
                }
            } else {
                prefabres.reuse();
            }
            let result = create(this.scene, key, this, prefabres) as T;
            prefabres.extendData = extendData;
            result.rayID = id;
            return result;
        }
    }
    clearPrefabs() {
        this.check();
        // return;
        this._dispose();
    }

    protected _dispose = () => {
        this.pools.forEach((list) => {
            list.forEach((item) => {
                item.dispose();
            });
        });
        this.others.forEach((item) => {
            item.dispose();
        });
        this.single_pools.forEach((item) => {
            item.dispose();
        });

        this.pools.clear();
        this.unuse_pools.clear();
        this.others.clear();
        this.single_pools.clear();
    }

    check() {
        console.warn(`Prefabs Is Clear: ${this.Prefabs.size == 0}`);
        let createCount = 0;
        this.pools.forEach((item) => {
            createCount += item.length || 0;
        });
        let cacheCount = 0;
        this.unuse_pools.forEach((item) => {
            cacheCount += item.length || 0;
        });
        cacheCount += this.single_pools.size || 0;
        console.warn(`Prefabs Create: ${createCount}, Cache: ${cacheCount}. Error: ${createCount - cacheCount}`);
    }

    hasObj = (key: IPrefabType): boolean => {
        if (this.unuse_pools.has(key)) {
            return this.unuse_pools.get(key).length > 0;
        }
        return false;
    }

    countObj = (key: IPrefabType): number => {
        if (this.unuse_pools.has(key)) {
            return this.unuse_pools.get(key).length
        }
        return 0;
    }
}
