import { BABYLON } from "../render3d/babylon";
import { MeshParticleSystem } from "../mesh_particle/mesh_particle_system";
import { IBakedKeyframes } from "../interface/index";

export function bakeMeshParticleSystem(particlesys: MeshParticleSystem, particleMatrix: number[], particleColorData: number[], frames: number[], frameTime: number, duration: number, bakeDuration?: number) {
    if (bakeDuration == undefined) {
        bakeDuration = particlesys.psTool.startDelay + particlesys.psTool.duration * 2;
    }
    if (duration > bakeDuration) {
        return Promise.resolve();
    } else {
        if (duration < particlesys.psTool.startDelay) {
            //
        } else if ((<any>particlesys.psTool)._isPlaying != true) {
            (<any>particlesys.psTool)._startCall();
            // particlesys._computeCall();
            {
                (<any>particlesys).computeParents();
                particlesys.psTool.varCompute(frameTime);
                (<any>particlesys)._mpDirty = true;
                (<any>particlesys)._recycleCall();
            }
        } else {
            // particlesys._computeCall();
            {
                (<any>particlesys).computeParents();
                particlesys.psTool.varCompute(frameTime);
                (<any>particlesys)._mpDirty = true;
                (<any>particlesys)._recycleCall();
            }
        }

        let tempCount = particlesys._updateCall();
    
        let mpMatrixList = particlesys.psTool.mpMatrixList;
        let mpColorData = particlesys.psTool.mpColorData;
    
        let frameCount = frames.length;
        if (frameCount == 0) {
            frames.push(0 + tempCount);
        } else {
            frames.push(frames[frameCount - 1] + tempCount);
        }
        
        for (let i = 0; i < tempCount * 16; i++) {
            particleMatrix.push(mpMatrixList[i]);
        }
        for (let i = 0; i < tempCount * 4; i++) {
            particleColorData.push(mpColorData[i]);
        }
        duration += frameTime;
        return bakeMeshParticleSystem(particlesys, particleMatrix, particleColorData, frames, frameTime, duration + frameTime, bakeDuration);
    }
}

interface IBakedData {
    positions: Float32Array,
    colors: Float32Array,
    uvs: Float32Array,
    normals: Float32Array,
    indices: Uint16Array,
    bakeFrames: IBakedKeyframes,
}
export function transformToMesh(particlesys: MeshParticleSystem, particleMatrix: number[], particleColorData: number[], frames: number[], frameTime: number, duration: number): IBakedData {
    let _sourceMesh = particlesys._sourceMesh;
    let matrix = BABYLON.Matrix.Identity();
    let tempVec3 = BABYLON.Vector3.Zero();
    let uv = BABYLON.Vector2.Zero();
    let sourcePosition = _sourceMesh.geometry.getVerticesData("position");
    let vertexCount = Math.floor(sourcePosition.length / 3);
    let sourceUV = _sourceMesh.geometry.getVerticesData("uv");
    let sourceNormal = _sourceMesh.geometry.getVerticesData("normal");
    let sourceIndices = _sourceMesh.geometry.getIndices();

    let positions: number[] = [];
    let colors: number[] = [];
    let uvs: number[] = [];
    let normals: number[] = sourceNormal ? [] : undefined;
    let indices: number[] = [];

    let frameCount = frames.length;
    let tempCount = frames[frameCount - 1];
    frames.push(tempCount);
    for (let i = 0; i < tempCount; i++) {
        let vertexOff = (0 + i) * vertexCount;
        BABYLON.Matrix.FromArrayToRef(particleMatrix, i * 16, matrix);

        for (let j = 0; j < vertexCount; j++) {
            BABYLON.Vector3.FromArrayToRef(sourcePosition, j * 3, tempVec3);
            BABYLON.Vector3.TransformCoordinatesToRef(tempVec3, matrix, tempVec3);
            let offset = (vertexOff + j) * 3;
            positions[offset + 0] = tempVec3.x;
            positions[offset + 1] = tempVec3.y;
            positions[offset + 2] = tempVec3.z;

            if (sourceNormal) {
                BABYLON.Vector3.FromArrayToRef(sourceNormal, j * 3, tempVec3);
                normals[offset + 0] = tempVec3.x;
                normals[offset + 1] = tempVec3.y;
                normals[offset + 2] = tempVec3.z;
            }
        }
    }
    for (let i = 0; i < tempCount; i++) {
        let r = particleColorData[i * 4 + 0];
        let g = particleColorData[i * 4 + 1];
        let b = particleColorData[i * 4 + 2];
        let a = particleColorData[i * 4 + 3];
        let vertexOff = (0 + i) * vertexCount;
        for (let j = 0; j < vertexCount; j++) {
            let offset = (vertexOff + j) * 4;
            let vr = Math.floor(r / 10) * 10;
            let vg = Math.floor(g / 10) * 10;
            let vb = Math.floor(b / 10) * 10;
            let va = Math.floor(a / 10) * 10;
            colors[offset + 0] = r - vr;
            colors[offset + 1] = g - vg;
            colors[offset + 2] = b - vb;
            colors[offset + 3] = a - va;
            let uvoffset = (vertexOff + j) * 2;
            let us = (vr) / 10 / 1000;
            let vs = (vg) / 10 / 1000;
            let uo = (vb) / 10 / 1000;
            let vo = (va) / 10 / 1000;
            if (sourceUV) {
                BABYLON.Vector2.FromArrayToRef(sourceUV, j * 2, uv);
                uvs[uvoffset + 0] = us * uv.x + uo;
                uvs[uvoffset + 1] = vs * uv.y + vo;
            } else {
                uvs[uvoffset + 0] = us * 1.0 + uo;
                uvs[uvoffset + 1] = vs * 1.0 + vo;
            }
        }
    }

    let indicesCount = sourceIndices.length;
    for (let i = 0; i < tempCount; i++) {
        sourceIndices.forEach((v) => {
            indices.push(v + vertexCount * i);
        });
    }

    let bakeFrames: IBakedKeyframes = {
        duration,
        keyframes: []
    };

    for (let i = 0; i < frameCount; i++) {
        let start = frames[i] * indicesCount;
        let end = frames[i + 1] * indicesCount;

        bakeFrames.keyframes.push({
            frames: i * frameTime,
            data: [{ start, count: end - start }]
        });
    }

    return {
        positions: new Float32Array(positions),
        colors: new Float32Array(colors),
        uvs: new Float32Array(uvs),
        normals: normals ? new Float32Array(normals) : undefined,
        indices: new Uint16Array(indices),
        bakeFrames,
    }
}