// import * as BABYLON from "../render3d/babylon";
import { BABYLON } from '../render3d/babylon';
import { IPITextureInfo } from '../material_info';
import { PI_material } from '../gltf/extensions/material';
import { LoadExtraMaterial } from '../render3d/extends/material';
import { GLTFLoader } from '../gltf/glTFLoader';

export interface IPIMaterialPBR {
    /**
     * 材质类型
     */
    type: string;

    /**
     * 共用材质的ID
     */
    instanceID?: string;

    alphaMode: number;
    needAlphaBlend?: 1;
    zWrite?: 1;

    disableLighting: boolean;

    metallic: number;
    roughness: number;
    metallicTexture?: IPITextureInfo;

    specularPower?: number;
    reflectivityColor: [number, number, number];

    alpha: number;
    cull: string;

    useAlphaFromDiffuseTexture: boolean;

    albedoColor?: [number, number, number];
    albedoTexture?: IPITextureInfo;

    opacityTexture?: IPITextureInfo;

    emissionColor: [number, number, number];
    emissionTexture?: IPITextureInfo;
    emissiveMapLevel?: number;

    bumpTexture?: IPITextureInfo;
    normalScale?: number;

    useLightmapAsShadowmap: boolean;
    lightmapTexture?: IPITextureInfo;

    // Fresnel
    EFBias?: number;
    EFPower?: number;
    EFLeft?: [number, number, number];
    EFRight?: [number, number, number];

    // alphaCutOff
    alphaCutOff?: number;
    /**
     * cutoffEdgeThickness
     */
    coET?: number;
    /**
     * cutoffEdgeColor
     */
    coEColor?: [number, number, number];
    /**
     * cutoffEdgeLevel
     */
    coEL?: number;
    /**
     * cutoffEdgeAlphaDiff
     */
    coEAD?: number;

    fog?: boolean;

    alphaClip?: boolean;
    renderQueue?: number;
    diffuseOU?: number;
    diffuseOV?: number;
    emissionOU?: number;
    emissionOV?: number;
    opacityOU?: number;
    opacityOV?: number;
    diffuseLevel?: number;
    opacityLevel?: number;

    directIntensity?: number;
    specularIntensity?: number;
    emissiveIntensity?: number;
    environmentIntensity?: number;

    ambientColor?: [number, number, number];
    ambientTexture?: IPITextureInfo;
    ambientMapLevel?: number;

    detailTexture?: IPITextureInfo;
    detailMapLevel?: number;
    detailDiffuseBlend?: number;
    detailBumpScale?: number;
    detailRoughnessBlend?: number;
}

//#region PBRSpecularGlossinessCreate
function pbrSpecularGlossinessCreate(context: string, babylonScene: any, materialName: string, index: number, materialInfo: any, babylonDrawMode: number): any {

    let mat = new BABYLON.PBRSpecularGlossinessMaterial(materialName, babylonScene);

    mat.glossiness         = materialInfo.glossiness;
    mat.specularColor      = new BABYLON.Color3(
        materialInfo.specularColor[0],
        materialInfo.specularColor[1],
        materialInfo.specularColor[2]
    );
    mat.diffuseColor       = new BABYLON.Color3(
        materialInfo.diffuseColor[0],
        materialInfo.diffuseColor[1],
        materialInfo.diffuseColor[2]
    );

    if (mat) {
        mat.sideOrientation = babylonScene.useRightHandedSystem ? BABYLON.Material.CounterClockWiseSideOrientation : BABYLON.Material.ClockWiseSideOrientation;
    }

    return mat;
}

function pbrSpecularGlossinessLoadAsync(context: string, material: any, mat: any, _loader: any): Promise<void> {
    let promises: Promise<any>[] = [];

    if (material.cull == "front") {
        mat.sideOrientation = !_loader.babylonScene.useRightHandedSystem ? BABYLON.Material.CounterClockWiseSideOrientation : BABYLON.Material.ClockWiseSideOrientation;
    }
    if (material.cull == "off") {
        mat.backFaceCulling = false;
    }

    if (material.diffuseTexture) {
        promises.push(_loader.loadTextureInfoAsync(context + "/PI_material/diffuseTexture", material.diffuseTexture, function(texture: any) {
            mat.diffuseTexture = texture;
            if (material.diffuseTexture.offset) {
                (<any>texture).uOffset = material.diffuseTexture.offset[0];
                (<any>texture).vOffset = material.diffuseTexture.offset[1];
            }
            if (material.diffuseTexture.scale) {
                (<any>texture).uScale = material.diffuseTexture.scale[0];
                (<any>texture).vScale = material.diffuseTexture.scale[1];
            }
            (<any>texture).level = material.diffuseTexture.level || 1.0;
        }));
    }

    return Promise.all(promises).then(function() { });
}
function pbrSpecularGlossinessLoadAsyncNew(context: string, material: any, mat: any, _loader: GLTFLoader): Promise<void> {
    let promises: Promise<any>[] = [];

    if (material.cull == "front") {
        mat.sideOrientation = !_loader.babylonScene.useRightHandedSystem ? BABYLON.Material.CounterClockWiseSideOrientation : BABYLON.Material.ClockWiseSideOrientation;
    }
    if (material.cull == "off") {
        mat.backFaceCulling = false;
    }

    if (material.diffuseTexture) {
        let texture = <BABYLON.Texture>_loader.loadTextureInfoAsync(context + "/PI_material/diffuseTexture", material.diffuseTexture);
            mat.diffuseTexture = texture;
            if (material.diffuseTexture.offset) {
                (<BABYLON.Texture>texture).uOffset = material.diffuseTexture.offset[0];
                (<BABYLON.Texture>texture).vOffset = material.diffuseTexture.offset[1];
            }
            if (material.diffuseTexture.scale) {
                (<BABYLON.Texture>texture).uScale = material.diffuseTexture.scale[0];
                (<BABYLON.Texture>texture).vScale = material.diffuseTexture.scale[1];
            }
            (<BABYLON.Texture>texture).level = material.diffuseTexture.level || 1.0;

    }

    return Promise.all(promises).then(function() { });
}
//#endregion

//#region PBRMetallicRoughnessCreate
function pbrMetallicRoughnessCreate(context: string, babylonScene: any, materialName: string, index: number, materialInfo: any, babylonDrawMode: number): any {

    let mat = new BABYLON.PBRMetallicRoughnessMaterial(materialName, babylonScene);
    mat.metallic         = materialInfo.metallic;
    mat.roughness        = materialInfo.roughness;

    mat.baseColor       = new BABYLON.Color3(
        materialInfo.baseColor[0],
        materialInfo.baseColor[1],
        materialInfo.baseColor[2]
    );

    if (mat) {
        mat.sideOrientation = babylonScene.useRightHandedSystem ? BABYLON.Material.CounterClockWiseSideOrientation : BABYLON.Material.ClockWiseSideOrientation;
    }

    return mat;
}

function pbrMetallicRoughnessLoadAsync(context: string, material: any, mat: any, _loader: any): Promise<void> {
    let promises: Promise<any>[] = [];

    if (material.cull == "front") {
        mat.sideOrientation = !_loader.babylonScene.useRightHandedSystem ? BABYLON.Material.CounterClockWiseSideOrientation : BABYLON.Material.ClockWiseSideOrientation;
    }
    if (material.cull == "off") {
        mat.backFaceCulling = false;
    }

    if (material.baseTexture) {
        promises.push(_loader.loadTextureInfoAsync(context + "/PI_material/baseTexture", material.baseTexture, function(texture: any) {
            mat.baseTexture = texture;
            if (material.baseTexture.offset) {
                (<any>texture).uOffset = material.baseTexture.offset[0];
                (<any>texture).vOffset = material.baseTexture.offset[1];
            }
            if (material.baseTexture.scale) {
                (<any>texture).uScale = material.baseTexture.scale[0];
                (<any>texture).vScale = material.baseTexture.scale[1];
            }
            (<any>texture).level = material.baseTexture.level || 1.0;
        }));
    }

    return Promise.all(promises).then(function() { });
}
function pbrMetallicRoughnessLoadAsyncNew(context: string, material: any, mat: any, _loader: GLTFLoader) {
    let promises: Promise<any>[] = [];

    if (material.cull == "front") {
        mat.sideOrientation = !_loader.babylonScene.useRightHandedSystem ? BABYLON.Material.CounterClockWiseSideOrientation : BABYLON.Material.ClockWiseSideOrientation;
    }
    if (material.cull == "off") {
        mat.backFaceCulling = false;
    }

    if (material.baseTexture) {
        let texture = <BABYLON.Texture>_loader.loadTextureInfoAsync(context + "/PI_material/baseTexture", material.baseTexture);
            mat.baseTexture = texture;
            if (material.baseTexture.offset) {
                (<any>texture).uOffset = material.baseTexture.offset[0];
                (<any>texture).vOffset = material.baseTexture.offset[1];
            }
            if (material.baseTexture.scale) {
                (<any>texture).uScale = material.baseTexture.scale[0];
                (<any>texture).vScale = material.baseTexture.scale[1];
            }
            (<any>texture).level = material.baseTexture.level || 1.0;
    }

    return Promise.all(promises).then(function() { });
}
//#endregion

//#region PBR
function pbrCreate(context: string, babylonScene: any, materialName: string, index: number, matInfo: any, babylonDrawMode: number): any {

    let materialInfo: IPIMaterialPBR = matInfo;
    let mat = new BABYLON.PBRMaterial(materialName, babylonScene);
    mat.usePhysicalLightFalloff = true;
    mat.useGLTFLightFalloff = false;
    mat.metallicF0Factor = 0.2; // ior = 1.5 时 计算结果 f0 = 0.04, 0.2*0.2 = 0.04
    mat.metallic         = materialInfo.metallic;
    mat.roughness        = materialInfo.roughness;
    mat.albedoColor      = new BABYLON.Color3(
            materialInfo.albedoColor[0],
            materialInfo.albedoColor[1],
            materialInfo.albedoColor[2]
        );
    mat.reflectivityColor  = new BABYLON.Color3(
        materialInfo.reflectivityColor[0],
        materialInfo.reflectivityColor[1],
        materialInfo.reflectivityColor[2]
    );
    if (materialInfo.emissionColor) {
        mat.emissiveColor  = new BABYLON.Color3(
            materialInfo.emissionColor[0],
            materialInfo.emissionColor[1],
            materialInfo.emissionColor[2]
        );
    }
    if (materialInfo.ambientColor) {
        mat.ambientColor  = new BABYLON.Color3(
            materialInfo.ambientColor[0],
            materialInfo.ambientColor[1],
            materialInfo.ambientColor[2]
        );
    }

    if (mat) {
        mat.sideOrientation = babylonScene.useRightHandedSystem ? BABYLON.Material.CounterClockWiseSideOrientation : BABYLON.Material.ClockWiseSideOrientation;
    }

    mat.forceIrradianceInFragment = true;

    return mat;
}

function pbrLoadAsync(context: string, materialInfo: IPIMaterialPBR, mat: BABYLON.PBRMaterial, _loader: any): Promise<void> {
    let promises: Promise<any>[] = [];

    if (materialInfo.cull == "front") {
        mat.sideOrientation = !_loader.babylonScene.useRightHandedSystem ? BABYLON.Material.CounterClockWiseSideOrientation : BABYLON.Material.ClockWiseSideOrientation;
    }
    if (materialInfo.cull == "off") {
        mat.backFaceCulling = false;
    }

    // if (material.HSB) {
    //     var curve = new BABYLON.ColorCurves();
    //     curve.globalHue = material.HSB[0];
    //     curve.globalSaturation = material.HSB[1];
    //     curve.globalExposure = material.HSB[2];

    //     mat.imageProcessingConfiguration = new BABYLON.ImageProcessingConfiguration();
    //     mat.imageProcessingConfiguration.colorCurvesEnabled = true;
    //     mat.imageProcessingConfiguration.colorCurves = curve;
    // }

    if (materialInfo.albedoTexture) {
        promises.push(_loader.loadTextureInfoAsync(context + "/PI_material/albedoTexture", materialInfo.albedoTexture, function(texture: any) {
            mat.albedoTexture = texture;
            if (materialInfo.albedoTexture.offset) {
                (<any>texture).uOffset = materialInfo.albedoTexture.offset[0];
                (<any>texture).vOffset = materialInfo.albedoTexture.offset[1];
            }
            if (materialInfo.albedoTexture.scale) {
                (<any>texture).uScale = materialInfo.albedoTexture.scale[0];
                (<any>texture).vScale = materialInfo.albedoTexture.scale[1];
            }
            (<any>texture).level = materialInfo.albedoTexture.level || 1.0;
        }));
    }

    if (materialInfo.metallicTexture) {
        promises.push(_loader.loadTextureInfoAsync(context + "/PI_material/metallicTexture", materialInfo.metallicTexture, function(texture: any) {
            mat.metallicTexture = texture;
            if (materialInfo.metallicTexture.offset) {
                (<any>texture).uOffset = materialInfo.metallicTexture.offset[0];
                (<any>texture).vOffset = materialInfo.metallicTexture.offset[1];
            }
            if (materialInfo.metallicTexture.scale) {
                (<any>texture).uScale = materialInfo.metallicTexture.scale[0];
                (<any>texture).vScale = materialInfo.metallicTexture.scale[1];
            }
            (<any>texture).level = materialInfo.metallicTexture.level || 1.0;
        }));

        mat.useRoughnessFromMetallicTextureAlpha = false;
        mat.useAmbientOcclusionFromMetallicTextureRed = true;
        mat.useRoughnessFromMetallicTextureGreen = true;
        mat.useMetallnessFromMetallicTextureBlue = true;
    }

    if (materialInfo.emissionTexture) {
        promises.push(_loader.loadTextureInfoAsync(context + "/PI_material/emissionTexture", materialInfo.emissionTexture, function(texture) {
            mat.emissiveTexture = texture;
            if (materialInfo.emissionTexture.offset) {
                (<BABYLON.Texture>texture).uOffset = materialInfo.emissionTexture.offset[0];
                (<BABYLON.Texture>texture).vOffset = materialInfo.emissionTexture.offset[1];
            }
            if (materialInfo.emissionTexture.scale) {
                (<BABYLON.Texture>texture).uScale = materialInfo.emissionTexture.scale[0];
                (<BABYLON.Texture>texture).vScale = materialInfo.emissionTexture.scale[1];
            }
            (<BABYLON.Texture>texture).level = materialInfo.emissionTexture.level || materialInfo.emissiveMapLevel || 1.0;
        }));
    }

    if (materialInfo.bumpTexture) {
        promises.push(_loader.loadTextureInfoAsync(context + "/PI_material/bumpTexture", materialInfo.bumpTexture, function(texture) {
            mat.bumpTexture = texture;
            if (materialInfo.bumpTexture.offset) {
                (<BABYLON.Texture>texture).uOffset = materialInfo.bumpTexture.offset[0];
                (<BABYLON.Texture>texture).vOffset = materialInfo.bumpTexture.offset[1];
            }
            if (materialInfo.bumpTexture.scale) {
                (<BABYLON.Texture>texture).uScale = materialInfo.bumpTexture.scale[0];
                (<BABYLON.Texture>texture).vScale = materialInfo.bumpTexture.scale[1];
            }
            (<BABYLON.Texture>texture).level = materialInfo.bumpTexture.level || materialInfo.normalScale || 1.0;
        }));
    }

    if (materialInfo.ambientTexture) {
        promises.push(_loader.loadTextureInfoAsync(context + "/PI_material/ambientTexture", materialInfo.ambientTexture, function(texture) {
            mat.ambientTexture = texture;
            if (materialInfo.ambientTexture.offset) {
                (<BABYLON.Texture>texture).uOffset = materialInfo.ambientTexture.offset[0];
                (<BABYLON.Texture>texture).vOffset = materialInfo.ambientTexture.offset[1];
            }
            if (materialInfo.ambientTexture.scale) {
                (<BABYLON.Texture>texture).uScale = materialInfo.ambientTexture.scale[0];
                (<BABYLON.Texture>texture).vScale = materialInfo.ambientTexture.scale[1];
            }
            (<BABYLON.Texture>texture).level = materialInfo.ambientTexture.level || 1.0;
        }));
    }

    if (materialInfo.detailTexture) {
        promises.push(_loader.loadTextureInfoAsync(context + "/PI_material/detailTexture", materialInfo.detailTexture, function(texture) {
            mat.detailMap.texture = texture;
            if (materialInfo.detailTexture.offset) {
                (<BABYLON.Texture>texture).uOffset = materialInfo.detailTexture.offset[0];
                (<BABYLON.Texture>texture).vOffset = materialInfo.detailTexture.offset[1];
            }
            if (materialInfo.detailTexture.scale) {
                (<BABYLON.Texture>texture).uScale = materialInfo.detailTexture.scale[0];
                (<BABYLON.Texture>texture).vScale = materialInfo.detailTexture.scale[1];
            }
            (<BABYLON.Texture>texture).level = materialInfo.detailTexture.level || 1.0;

            mat.detailMap.bumpLevel = materialInfo.detailBumpScale;
            mat.detailMap.diffuseBlendLevel = materialInfo.detailDiffuseBlend;
            mat.detailMap.roughnessBlendLevel = materialInfo.detailRoughnessBlend;
        }));
    }

    if (materialInfo.lightmapTexture) {
        promises.push(_loader.loadTextureInfoAsync(context + "/PI_material/lightmapTexture", materialInfo.lightmapTexture, function(texture) {
            mat.lightmapTexture = texture;
            if (materialInfo.lightmapTexture.offset) {
                (<BABYLON.Texture>texture).uOffset = materialInfo.lightmapTexture.offset[0];
                (<BABYLON.Texture>texture).vOffset = materialInfo.lightmapTexture.offset[1];
            }
            if (materialInfo.lightmapTexture.scale) {
                (<BABYLON.Texture>texture).uScale = materialInfo.lightmapTexture.scale[0];
                (<BABYLON.Texture>texture).vScale = materialInfo.lightmapTexture.scale[1];
            }
            (<BABYLON.Texture>texture).level = materialInfo.lightmapTexture.level || 1.0;
            (<BABYLON.Texture>texture).coordinatesIndex = 1;
        }));
    }

    mat.directIntensity = materialInfo.directIntensity || 0;
    mat.emissiveIntensity = materialInfo.emissiveIntensity || 0;
    mat.environmentIntensity = materialInfo.environmentIntensity || 0;
    mat.specularIntensity = materialInfo.specularIntensity || 0;

    if (materialInfo.ambientMapLevel) {
        mat.ambientTextureStrength = materialInfo.ambientMapLevel;
    }

    return Promise.all(promises).then(function() { });
}
function pbrLoadAsyncNew(context: string, matInfo: any, mat: BABYLON.PBRMaterial, _loader: GLTFLoader) {
    let promises: Promise<any>[] = [];
    let materialInfo: IPIMaterialPBR = matInfo;

    if (materialInfo.cull == "front") {
        mat.sideOrientation = !_loader.babylonScene.useRightHandedSystem ? BABYLON.Material.CounterClockWiseSideOrientation : BABYLON.Material.ClockWiseSideOrientation;
    }
    if (materialInfo.cull == "off") {
        mat.backFaceCulling = false;
    }

    // if (material.HSB) {
    //     var curve = new BABYLON.ColorCurves();
    //     curve.globalHue = material.HSB[0];
    //     curve.globalSaturation = material.HSB[1];
    //     curve.globalExposure = material.HSB[2];

    //     mat.imageProcessingConfiguration = new BABYLON.ImageProcessingConfiguration();
    //     mat.imageProcessingConfiguration.colorCurvesEnabled = true;
    //     mat.imageProcessingConfiguration.colorCurves = curve;
    // }

    if (materialInfo.albedoTexture) {
        let texture = <BABYLON.Texture>_loader.loadTextureInfoAsync(context + "/PI_material/albedoTexture", materialInfo.albedoTexture);
            mat.albedoTexture = texture;
            if (materialInfo.albedoTexture.offset) {
                (<any>texture).uOffset = materialInfo.albedoTexture.offset[0];
                (<any>texture).vOffset = materialInfo.albedoTexture.offset[1];
            }
            if (materialInfo.albedoTexture.scale) {
                (<any>texture).uScale = materialInfo.albedoTexture.scale[0];
                (<any>texture).vScale = materialInfo.albedoTexture.scale[1];
            }
            (<any>texture).level = materialInfo.albedoTexture.level || 1.0;
    }

    if (materialInfo.metallicTexture) {
        let texture = <BABYLON.Texture>_loader.loadTextureInfoAsync(context + "/PI_material/metallicTexture", materialInfo.metallicTexture);
            mat.metallicTexture = texture;
            if (materialInfo.metallicTexture.offset) {
                (<any>texture).uOffset = materialInfo.metallicTexture.offset[0];
                (<any>texture).vOffset = materialInfo.metallicTexture.offset[1];
            }
            if (materialInfo.metallicTexture.scale) {
                (<any>texture).uScale = materialInfo.metallicTexture.scale[0];
                (<any>texture).vScale = materialInfo.metallicTexture.scale[1];
            }
            (<any>texture).level = materialInfo.metallicTexture.level || 1.0;

        mat.useRoughnessFromMetallicTextureAlpha = false;
        mat.useAmbientOcclusionFromMetallicTextureRed = true;
        mat.useRoughnessFromMetallicTextureGreen = true;
        mat.useMetallnessFromMetallicTextureBlue = true;
    }

    if (materialInfo.emissionTexture) {
        let texture = <BABYLON.Texture>_loader.loadTextureInfoAsync(context + "/PI_material/emissionTexture", materialInfo.emissionTexture);
            mat.emissiveTexture = texture;
            if (materialInfo.emissionTexture.offset) {
                (<BABYLON.Texture>texture).uOffset = materialInfo.emissionTexture.offset[0];
                (<BABYLON.Texture>texture).vOffset = materialInfo.emissionTexture.offset[1];
            }
            if (materialInfo.emissionTexture.scale) {
                (<BABYLON.Texture>texture).uScale = materialInfo.emissionTexture.scale[0];
                (<BABYLON.Texture>texture).vScale = materialInfo.emissionTexture.scale[1];
            }
            (<BABYLON.Texture>texture).level = materialInfo.emissionTexture.level || materialInfo.emissiveMapLevel || 1.0;
    }

    if (materialInfo.bumpTexture) {
        let texture = <BABYLON.Texture>_loader.loadTextureInfoAsync(context + "/PI_material/bumpTexture", materialInfo.bumpTexture);
            mat.bumpTexture = texture;
            if (materialInfo.bumpTexture.offset) {
                (<BABYLON.Texture>texture).uOffset = materialInfo.bumpTexture.offset[0];
                (<BABYLON.Texture>texture).vOffset = materialInfo.bumpTexture.offset[1];
            }
            if (materialInfo.bumpTexture.scale) {
                (<BABYLON.Texture>texture).uScale = materialInfo.bumpTexture.scale[0];
                (<BABYLON.Texture>texture).vScale = materialInfo.bumpTexture.scale[1];
            }
            (<BABYLON.Texture>texture).level = materialInfo.bumpTexture.level || materialInfo.normalScale || 1.0;
    }

    if (materialInfo.ambientTexture) {
        let texture = <BABYLON.Texture>_loader.loadTextureInfoAsync(context + "/PI_material/ambientTexture", materialInfo.ambientTexture);
            mat.ambientTexture = texture;
            if (materialInfo.ambientTexture.offset) {
                (<BABYLON.Texture>texture).uOffset = materialInfo.ambientTexture.offset[0];
                (<BABYLON.Texture>texture).vOffset = materialInfo.ambientTexture.offset[1];
            }
            if (materialInfo.ambientTexture.scale) {
                (<BABYLON.Texture>texture).uScale = materialInfo.ambientTexture.scale[0];
                (<BABYLON.Texture>texture).vScale = materialInfo.ambientTexture.scale[1];
            }
            (<BABYLON.Texture>texture).level = materialInfo.ambientTexture.level || 1.0;
    }

    if (materialInfo.detailTexture) {
        let texture = <BABYLON.Texture>_loader.loadTextureInfoAsync(context + "/PI_material/detailTexture", materialInfo.detailTexture);
            mat.detailMap.texture = texture;
            if (materialInfo.detailTexture.offset) {
                (<BABYLON.Texture>texture).uOffset = materialInfo.detailTexture.offset[0];
                (<BABYLON.Texture>texture).vOffset = materialInfo.detailTexture.offset[1];
            }
            if (materialInfo.detailTexture.scale) {
                (<BABYLON.Texture>texture).uScale = materialInfo.detailTexture.scale[0];
                (<BABYLON.Texture>texture).vScale = materialInfo.detailTexture.scale[1];
            }
            (<BABYLON.Texture>texture).level = materialInfo.detailTexture.level || 1.0;

            mat.detailMap.bumpLevel = materialInfo.detailBumpScale;
            mat.detailMap.diffuseBlendLevel = materialInfo.detailDiffuseBlend;
            mat.detailMap.roughnessBlendLevel = materialInfo.detailRoughnessBlend;
    }

    if (materialInfo.lightmapTexture) {
        let texture = <BABYLON.Texture>_loader.loadTextureInfoAsync(context + "/PI_material/lightmapTexture", materialInfo.lightmapTexture);
            mat.lightmapTexture = texture;
            if (materialInfo.lightmapTexture.offset) {
                (<BABYLON.Texture>texture).uOffset = materialInfo.lightmapTexture.offset[0];
                (<BABYLON.Texture>texture).vOffset = materialInfo.lightmapTexture.offset[1];
            }
            if (materialInfo.lightmapTexture.scale) {
                (<BABYLON.Texture>texture).uScale = materialInfo.lightmapTexture.scale[0];
                (<BABYLON.Texture>texture).vScale = materialInfo.lightmapTexture.scale[1];
            }
            (<BABYLON.Texture>texture).level = materialInfo.lightmapTexture.level || 1.0;
            (<BABYLON.Texture>texture).coordinatesIndex = 1;
    }

    mat.directIntensity = materialInfo.directIntensity || 0;
    mat.emissiveIntensity = materialInfo.emissiveIntensity || 0;
    mat.environmentIntensity = materialInfo.environmentIntensity || 0;
    mat.specularIntensity = materialInfo.specularIntensity || 0;

    if (materialInfo.ambientMapLevel) {
        mat.ambientTextureStrength = materialInfo.ambientMapLevel;
    }

    return Promise.all(promises).then(function() { });
}
//#endregion

export function registPBR() {
    BABYLON.GLTF2.Loader.Extensions.PI_material.regist(
        "PBR",
        pbrCreate,
        pbrLoadAsync
    );
    
    LoadExtraMaterial.loadCalls.set("PBR", pbrCreate);
    LoadExtraMaterial.loadPropertyCalls.set("PBR", pbrLoadAsyncNew);
}

export function registPBRMetallicRoughness() {
    BABYLON.GLTF2.Loader.Extensions.PI_material.regist(
        "PBRMetallicRoughness",
        pbrMetallicRoughnessCreate,
        pbrMetallicRoughnessLoadAsync
    );
    LoadExtraMaterial.loadCalls.set("PBRMetallicRoughness", pbrMetallicRoughnessCreate);
    LoadExtraMaterial.loadPropertyCalls.set("PBRMetallicRoughness", pbrMetallicRoughnessLoadAsyncNew);
}

export function registPBRSpecularGlossiness() {
    BABYLON.GLTF2.Loader.Extensions.PI_material.regist(
        "PBRSpecularGlossiness",
        pbrSpecularGlossinessCreate,
        pbrSpecularGlossinessLoadAsync
    );
    LoadExtraMaterial.loadCalls.set("PBRSpecularGlossiness", pbrSpecularGlossinessCreate);
    LoadExtraMaterial.loadPropertyCalls.set("PBRSpecularGlossiness", pbrSpecularGlossinessLoadAsyncNew);
}