import { RenderTarget } from "./render_target";

export enum BlendMode {
    None,
    Normal,
    Additive,
    Multiply,
    Screen
}

export enum RenderMode {
    Local = 1,
    Canvas,
    Temp,
    TempScreen,
    TargetFBO
}

export interface IRenderData {
    /**
     * 源 纹理
     */
    sourceTex: WebGLTexture;
    /**
     * 源 FBO - 当 renderMode = RenderMode.Locel  时必须需要 sourceFBO
     */
    sourceFBO?: WebGLFramebuffer;
    /**
     * 源 FBO - 当 renderMode = RenderMode.TargetFBO 时必须需要 targetFBO
     */
    targetFBO?: WebGLFramebuffer;
    /**
     * 源纹理宽度
     */
    texWidth: number;
    /**
     * 源纹理高度
     */
    texHeight: number;
    /**
     * 渲染模式
     */
    renderMode?: RenderMode;
    /**
     * 渲染目标缩放值 - 相比源目标
     */
    scale?: number;
    sourceTex2?: WebGLTexture;
}

export interface IRenderDataNew {
    /**
     * 源 纹理
     */
    sourceRT: RenderTarget;
    /**
     * 源 纹理
     */
    targetRT?: RenderTarget;
    canvas?: HTMLCanvasElement;
}

export interface Restorable {
    restore(): void;
}
