import { PostProcess } from "./postprocess";
import { ShaderCompiler } from "./shader_compiler";

export enum EIrisAreaMode {
    None = 0,
    Circle = 1,
    Rectangle = 2
}
export enum EIrisMaskFuncName {
    None = ``,
    Circle = `IrisMaskCircle`,
    Rectangle = `IrisMaskRectangle`
}

export class Blur extends PostProcess {

    // 是否应用清晰窗口
    irisAreaMode: EIrisAreaMode = EIrisAreaMode.Circle;
    // 清晰窗口范围尺寸
    irisAreaSizeX: number = 0;
    // 清晰窗口范围尺寸
    irisAreaSizeY: number = 0;
    // 清晰窗口影响的过渡范围尺寸
    irisAreaTransition: number = 0;
    // 清晰窗口中心点的水平偏移
    irisAreaOffsetX: number = 0;
    // 清晰窗口中心点的垂直偏移
    irisAreaOffsetY: number = 0;

    // public static IrisAreaActive = "u_iris_work";
    public static IrisTransition = "u_iris_transition";
    public static IrisAreaMode = "u_iris_mode";
    public static IrisInfo = "u_iris";
    public static IrisMaskFuncName = "IrisMask";

    public static GetIrisMaskFragmentCall(mode: EIrisAreaMode, input: string) {
        switch (mode) {
            case EIrisAreaMode.Circle: {
                return `IrisMask(${input})`; // `${EIrisMaskFuncName.Circle}(${input})`;
            }
            case EIrisAreaMode.Rectangle: {
                return `IrisMask(${input})`; // `${EIrisMaskFuncName.Rectangle}(${input})`;
            }
            default: {
                return `IrisMask(${input})`; // `1.0`;
            }
        }
    }
    public static GetIrisMaskFragmentDefined(mode: EIrisAreaMode) {
        switch (mode) {
            case EIrisAreaMode.Circle: {
                return Blur.IrisFragmentIrisMask(); // Blur.IrisFragmentIrisMaskCircle();
            }
            case EIrisAreaMode.Rectangle: {
                return Blur.IrisFragmentIrisMask(); // Blur.IrisFragmentIrisMaskRectangle();
            }
            default: {
                return Blur.IrisFragmentIrisMask(); // ``;
            }
        }
    }

    protected applyIris(shader: ShaderCompiler) {
        // this.shader.setUniformi(Blur.IrisAreaActive, this.irisAreaActive ? 1 : 0);
        // X - irisAreaSize
        // Y - irisAreaTransition
        // Z - irisAreaOffsetX
        // W - irisAreaOffsetY
        // switch (this.irisAreaMode) {
        //     case EIrisAreaMode.Circle:
        //     case EIrisAreaMode.Rectangle: {
        //         this.shader.setUniform4f(Blur.IrisInfo, this.irisAreaSizeX, this.irisAreaSizeY, this.irisAreaOffsetX, this.irisAreaOffsetY);
        //         this.shader.setUniformf(Blur.IrisTransition, this.irisAreaTransition <= 0 ? 0.01 : this.irisAreaTransition);
        //         return;
        //     }
        //     default: {
        //         return;
        //     }
        // }
        switch (this.irisAreaMode) {
            case EIrisAreaMode.Circle: {
                shader.setUniformi(Blur.IrisAreaMode, 1);
                shader.setUniform4f(Blur.IrisInfo, this.irisAreaSizeX, this.irisAreaSizeY, this.irisAreaOffsetX, this.irisAreaOffsetY);
                shader.setUniformf(Blur.IrisTransition, this.irisAreaTransition <= 0 ? 0.01 : this.irisAreaTransition);
                return;
            }
            case EIrisAreaMode.Rectangle: {
                shader.setUniformi(Blur.IrisAreaMode, 2);
                shader.setUniform4f(Blur.IrisInfo, this.irisAreaSizeX, this.irisAreaSizeY, this.irisAreaOffsetX, this.irisAreaOffsetY);
                shader.setUniformf(Blur.IrisTransition, this.irisAreaTransition <= 0 ? 0.01 : this.irisAreaTransition);
                return;
            }
            default: {
                shader.setUniformi(Blur.IrisAreaMode, 0);
                shader.setUniform4f(Blur.IrisInfo, 0, 0, 2.0, 2.0);
                shader.setUniformf(Blur.IrisTransition, this.irisAreaTransition <= 0 ? 0.01 : this.irisAreaTransition);
                return;
            }
        }
    }

    protected static IrisFragmentDefined() {
        return `
        uniform vec4 ${Blur.IrisInfo};
        uniform float ${Blur.IrisTransition};
        uniform int ${Blur.IrisAreaMode};
        `;
    }

    protected static IrisFragmentIrisMask() {
        return `
        float IrisMask(vec2 uv)
        {
            if (${Blur.IrisAreaMode} >= 2) {
                vec2 center = uv * 2.0 - 1.0 - vec2(${Blur.IrisInfo}.z, ${Blur.IrisInfo}.w); // [0,1] -> [-1,1]
                float x = clamp(abs(center.x) - ${Blur.IrisInfo}.x, 0.0, ${Blur.IrisTransition}) / ${Blur.IrisTransition};
                float y = clamp(abs(center.y) - ${Blur.IrisInfo}.y, 0.0, ${Blur.IrisTransition}) / ${Blur.IrisTransition};
                return max(x, y);
            }
            else if (${Blur.IrisAreaMode} >= 1) {
                vec2 center = uv * 2.0 - 1.0 - vec2(${Blur.IrisInfo}.z, ${Blur.IrisInfo}.w); // [0,1] -> [-1,1]
                return clamp(abs(length(center)) - ${Blur.IrisInfo}.x, 0.0, ${Blur.IrisTransition}) / ${Blur.IrisTransition};
            }
            else {
                return 1.0;
            }
        }
        `;
    }

    protected static IrisFragmentIrisMaskCircle() {
        return `
        float ${EIrisMaskFuncName.Circle}(vec2 uv)
        {
            vec2 center = uv * 2.0 - 1.0 - vec2(${Blur.IrisInfo}.z, ${Blur.IrisInfo}.w); // [0,1] -> [-1,1]
            return clamp(abs(length(center)) - ${Blur.IrisInfo}.x, 0.0, ${Blur.IrisTransition}) / ${Blur.IrisTransition};
        }
        `;
    }

    protected static IrisFragmentIrisMaskRectangle() {
        return `
        float ${EIrisMaskFuncName.Rectangle}(vec2 uv)
        {
            vec2 center = uv * 2.0 - 1.0 - vec2(${Blur.IrisInfo}.z, ${Blur.IrisInfo}.w); // [0,1] -> [-1,1]
            float x = clamp(abs(center.x) - ${Blur.IrisInfo}.x, 0.0, ${Blur.IrisTransition}) / ${Blur.IrisTransition};
            float y = clamp(abs(center.y) - ${Blur.IrisInfo}.y, 0.0, ${Blur.IrisTransition}) / ${Blur.IrisTransition};
            return max(x, y);
        }
        `;
    }

}