
import { PostProcessDataManager } from "./data_manager";
import { PostProcess } from "./postprocess";
import { RenderTexture, RenderTarget } from "./render_target";
import { ShaderCompiler } from "./shader_compiler";
import { ManagedWebGLRenderingContext } from "./webgl";

export class CircleOfConfusion extends PostProcess {

    public static shaderName = `CircleOfConfusion`;
    public static getShader(context: ManagedWebGLRenderingContext) {
        let manager = PostProcessDataManager.get(context);
        let shaderKey = `${CircleOfConfusion.shaderName}`;
        let result = manager.getShaderCompiler(shaderKey);
        if (!result) {
            result = new ShaderCompiler(context, PostProcess.getVertexString(), CircleOfConfusion.getFragmentString());
            manager.setShaderCompiler(shaderKey, result);
        }

        return result;
    }
    private static CameraMinMaxZ = 'cameraMinMaxZ';
    private static FocusDistance = 'focusDistance';
    private static COCPrecalculation = 'cocPrecalculation';
    /**
     * 颜色变换系数
     */
    degree: number = 0.0;

    constructor(canvasOrContext: ManagedWebGLRenderingContext) {

        super(canvasOrContext);
    }

    public apply(source: RenderTexture[], target: RenderTarget): ShaderCompiler {
        super.apply(source, target);

        const gl = this.canvasOrContext.gl;
        let shader = CircleOfConfusion.getShader(this.canvasOrContext);
        shader.bind();

        gl.activeTexture(gl.TEXTURE0);
        gl.bindTexture(gl.TEXTURE_2D, source[0]._webGLTexture);

        shader.setUniformi(ShaderCompiler.TEXTURE_SAMPLER, 0);
        shader.setUniformf(ShaderCompiler.DEGREE, this.degree);

        PostProcess.bindData(this.canvasOrContext, this.manager.getDataBufferCfg(PostProcessDataManager.baseDataBufferCfgKey), shader);

        return shader;
    }

    protected static getFragmentString() {
        return `
        #define SHADER_NAME fragment:blackAndWhite

        // samplers
        uniform sampler2D depthSampler;

        // varyings
        varying vec2 vUV;

        // preconputed uniforms (not effect parameters)
        uniform vec2 cameraMinMaxZ;

        // uniforms
        uniform float focusDistance;
        uniform float cocPrecalculation;

        void main(void)
        {
            float depth = texture2D(depthSampler, vUV).r;
            float pixelDistance = (cameraMinMaxZ.x + (cameraMinMaxZ.y - cameraMinMaxZ.x)*depth)*1000.0;	// actual distance from the lens in scene units/1000 (eg. millimeter)
            float coc = abs(cocPrecalculation* ((focusDistance - pixelDistance)/pixelDistance));
            coc = clamp(coc, 0.0, 1.0);
            gl_FragColor = vec4(coc, depth, coc, 1.0);
        }
        `;
    }
}