
import { PostProcessDataManager } from "./data_manager";
import { IHueDegree, KeyHueDegree } from "../interface/postprocess";
import { PostProcess } from "./postprocess";
import { RenderTexture, RenderTarget } from "./render_target";
import { ShaderCompiler } from "./shader_compiler";
import { ManagedWebGLRenderingContext } from "./webgl";

export class HueDegree extends PostProcess {
    public static shaderName = KeyHueDegree;
    public static getShader(context: ManagedWebGLRenderingContext) {
        let manager = PostProcessDataManager.get(context);
        let shaderKey = `${HueDegree.shaderName}`;
        let result = manager.getShaderCompiler(shaderKey);
        if (!result) {
            result = new ShaderCompiler(context, PostProcess.getVertexString(), HueDegree.getFragmentString());
            manager.setShaderCompiler(shaderKey, result);
        }

        return result;
    }

    /**
     * 颜色变换系数
     */
    degree: number = 0.0;

    public apply(param: IHueDegree, sources: RenderTexture[], target: RenderTarget, ctx: ManagedWebGLRenderingContext) {
        super.apply(param, sources, target, ctx);

        const gl = this.canvasOrContext.gl;
        let shader = HueDegree.getShader(this.canvasOrContext);
        shader.bind();

        gl.activeTexture(gl.TEXTURE0);
        gl.bindTexture(gl.TEXTURE_2D, sources[0]._webGLTexture);
        shader.setUniformi(ShaderCompiler.TEXTURE_SAMPLER, 0);
        shader.setUniformf(ShaderCompiler.DEGREE, param.degree);

        let dataBuffer = this.manager.getDataBufferCfg(PostProcessDataManager.baseDataBufferCfgKey);
        PostProcess.bindData(this.canvasOrContext, dataBuffer, shader);
        PostProcess.drawData(this.canvasOrContext, dataBuffer);
        this.afterDraw();
    }

    protected static getFragmentString() {
        return `

        #define SHADER_NAME fragment:hueDegree
        precision highp float;

        varying vec2 sampleCenter;
        uniform sampler2D textureSampler;
        uniform float degree;

        void main(void) {
            vec3 color = texture2D(textureSampler, sampleCenter).rgb;
            float luminance = dot(color, vec3(0.3, 0.59, 0.11));
            vec3 hueDegree = vec3(luminance, luminance, luminance);

            gl_FragColor = vec4(color-((color-hueDegree)*degree), 1.0);
        }
        `;
    }
}