
import { PostProcessDataManager } from "./data_manager";
import { PostProcess } from "./postprocess";
import { RenderTarget, RenderTexture } from "./render_target";
import { ShaderCompiler } from "./shader_compiler";
import { ManagedWebGLRenderingContext } from "./webgl";

export class KawaseBlur extends PostProcess {
    public static shaderName = `KawaseBlur`;
    public static getShader(context: ManagedWebGLRenderingContext) {
        let manager = PostProcessDataManager.get(context);
        let shaderKey = `${KawaseBlur.shaderName}`;
        let result = manager.getShaderCompiler(shaderKey);
        if (!result) {
            result = new ShaderCompiler(context, KawaseBlur.getVertexString(), KawaseBlur.getFragmentString());
            manager.setShaderCompiler(shaderKey, result);
        }

        return result;
    }

    public static readonly maxBlurRenderIteration = 4;
    public blurRadius: number = 1;
    public blurFactor: number = 1;
    public apply(sources: RenderTexture[], target: RenderTarget): ShaderCompiler {
        super.apply(sources, target);

        const gl = this.canvasOrContext.gl;
        let shader = KawaseBlur.getShader(this.canvasOrContext);
        shader.bind();

        gl.activeTexture(gl.TEXTURE0);
        gl.bindTexture(gl.TEXTURE_2D, sources[0]._webGLTexture);
        shader.setUniformi(ShaderCompiler.TEXTURE_SAMPLER, 0);
        shader.setUniform2f("u_blurOffset", this.blurRadius * this.blurFactor / sources[0].width, this.blurRadius * this.blurFactor / sources[0].height);

        PostProcess.bindData(this.canvasOrContext, this.manager.getDataBufferCfg(PostProcessDataManager.baseDataBufferCfgKey), shader);

        return shader;
    }

    // https://zhuanlan.zhihu.com/p/125744132
    protected static getFragmentString() {
        return `
        #define SHADER_NAME fragment:KawaseBlur
        precision highp float;

        varying vec2 sampleCenter;

        uniform sampler2D textureSampler;

        uniform vec2 u_blurOffset;

        vec4 KawaseBlur( vec2 uv , vec2 offset, sampler2D tex )
        {
            vec4 color =  texture2D( tex, uv + vec2( offset.x,  offset.y) )
                        + texture2D( tex, uv + vec2(-offset.x,  offset.y) )
                        + texture2D( tex, uv + vec2( offset.x, -offset.y) )
                        + texture2D( tex, uv + vec2(-offset.x, -offset.y) );
            return color * 0.25;
        }

        void main(void)
        {
            gl_FragColor = KawaseBlur(sampleCenter, u_blurOffset, textureSampler);;
        }
        `;
    }
}