import { IRenderData } from "./base";
import { HueDegree } from "./hue_degree";
import { GaussianBlur, IBlurOption } from "./gaussian_blur";
import { BokehBlur } from "./bokeh_blur";
import { RenderTargetFactory } from "./render_target";
import { ManagedWebGLRenderingContext } from "./webgl";
import { RadialBlur } from "./radial_blur";
import { DirectionalBlur } from "./directional_blur";
import { WaveJitter } from "./wave_jitter";
import { PointCloud } from "./point_cloud";
import { IrisBlur } from "./iris_blur";
import { RadiusWave } from "./radius_wave";
import { AnalogGlitch } from "./analog_glitch";
import { GlitchImageBlock } from "./glitch_image_block";
import { AnimeGlitchImageBlock } from "./glitch_image_block_anim";
import { GlitchImagePoint } from "./glitch_image_point";
import { AnimeGlitchImagePoint } from "./glitch_image_point_anim";
import { GlitchColorBlock } from "./glitch_color_block";
import { AnimeGlitchColorBlock } from "./glitch_color_block_anim";
import { AnimeGlitchImageBlockMix } from "./glitch_image_block_mix_anim";
import { GlitchImageBlockMix } from "./glitch_image_block_mix";
import { GlitchBlockOffset } from "./glitch_block_offset";
import { AnimeGlitchBlockOffset } from "./glitch_block_offset_anim";
import { LineFlow } from "./line_flow";
import { MultiBlendPP } from "./multi_blend";
import { Vignette } from "./vignette";
import { HSV } from "./hsv";
import { Noise3D } from "./noise3d";

export class RenderFactory {
    public static Context: ManagedWebGLRenderingContext;

    private static Canvas: HTMLCanvasElement;

    public static init(canvas: HTMLCanvasElement) {
        this.Context = new ManagedWebGLRenderingContext(canvas);

        this.Canvas = canvas;

        RenderTargetFactory.init(this.Canvas, this.Context);
    }

    public static createBlur(opt: IBlurOption, data?: IRenderData) {
        const blur = new GaussianBlur(this.Context);

        return blur;
    }

    public static createHueDegree(degree: number, data?: IRenderData) {
        const pp = new HueDegree(this.Context);

        return pp;
    }

    public static createBokehBlur(data?: IRenderData) {
        const pp = new BokehBlur(this.Context);

        return pp;
    }
    public static createRadialBlur(data?: IRenderData) {
        const pp = new RadialBlur(this.Context);

        return pp;
    }
    public static createDirectionalBlur(data?: IRenderData) {
        const pp = new DirectionalBlur(this.Context);

        return pp;
    }
    public static createWaveJitter(data?: IRenderData) {
        const pp = new WaveJitter(this.Context);

        return pp;
    }
    public static createPointCloud(data?: IRenderData) {
        const pp = new PointCloud(this.Context);

        return pp;
    }
    public static createIrisBlur(data?: IRenderData) {
        const pp = new IrisBlur(this.Context);

        return pp;
    }
    public static createRadiusWave(data?: IRenderData) {
        const pp = new RadiusWave(this.Context);

        return pp;
    }
    public static createAnalogGlitch(data?: IRenderData) {
        const pp = new AnalogGlitch(this.Context);

        return pp;
    }
    public static createGlitchImageBlock(data?: IRenderData) {
        const pp = new GlitchImageBlock(this.Context);

        return pp;
    }
    public static createGlitchImageBlockAnime(data?: IRenderData) {
        const pp = new AnimeGlitchImageBlock(this.Context);

        return pp;
    }
    public static createGlitchColorBlock(data?: IRenderData) {
        const pp = new GlitchColorBlock(this.Context);

        return pp;
    }
    public static createGlitchColorBlockAnime(data?: IRenderData) {
        const pp = new AnimeGlitchColorBlock(this.Context);

        return pp;
    }
    public static createGlitchImagePoint(data?: IRenderData) {
        const pp = new GlitchImagePoint(this.Context);

        return pp;
    }
    public static createGlitchImagePointAnime(data?: IRenderData) {
        const pp = new AnimeGlitchImagePoint(this.Context);

        return pp;
    }
    public static createGlitchImageBlockMix(data?: IRenderData) {
        const pp = new GlitchImageBlockMix(this.Context);

        return pp;
    }
    public static createGlitchImageBlockMixAnime(data?: IRenderData) {
        const pp = new AnimeGlitchImageBlockMix(this.Context);

        return pp;
    }
    public static createGlitchBlockOffset(data?: IRenderData) {
        const pp = new GlitchBlockOffset(this.Context);

        return pp;
    }
    public static createGlitchBlockOffsetAnime(data?: IRenderData) {
        const pp = new AnimeGlitchBlockOffset(this.Context);

        return pp;
    }
    public static createLineFlow(data?: IRenderData) {
        const pp = new LineFlow(this.Context);

        return pp;
    }
    public static createVignette(data?: IRenderData) {
        const pp = new Vignette(this.Context);

        return pp;
    }
    public static createMultiBlendPP(data?: IRenderData) {
        const pp = new MultiBlendPP(this.Context);

        return pp;
    }
    public static createHSV(data?: IRenderData) {
        const pp = new HSV(this.Context);

        return pp;
    }
    public static createNoise3D(data?: IRenderData) {
        const pp = new Noise3D(this.Context);

        return pp;
    }
}