/**
 *
 */

// ###################################################
import { BABYLON } from './babylon';
import { ResLoader } from './loader';
import { OfflineProvider } from './offline_provider';
import { ExtensionTools, Res, ResTab } from '../interface/index';

// ###################################################
declare let module;
let logger;
ExtensionTools.RegistStartUp(() => { logger = ExtensionTools.Logger(module.name, "BABYLON Engine"); });

// ###################################################
const compressTextureFormatInfo = {
    astc: 'ktx',
    s3tc: 'ktx'
};

export class Engine extends BABYLON.Engine {

    public static ofl: OfflineProvider;
    public piEngine: boolean;

    constructor(canvasOrContext: BABYLON.Nullable<HTMLCanvasElement | WebGLRenderingContext>, antialias?: boolean, options?: BABYLON.EngineOptions, adaptToDeviceRatio: boolean = false) {

        super(<any>canvasOrContext, antialias, options, adaptToDeviceRatio);
        this.piEngine = true

        // 压缩纹理适配
        const compressed_tex_format = ExtensionTools.QuerySupportCompressedTextureFormat();
        this._textureFormatInUse = compressTextureFormatInfo[compressed_tex_format];
        console.warn(`TextureFormatInUse : ${compressed_tex_format}`);
    }

    public static prepare(resTab: ResTab) {
        if (!Engine.ofl) {
            Engine.ofl = new OfflineProvider(resTab);
        }
        // 提供脱机资源加载
        BABYLON.Engine.OfflineProviderFactory = (_urlToScene: string, _callbackManifestChecked: (checked: boolean) => any, _disableManifestCheck: boolean) => {
            // BABYLON.Tools.SetImmediate(() => {
            //     _callbackManifestChecked(true);
            // });

            ExtensionTools.AddTask(_callbackManifestChecked, [true], 99999, 0);

            // return new OfflineProvider(new ResTab()); // 每场景一个OfflineProvider
            return Engine.ofl;
        };
        // GLTF2Loader.resLoader = Engine.loader;
        // 修改默认的gltf文件加载类
        BABYLON.GLTFFileLoader._CreateGLTF2Loader = (parent) => new BABYLON.GLTF2.GLTFLoader(parent);

    }

    // 创建Texture
    public createTexture(urlArg: BABYLON.Nullable<string>, noMipmap: boolean, invertY: boolean, scene: BABYLON.Nullable<BABYLON.Scene>, samplingMode: number = Engine.TEXTURE_TRILINEAR_SAMPLINGMODE,
        onLoad: BABYLON.Nullable<() => void> = null, onError: BABYLON.Nullable<(message: string, exception: any) => void> = null,
        buffer: BABYLON.Nullable<string | ArrayBuffer | HTMLImageElement | Blob> = null, fallback: BABYLON.Nullable<BABYLON.InternalTexture> = null, format: BABYLON.Nullable<number> = null,
        forcedExtension: BABYLON.Nullable<string> = null): BABYLON.InternalTexture {

        const tex = BABYLON.Engine.prototype.createTexture.call(this, urlArg, noMipmap, invertY, scene, samplingMode, onLoad, onError, buffer, fallback, format, forcedExtension);

        // 当使用压缩纹理时底层置换url为压缩纹理路径，但上层只管理 jpg/png,所以这里确保url为上层路径,修改不影响底层
        tex.url = urlArg;

        // 将tex添加到资源表
        // Engine.loader.createTexture(tex);

        return tex;
    }

    public _loadFile(url: string, onSuccess: (data: string | ArrayBuffer, responseURL?: string) => void, onProgress?: (data: any) => void,
        offlineProvider?: BABYLON.IOfflineProvider, useArrayBuffer?: boolean, onError?: (request?: BABYLON.IWebRequest, exception?: any) => void): any {
        ExtensionTools.LoadFile(Engine.ofl.resTab, url).then((r) => {
            if (ArrayBuffer.isView(r)) {
                r = (<Uint8Array>r).slice().buffer;
            }
            onSuccess(r);
        });
    }
}

export class BabylonLogTool {
    public static logFlag: boolean = false;
    public static warnFlag: boolean = false;
    public static log(...str: any[]) {
        if (BabylonLogTool.logFlag) {
            console.log(...str);
        }
    }
    public static warn(...str: any[]) {
        if (BabylonLogTool.logFlag) {
            console.warn(...str);
        }
    }
}