import { BABYLON } from "./babylon";
import { initAbstractMesh } from "./extends/abstruct_mesh";
import { initAnimation } from "./extends/animation";
import { initAnimationGroup } from "./extends/animation_group";
import { initBaseTexture } from "./extends/base_texture";
import { initBRDFTextureTools } from "./extends/brdf_texture_tools";
import { initCustomMaterial } from "./extends/custom_material";
import { initFileTools } from "./extends/file_tools";
import { initGeometry } from "./extends/geometry";
import { initGLTFLoader } from "./extends/gltf_loader";
import { initGLTFLoadAnimaitonChannel } from "./extends/gltf_loader_animation";
import { initGLTFLoadAnimaitonChannelNew } from "./extends/gltf_loader_animation_new";
import { initInstancedMesh } from "./extends/instanced_mesh";
import { initKTXTextureLoader } from "./extends/ktx_texture_loader";
import { LoadExtraMaterial } from "./extends/material";
import { initMesh } from "./extends/mesh";
import { initParticleSystem } from "./extends/particle_system";
import { initRuntimeAnimation } from "./extends/runtime_animation";
import { initScene } from "./extends/scene";
import { initTargetCamera } from "./extends/target_camera";
import { initTexture } from "./extends/texture";
import { initThinEngine } from "./extends/thin_engine";
import { OfflineProvider } from "./offline_provider";

// tslint:disable-next-line:max-func-body-length
export const initExtends = (usePBR: boolean = false) => {

    (<any>OfflineProvider.piBABYLONTextureRecord).format = () => {
        var aa = {}; OfflineProvider.piBABYLONTextureRecord.forEach((v,ke) => { var k = v[0] + "x" + v[1]; if (!aa[k]) {aa[k]=[]} aa[k].push(ke) }); for (let ll in aa) { aa[ll].sort() };
        return aa;
    }

    initAbstractMesh();
    initAnimationGroup();
    initAnimation();
    initBaseTexture();
    if (!usePBR) {
        initBRDFTextureTools();
    }
    initGeometry();
    initFileTools();
    initGLTFLoader();
    initInstancedMesh();
    initKTXTextureLoader();
    initMesh();
    initParticleSystem();
    initRuntimeAnimation();
    initScene();
    initTargetCamera();
    initTexture();
    // initThinEngine 在 initFileTools 之后
    initThinEngine();
    initCustomMaterial();
    initGLTFLoadAnimaitonChannel();
    initGLTFLoadAnimaitonChannelNew();
    LoadExtraMaterial.init();

    BABYLON.SkeletonMatrixCacheTable.prototype.getCache = function(actionName: string, frame: number) {
        let cacheDataInfo = this._cacheInfo[actionName];
        if (this._matrixData && cacheDataInfo) {
            const frameSize = cacheDataInfo.dataLength / cacheDataInfo.totalFrameCount;
            let start = (cacheDataInfo.dataStart + frameSize * frame) * 4;
            let end = start + frameSize * 4;
            let floatArray: Float32Array;
            // if (ArrayBuffer.isView(this._matrixData)) {
            //     floatArray = new Float32Array(this._matrixData.slice(start, end).buffer);
            // } else {
            //     floatArray = new Float32Array((<ArrayBuffer>this._matrixData).slice(start, end));
            // }
            let count = (end - start) / 4;
            floatArray = new Float32Array(count + 16);
            let temp = new Float32Array(this._matrixData.buffer, this._matrixData.byteOffset + start, count); // this._matrixData.slice(start, end);
            for (let i = 0; i < count; i++) {
                floatArray[i] = temp[i];
            }
            return floatArray;
        }
        else {
            return;
        }
    }

    BABYLON.Geometry.prototype.setVerticesBuffer = function (buffer, totalVertices) {
        if (totalVertices === void 0) { totalVertices = null; }
        var kind = buffer.getKind();
        if (this._vertexBuffers[kind]) {
            this._vertexBuffers[kind].dispose();
        }
        this._vertexBuffers[kind] = buffer;
        if (kind === BABYLON.VertexBuffer.PositionKind) {
            var data = <any>buffer.getData();
            if (totalVertices != null) {
                this._totalVertices = totalVertices;
            }
            else {
                if (data != null) {
                    this._totalVertices = data.length / (buffer.byteStride / 4);
                }
            }
            this._updateExtend(data);
            this._resetPointsArrayCache();
            var meshes = this._meshes;
            var numOfMeshes = meshes.length;
            for (var index = 0; index < numOfMeshes; index++) {
                var mesh: BABYLON.Mesh = meshes[index];
                if (!mesh.getBoundingInfo()?.isImport) {
                    mesh._boundingInfo = new BABYLON.BoundingInfo(this._extend.minimum, this._extend.maximum);
                }
                mesh._createGlobalSubMesh(false);
                mesh.computeWorldMatrix(true);
            }
        }
        this.notifyUpdate(kind);
        if (this._vertexArrayObjects) {
            this._disposeVertexArrayObjects();
            this._vertexArrayObjects = {}; // Will trigger a rebuild of the VAO if supported
        }
    }; 
};