import { ExtensionTools } from '../../interface/index';
import { BABYLON } from '../babylon';
import { BoundingInfo } from './bounding_info';
import { MathDataPool } from './math';

export const initAbstractMesh = () => {

    BABYLON.AbstractMesh.prototype._updateBoundingInfo = function() {
        const effectiveMesh = this._effectiveMesh;
        let matrix = effectiveMesh.worldMatrixFromCache;
        if (this._boundingInfo) {
            // PI_BEGIN - 非正常手段处理 Unity 中 SkinMesh 节点有旋转，但导入后生成的 Mesh 无旋转, gltf 规范要求蒙皮Mesh实际控制者为骨骼根节点
            // this._boundingInfo.update(effectiveMesh.worldMatrixFromCache);
            // /////////////////////////////
            if (this._scene.skeletonMeshHasRotate && effectiveMesh != this) {
                if ((<BABYLON.BoundingInfo>this._boundingInfo).isImport) {
                    if (ExtensionTools.Option.optimizeBoundingInfoOnlySphere) {
                        let center = (<BABYLON.BoundingInfo>this._boundingInfo).boundingSphere.center;
                        let extend = (<BABYLON.BoundingInfo>this._boundingInfo).boundingSphere.radius;
    
                        let newCenter = MathDataPool.Vector3(center.x, center.z, -center.y);
                        let newExtend = MathDataPool.Vector3(extend, extend, -extend);
    
                        (<BABYLON.BoundingInfo>this._boundingInfo).centerOn(newCenter, newExtend);
    
                        (<BABYLON.BoundingInfo>this._boundingInfo).isImport = undefined;
                    } else {
                        let center = (<BABYLON.BoundingInfo>this._boundingInfo).boundingBox.center;
                        let extend = (<BABYLON.BoundingInfo>this._boundingInfo).boundingBox.extendSize;
    
                        let newCenter = MathDataPool.Vector3(center.x, center.z, -center.y);
                        let newExtend = MathDataPool.Vector3(extend.x, extend.z, -extend.y);
    
                        (<BABYLON.BoundingInfo>this._boundingInfo).centerOn(newCenter, newExtend);
    
                        (<BABYLON.BoundingInfo>this._boundingInfo).isImport = undefined;
                    }
                }
            }

            // (<BABYLON.AbstractMesh>this).showBoundingBox = true;

            this._boundingInfo.update(effectiveMesh.worldMatrixFromCache);
            // PI_END
        }
        else {
            if (ExtensionTools.Option.optimizeBoundingInfoOnlySphere) {
                this._boundingInfo = new BoundingInfo(this.absolutePosition, this.absolutePosition, matrix);
            } else {
                this._boundingInfo = new BABYLON.BoundingInfo(this.absolutePosition, this.absolutePosition, matrix);
            }
        }
        this._updateSubMeshesBoundingInfo(matrix);
        return this;
    }
};