import { Engine } from '../engine';
import { BABYLON } from '../babylon';
import { ExtensionTools, Res, ResTab } from '../../interface/index';

export interface AnimationGroup extends BABYLON.AnimationGroup {
    keysResList: Res<any, BABYLON.IAnimationKey[]>[];
    piRecordRes(res: Res<any, BABYLON.IAnimationKey[]>);
}

// tslint:disable-next-line:max-func-body-length
export const initAnimationGroup = () => {
    BABYLON.AnimationGroup.prototype.goto = function(amount: number) {
        let frame = this._from + (this._to - this._from) * amount;
        this.goToFrame(frame);
        return this;
    }
    /**
     * Add an animation (with its target) in the group
     * @param animation defines the animation we want to add
     * @param target defines the target of the animation
     * @returns the TargetedAnimation object
     */
    // tslint:disable-next-line:typedef
    BABYLON.AnimationGroup.prototype.addTargetedAnimation = function (animation, target) {
        // PI_BEGIN  _targetedAnimations.animation变数组 相同节点的动画放一起
        const targetedAnimation: BABYLON.TargetedAnimation = <any>{
            animation: animation,
            target: target
        };

        const keys = animation.getKeys();
        // tslint:disable-next-line:no-invalid-this
        if (this._from > keys[0].frame) {
            // tslint:disable-next-line:no-invalid-this
            this._from = keys[0].frame;
        }
        // tslint:disable-next-line:no-invalid-this
        if (this._to < keys[keys.length - 1].frame) {
            // tslint:disable-next-line:no-invalid-this
            this._to = keys[keys.length - 1].frame;
        }
        // tslint:disable-next-line:no-invalid-this
        this._targetedAnimations.push(targetedAnimation);

        return targetedAnimation;
    };

    BABYLON.AnimationGroup.prototype.dispose = function (this: BABYLON.AnimationGroup) {
        (<any>this)._targetedAnimations = [];
        (<any>this)._animatables = [];
        var index = (<any>this)._scene.animationGroups.indexOf(this);
        if (index > -1) {
            (<any>this)._scene.animationGroups.splice(index, 1);
        }
        this.onAnimationEndObservable.clear();
        this.onAnimationGroupEndObservable.clear();
        this.onAnimationGroupPauseObservable.clear();
        this.onAnimationGroupPlayObservable.clear();
        this.onAnimationLoopObservable.clear();
        this.onAnimationGroupLoopObservable.clear();
        (<AnimationGroup>this).keysResList?.forEach((res) => {
            Engine.ofl.resTab.delete(res);
        });
        (<AnimationGroup>this).keysResList = undefined;
    };

    (<AnimationGroup>BABYLON.AnimationGroup.prototype).piRecordRes = function (this: AnimationGroup, res: Res<any, BABYLON.IAnimationKey[]>) {
        if (!this.keysResList) {
            this.keysResList = [];
        }
        this.keysResList.push(res);
    };
};