import { BABYLON } from '../babylon';
import { BABYLON_TEXTURE_TYPE, ITextureArgCfg, PiBABYLONTexture } from '../../util/res_mgr';
import { Scene } from '../scene';
import { Engine } from '../engine';
import { ExtensionTools, Res, ResTab } from '../../interface/index';

declare let module;
let logger;
ExtensionTools.RegistStartUp(() => { logger = ExtensionTools.Logger(module.name, "BABYLON"); });

export const initBaseTexture = () => {

    BABYLON.BaseTexture.prototype.releaseInternalTexture = function (this: BABYLON.BaseTexture) {
        if ((<any>this)._pi_res) {
            const resTab = Engine.ofl.resTab;
            resTab.delete((<any>this)._pi_res);
            (<any>this)._pi_res = null;
        }
        if (this._texture) {
            // this._texture.dispose();
            this._texture = null;
        }
    };

    BABYLON.BaseTexture.prototype.delayLoad = function (this: BABYLON.BaseTexture, res?: Res<ITextureArgCfg, PiBABYLONTexture>) {
        if ((<any>this)._pi_res) {
            // debugger;
            console.error('BaseTexture.prototype.delayLoad');
        }
        (<any>this)._pi_res = res;
    };

    /**
     * @hidden
     */
    // tslint:disable-next-line:typedef
    BABYLON.BaseTexture.prototype._getFromCache = function (this: BABYLON.BaseTexture, url, noMipmap, sampling?, invertY?) {
        // tslint:disable-next-line:no-invalid-this
        if (!this._scene) {
            return null;
        }

        if (!!url && this.getClassName() == "Texture") {
            const scene = (<Scene>this._scene);
            const engine = scene.getEngine();
            const resTab: ResTab = Engine.ofl.resTab;

            const noMipmap = this.noMipmap;
            const invertY = (<any>this).invertY;
            const samplingMode = (<any>this).samplingMode;

            resTab.load(BABYLON_TEXTURE_TYPE, url, [<ITextureArgCfg>{ engine, invertY, scene, noMipmap, samplingMode, internalTexture: undefined }])
                .then((res: Res<ITextureArgCfg, PiBABYLONTexture>) => {
                    (<any>this).delayLoad(res);
                })
                .catch((e) => {
                    console.error(`BaseTexture getFromCache error: ${e}`);
                });
        }

        // PI_END
        return null;
    };
};