
import { BABYLON } from "../babylon";
import { BABYLON_GEOMETRY_TYPE } from "../../util/res_mgr";
import { IAccessor } from "../../gltf/glTF2Interface";
import { ExtensionTools, Res, ResTab } from '../../interface/index';

export class VertexBuffer extends BABYLON.VertexBuffer {
    public static resTab: ResTab;
    public res: Res<any, BABYLON.Buffer>;
    
    public dispose(): void {
        if ((<any>this)._ownsBuffer) {
            this._buffer.dispose();
        } else if (this.res) {
            VertexBuffer.resTab.delete(this.res);
            this.res = undefined;
        }
    }
}

export interface IBufferArg {
    kind: string,
    callback?: (accessor: IAccessor) => void,
    data: Float32Array,
    engine: BABYLON.Engine,
}
function loadBuffer(
    tab: ResTab, _type: 'BABYLON_GEOMETRY_TYPE', _name: string, args: IBufferArg
): Promise<BABYLON.Buffer> {
    return Promise.resolve(new BABYLON.Buffer(args.engine, args.data, false));
}

ExtensionTools.RegistStartUp(() => {
    ExtensionTools.ResLoadRegist<typeof BABYLON_GEOMETRY_TYPE, IBufferArg, BABYLON.Buffer>(
        BABYLON_GEOMETRY_TYPE,
        loadBuffer,
        (link: BABYLON.Buffer) => link.dispose(),
        (link: BABYLON.Buffer) => {
            if ((<ArrayBuffer | ArrayBufferView>link._data).byteLength) {
                return (<ArrayBuffer | ArrayBufferView>link._data).byteLength;
            } else {
                return (<number[]>link._data).length * 4;
            }
        }
    );
    
    VertexBuffer.resTab = ExtensionTools.CreateResTab();
})
