import { BABYLON } from '../babylon';
import { ExtensionTools } from '../../interface/index';

// tslint:disable-next-line:max-func-body-length
export const initFileTools = () => {
    BABYLON.FileTools.LoadImage = (input: string | ArrayBuffer | Blob, onLoad: (img: HTMLImageElement) => void, onError: (message?: string, exception?: any) => void, offlineProvider: BABYLON.Nullable<BABYLON.IOfflineProvider>): HTMLImageElement => {

        const resTab = ExtensionTools.CreateResTab();

        if ((<string>input).startsWith("data:image")) {
            let image = new Image();
            image.onload = () => {
                onLoad(image);
                image.onerror = undefined;
                image.src = "";
            };
            image.onerror = (e) => {
                onError(`${input} : ${e}`);
            };
            image.src = <string>input;
        } else {
            resTab.load(ExtensionTools.RES_IMAGE, <string>input, []).then((res) => {
                onLoad(res.link);
                Promise.resolve().then(() => { resTab.clear() });
            }).catch((e) => {
                onError(`${input} : ${e}`);
            });
        }

        return;
    };
};