
import { IMaterial } from "../../gltf/glTF2Interface";
import { PI_material } from "../../gltf/extensions/material";
import { BABYLON } from "../babylon";
import { GLTFLoader } from "../../gltf/glTFLoader";
import { IPITextureInfo } from "../../material_info";

export class LoadExtraMaterial {
    public static init() {
        PI_material.prototype.createMaterial = function (this: PI_material, context: string, indexMat: number, indexNode: number, material: IMaterial, babylonDrawMode: number, notParticleSys?: boolean) {
            if (material.extensions && material.extensions.PI_material) {
                let materialMap = this._loader.babylonScene.pi_material_map;

                let materialType = material.extensions.PI_material.type;
                let materialInfo = material.extensions.PI_material;
                let instanceID = materialInfo.instanceID;

                if (instanceID && materialMap.has(instanceID)) {
                    return materialMap.get(instanceID);
                }


                let key = materialType;
                let call = LoadExtraMaterial.loadCalls.get(key);
                if (call) {
                    let mat = call(context, this._loader.babylonScene, material.name, indexMat, materialInfo, babylonDrawMode, notParticleSys);
                    return mat;
                }

                // const factory = PI_material.MaterialFactoryMap.get(materialType);

                // if (factory) {
                //     let mat = factory(context, this._loader.babylonScene, material.name, indexMat, materialInfo, babylonDrawMode);
                //     if (instanceID && mat) {
                //         materialMap.set(instanceID, mat);
                //     }
                //     if (materialInfo.zWrite) {
                //         (<BABYLON.StandardMaterial>mat).forceDepthWrite = true;
                //     }

                //     return mat;
                // }
                else {
                    let mat = new BABYLON.StandardMaterial(material.name, this._loader.babylonScene); // PI_material._createMaterialPiShader(context, this._loader.babylonScene, material.name, indexMat, materialInfo, babylonDrawMode);
                    if (instanceID && mat) {
                        materialMap.set(instanceID, mat);
                    }
                    if (materialInfo.zWrite) {
                        (<BABYLON.StandardMaterial>mat).forceDepthWrite = true;
                    }

                    return mat;
                }
            }
            else if (material.extras) {
                let matInfo = material;
                let flag = false;
                for (let key in matInfo.extras) {
                    let call = LoadExtraMaterial.loadCalls.get(key);
                    if (call) {
                        let mat = call(context, this._loader.babylonScene, material.name, indexMat, matInfo.extras[key], babylonDrawMode, notParticleSys);
                        return mat;
                        // if (mat != undefined) {
                        //     flag = true;
                        // }
                    } else if (matInfo.extras[key].type) {
                        let key2 = matInfo.extras[key].type;
                        let call = LoadExtraMaterial.loadCalls.get(key2);
                        if (call) {
                            let mat = call(context, this._loader.babylonScene, material.name, indexMat, matInfo.extras[key], babylonDrawMode, notParticleSys);
                            return mat;
                            // if (mat != undefined) {
                            //     flag = true;
                            // }
                        }
                    }
                }

                // if (!flag) {
                console.error("Cannt Load Material !");
                let mat = new BABYLON.StandardMaterial(matInfo.name, this._loader.babylonScene);
                return mat;
                // }

            }
            else {
                return null;
            }
        }
        PI_material.prototype.loadMaterialPropertiesAsync = function (this: PI_material, context: string, material: IMaterial, mat: BABYLON.Material, notParticleSys?: boolean) {
            if (material.extensions && material.extensions.PI_material) {
                let materialType = material.extensions.PI_material.type;
                let materialInfo = material.extensions.PI_material;

                // let instanceID   = materialInfo.instanceID;
                // if (instanceID && mat.pi_instance_id == instanceID) {
                //     return Promise.resolve();
                // }
                // mat.pi_instance_id = instanceID;

                let key = materialType;
                let call = LoadExtraMaterial.loadPropertyCalls.get(key);
                if (call) {
                    return call(context, materialInfo, mat, this._loader, notParticleSys);
                }

                // let loadAsync = PI_material.MaterialPropertiesLoadAsyncMap.get(materialType);
                // if (loadAsync) {
                //     return loadAsync(context, materialInfo, <BABYLON.StandardMaterial>mat, this._loader);
                // }
                // else {
                //     return ;// PI_material._loadMaterialPropertiesAsync(context, material, <BABYLON.StandardMaterial>mat, this._loader);
                // }
            }
            else if (material.extras) {
                let matInfo = material;
                for (let key in matInfo.extras) {
                    let call = LoadExtraMaterial.loadPropertyCalls.get(key);
                    if (call) {
                        return call(context, matInfo.extras[key], mat, this._loader, notParticleSys);
                    } else if (matInfo.extras[key].type) {
                        let key2 = matInfo.extras[key].type;
                        let call = LoadExtraMaterial.loadPropertyCalls.get(key2);
                        if (call) {
                            return call(context, matInfo.extras[key], mat, this._loader, notParticleSys);
                        }
                    }
                }
            }
            else {
                return null;
            }
        }
    }
    public static loadCalls: Map<string, (context: string, babylonScene: BABYLON.Scene, materialName: string, index: number, materialInfo: BABYLON.GLTF2.Loader.IPIMaterial, babylonDrawMode: number, notParticleSys?: boolean) => BABYLON.Material> = new Map();
    public static loadPropertyCalls: Map<string, (context: string, material: BABYLON.GLTF2.Loader.IPIMaterial, mat: BABYLON.Material, _loader: GLTFLoader, notParticleSys?: boolean) => void> = new Map();

    public static diffuseTexture(mat: { diffuseTexture: BABYLON.Texture, useAlphaFromDiffuseTexture?: boolean, diffuseTexturePremultiplyAlpha?: boolean }, _loader: GLTFLoader, context: string,
        info: {
            diffuseTexture: IPITextureInfo, diffuseLevel?: number, diffuseOU?: number, diffuseOV?: number,
            useAlphaFromDiffuseTexture?: boolean, diffusePreAlpha?: boolean
        }
    ) {

        if (info.diffuseTexture) {
            let texture = <BABYLON.Texture>_loader.loadTextureInfoAsync(context + "/PI_material/emissionTexture", info.diffuseTexture);
            if (mat.diffuseTexture) {
                mat.diffuseTexture.dispose();
            }
            mat.diffuseTexture = texture;
            let float = info.diffuseLevel;
            if (float) {
                texture.level = float;
            }
            float = info.diffuseOU;
            if (float != undefined) {
                texture.offsetSpeedU = float; // 1000 / float;
            }
            float = info.diffuseOV;
            if (float != undefined) {
                texture.offsetSpeedV = float; // 1000 / float;
            }
            let numbers = info.diffuseTexture.scale;
            if (numbers) {
                texture.uScale = numbers[0];
                texture.vScale = numbers[1];
            }
            numbers = info.diffuseTexture.offset;
            if (numbers) {
                texture.uOffset = numbers[0];
                texture.vOffset = numbers[1];
            }
            let tilloff = info.diffuseTexture.extras?.atlasTilloff;
            if (tilloff) {
                texture.atlasTilloff = tilloff;
            }
            if (info.useAlphaFromDiffuseTexture) {
                mat.useAlphaFromDiffuseTexture = info.useAlphaFromDiffuseTexture;
            }
            if (info.diffuseTexture.hasAlpha) {
                texture.hasAlpha = true;
            }
            if (info.diffusePreAlpha) {
                mat.diffuseTexturePremultiplyAlpha = true;
            }
        } else {
            // return Promise.resolve(null);
        }
    }
    public static emissiveTexture(mat: { emissiveTexture: BABYLON.Texture }, _loader: GLTFLoader, context: string, info: { emissionTexture?: any, emissiveMapLevel?: number, emissionOU?: number, emissionOV?: number }) {

        if (info.emissionTexture) {
            let texture = <BABYLON.Texture>_loader.loadTextureInfoAsync(context + "/PI_material/emissionTexture", info.emissionTexture);
            if (mat.emissiveTexture) {
                mat.emissiveTexture.dispose();
            }
            mat.emissiveTexture = texture;
            let float = info.emissiveMapLevel;
            if (float) {
                texture.level = float;
            }
            float = info.emissionOU;
            if (float != undefined) {
                texture.offsetSpeedU = float; // 1000 / float;
            }
            float = info.emissionOV;
            if (float != undefined) {
                texture.offsetSpeedV = float; // 1000 / float;
            }
            let numbers = info.emissionTexture.scale;
            if (numbers) {
                texture.uScale = numbers[0];
                texture.vScale = numbers[1];
            }
            numbers = info.emissionTexture.offset;
            if (numbers) {
                texture.uOffset = numbers[0];
                texture.vOffset = numbers[1];
            }
            let tilloff = info.emissionTexture.extras?.atlasTilloff;
            if (tilloff) {
                texture.atlasTilloff = tilloff;
            }
        } else {
            // return Promise.resolve(null);
        }
    }

    public static opacityTexture(mat: { opacityTexture: BABYLON.Texture, opacityTextureApplyRGB?: boolean }, _loader: GLTFLoader, context: string, info: { opacityTexture?: any, opacityLevel?: number, opacityOU?: number, opacityOV?: number, opacityApplyRGB?: boolean }) {
        if (info.opacityTexture) {
            let texture = <BABYLON.Texture>_loader.loadTextureInfoAsync(context + "/PI_material/opacityTexture", info.opacityTexture);
            if (mat.opacityTexture) {
                mat.opacityTexture.dispose();
            }
            mat.opacityTexture = texture;
            let float = info.opacityLevel;
            if (float) {
                texture.level = float;
            }
            float = info.opacityOU;
            if (float != undefined) {
                texture.offsetSpeedU = float; // 1000 / float;
            }
            float = info.opacityOV;
            if (float != undefined) {
                texture.offsetSpeedV = float; // 1000 / float;
            }
            let numbers = info.opacityTexture.scale;
            if (numbers) {
                texture.uScale = numbers[0];
                texture.vScale = numbers[1];
            }
            numbers = info.opacityTexture.offset;
            if (numbers) {
                texture.uOffset = numbers[0];
                texture.vOffset = numbers[1];
            }
            let tilloff = info.opacityTexture.extras?.atlasTilloff;
            if (tilloff) {
                texture.atlasTilloff = tilloff;
            }
            if (info.opacityApplyRGB) {
                mat.opacityTextureApplyRGB = true;
            }
        } else {
            // return Promise.resolve(null);
        }
    }

    public static maskTexture(mat: { maskTexture: any }, _loader: GLTFLoader, context: string, info: { maskTexture?: IPITextureInfo, emissiveMapLevel?: number, emissionOU?: number, emissionOV?: number }) {
        if (info.maskTexture) {
            let texture = <BABYLON.Texture>_loader.loadTextureInfoAsync(context + "/PI_material/maskTexture", info.maskTexture);
            if (mat.maskTexture) {
                mat.maskTexture.dispose();
            }
            (<BABYLON.StandardMaterial>mat).maskTexture = texture;
            let numbers = info.maskTexture.scale;
            if (numbers) {
                texture.uScale = numbers[0];
                texture.vScale = numbers[1];
            }
            numbers = info.maskTexture.offset;
            if (numbers) {
                texture.uOffset = numbers[0];
                texture.vOffset = numbers[1];
            }
            let tilloff = info.maskTexture.extras?.atlasTilloff;
            if (tilloff) {
                texture.atlasTilloff = tilloff;
            }
            // });
        } else {
            // return Promise.resolve(null);
        }
    }
    
    public static mask2Texture(mat: { mask2Texture: any }, _loader: GLTFLoader, context: string, info: { mask2Texture?: IPITextureInfo, emissiveMapLevel?: number, emissionOU?: number, emissionOV?: number }) {
        if (info.mask2Texture) {
            let texture = <BABYLON.Texture>_loader.loadTextureInfoAsync(context + "/PI_material/mask2Texture", info.mask2Texture);
            if (mat.mask2Texture) {
                mat.mask2Texture.dispose();
            }
            (<any>mat).mask2Texture = texture;
            let numbers = info.mask2Texture.scale;
            if (numbers) {
                texture.uScale = numbers[0];
                texture.vScale = numbers[1];
            }
            numbers = info.mask2Texture.offset;
            if (numbers) {
                texture.uOffset = numbers[0];
                texture.vOffset = numbers[1];
            }
            let tilloff = info.mask2Texture.extras?.atlasTilloff;
            if (tilloff) {
                texture.atlasTilloff = tilloff;
            }
            // });
        } else {
            // return Promise.resolve(null);
        }
    }

    public static state(mat: BABYLON.StandardMaterial, info: any, _loader: GLTFLoader) {
        let float = info.alpha;
        if (float != undefined) {
            mat.alpha = float;
        }
        float = info.alphaCutOff;
        if (float != undefined) {
            mat.alphaCutOff = float;
        }

        mat.backFaceCulling = true;
        if (info.cull == "off") {
            mat.backFaceCulling = false;
            // cullMode = CullMode.Off;
        } else if (info.cull == "front") {
            // cullMode = CullMode.Front;
            mat.sideOrientation = !_loader.babylonScene.useRightHandedSystem ? BABYLON.Material.CounterClockWiseSideOrientation : BABYLON.Material.ClockWiseSideOrientation;
        }

        mat.forceDepthWrite
        let depthTest = true;
        let depthWrite = false;
        if (info.zWrite) {
            depthWrite = true;
            mat.forceDepthWrite = true;
        }

        let alphaMode = info.alphaMode == undefined ? 0 : info.alphaMode;
        let renderQueue = info.renderQueue == undefined ? 3000 : info.renderQueue;

        if (info.renderQueue) {
            if (info.renderQueue < 2450) {
                mat.transparencyMode = BABYLON.Material.MATERIAL_OPAQUE;
            }
            else if (info.renderQueue < 3000) {
                mat.transparencyMode = BABYLON.Material.MATERIAL_ALPHATEST;
            }
            else {
                if (info.alphaClip) {
                    mat.transparencyMode = BABYLON.Material.MATERIAL_ALPHATESTANDBLEND;
                } else {
                    mat.transparencyMode = BABYLON.Material.MATERIAL_ALPHABLEND;
                }
                mat.needAlphaBlending = () => true;
            }
        }

        mat.alphaMode = alphaMode;
    }
}