import { BABYLON } from '../babylon';

// tslint:disable-next-line:max-func-body-length
export const initParticleSystem = () => {

    // tslint:disable-next-line:max-func-body-length
    (<any>BABYLON.ParticleSystem.prototype)._createVertexBuffers = function() {
        // tslint:disable-next-line:no-invalid-this
        this._vertexBufferSize = this._useInstancing ? 10 : 12;
        // tslint:disable-next-line:no-invalid-this
        if (this._isAnimationSheetEnabled) {
            // tslint:disable-next-line:no-invalid-this
            this._vertexBufferSize += 1;
        }

        // tslint:disable-next-line:no-invalid-this
        if (!this._isBillboardBased || this.billboardMode === BABYLON.ParticleSystem.BILLBOARDMODE_STRETCHED) {
            // tslint:disable-next-line:no-invalid-this
            this._vertexBufferSize += 3;
        }

        // tslint:disable-next-line:no-invalid-this
        if (this._useRampGradients) {
            // tslint:disable-next-line:no-invalid-this
            this._vertexBufferSize += 4;
        }

        // tslint:disable-next-line:no-invalid-this
        const engine = this._scene.getEngine();
        // tslint:disable-next-line:no-invalid-this
        this._vertexData = new Float32Array(this._capacity * this._vertexBufferSize * (this._useInstancing ? 1 : 4));
        // tslint:disable-next-line:no-invalid-this
        this._vertexBuffer = new BABYLON.Buffer(engine, this._vertexData, true, this._vertexBufferSize);

        let dataOffset = 0;
        // tslint:disable-next-line:no-invalid-this
        const positions = this._vertexBuffer.createVertexBuffer(BABYLON.VertexBuffer.PositionKind, dataOffset, 3, this._vertexBufferSize, this._useInstancing);
        // tslint:disable-next-line:no-invalid-this
        this._vertexBuffers[BABYLON.VertexBuffer.PositionKind] = positions;
        dataOffset += 3;

        // tslint:disable-next-line:no-invalid-this
        const colors = this._vertexBuffer.createVertexBuffer(BABYLON.VertexBuffer.ColorKind, dataOffset, 4, this._vertexBufferSize, this._useInstancing);
        // tslint:disable-next-line:no-invalid-this
        this._vertexBuffers[BABYLON.VertexBuffer.ColorKind] = colors;
        dataOffset += 4;

        // tslint:disable-next-line:no-invalid-this
        const options = this._vertexBuffer.createVertexBuffer('angle', dataOffset, 1, this._vertexBufferSize, this._useInstancing);
        // tslint:disable-next-line:no-invalid-this
        this._vertexBuffers.angle = options;
        dataOffset += 1;

        // tslint:disable-next-line:no-invalid-this
        const size = this._vertexBuffer.createVertexBuffer('size', dataOffset, 2, this._vertexBufferSize, this._useInstancing);
        // tslint:disable-next-line:no-invalid-this
        this._vertexBuffers.size = size;
        dataOffset += 2;

        // tslint:disable-next-line:no-invalid-this
        if (this._isAnimationSheetEnabled) {
            // tslint:disable-next-line:no-invalid-this
            const cellIndexBuffer = this._vertexBuffer.createVertexBuffer('cellIndex', dataOffset, 1, this._vertexBufferSize, this._useInstancing);
            // tslint:disable-next-line:no-invalid-this
            this._vertexBuffers.cellIndex = cellIndexBuffer;
            dataOffset += 1;
        }

        // tslint:disable-next-line:no-invalid-this
        if (!this._isBillboardBased || this.billboardMode === BABYLON.ParticleSystem.BILLBOARDMODE_STRETCHED) {
            // tslint:disable-next-line:no-invalid-this
            const directionBuffer = this._vertexBuffer.createVertexBuffer('direction', dataOffset, 3, this._vertexBufferSize, this._useInstancing);
            // tslint:disable-next-line:no-invalid-this
            this._vertexBuffers.direction = directionBuffer;
            dataOffset += 3;
        }

        // tslint:disable-next-line:no-invalid-this
        if (this._useRampGradients) {
            // tslint:disable-next-line:no-invalid-this
            const rampDataBuffer = this._vertexBuffer.createVertexBuffer('remapData', dataOffset, 4, this._vertexBufferSize, this._useInstancing);
            // tslint:disable-next-line:no-invalid-this
            this._vertexBuffers.remapData = rampDataBuffer;
            dataOffset += 4;
        }

        let offsets: BABYLON.VertexBuffer;
        // tslint:disable-next-line:no-invalid-this
        if (this._useInstancing) {
            // PI_BEGIN - 粒子系统 非固定 UV
            // let spriteData = new Float32Array([0, 0, 1, 0, 1, 1, 0, 1]);
            // //////////////////////////////
            // tslint:disable-next-line:no-invalid-this
            const uScale = this.particleTexture ? this.particleTexture.uScale : 1;
            // tslint:disable-next-line:no-invalid-this
            const vScale = this.particleTexture ? this.particleTexture.vScale : 1;
            // tslint:disable-next-line:no-invalid-this
            const uOffset = this.particleTexture ? this.particleTexture.uOffset : 0;
            // tslint:disable-next-line:no-invalid-this
            const vOffset = this.particleTexture ? this.particleTexture.vOffset : 0;
            // tslint:disable-next-line:binary-expression-operand-order
            const spriteData = new Float32Array([0 + uOffset, 0 + vOffset, uScale + uOffset, 0 + vOffset, uScale + uOffset, vScale + vOffset, uOffset, vScale + vOffset]);
            // PI_END
            // tslint:disable-next-line:no-invalid-this
            this._spriteBuffer = new BABYLON.Buffer(engine, spriteData, false, 2);
            // tslint:disable-next-line:no-invalid-this
            offsets = this._spriteBuffer.createVertexBuffer('offset', 0, 2);
        } else {
            // tslint:disable-next-line:no-invalid-this
            offsets = this._vertexBuffer.createVertexBuffer('offset', dataOffset, 2, this._vertexBufferSize, this._useInstancing);
            dataOffset += 2;
        }
        // tslint:disable-next-line:no-invalid-this
        this._vertexBuffers.offset = offsets;
    };

    if (BABYLON.BaseParticleSystem) {
        // 调整默认速率
        BABYLON.BaseParticleSystem.BASE_UPDATE_SPEED = 1 / 60;
    }
};