/**
 *
 */

// ###################################################
// import { RES_FILE, RES_IMAGE } from 'pi_utils/res/res_load';
// import { PISYS } from 'pi_sys/index';
import { BABYLON_ANIMATION_TYPE } from '../util/res_mgr';
import { BABYLON } from './babylon';
import { ExtensionTools, Res, ResTab } from '../interface/index';
// ###################################################
declare let module;
let logger;
ExtensionTools.RegistStartUp(() => { logger = ExtensionTools.Logger(module.name, "BABYLON"); });

// ###################################################

// 实现babylon的脱机提供者， 从外部ResTab获取数据
export class OfflineProvider implements BABYLON.IOfflineProvider {
    public static onoffCollectInfo = false;
    public static piBABYLONResCollect = {};
    public static piBABYLONTextureRecord = new Map();
    public static piBABYLONGeometryRecord = new Map();
    public static piBABYLONFileRecord = new Map();

    /**
     * 外部可拓展的gltf预处理接口
     * @param json GLTF 数据
     * @returns
     */
    public static GLTFFormat: (path: string, json: BABYLON.GLTF2.Loader.IGLTF) => BABYLON.GLTF2.Loader.IGLTF = function (path: string, json: BABYLON.GLTF2.Loader.IGLTF) {
        return json;
    };
    public enableSceneOffline: boolean = true;
    public enableTexturesOffline: boolean = true;
    public resTab: ResTab = ExtensionTools.CreateResTab();

    public animMap: Map<string, Res<void, BABYLON.IAnimationKey[]>> = new Map();
    // public animCounter: {[index: string]: number} = {};

    constructor(resTab: ResTab) {
        this.resTab = resTab;
    }

    public open(successCallback: () => void, _errorCallback: () => void) {
        successCallback();
    }

    public loadImage(url: string, image: HTMLImageElement, callBack?: (img: HTMLImageElement) => void) {
        // TODO：将url转成小游戏
        const _u = url;
        const _window = <any>window;

        this.resTab.load(ExtensionTools.RES_IMAGE, url, [image])
            // loadImageRes(this.resTab, url, [image])
            .then((res) => {
                // if (res.link !== image) {
                //     image.src = res.link.src;
                //     BabylonLogTool.log(`wrong with image ${url} ${res.link === image}`);
                // }
                if (res.link) {
                    callBack && callBack(res.link);
                    this.resTab.delete(res);
                } else {
                    console.error("OfflineProvider.loadImage Error");
                }
            });
        // if (_window.pi_modules && _window.pi_modules.load && _window.pi_modules.load && _window.pi_modules.load.exports.isWXMinigame && _window.pi_modules.load.exports.isWXMinigame()) {
        //     _window.pi_modules['pi/minigame/adapter'].exports.replaceImagePath(url,(a) => {
        //         image.src = a;
        //     },() => {
        //         image.src = url;
        //     });
        // } else {
        // image.src = _u;
        // }
    }

    public loadFile(url: string, sceneLoaded: (data: any) => void, _progressCallBack?: (data: any) => void, errorCallback?: (e) => void, _useArrayBuffer?: boolean) {
        const restype = ExtensionTools.RES_FILE;
        // if (url.endsWith('.png') || url.endsWith('.jpg')) {
        //     restype = 'image';
        // } else {
        //     restype = RES_TYPE_FILE;
        // }
        let res = this.resTab.get<any, ArrayBuffer>(restype, url);

        if (res) {
            Promise.resolve().then(() => {
                if (url.endsWith('.gltf')) {
                    const str = ExtensionTools.UTF8Decode(res.link);
                    let json = JSON.parse(str);
                    json = OfflineProvider.GLTFFormat(url, json);
                    // sceneLoaded(str); // 4.0.0
                    sceneLoaded({ json }); // 4.2.0
                } else {
                    if (ArrayBuffer.isView(res.link)) {
                        res.link = (<Uint8Array>res.link).slice().buffer;
                    }
                    sceneLoaded(res.link);
                }
                this.resTab.delete(res);
            });
        } else {
            this.resTab.load(restype, url, [url])
                .then(
                    (r) => {
                        if (OfflineProvider.onoffCollectInfo) {
                            OfflineProvider.piBABYLONFileRecord.set(url, (<Uint8Array>r.link).byteLength);
                        }
                        if (url.endsWith('.gltf')) {
                            const str = ExtensionTools.UTF8Decode(r.link);
                            let json = JSON.parse(str);
                            json = OfflineProvider.GLTFFormat(url, json);
                            // sceneLoaded(str); // 4.0.0
                            sceneLoaded({ json }); // 4.2.0
                        } else {
                            if (ArrayBuffer.isView(r.link)) {
                                r.link = (<Uint8Array>r.link).slice().buffer;
                            }
                            sceneLoaded(r.link);
                        }
                        this.resTab.delete(r);
                    }
                )
                .catch((e) => {
                    console.error(`${url} : ${e}`);
                    errorCallback(e);
                });
        }
    }

    public loadAnimationKeys(name: string): Res<void, BABYLON.IAnimationKey[]> {
        // const res = <any>this.animMap.get(name);

        // if (res) {
        //     // this.animCounter[name]++;
        //     return res.link;
        // } else {
        //     return null;
        // }

        // this.resTab.load(BABYLON_ANIMATION_TYPE, name, []);

        // const r = this.resTab.get(BABYLON_ANIMATION_TYPE, name);
        // if (r) {
        //     return r.link;
        // } else {
        //     return null;
        // }

        return this.resTab.get(BABYLON_ANIMATION_TYPE, name);
    }

    public createAnimationKeys(name: string, keys: BABYLON.IAnimationKey[]): Promise<Res<any, BABYLON.IAnimationKey[]>> {

        // const res = <Res<any, BABYLON.IAnimationKey[]>>(this.resTab).createRes(name, BABYLON_ANIMATION_TYPE, keys, []);//, (keys: any[]) => {  console.log(`destroy animation keys`)}, () => { return 10240 });
        // this.animCounter[name] = 1;
        // this.animMap.set(name, res);

        // (<any>this.resTab).addRes(res.key, res);

        // this.resTab.load(BABYLON_ANIMATION_TYPE, name, [keys])
        //     .catch(() => {
        //         BabylonLogTool.log('createAnimationKeys fail!, name:', name);
        //     });

        return this.resTab.load(
            BABYLON_ANIMATION_TYPE,
            name,
            [keys]
        );
        // return res;
    }

    public releaseAnimationKeys(name) {
        //
        console.info(`releaseAnimationKeys ${name}`);
    }

    public release() {
        let resTab = this.resTab;
        Promise.resolve().then(() => { resTab.clear() });
    }
}
