/**
 *
 */
import { BABYLON } from './render3d/babylon';
import { ITransformCfg } from './interface/modelobj';
import { ModelObj } from './model_obj';
import { OfflineProvider } from './render3d/offline_provider';
import { GLTFLoader, GLTFLodResult } from './gltf/glTFLoader';
import { ITransformNode } from './interface/transform';
import { ExtensionTools } from './interface/index';
import { MathDataPool } from './render3d/extends/math';

declare let module;
let logger;
ExtensionTools.RegistStartUp(() => { logger = ExtensionTools.Logger(module.name, "BABYLON"); });

// tslint:disable-next-line:no-unnecessary-class
export class NodeTools {
    public static sceneAsRightHand: boolean = false;
    /**
     * 旋转 Mesh - 自转
     * @param mesh 目标mesh
     * @param rotate 旋转参数： [ x, y, z ]
     */
    public static rotateNode(node: ITransformNode, rotate: number[]) {
        if (!node.rotationQuaternion) {
            node.rotationQuaternion = new BABYLON.Quaternion();
        }

        NodeTools.rotateQuaternion(<BABYLON.Quaternion>node.rotationQuaternion, rotate);
    }
    /**
     * 旋转 Mesh - 自转
     * @param quaternion 目标 Quaternion
     * @param rotate 旋转参数： [ x, y, z ]
     */
    public static rotateQuaternion(quaternion: BABYLON.Quaternion, rotate: number[]) {
        // YXZ
        if (NodeTools.sceneAsRightHand) {
            BABYLON.Quaternion.RotationYawPitchRollToRef(rotate[1] - Math.PI, rotate[0], -rotate[2], quaternion);
        } else {
            BABYLON.Quaternion.RotationYawPitchRollToRef(rotate[1], rotate[0], rotate[2], quaternion);
        }
    }
    public static positNode(node: ITransformNode, data: number[]) {
        if (NodeTools.sceneAsRightHand) {
            (<BABYLON.Vector3>node.position).copyFromFloats(-data[0], data[1], -data[2]);
        } else {
            (<BABYLON.Vector3>node.position).copyFromFloats(data[0], data[1], data[2]);
        }
    }
    public static scaleNode(node: ITransformNode, data: number[]) {
        if (NodeTools.sceneAsRightHand) {
            (<BABYLON.Vector3>node.scaling).copyFromFloats(data[0], data[1], -data[2]);
        } else {
            (<BABYLON.Vector3>node.scaling).copyFromFloats(data[0], data[1], data[2]);
        }
    }
    public static nodeLookAt(node: BABYLON.TransformNode, data: number[]) {
        if (NodeTools.sceneAsRightHand) {
            node.lookAt(MathDataPool.Vector3(data[0], data[1], -data[2]));
        } else {
            node.lookAt(MathDataPool.Vector3(data[0], data[1], data[2]));
        }
    }
    public static nodeTransform(node: BABYLON.TransformNode, transform: ITransformCfg) {
        if (transform !== undefined) {
            if (transform.position) {
                NodeTools.positNode(node, transform.position);
            }
            if (transform.scale) {
                NodeTools.scaleNode(node, transform.scale);
            }
            if (transform.rotate) {
                NodeTools.rotateNode(node, transform.rotate);
            }
        }
    }
}

// tslint:disable-next-line:no-unnecessary-class
export class CameraTool {
    /**
     * 修改 正交相机视角
     * @param size 正交相机视角大小
     */
    public static changeCameraOrth(camera: BABYLON.FreeCamera, size: number, width: number, height: number, sizeForWidth: boolean = false) {
        const yDistance = sizeForWidth ? Math.round(size / width / height * 100) / 100 : size;
        const xDistance = sizeForWidth ? size : Math.round(size * width / height * 100) / 100;
        camera.orthoLeft = -xDistance;
        camera.orthoRight = xDistance;
        camera.orthoTop = yDistance;
        camera.orthoBottom = -yDistance;
    }

    public static getAccurateSize(xDistance: number, w: number, h: number) {
        return xDistance / (w / h);
    }
}

// tslint:disable-next-line:no-unnecessary-class
export class LoaderTool {
    private static _ResPath        = 'app/scene_res/res/';
    private static _SceneResPath   = 'scenes/';
    private static _ModelResPath   = 'models/';
    private static _NodeResPath    = 'models/';
    private static _EffectResPath  = 'effects/';
    public static UseNewGLTFLoader = false;

    public static getResPath() {
        return this._ResPath;
    }
    public static getSceneResPath() {
        return this._SceneResPath;
    }
    public static getModelResPath() {
        return this._ModelResPath;
    }
    public static getNodeResPath() {
        return this._NodeResPath;
    }
    public static getEffectResPath() {
        return this._EffectResPath;
    }
    /**
     *
     * @param model 目标模型
     * @param targetPath 如果赋值(完整路径)将忽略 model 配置的路径
     * @param targetFile 如果赋值(完整路径)将忽略 model 配置的路径
     */
    public static loadScene(model: any, targetPath?: string, targetFile?: string) {
        // const name = model.name;
        // const path = model.path;
        // const fileName = model.fileName;
        // const modelName = model.modelName;
        // const sceneImpl = model.scene.impl;

        // targetPath = targetPath ? targetPath : `${this.getResPath()}${this.getSceneResPath()}${path}`;
        // targetFile = targetFile ? targetFile : `${fileName}.gltf`;

        // BABYLON.SceneLoader.Append(targetPath, targetFile, sceneImpl, model.appened, (e) => {
        //     logger.info('%o', e);
        // });
    }
    /**
     *
     * @param model 目标模型
     * @param targetPath 如果赋值(完整路径)将忽略 model 配置的路径
     * @param targetFile 如果赋值(完整路径)将忽略 model 配置的路径
     */
    public static loadSceneAsync(model: any, targetPath?: string, targetFile?: string) {
        // const name = model.name;
        // const path = model.path;
        // const fileName = model.fileName;
        // const modelName = model.modelName;
        // const sceneImpl = model.scene.impl;

        // targetPath = targetPath ? targetPath : `${this.getResPath()}${this.getSceneResPath()}${path}`;
        // targetFile = targetFile ? targetFile : `${fileName}.gltf`;

        // return BABYLON.SceneLoader.AppendAsync(targetPath, targetFile, sceneImpl, (e) => {
        //     logger.info('%o', e);
        // }).then(
        //         (res) => {
        //             model.appened(res);
        //         }
        //     );
    }
    /**
     * @deprecated
     * @param model 目标模型
     * @param targetPath 如果赋值(完整路径)将忽略 model 配置的路径
     * @param targetFile 如果赋值(完整路径)将忽略 model 配置的路径
     */
    public static insertMesh(model: ModelObj, targetPath?: string, targetFile?: string) {
        (<any>model)._loadPromise = this.insertMeshAsync(model, targetPath, targetFile);
    }
    /**
     *
     * @param model 目标模型
     * @param targetPath 如果赋值(完整路径)将忽略 model 配置的路径
     * @param targetFile 如果赋值(完整路径)将忽略 model 配置的路径
     */
    public static insertMeshAsync(model: ModelObj, targetPath?: string, targetFile?: string): Promise<GLTFLodResult> {
        const name = model.name;
        const path = model.path;
        const fileName = model.fileName;
        const modelName = model.modelName;
        const sceneImpl = model.scene.impl;

        targetPath = targetPath ? targetPath : `${this.getResPath()}${this.getModelResPath()}${path}`;
        targetFile = targetFile ? targetFile : `${fileName}.gltf`;

        if (LoaderTool.UseNewGLTFLoader) {
            let loader = new GLTFLoader(sceneImpl, model.resTab, targetPath, targetFile, fileName, undefined, model.gltfModifyConfig);
            model._loader = loader;
            return loader.loadAsync();
        } else {

            return new Promise(
                function (resolve, reject) {
                    BABYLON.SceneLoader.ImportMesh(
                        modelName,
                        targetPath,
                        targetFile,
                        sceneImpl,
                        (meshes, particleSystems, skeletons, animationGroups, transformNodes, geometries, lights) => {
                            // resolve({
                            //     meshes: meshes,
                            //     particleSystems: particleSystems,
                            //     skeletons: skeletons,
                            //     animationGroups: animationGroups,
                            //     transformNodes: transformNodes,
                            //     geometries: geometries,
                            //     lights: lights
                            // });
                            resolve({
                                _root_: meshes[0],
                                nodes: transformNodes,
                                meshes: meshes,
                                particles: particleSystems,
                                materials: [],
                                animationGroups: animationGroups,
                                // skeletons: skeletons,
                                // gltf: (<any>res).gltf
                            })
                        },
                        (e) => {
                            console.info('%o', e);
                        }, function (scene, message, exception) {
                            reject(exception || new Error(message));
                        }, undefined, model.gltfModifyConfig
                    );
                }
            );

            return BABYLON.SceneLoader.ImportMeshAsync(
                modelName,
                targetPath,
                targetFile,
                sceneImpl,
                (e) => {
                    console.info('%o', e);
                },
                undefined,
                model.gltfModifyConfig
            )
            .then(
                    (res) => {
                        return { _root_: res.meshes[0], nodes: res.transformNodes, meshes: res.meshes, particles: <any>res.particleSystems, materials: [], animationGroups: res.animationGroups, skeletons: res.skeletons, gltf: (<any>res).gltf };
                    }
                );
        }
    }
    /**
     *
     * @param model 目标模型
     * @param targetPath 如果赋值(完整路径)将忽略 model 配置的路径
     * @param targetFile 如果赋值(完整路径)将忽略 model 配置的路径
     */
    public static insertEffect(model: ModelObj, targetPath?: string, targetFile?: string) {
        // const name = model.name;
        // const path = model.path;
        // const fileName = model.fileName;
        // const modelName = model.modelName;
        // const sceneImpl = model.scene.impl;

        // targetPath = targetPath ? targetPath : `${this.getResPath()}${this.getEffectResPath()}${path}`;
        // targetFile = targetFile ? targetFile : `${fileName}.gltf`;

        // BABYLON.SceneLoader.ImportMesh(modelName, targetPath, targetFile, sceneImpl, model.loaded, undefined, undefined, undefined, model.gltfModifyConfig);
    }
    /**
     *
     * @param model 目标模型
     * @param targetPath 如果赋值(完整路径)将忽略 model 配置的路径
     * @param targetFile 如果赋值(完整路径)将忽略 model 配置的路径
     */
    public static insertEffectAsync(model: ModelObj, targetPath?: string, targetFile?: string) {
        // const name = model.name;
        // const path = model.path;
        // const fileName = model.fileName;
        // const modelName = model.modelName;
        // const sceneImpl = model.scene.impl;

        // targetPath = targetPath ? targetPath : `${this.getResPath()}${this.getEffectResPath()}${path}`;
        // targetFile = targetFile ? targetFile : `${fileName}.gltf`;

        // return BABYLON.SceneLoader.ImportMeshAsync(modelName, targetPath, targetFile, sceneImpl, undefined, undefined, model.gltfModifyConfig)
        //     .then(
        //         (res) => {
        //             return model.loaded(res.meshes, res.particleSystems, res.skeletons, res.animationGroups);
        //         }
        //     );
    }
    public static loadSklai(model: ModelObj, targetPath?: string, targetFile?: string): Promise<void> {
        const name = model.name;
        const path = model.path;
        const fileName = model.fileName;
        const modelName = model.modelName;
        const sceneImpl = model.scene.impl;

        targetPath = targetPath ? targetPath : `${this.getResPath()}${this.getModelResPath()}${path}`;
        targetFile = targetFile ? targetFile : `${fileName}.sklai`;

        let url = targetPath + targetFile;
        return (<OfflineProvider>sceneImpl.offlineProvider).resTab.load(ExtensionTools.RES_FILE, url, ["Uint8Array"])
            .then(
                (r) => {
                    if (ArrayBuffer.isView(r.link)) {
                        r.link = new Uint8Array((<Uint8Array>r.link).slice().buffer);
                    }
                    (<any>model)._sklaiPath = url;
                    (<any>model)._sklaiRes = r;
                    // return [url, r];
                }
            );
    }
}