import { EAtlasRenderMode, ITexturePackAtlasCompact, ITexturePackFrameCompact } from "../interface/texturepacker";
import { Constants, ExtensionTools, KTXInfo, ResTab } from "../interface/index";
import { ISpriteAtlas, ISpriteAtlasCells, ISpriteAtlasRects, ITexturePackAtlas, ITexturePackFrame } from "../interface/sprite";
import { BABYLON } from "../render3d/babylon";
import { IOption, MaxRectsPacker } from "../maxrects_packer/maxrects-packer";
import { IRectangle } from "../maxrects_packer/geom/Rectangle";
import { EFilter, EWrap, RenderTarget, RenderTargetFactory } from "../postprocess/render_target";

export class TextureFrame implements ITexturePackFrame {
    // frame: { x: number; y: number; w: number; h: number; };
    get rotated() {
        return this.data[0] > 0;
    }
    get trimmed() {
        return this.data[1] > 0;
    }
    get spriteSourceSize() {
        return { x: this.data[4], y: this.data[5], w: this.data[6], h: this.data[7] };
    }
    get sourceSize() {
        return { w: this.data[2], h: this.data[3] };
    }
    get frame() {
        return { x: this.data[8], y: this.data[9], w: this.data[10], h: this.data[11] };
    }
    data: ITexturePackFrameCompact;
    constructor(data: ITexturePackFrameCompact) {
        this.data = data;
        // this.rotated = data[0] > 0;
        // this.trimmed = data[1] > 0;
        // this.sourceSize = { w: data[2], h: data[3] };
        // this.spriteSourceSize = { x: data[4], y: data[5], w: data[6], h: data[7] };
        // this.frame = { x: data[8], y: data[9], w: data[10], h: data[11] };
    }
}

export class TexturePackAtlas implements ITexturePackAtlas {
    frames: { [index: string]: ITexturePackFrame; };
    animations?: { [index: string]: string[]; };
    meta: { app: "https://www.codeandweb.com/texturepacker" | string; version: string; image: string; format: string; size: { w: number; h: number; }; smartupdate: string; };
    image: string;
    samplerMode?: number;
    alphaMode?: number;
    noMipmap?: boolean;
    renderMode?: EAtlasRenderMode | number;
    isInvertY?: boolean;
    scale?: [number, number];
    public static fromJson(input: ITexturePackAtlas[] | ITexturePackAtlasCompact[]) {
        if (input instanceof Array) {
            let items = [];
            input.forEach((item) => {
                let data: ITexturePackAtlas | ITexturePackAtlasCompact = <ITexturePackAtlas | ITexturePackAtlasCompact>item;
                if ((<ITexturePackAtlasCompact>data).compact) {
                    let result = new TexturePackAtlas();
                    result.frames = {};
                    result.animations = data.animations;
                    result.meta = data.meta;
                    result.image = data.image;
                    for (let key in data.frames) {
                        result.frames[key] = new TextureFrame((<ITexturePackAtlasCompact>data).frames[key]);
                    }
                    result.samplerMode = data.samplerMode;
                    result.alphaMode = data.alphaMode;
                    result.noMipmap = data.noMipmap;
                    result.renderMode = data.renderMode;
                    result.isInvertY = data.isInvertY;
                    result.scale = data.scale;
                    items.push(result)
                } else {
                    let result = new TexturePackAtlas();
                    result.frames = (<ITexturePackAtlas>data).frames;
                    result.animations = data.animations;
                    result.meta = data.meta;
                    result.image = data.image;
                    result.samplerMode = data.samplerMode;
                    result.alphaMode = data.alphaMode;
                    result.noMipmap = data.noMipmap;
                    result.renderMode = data.renderMode;
                    result.isInvertY = data.isInvertY;
                    result.scale = data.scale;
                    items.push(result)
                }
            })
            return items;
        } else {
            let data: ITexturePackAtlas | ITexturePackAtlasCompact = <ITexturePackAtlas | ITexturePackAtlasCompact>input;
            if ((<ITexturePackAtlasCompact>data).compact) {
                let result = new TexturePackAtlas();
                result.frames = {};
                result.animations = data.animations;
                result.meta = data.meta;
                result.image = data.image;
                for (let key in data.frames) {
                    result.frames[key] = new TextureFrame((<ITexturePackAtlasCompact>data).frames[key]);
                }

                result.samplerMode = data.samplerMode;
                result.alphaMode = data.alphaMode;
                result.noMipmap = data.noMipmap;
                result.renderMode = data.renderMode;
                result.isInvertY = data.isInvertY;
                result.scale = data.scale;
                return result;
            } else {
                let result = new TexturePackAtlas();
                result.frames = (<ITexturePackAtlas>data).frames;
                result.animations = data.animations;
                result.meta = data.meta;
                result.image = data.image;
                result.samplerMode = data.samplerMode;
                result.alphaMode = data.alphaMode;
                result.noMipmap = data.noMipmap;
                result.renderMode = data.renderMode;
                result.isInvertY = data.isInvertY;
                result.scale = data.scale;
                return result;
            }
        }
    }
}

export interface IAtlasParam {
    spriteLimit?: number,
    samplerMode?: number,
    alphaMode?: number,
    noMipmap?: boolean,
    renderMode?: EAtlasRenderMode | number;
}

export class GlobalAtlasManager {
    public static combineAtlasLink: Map<string, string> = new Map();
    public static combineAtlasRect: Map<string, string> = new Map();
    public static atlasMap: Map<string, Atlas> = new Map();
    public static atlasSpriteLimit: Map<string, IAtlasParam> = new Map();
    private static atlasLoadMap: Map<string, Promise<Atlas[]>> = new Map();
    public static readonly Default: [number, number, number, number] = [0, 0, 0, 0]
    public static load(atlasConfigPath: string, spriteLimit?: number, samplerMode?: number, alphaMode?: number, noMipmap: boolean = true, renderMode: EAtlasRenderMode = 0) {
        let result = GlobalAtlasManager.atlasLoadMap.get(atlasConfigPath);
        if (result) {
            return result;
        } else {
            let res = ExtensionTools.CreateResTab();
            result = ExtensionTools.LoadFile(res, atlasConfigPath).then((data) => {
                let json = ExtensionTools.UTF8Decode((<ArrayBuffer>data));
                let atlas = TexturePackAtlas.fromJson(JSON.parse(json));
                let result = GlobalAtlasManager.regist(atlas, spriteLimit, samplerMode, alphaMode, noMipmap, renderMode);
                res.clear();
                return result;
            });
            GlobalAtlasManager.atlasLoadMap.set(atlasConfigPath, result);
            return result;
        }
    }
    public static registImageWithBox(
        image: string, frame: string,
        frameData: {
            frame: { x: number, y: number, w: number, h: number },
            spriteSourceSize: { x: number, y: number, w: number, h: number },
            sourceSize: { w: number, h: number }
        }, atlasWidth: number, atlasHeight: number,
        spriteLimit?: number, samplerMode?: number, alphaMode?: number, noMipmap: boolean = true, renderMode: EAtlasRenderMode = 0, isInvertY: boolean = false, scale?: [number, number]
    ) {
        let atlas = GlobalAtlasManager.atlasMap.get(image);
        (<ITexturePackFrame>frameData).rotated = false;
        (<ITexturePackFrame>frameData).trimmed = true;
        if (atlas) {
            (<TexturePackAtlas>atlas.atlas).frames[frame] = <ITexturePackFrame>frameData;
            atlas.extendAtlasFrameBySize(frame, frameData.frame.x, frameData.frame.y, frameData.frame.w, frameData.frame.h, atlasWidth, atlasHeight);
            return;
        }

        let result = new TexturePackAtlas();

        result.frames = {};
        result.frames[frame] = <ITexturePackFrame>frameData;
        result.meta = {
            app: "",
            version: "string",
            image: image,
            format: "",
            size: {
                w: atlasWidth,
                h: atlasHeight,
            },
            smartupdate: "",
        };
        result.image = image;
        result.samplerMode = samplerMode;
        result.alphaMode = alphaMode;
        result.noMipmap = noMipmap;
        result.renderMode = renderMode;
        result.isInvertY = isInvertY;
        result.scale = scale;

        GlobalAtlasManager.regist(result, spriteLimit, samplerMode, alphaMode, noMipmap, renderMode);
    }
    public static registImage(
        image: string, frame: string,
        x: number, y: number, w: number, h: number, width: number, height: number,
        spriteLimit?: number, samplerMode?: number, alphaMode?: number, noMipmap: boolean = true, renderMode: EAtlasRenderMode | number = 0, isInvertY: boolean = false, scale?: [number, number]
    ) {
        let atlas = GlobalAtlasManager.atlasMap.get(image);
        if (atlas) {
            (<TexturePackAtlas>atlas.atlas).frames[frame] = {
                frame: { x, y, w, h },
                rotated: false,
                trimmed: true,
                spriteSourceSize: { x: 0, y: 0, w, h },
                sourceSize: { w: w, h: h }
            }
            atlas.extendAtlasFrameBySize(frame, x, y, w, h, width, height);
            return;
        }

        let result = new TexturePackAtlas();

        result.frames = {};
        result.frames[frame] = {
            frame: { x, y, w, h },
            rotated: false,
            trimmed: true,
            spriteSourceSize: { x: 0, y: 0, w, h },
            sourceSize: { w: w, h: h }
        };
        result.meta = {
            app: "",
            version: "string",
            image: image,
            format: "",
            size: {
                w: width,
                h: height,
            },
            smartupdate: "",
        };
        result.image = image;
        result.samplerMode = samplerMode;
        result.alphaMode = alphaMode;
        result.noMipmap = noMipmap;
        result.renderMode = renderMode;
        result.isInvertY = isInvertY;
        result.scale = scale;

        GlobalAtlasManager.regist(result, spriteLimit, samplerMode, alphaMode, noMipmap, renderMode);
    }
    public static registImageWithAnim(
        image: string, animName: string, column: number, row: number, ceilWidth: number, ceilHeight: number, width: number, height: number,
        spriteLimit?: number, samplerMode?: number, alphaMode?: number, noMipmap: boolean = true, renderMode: EAtlasRenderMode = 0, isInvertY: boolean = false, scale?: [number, number]
    ) {
        let result = new TexturePackAtlas();

        result.frames = {};
        result.meta = {
            app: "",
            version: "string",
            image: image,
            format: "",
            size: {
                w: width,
                h: height,
            },
            smartupdate: "",
        };
        result.image = image;
        result.samplerMode = samplerMode;
        result.alphaMode = alphaMode;
        result.noMipmap = noMipmap;
        result.renderMode = renderMode;
        result.isInvertY = isInvertY;
        result.scale = scale;
        result.animations = {};

        let idx = 0;
        if (isInvertY) {
            for (let y = 0; y < row; y++) {
                for (let x = 0; x < column; x++) {
                    let cx = x * ceilWidth;
                    let cy = (row - y - 1) * ceilHeight;
                    let frame: ITexturePackFrameCompact = [
                        0,
                        0,
                        ceilWidth, ceilHeight,
                        0, 0, ceilWidth, ceilHeight,
                        cx, cy, ceilWidth, ceilHeight,
                    ];
                    result.frames[idx] = new TextureFrame(frame);
                    idx ++;
                }
            }
        } else {
            for (let y = 0; y < row; y++) {
                for (let x = 0; x < column; x++) {
                    let cx = x * ceilWidth;
                    let cy = y * ceilHeight;
                    let frame: ITexturePackFrameCompact = [
                        0,
                        0,
                        ceilWidth, ceilHeight,
                        0, 0, ceilWidth, ceilHeight,
                        cx, cy, ceilWidth, ceilHeight,
                    ];
                    result.frames[idx] = new TextureFrame(frame);
                    idx ++;
                }
            }
        }
        let animation = [];
        for (let i = 0; i < idx; i++) {
            animation.push(`${i}`);
        }
        result.animations[animName] = animation;

        GlobalAtlasManager.regist(result, spriteLimit, samplerMode, alphaMode, noMipmap, renderMode);
    }
    public static regist(atlas: ISpriteAtlas | ISpriteAtlas[], spriteLimit?: number, samplerMode?: number, alphaMode?: number, noMipmap: boolean = true, renderMode: EAtlasRenderMode | number = 0): Atlas[] { 
        let result = [];
        if (atlas instanceof Array) {
            atlas.forEach((item) => {
                spriteLimit = GlobalAtlasManager.atlasSpriteLimit.get(item.image)?.spriteLimit || spriteLimit;
                samplerMode = GlobalAtlasManager.atlasSpriteLimit.get(item.image)?.samplerMode || samplerMode || item.samplerMode;
                alphaMode = GlobalAtlasManager.atlasSpriteLimit.get(item.image)?.alphaMode || alphaMode || item.alphaMode;
                noMipmap = GlobalAtlasManager.atlasSpriteLimit.get(item.image)?.noMipmap || noMipmap || item.noMipmap;
                renderMode = GlobalAtlasManager.atlasSpriteLimit.get(item.image)?.renderMode || renderMode || item.renderMode;
                let atlas = new Atlas(item, samplerMode, alphaMode, noMipmap, renderMode);
                GlobalAtlasManager.atlasMap.set(item.image, atlas);
                GlobalAtlasManager.atlasSpriteLimit.set(item.image, { spriteLimit, samplerMode, alphaMode, noMipmap, renderMode });
                result.push(atlas);
            });
        } else {
            spriteLimit = GlobalAtlasManager.atlasSpriteLimit.get(atlas.image)?.spriteLimit || spriteLimit;
            samplerMode = GlobalAtlasManager.atlasSpriteLimit.get(atlas.image)?.samplerMode || samplerMode || atlas.samplerMode;
            alphaMode = GlobalAtlasManager.atlasSpriteLimit.get(atlas.image)?.alphaMode || alphaMode || atlas.alphaMode;
            noMipmap = GlobalAtlasManager.atlasSpriteLimit.get(atlas.image)?.noMipmap || noMipmap || atlas.noMipmap;
            renderMode = GlobalAtlasManager.atlasSpriteLimit.get(atlas.image)?.renderMode || renderMode || atlas.renderMode;
            let _atlas = new Atlas(atlas, samplerMode, alphaMode, noMipmap, renderMode)
            GlobalAtlasManager.atlasMap.set(atlas.image, _atlas);
            GlobalAtlasManager.atlasSpriteLimit.set(atlas.image, { spriteLimit, samplerMode, alphaMode, noMipmap, renderMode });
            result.push(_atlas);
        }
        return result;
    }
    public static getAtlas(atlas: ISpriteAtlas | string): Atlas {
        if ((<ISpriteAtlas>atlas).image) {
            let result = GlobalAtlasManager.atlasMap.get(<string>(<ISpriteAtlas>atlas).image);
            if (result == undefined) {
                GlobalAtlasManager.regist(<ISpriteAtlas>atlas);
            }

            return GlobalAtlasManager.atlasMap.get(<string>(<ISpriteAtlas>atlas).image);
        } else {
            return GlobalAtlasManager.atlasMap.get(<string>atlas);
        }
    }
    public static queryRealAtlasName(atlas: string): string {
        return GlobalAtlasManager.combineAtlasLink.get(atlas) || atlas;
    }
    /**
     * 
     * @param spriteFrame 
     * @param atlasKey 如果知道散图属于哪个图集可以加快运行速度
     * @returns 
     */
    public static getFrame(spriteFrame: string | number, atlasKey?: string): [number, number, number, number] {
        if (atlasKey) {
            let realAtlasKey = GlobalAtlasManager.queryRealAtlasName(atlasKey);
            let atlas = this.atlasMap.get(realAtlasKey);
            if (realAtlasKey != atlasKey) {
                return atlas.getFrame(`${atlasKey}#${spriteFrame}`) || GlobalAtlasManager.Default;
            } else {
                return atlas.getFrame(spriteFrame) || GlobalAtlasManager.Default;
            }
        } else {
            let keys = this.atlasMap.keys();
            let key = keys.next().value;
            let result = undefined;
            while (key) {
                let atlas = this.atlasMap.get(key);

                if (atlas) {
                    result = atlas.getFrame(spriteFrame);
                    if (result) {
                        break;
                    }
                }
                key = keys.next().value;
            }

            return result || GlobalAtlasManager.Default;
        }
    }
    public static getTexturePackFrame(spriteFrame: string | number, atlasKey: string): ITexturePackFrame {
        let realAtlasKey = GlobalAtlasManager.queryRealAtlasName(atlasKey);
        let atlas = this.atlasMap.get(realAtlasKey);
        if ((<ITexturePackAtlas>atlas.atlas)?.frames) {
            if (realAtlasKey != atlasKey) {
                return (<ITexturePackAtlas>atlas.atlas)?.frames[`${atlasKey}#${spriteFrame}`];
            } else {
                return (<ITexturePackAtlas>atlas.atlas)?.frames[spriteFrame];
            }
        } else {
            return undefined;
        }
    }
    public static getTexture(atlasKey: string, scene: BABYLON.Scene): BABYLON.Texture | undefined {
        let realAtlasKey = GlobalAtlasManager.queryRealAtlasName(atlasKey);
        let atlas = this.atlasMap.get(realAtlasKey);
        return atlas?.getTexture(scene);
    }
    /**
     * 
     * @param spriteFrame 
     * @returns 
     */
    public static getAtlasByFrame(spriteFrame: string | number): Atlas {
        let keys = this.atlasMap.keys();
        let key = keys.next().value;
        let result: Atlas = undefined;
        while (key) {
            let atlas = this.atlasMap.get(key);

            if (atlas) {
                if (atlas.getFrame(spriteFrame)) {
                    result = atlas;
                    break;
                }
            }
            key = keys.next().value;
        }

        return result;
    }
    public static extendAtlasFrame(atlasKey: string, key: string | number, rect: [number, number, number, number]): void {
        let atlas = this.atlasMap.get(atlasKey);
        if (atlas) {
            atlas.extendAtlasFrame(key, rect);
        }
    }
}

export class Atlas {
    protected _atlasFrames: { [index: string | number]: [number, number, number, number] };
    public readonly atlasImage: string;
    constructor(
        public readonly atlas: ISpriteAtlas,
        public readonly samplerMode: number = BABYLON.Texture.LINEAR_LINEAR,
        public readonly alphaMode: number = Constants.ALPHA_COMBINE,
        public readonly noMipmap: boolean = true,
        public readonly renderMode: EAtlasRenderMode | number = 0
    ) {
        if ((<ITexturePackAtlas>atlas).meta) {
            this.atlasImage = (<ITexturePackAtlas>atlas).image;
        } else {
            this.atlasImage = (<ISpriteAtlasRects>atlas).image;
        }

        if ((<ISpriteAtlasRects>this.atlas).rects) {
            this._atlasFrames = (<ISpriteAtlasRects>this.atlas).rects;
        } else if ((<ITexturePackAtlas>atlas).meta) {
            let atlasFrames: { [index: string]: [number, number, number, number] } = {};

            let w = (<ITexturePackAtlas>atlas).meta.size.w;
            let h = (<ITexturePackAtlas>atlas).meta.size.h;
            for (let key in (<ITexturePackAtlas>atlas).frames) {
                let frame = (<ITexturePackAtlas>atlas).frames[key];
                if (frame.rotated) {
                    let uoff = frame.frame.x / w;
                    let voff = frame.frame.y / h;
                    let hsize = frame.frame.h / w;
                    let vsize = frame.frame.w / h;
                    atlasFrames[key] = [hsize, vsize, uoff, voff];
                } else {
                    let uoff = frame.frame.x / w;
                    let voff = frame.frame.y / h;
                    let hsize = frame.frame.w / w;
                    let vsize = frame.frame.h / h;
                    atlasFrames[key] = [hsize, vsize, uoff, voff];
                }
            }
            this._atlasFrames = <{ [index: string]: [number, number, number, number] }>atlasFrames;
        } else {
            let atlasFrames: [number, number, number, number][] = [];

            let hCount = (<ISpriteAtlasCells>this.atlas).hCount;
            let vCount = (<ISpriteAtlasCells>this.atlas).vCount;
            let hsize = (<ISpriteAtlasCells>this.atlas).hsize;
            let vsize = (<ISpriteAtlasCells>this.atlas).vsize;
            let frames = hCount * vCount;
            for (let i = 0; i < frames; i++) {
                let uoff = i % hCount * hsize;
                let voff = Math.floor(i / hCount) * vsize;
                atlasFrames[i] = [hsize, vsize, uoff, voff];
            }

            this._atlasFrames = <{ [index: number]: [number, number, number, number] }>atlasFrames;
        }
    }
    public getTexture(scene: BABYLON.Scene): BABYLON.Texture | undefined {
        return;
    }
    /**
     * 
     * @param key 帧名称
     * @param x 帧在整图中的 位置
     * @param y 帧在整图中的 位置
     * @param w 帧在整图中的 宽度
     * @param h 帧在整图中的 高度
     * @param width 帧所在整图的 宽度
     * @param height 帧所在整图的 高度
     */
    extendAtlasFrameBySize(key: string | number, x: number, y: number, w: number, h: number, width: number, height: number): void {
        this._atlasFrames[key] = [w / width, h / height,x / width, y / height];
    }
    extendAtlasFrame(key: string | number, rect: [number, number, number, number]): void {
        this._atlasFrames[key] = rect;
    }
    getFrame(spriteFrame: string | number): [number, number, number, number] {
        return this._atlasFrames[spriteFrame];
    }
}

export interface ICustomRect extends IRectangle {
    key: string,
    url: string,
    rot?: boolean,
    /**
     * all input with same will be packed in the same bin.
     */
    tag?: string,
    spriteSourceSize: { x: number, y: number, w: number, h: number },
    sourceSize: { w: number, h: number }
}


export class CombinedAtlas extends Atlas {
    constructor(
        atlas: ITexturePackAtlas,
        samplerMode?: number, alphaMode?: number, noMipmap?: boolean, renderMode?: EAtlasRenderMode
    ) {
        super(atlas, samplerMode, alphaMode, noMipmap, renderMode);
    }
    public getTexture(scene: BABYLON.Scene): BABYLON.Texture | undefined {
        let atlas = <ITexturePackAtlas>this.atlas;
        if (AtlasCombine._internalTexture == undefined) {
            AtlasCombine._internalTexture = new BABYLON.InternalTexture(scene.getEngine(), 0);
            AtlasCombine._internalTexture.width = atlas.meta.size.w;
            AtlasCombine._internalTexture.height = atlas.meta.size.h;
            AtlasCombine._internalTexture.baseWidth = atlas.meta.size.w;
            AtlasCombine._internalTexture.baseHeight = atlas.meta.size.h;
            AtlasCombine._internalTexture.isReady = true;
            AtlasCombine._internalTexture._webGLTexture = AtlasCombine._webglTexture;
        }

        let texture = new BABYLON.Texture(null, scene, false, false, BABYLON.Texture.LINEAR_LINEAR);
        texture.wrapU = 0;
        texture.wrapV = 0;
        texture.delayLoadState = BABYLON.Constants.DELAYLOADSTATE_LOADED;
        texture._texture = AtlasCombine._internalTexture;
        texture.dispose = () => { };
        (<any>texture)._delayedOnLoad = null;
        (<any>texture)._delayedOnError = null;
        texture.level = 1;
        texture.hasAlpha = true;

        return texture;
    }
}
export class AtlasCombine {
    private static CombineUnique: number = 0;
    combinedID: string;
    static _webglTexture: WebGLTexture;
    static _internalTexture: BABYLON.InternalTexture;
    public loadPromise: Promise<void[]>;
    private _isDisposed: boolean = false;
    constructor(
        protected atlasList: string[],
        spriteLimit?: number, samplerMode?: number, alphaMode?: number, noMipmap?: boolean, renderMode?: EAtlasRenderMode
    ) {
        let maxWidth = 4096;
        let maxHeight = 4096;
        const options: IOption = {
            smart: true,
            pot: false,
            square: false,
            allowRotation: false,
            tag: false,
            border: 0,
        }; // Set packing options
        let packer = new MaxRectsPacker<ICustomRect>(maxWidth, maxHeight, 0, options);
        let inputs: ICustomRect[] = [];
        atlasList.forEach((atlasName) => {
            let atlas = <ITexturePackAtlas | ITexturePackAtlasCompact>GlobalAtlasManager.atlasMap.get(atlasName).atlas;
            inputs.push(<ICustomRect>{
                key: atlas.image,
                url: atlas.image,
                x: 0,
                y: 0,
                width: atlas.meta.size.w,
                height: atlas.meta.size.h,
                spriteSourceSize: { x: 0, y: 0, w: atlas.meta.size.w, h: atlas.meta.size.h },
                sourceSize: { w: atlas.meta.size.w, h: atlas.meta.size.h }
            })
        });

        let promiseArray: Promise<void>[] = [];
        packer.addArray(inputs); // Start packing with input array
        packer.next(); // Start a new packer bin
        if (packer.bins.length == 1) {
            let bin =  packer.bins[0];
            let id = AtlasCombine.CombineUnique++;
            let result: ITexturePackAtlas = {
                frames: {},
                meta: {
                    app: "pi_texturepacker",
                    version: "",
                    image: `${id}`,
                    format: "",
                    size: { w: maxWidth, h: maxHeight },
                    smartupdate: "",
                },
                animations: {},
                image: `${id}`,
            };
            bin.rects.forEach((rect, idx) => {
                let sourceAtlas = <ITexturePackAtlas>GlobalAtlasManager.atlasMap.get(rect.key).atlas;
                for (let key in sourceAtlas.frames) {
                    let sourceFrame = sourceAtlas.frames[key];
                    let sourceFrameData = sourceAtlas.frames[key].frame;
                    let newKey = `${sourceAtlas.image}#${key}`;
                    result.frames[newKey] = {
                        frame: { x: sourceFrameData.x + rect.x, y: sourceFrameData.y + rect.y, w: sourceFrameData.w, h: sourceFrameData.h },
                        rotated: sourceFrame.rotated,
                        trimmed: sourceFrame.trimmed,
                        spriteSourceSize: sourceFrame.spriteSourceSize,
                        sourceSize: sourceFrame.sourceSize
                    }
                }


                let gl: WebGLRenderingContext = RenderTargetFactory.getContext().gl;
                const resTab = ExtensionTools.CreateResTab();
                let promise = resTab.load(ExtensionTools.RES_IMAGE, <string>sourceAtlas.image, []).then((res) => {
                    if (this._isDisposed == false) {
                        let image: HTMLImageElement | KTXInfo = res.link;
                        if ((<KTXInfo>image).mipmaps) {
                            if (AtlasCombine._webglTexture == undefined) {
                                            // 创建纹理
                                AtlasCombine._webglTexture = gl.createTexture();
                                gl.activeTexture(gl.TEXTURE0);
                                gl.bindTexture(gl.TEXTURE_2D, AtlasCombine._webglTexture);
                                gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_MAG_FILTER, RenderTargetFactory.GLMagFilter(EFilter.LINEAR, gl));
                                gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_MIN_FILTER, RenderTargetFactory.GLMinFilter(EFilter.LINEAR, gl));
                                gl.texParameteri(
                                    gl.TEXTURE_2D,
                                    gl.TEXTURE_WRAP_S,
                                    RenderTargetFactory.GLWrap(EWrap.CLAMP_TO_EDGE, gl),
                                );
                                gl.texParameteri(
                                    gl.TEXTURE_2D,
                                    gl.TEXTURE_WRAP_T,
                                    RenderTargetFactory.GLWrap(EWrap.CLAMP_TO_EDGE, gl)
                                );
                                // gl.compressedTexImage2D(gl.TEXTURE_2D, 0, (<KTXInfo>image).format, maxWidth, maxHeight, 0, null);
                            }

                            gl.bindTexture(gl.TEXTURE_2D, AtlasCombine._webglTexture);
                            for (let i = 0; i < (<KTXInfo>image).mipmapCount; i++) {
                                const data = (<KTXInfo>image).mipmaps[0];
                                gl.compressedTexSubImage2D(gl.TEXTURE_2D, 0, rect.x, rect.y, data.width, data.height, (<KTXInfo>image).format, data.data);
                            }
                            gl.bindTexture(gl.TEXTURE_2D, null);
                        } else {
                            if (AtlasCombine._webglTexture == undefined) {
                                            // 创建纹理
                                AtlasCombine._webglTexture = gl.createTexture();
                                gl.activeTexture(gl.TEXTURE0);
                                gl.bindTexture(gl.TEXTURE_2D, AtlasCombine._webglTexture);
                                gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_MAG_FILTER, RenderTargetFactory.GLMagFilter(EFilter.LINEAR, gl));
                                gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_MIN_FILTER, RenderTargetFactory.GLMinFilter(EFilter.LINEAR, gl));
                                gl.texParameteri(
                                    gl.TEXTURE_2D,
                                    gl.TEXTURE_WRAP_S,
                                    RenderTargetFactory.GLWrap(EWrap.CLAMP_TO_EDGE, gl),
                                );
                                gl.texParameteri(
                                    gl.TEXTURE_2D,
                                    gl.TEXTURE_WRAP_T,
                                    RenderTargetFactory.GLWrap(EWrap.CLAMP_TO_EDGE, gl)
                                );
                                gl.texImage2D(gl.TEXTURE_2D, 0, gl.RGBA, maxWidth, maxHeight, 0, gl.RGBA, gl.UNSIGNED_BYTE, null);
                            }

                            gl.bindTexture(gl.TEXTURE_2D, AtlasCombine._webglTexture);
                            gl.texSubImage2D(gl.TEXTURE_2D, 0, rect.x, rect.y, gl.RGBA, gl.UNSIGNED_BYTE, <HTMLImageElement>image);
                            gl.bindTexture(gl.TEXTURE_2D, null);
                        }
                    }
                    Promise.resolve().then(() => { resTab.clear() });
                }).catch((e) => {
                    
                });
                promiseArray.push(promise);
            });
            this.combinedID = result.image;
            let _atlas = new CombinedAtlas(result, samplerMode, alphaMode, noMipmap, renderMode)
            GlobalAtlasManager.atlasMap.set(result.image, _atlas);
            GlobalAtlasManager.atlasSpriteLimit.set(result.image, { spriteLimit, samplerMode, alphaMode, noMipmap, renderMode });
            
            bin.rects.forEach((rect, idx) => {
                GlobalAtlasManager.combineAtlasLink.set(rect.key, result.image);
            });
        } else {
            atlasList.forEach((atlasName) => {
                let atlas = <ITexturePackAtlas>GlobalAtlasManager.atlasMap.get(atlasName).atlas;
                let _atlas = new CombinedAtlas(atlas, samplerMode, alphaMode, noMipmap, renderMode)
                GlobalAtlasManager.atlasMap.set(atlas.image, _atlas);
                GlobalAtlasManager.atlasSpriteLimit.set(atlas.image, { spriteLimit, samplerMode, alphaMode, noMipmap, renderMode });
            });
            promiseArray.push(Promise.resolve());
        }
        this.loadPromise = Promise.all(promiseArray);
    }
    destroy() {
        this.atlasList.forEach((key) => {
            GlobalAtlasManager.combineAtlasLink.delete(key);
        });
        this._isDisposed = true;
        GlobalAtlasManager.atlasMap.delete(this.combinedID);
    }
}