import { CustomScene, IModelOpt, IPrefabType, Model, ModelObj, ModelPrefab, NodeObj, PrefabModelObjPool } from "../index";
import { BABYLON } from "../render3d/babylon";

interface ISceneWithModelRectTextureProgress extends CustomScene {
    __RectTextureProgressMeshes: Map<string, BABYLON.Mesh>;
}

interface ICreation extends IModelOpt {
    texPath: string,
}

export class ModelRectTextureProgress extends ModelObj {
    public static regist(scene: CustomScene, texPath: string) {
        scene.prefabPool.registPrefabCfg(texPath, { path: "null", fileName: "", single: false, cacheCount: 8, texPath });
        PrefabModelObjPool.registPrefabFactory(texPath, ModelPrefab.Creation);
        PrefabModelObjPool.registPrefabResFactory(texPath, ModelRectTextureProgress.Creation);

        let tempscene = <ISceneWithModelRectTextureProgress>scene;
        if (!tempscene.__RectTextureProgressMeshes) {
            tempscene.__RectTextureProgressMeshes = new Map();
        }
        scene.impl.onDisposeObservable.add(() => {
            tempscene.__RectTextureProgressMeshes.forEach(source => {
                source.dispose(false, true);
                (<any>source).__instance_keep = undefined;
            });
            tempscene.__RectTextureProgressMeshes.clear();
        });
        if (!tempscene.__RectTextureProgressMeshes.get(texPath)) {
            let source = createSourceMesh(scene.impl, texPath);
            let ins = source.createInstance(`Keep_ModelRectTextureProgress`);
            ins.setEnabled(false);
            (<any>source).__instance_keep = ins;
            tempscene.__RectTextureProgressMeshes.set(texPath, source);
        }
    }
    // public static Creation(scene: CustomScene, info: IPrefabType, pool: PrefabModelObjPool, model: ModelRectTextureProgress) {
    //     let result = new NodeObj(scene, info, pool, model);
    //     return result;
    // }
    public static Creation(name: string, scene: CustomScene, opt?: ICreation) {
        let model = new ModelRectTextureProgress(name, scene);
        let tempscene = <ISceneWithModelRectTextureProgress>scene;
        let source = tempscene.__RectTextureProgressMeshes.get(opt.texPath);
        let instance = source.createInstance(name);

        let rootImpl = instance;
        model.rootImpl = rootImpl;
        model.instance = instance;
        model._loadPromise = Promise.resolve().then(() => {
            model.loaded(rootImpl, [rootImpl], [], [], [], [], []);
            return model;
        });

        return model;
    }
    protected instance: BABYLON.InstancedMesh;
    public set progress(val: number) {
        this.instance.instancedBuffers.a_progress.x = val;
    }
    public setAlpha(val: number) {
        super.setAlpha(val);
        this.instance.instancedBuffers.a_progress.x = val;
    }
    public set alphaIndex(val: number) {
        super.alphaIndex = val;
        this.instance.alphaIndex = val;
    }
    constructor(name: string, scene: CustomScene) {
        super(name, scene);
    }

    _dispose(disposeMaterialAndTextures: boolean = true) {
        this.instance.parent = null;
        this.instance.dispose();
        super._dispose(disposeMaterialAndTextures);
    }
}

function createSourceMesh(scene: BABYLON.Scene, texPath: string): BABYLON.Mesh {

    // 创建自定义材质
    let custom_material = new BABYLON.CustomMaterial(`ProgressMat`, scene);
    custom_material.AddAttribute(`a_progress`);
    custom_material.Vertex_Definitions(
        `
                #define PI_2 6.283185307179586
                attribute vec2 a_progress;
                varying vec2 v_progress;
                `
    );
    custom_material.Vertex_MainBegin(
        `
                v_progress  = a_progress;
                `
    );
    custom_material.Fragment_Definitions(
        `
                varying vec2 v_progress;
                `
    );
    custom_material.Fragment_Custom_Diffuse(
        `
                vec2 progress = v_progress;

                vec2 bgUV           = vec2(vDiffuseUV.x, vDiffuseUV.y * 0.5);
                vec4 bgColor        = texture2D(diffuseSampler, bgUV);

                vec2 gressUV        = vec2(vDiffuseUV.x, vDiffuseUV.y * 0.5 + 0.5);
                vec4 gressColor     = texture2D(diffuseSampler, gressUV);

                float gressAlpha     = progress.x > vDiffuseUV.x ? 1.0 : 0.0;

                bgColor.rgb *= bgColor.a;
                gressColor.rgb *= gressColor.a;

                baseColor       = bgColor * (1.0 - gressAlpha * gressColor.a) + gressColor * gressAlpha;

                alpha = baseColor.a;
                `
    );
    let texture = new BABYLON.Texture(texPath, scene, false, false);
    texture.updateURL(texPath);
    custom_material.diffuseTexture = texture;
    custom_material.disableLighting = true;
    custom_material.specularColor = BABYLON.Color3.Black();
    custom_material.diffuseColor = BABYLON.Color3.White();
    custom_material.ambientColor = BABYLON.Color3.White();
    custom_material.emissiveColor.copyFromFloats(1, 1, 1);
    custom_material.useAlphaFromDiffuseTexture = true;
    custom_material.alphaMode = 2;
    custom_material.diffuseTexture.hasAlpha = true;
    custom_material.diffuseTexture.wrapU = BABYLON.Texture.CLAMP_ADDRESSMODE;
    custom_material.diffuseTexture.wrapV = BABYLON.Texture.CLAMP_ADDRESSMODE;

    let _mesh = BABYLON.MeshBuilder.CreatePlane(`Progress`, { size: 1 }, scene);
    _mesh.material = custom_material;
    // _mesh.billboardMode = BABYLON.Mesh.BILLBOARDMODE_ALL;
    _mesh.registerInstancedBuffer(`a_progress`, 2);
    _mesh.instancedBuffers.a_progress = new BABYLON.Vector2(0.0, 0.0);
    _mesh.setEnabled(false);

    // custom_material.needDepthPrePass = true;

    _mesh.alphaIndex = 3010;

    return _mesh;
}