export enum EUVMode {
    UpIsAZ = 1,
    UpIsNX = 2,
    UpIsNZ = 3,
    UpIsAX = 4,
}

export class Tile {
    public static UV_MODE = 1;
    public static Y_POSITION_SCALE = 1;
    public static X_POSITION_SCALE = 1;
    public static POINT_COUNT: number = 4;
    public static POINT_POSITION_COUNT: number = 3;
    public static POINT_UV_COUNT: number = 2;
    public static POINT_COLOR_COUNT: number = 4;
    public static TRIANGLE_COUNT: number = 2;
    public static TRIANGLE_INDICES_COUNT: number = 3;
    public static positionsWithRotation(index: number, row: number, col: number, width: number, height: number, tileWidth: number, tileHeight: number, data: Float32Array, rotation: number, strip: number = Tile.POINT_POSITION_COUNT, offset: number = 0) {
        let pointIndex = index * Tile.POINT_COUNT;
        let x = (col + 0.5) * tileWidth;
        let y = (height - row - 1 + 0.5) * tileHeight;

        let halfW = tileWidth * 0.5;
        let halfH = tileHeight * 0.5;

        data[pointIndex * strip + offset  + 0] = (x - halfW) * Tile.X_POSITION_SCALE;
        data[pointIndex * strip + offset  + 1] = (y - halfH) * Tile.Y_POSITION_SCALE;
        data[pointIndex * strip + offset  + 2] = 0;
        pointIndex += 1;

        data[pointIndex * strip + offset  + 0] = (x + halfW) * Tile.X_POSITION_SCALE;
        data[pointIndex * strip + offset  + 1] = (y - halfH) * Tile.Y_POSITION_SCALE;
        data[pointIndex * strip + offset  + 2] = 0;
        pointIndex += 1;

        data[pointIndex * strip + offset  + 0] = (x + halfW) * Tile.X_POSITION_SCALE;
        data[pointIndex * strip + offset  + 1] = (y + halfH) * Tile.Y_POSITION_SCALE;
        data[pointIndex * strip + offset  + 2] = 0;
        pointIndex += 1;

        data[pointIndex * strip + offset  + 0] = (x - halfW) * Tile.X_POSITION_SCALE;
        data[pointIndex * strip + offset  + 1] = (y + halfH) * Tile.Y_POSITION_SCALE;
        data[pointIndex * strip + offset  + 2] = 0;
        pointIndex += 1;
    }
    public static positions(index: number, row: number, col: number, width: number, height: number, tileWidth: number, tileHeight: number, data: Float32Array, strip: number = Tile.POINT_POSITION_COUNT, offset: number = 0) {
        let pointIndex = index * Tile.POINT_COUNT;
        let x = (col) * tileWidth;
        let y = (height - row - 1) * tileHeight;
        let halfW = tileWidth * 0.5;
        let halfH = tileHeight * 0.5;

        data[pointIndex * strip + offset  + 0] = (x - halfW) * Tile.X_POSITION_SCALE;
        data[pointIndex * strip + offset  + 1] = (y - halfH) * Tile.Y_POSITION_SCALE;
        data[pointIndex * strip + offset  + 2] = 0;
        pointIndex += 1;

        data[pointIndex * strip + offset  + 0] = (x + halfW) * Tile.X_POSITION_SCALE;
        data[pointIndex * strip + offset  + 1] = (y - halfH) * Tile.Y_POSITION_SCALE;
        data[pointIndex * strip + offset  + 2] = 0;
        pointIndex += 1;

        data[pointIndex * strip + offset  + 0] = (x + halfW) * Tile.X_POSITION_SCALE;
        data[pointIndex * strip + offset  + 1] = (y + halfH) * Tile.Y_POSITION_SCALE;
        data[pointIndex * strip + offset  + 2] = 0;
        pointIndex += 1;

        data[pointIndex * strip + offset  + 0] = (x - halfW) * Tile.X_POSITION_SCALE;
        data[pointIndex * strip + offset  + 1] = (y + halfH) * Tile.Y_POSITION_SCALE;
        data[pointIndex * strip + offset  + 2] = 0;
        pointIndex += 1;
    }
    public static positions_2(index: number, row: number, col: number, width: number, height: number, tileWidth: number, tileHeight: number, data: Float32Array, edgeScale: number, strip: number = Tile.POINT_POSITION_COUNT, offset: number = 0) {
        let pointIndex = index * Tile.POINT_COUNT;
        row = (height - row - 1);
        let x = Math.floor(col / 2) * tileWidth;
        let y = Math.floor(row / 2) * tileHeight;

        let sX = tileWidth * (col % 2 == 1 ? edgeScale : 1.0 - edgeScale);
        let sY = tileHeight * (row % 2 == 1 ? edgeScale : 1.0 - edgeScale);
        let dX = tileWidth * (col % 2 == 1 ? 1.0 - edgeScale : 0);
        let dY = tileHeight * (row % 2 == 1 ? 1.0 - edgeScale : 0);

        data[pointIndex * strip + offset  + 0] = (x + dX) * Tile.X_POSITION_SCALE;
        data[pointIndex * strip + offset  + 1] = (y + dY) * Tile.Y_POSITION_SCALE;
        data[pointIndex * strip + offset  + 2] = 0;
        pointIndex += 1;

        data[pointIndex * strip + offset  + 0] = (x + dX + sX) * Tile.X_POSITION_SCALE;
        data[pointIndex * strip + offset  + 1] = (y + dY) * Tile.Y_POSITION_SCALE;
        data[pointIndex * strip + offset  + 2] = 0;
        pointIndex += 1;

        data[pointIndex * strip + offset  + 0] = (x + dX + sX) * Tile.X_POSITION_SCALE;
        data[pointIndex * strip + offset  + 1] = (y + dY + sY) * Tile.Y_POSITION_SCALE;
        data[pointIndex * strip + offset  + 2] = 0;
        pointIndex += 1;

        data[pointIndex * strip + offset  + 0] = (x + dX) * Tile.X_POSITION_SCALE;
        data[pointIndex * strip + offset  + 1] = (y + dY + sY) * Tile.Y_POSITION_SCALE;
        data[pointIndex * strip + offset  + 2] = 0;
        pointIndex += 1;
    }
    public static indices(index: number, data: Uint16Array) {
        let pointIndex = index * Tile.POINT_COUNT;
        let indicesIndex = index * Tile.TRIANGLE_COUNT;

        data[indicesIndex * Tile.TRIANGLE_INDICES_COUNT + 0] = pointIndex;
        data[indicesIndex * Tile.TRIANGLE_INDICES_COUNT + 1] = pointIndex + 1;
        data[indicesIndex * Tile.TRIANGLE_INDICES_COUNT + 2] = pointIndex + 2;
        indicesIndex += 1;

        data[indicesIndex * Tile.TRIANGLE_INDICES_COUNT + 0] = pointIndex;
        data[indicesIndex * Tile.TRIANGLE_INDICES_COUNT + 1] = pointIndex + 2;
        data[indicesIndex * Tile.TRIANGLE_INDICES_COUNT + 2] = pointIndex + 3;
        indicesIndex += 1;
    }
    public static colors(index: number, data: Float32Array, r: number, g: number, b: number, a: number, strip: number = Tile.POINT_COLOR_COUNT, offset: number = 0) {
        let pointIndex = index * Tile.POINT_COUNT;

        let i = pointIndex * strip;
        data[i + offset + 0] = r;
        data[i + offset + 1] = g;
        data[i + offset + 2] = b;
        data[i + offset + 3] = a;
        pointIndex += 1;

        i = pointIndex * strip;
        data[i + offset + 0] = r;
        data[i + offset + 1] = g;
        data[i + offset + 2] = b;
        data[i + offset + 3] = a;
        pointIndex += 1;

        i = pointIndex * strip;
        data[i + offset + 0] = r;
        data[i + offset + 1] = g;
        data[i + offset + 2] = b;
        data[i + offset + 3] = a;
        pointIndex += 1;

        i = pointIndex * strip;
        data[i + offset + 0] = r;
        data[i + offset + 1] = g;
        data[i + offset + 2] = b;
        data[i + offset + 3] = a;
        pointIndex += 1;
    }
    public static changeUVMode(mode: EUVMode) {
        if (mode == EUVMode.UpIsAX) {
            Tile.uvs = Tile._uvsAX;
        } else if (mode == EUVMode.UpIsNX) {
            Tile.uvs = Tile._uvsNX;
        } else if (mode == EUVMode.UpIsNZ) {
            Tile.uvs = Tile._uvsNZ;
        } else {
            Tile.uvs = Tile._uvsAZ;
        }
    }
    public static uvs = Tile._uvsAZ;
    public static listUVs = [Tile._uvsAZ, Tile._uvsNX, Tile._uvsNZ, Tile._uvsAX];
    public static _uvsAZ(index: number, data: Float32Array, imageColumns: number, cellID: number, imageUSize: number, imageVSize: number, strip: number = Tile.POINT_UV_COUNT, offset: number = 0) {
        let pointIndex = index * Tile.POINT_COUNT;

        let cellRow = Math.floor(cellID / imageColumns);
        let cellCol = cellID % imageColumns;
        let u = imageUSize * cellCol;
        let v = imageVSize * cellRow;

        // 4
        data[pointIndex * strip + offset + 0] = u;
        data[pointIndex * strip + offset + 1] = v + imageVSize;
        pointIndex += 1;

        // 3
        data[pointIndex * strip + offset + 0] = u + imageUSize;
        data[pointIndex * strip + offset + 1] = v + imageVSize;
        pointIndex += 1;

        // 2
        data[pointIndex * strip + offset + 0] = u + imageUSize;
        data[pointIndex * strip + offset + 1] = v;
        pointIndex += 1;

        // 1
        data[pointIndex * strip + offset + 0] = u;
        data[pointIndex * strip + offset + 1] = v;
        pointIndex += 1;
    }
    public static _uvsNZ(index: number, data: Float32Array, imageColumns: number, cellID: number, imageUSize: number, imageVSize: number, strip: number = Tile.POINT_UV_COUNT, offset: number = 0) {
        let pointIndex = index * Tile.POINT_COUNT;

        let cellRow = Math.floor(cellID / imageColumns);
        let cellCol = cellID % imageColumns;
        let u = imageUSize * cellCol;
        let v = imageVSize * cellRow;

        // 2
        data[pointIndex * strip + offset + 0] = u + imageUSize;
        data[pointIndex * strip + offset + 1] = v;
        pointIndex += 1;

        // 1
        data[pointIndex * strip + offset + 0] = u;
        data[pointIndex * strip + offset + 1] = v;
        pointIndex += 1;

        // 4
        data[pointIndex * strip + offset + 0] = u;
        data[pointIndex * strip + offset + 1] = v + imageVSize;
        pointIndex += 1;

        // 3
        data[pointIndex * strip + offset + 0] = u + imageUSize;
        data[pointIndex * strip + offset + 1] = v + imageVSize;
        pointIndex += 1;

    }
    public static _uvsNX(index: number, data: Float32Array, imageColumns: number, cellID: number, imageUSize: number, imageVSize: number, strip: number = Tile.POINT_UV_COUNT, offset: number = 0) {
        let pointIndex = index * Tile.POINT_COUNT;

        let cellRow = Math.floor(cellID / imageColumns);
        let cellCol = cellID % imageColumns;
        let u = imageUSize * cellCol;
        let v = imageVSize * cellRow;

        // 3
        data[pointIndex * strip + offset + 0] = u + imageUSize;
        data[pointIndex * strip + offset + 1] = v + imageVSize;
        pointIndex += 1;

        // 2
        data[pointIndex * strip + offset + 0] = u + imageUSize;
        data[pointIndex * strip + offset + 1] = v;
        pointIndex += 1;

        // 1
        data[pointIndex * strip + offset + 0] = u;
        data[pointIndex * strip + offset + 1] = v;
        pointIndex += 1;

        // 4
        data[pointIndex * strip + offset + 0] = u;
        data[pointIndex * strip + offset + 1] = v + imageVSize;
        pointIndex += 1;

    }
    public static _uvsAX(index: number, data: Float32Array, imageColumns: number, cellID: number, imageUSize: number, imageVSize: number, strip: number = Tile.POINT_UV_COUNT, offset: number = 0) {
        let pointIndex = index * Tile.POINT_COUNT;

        let cellRow = Math.floor(cellID / imageColumns);
        let cellCol = cellID % imageColumns;
        let u = imageUSize * cellCol;
        let v = imageVSize * cellRow;

        // 1
        data[pointIndex * strip + offset + 0] = u;
        data[pointIndex * strip + offset + 1] = v;
        pointIndex += 1;

        // 4
        data[pointIndex * strip + offset + 0] = u;
        data[pointIndex * strip + offset + 1] = v + imageVSize;
        pointIndex += 1;

        // 3
        data[pointIndex * strip + offset + 0] = u + imageUSize;
        data[pointIndex * strip + offset + 1] = v + imageVSize;
        pointIndex += 1;

        // 2
        data[pointIndex * strip + offset + 0] = u + imageUSize;
        data[pointIndex * strip + offset + 1] = v;
        pointIndex += 1;
    }

    public static updateUVs(mode: EUVMode, cellID: number) {
        return Tile.listUVs[Tile.updateBuildMode(mode, cellID)];
    }

    public static uvModeCallIndex(mode: EUVMode) {
        let funcIndex = 0;
        if (mode == EUVMode.UpIsAX) {
            funcIndex = 3;
        } else if (mode == EUVMode.UpIsNX) {
            funcIndex = 1;
        } else if (mode == EUVMode.UpIsNZ) {
            funcIndex = 2;
        } else {
            funcIndex = 0;
        }

        return funcIndex;
    }

    public static updateBuildMode(mode: EUVMode, cellID: number) {
        let funcIndex = Tile.uvModeCallIndex(mode);
        funcIndex += 0;
        funcIndex = funcIndex % 4;
        return funcIndex;
    }
}