
import { Tile } from "./tile";
import { ITileObjectKey, ITileState } from "./interface";

export class TileObject {
    public static key(layer: number, assetSource: string) {
        return `${assetSource}#${layer}`;
    }
    public _layer: number = 0;
    protected _x: number = 0;
    protected _y: number = 0;
    protected _state: ITileState;
    protected _width: number = 1;
    protected _height: number = 1;
    protected _visible: boolean = true;
    protected _dirty: boolean = true;
    public key: ITileObjectKey;
    protected _r: number = 1;
    protected _g: number = 1;
    protected _b: number = 1;
    protected _a: number = 1;
    constructor(
        public readonly id: any,
        state: ITileState,
        layer: number,
    ) {
        this._state = state;
        this._layer = layer;
        this.key = TileObject.key(this._layer, this.state.assetSource);
    }
    set state(value: ITileState) {
        this._state = value;
        this._dirty = true;
        this.key = TileObject.key(this._layer, this.state.assetSource);
    }
    get state() {
        return this._state;
    }
    set width(value: number) {
        this._width = value;
        this._dirty = true;
    }
    get width() {
        return this._width;
    }
    set height(value: number) {
        this._height = value;
        this._dirty = true;
    }
    get height() {
        return this._height;
    }
    set layer(value: number) {
        this._layer = value;
        this._dirty = true;
        this.key = TileObject.key(this._layer, this.state.assetSource);
    }
    get layer() {
        return this._layer;
    }
    set x(value: number) {
        this._x = value;
        this._dirty = true;
    }
    get x() {
        return this._x;
    }
    set y(value: number) {
        this._y = value;
        this._dirty = true;
    }
    get y() {
        return this._y;
    }
    set visible(value: boolean) {
        this._visible = value;
        this._dirty = true;
    }
    get visible() {
        return this._visible;
    }
    set r(value: number) {
        this._r = value;
        this._dirty = true;
    }
    get r() {
        return this._r;
    }
    set g(value: number) {
        this._g = value;
        this._dirty = true;
    }
    get g() {
        return this._g;
    }
    set b(value: number) {
        this._b = value;
        this._dirty = true;
    }
    get b() {
        return this._b;
    }
    set a(value: number) {
        this._a = value;
        this._dirty = true;
    }
    get a() {
        return this._a;
    }
    public _uvs = Tile.uvs;

    public eventsFilter(events?: [number, number, number, number][]): [number, number, number, number][] {
        return events;
    }

    public static positions(index: number, item: TileObject, data: Float32Array, strip: number = Tile.POINT_POSITION_COUNT, offset: number = 0, events?: [number, number, number, number][]) {
        let pointIndex = index * Tile.POINT_COUNT;
        let x = item.x;
        let y = item.y;
        let tileWidth = item.width;
        let tileHeight = item.height;
        let halfW = item.width / 2;
        let halfH = item.height / 2;

        let x1 = (x - halfW) * Tile.X_POSITION_SCALE;
        let x2 = (x + halfW) * Tile.X_POSITION_SCALE;
        let x3 = (x + halfW) * Tile.X_POSITION_SCALE;
        let x4 = (x - halfW) * Tile.X_POSITION_SCALE;
        let y1 = (y - halfH) * Tile.Y_POSITION_SCALE;
        let y2 = (y - halfH) * Tile.Y_POSITION_SCALE;
        let y3 = (y + halfH) * Tile.Y_POSITION_SCALE;
        let y4 = (y + halfH) * Tile.Y_POSITION_SCALE;

        let distance = 0;
        let dx1 = 0;
        let dy1 = 0;
        let dx2 = 0;
        let dy2 = 0;
        let dx3 = 0;
        let dy3 = 0;
        let dx4 = 0;
        let dy4 = 0;
        let ddx1 = 0;
        let ddy1 = 0;
        let ddx2 = 0;
        let ddy2 = 0;
        let ddx3 = 0;
        let ddy3 = 0;
        let ddx4 = 0;
        let ddy4 = 0;
        events?.forEach((item) => {
            let cx = Math.abs(x - item[0]);
            let cy = Math.abs(y - item[1]);
            if (cx > 0.55 || cy > 0.55) {
                if (cx > (cy + 0.2)) { cx = 1; cy = 0; } else if ((cx + 0.2) < cy) { cx = 0; cy = 1; } else { cx = 0.5; cy = 0.5; }
                let diff = item[3] * 0.5;
                dx1 = x1 - item[0]; dy1 = y1 - item[1]; dx1 *= cx; dy1 *= cy;
                distance = Math.sqrt(dx1 * dx1 + dy1 * dy1);
                ddx1 += distance > 0 ? dx1 / distance * (1.0 - Math.min(1., distance / item[2])) * diff : 0;
                ddy1 += distance > 0 ? dy1 / distance * (1.0 - Math.min(1., distance / item[2])) * diff : 0;
                dx2 = x2 - item[0]; dy2 = y2 - item[1]; dx2 *= cx; dy2 *= cy;
                distance = Math.sqrt(dx2 * dx2 + dy2 * dy2);
                ddx2 += distance > 0 ? dx2 / distance * (1.0 - Math.min(1., distance / item[2])) * diff : 0;
                ddy2 += distance > 0 ? dy2 / distance * (1.0 - Math.min(1., distance / item[2])) * diff : 0;
                dx3 = x3 - item[0]; dy3 = y3 - item[1]; dx3 *= cx; dy3 *= cy;
                distance = Math.sqrt(dx3 * dx3 + dy3 * dy3);
                ddx3 += distance > 0 ? dx3 / distance * (1.0 - Math.min(1., distance / item[2])) * diff : 0;
                ddy3 += distance > 0 ? dy3 / distance * (1.0 - Math.min(1., distance / item[2])) * diff : 0;
                dx4 = x4 - item[0]; dy4 = y4 - item[1]; dx4 *= cx; dy4 *= cy;
                distance = Math.sqrt(dx4 * dx4 + dy4 * dy4);
                ddx4 += distance > 0 ? dx4 / distance * (1.0 - Math.min(1., distance / item[2])) * diff : 0;
                ddy4 += distance > 0 ? dy4 / distance * (1.0 - Math.min(1., distance / item[2])) * diff : 0;
            }
        });

        data[pointIndex * strip + offset + 0] = x1 + ddx1;
        data[pointIndex * strip + offset + 1] = 0;
        data[pointIndex * strip + offset + 2] = y1 + ddy1;
        pointIndex += 1;

        data[pointIndex * strip + offset + 0] = x2 + ddx2;
        data[pointIndex * strip + offset + 1] = 0;
        data[pointIndex * strip + offset + 2] = y2 + ddy2;
        pointIndex += 1;

        data[pointIndex * strip + offset + 0] = x3 + ddx3;
        data[pointIndex * strip + offset + 1] = 0;
        data[pointIndex * strip + offset + 2] = y3 + ddy3;
        pointIndex += 1;

        data[pointIndex * strip + offset + 0] = x4 + ddx4;
        data[pointIndex * strip + offset + 1] = 0;
        data[pointIndex * strip + offset + 2] = y4 + ddy4;
        pointIndex += 1;
    }
    dispose() {
        //
    }
}